package ant.agent;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.messages.Command;
import rescuecore2.messages.control.KASense;
import rescuecore2.misc.collections.LazyMap;
import rescuecore2.standard.components.StandardAgent;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import ant.AbstractAntAgent;
import ant.AbstractAntCommunicationKit;
import ant.AgentContext;
import ant.AntComponentKit;
import ant.AntProperties;
import ant.agent.command.AntClear;
import ant.agent.command.AntCommand;
import ant.agent.command.AntExtinguish;
import ant.agent.command.AntLoad;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.command.AntRescue;
import ant.agent.command.AntRest;
import ant.agent.command.AntSpeak;
import ant.agent.command.AntSubscribe;
import ant.agent.command.AntUnload;
import ant.exception.AntComponentRegisterException;
import ant.liaison.LiaisonCommunicationKit;
import ant.liaison.SimpleAntCommunicationKit;

/**
 * 
 * @author Ivan
 * 
 * @param <E>
 */
public abstract class CommonAgent<E extends Human> extends StandardAgent<E> {
	private static final Log log = LogFactory.getLog(CommonAgent.class);
	
	private boolean hasException = false; 
	private AgentContext context;
	private AbstractAntAgent<E> ant;
	private AbstractAntCommunicationKit cmu;
	private AntComponentKit kit;
	private int ignoreTime;
	
	public CommonAgent(AbstractAntAgent<E> ant) {
		this.ant = ant;
	}

	@Override
	protected final void postConnect() {
		super.postConnect();
		
		try {
			postConnectInternal();
		} catch (Throwable e) {
			System.err.println("*****************************************");
			System.err.println("***********Initial  Exception************");
			System.err.println("*****************************************");
			e.printStackTrace();
			hasException = true;
			
			exceptionInitial();
		}
	}

	private void postConnectInternal() {
		try {
			this.kit = new AntComponentKit(getAgentType());

			if (AntProperties.isCommKitSimple())
				this.cmu = new SimpleAntCommunicationKit();
			else
				this.cmu = new LiaisonCommunicationKit();

		} catch (AntComponentRegisterException e) {
			throw new RuntimeException(e);
		}

		model.indexClass(StandardEntityURN.AMBULANCE_TEAM, StandardEntityURN.FIRE_BRIGADE,
				StandardEntityURN.POLICE_FORCE, StandardEntityURN.BUILDING,
				StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.FIRE_STATION,
				StandardEntityURN.POLICE_OFFICE, StandardEntityURN.ROAD, StandardEntityURN.REFUGE,
				StandardEntityURN.BUILDING, StandardEntityURN.BLOCKADE, StandardEntityURN.CIVILIAN);

		this.context = new AgentContext(this, this.model, this.cmu, this.kit, this.config);
		this.cmu.initial(context, this.model);
		this.kit.initial(context, this.model);
		this.ant.initial(context);
	}

	@Override
	protected final void processSense(KASense sense) {
		model.merge(sense.getChangeSet());
		
		if (!hasException) {
			try {
				processSenseInternal(sense);
			} catch (Throwable e) {
				System.err.println("*****************************************");
				System.err.println("************Update Exception*************");
				System.err.println("*****************************************");
				e.printStackTrace();
				hasException = true;
				
				exceptionInitial();
			}
		} else {
			Collection<Command> heard = sense.getHearing();
			exceptionThink(sense.getTime(), sense.getChangeSet(), heard);
		}
	}
	
	protected void exceptionInitial() {
		buildingIDs = new ArrayList<EntityID>();
        roadIDs = new ArrayList<EntityID>();
        refugeIDs = new ArrayList<EntityID>();
        for (StandardEntity next : model) {
            if (next instanceof Building) {
                buildingIDs.add(next.getID());
            }
            if (next instanceof Road) {
                roadIDs.add(next.getID());
            }
            if (next instanceof Refuge) {
                refugeIDs.add(next.getID());
            }
        }
        search = new SampleSearch(model);
        neighbours = search.getGraph();
        this.rand = new Random(1);
	}

	protected abstract void exceptionThink(int time, ChangeSet changed, Collection<Command> heard);

	private void processSenseInternal(KASense sense) {
		long st_time = System.currentTimeMillis();
		{
			this.context.setSense(sense);
			this.cmu.update(context);
			handleMessages(sense.getTime(), cmu.pullpreSubscribe(), cmu.pullpreSpeaks());
			this.kit.update(context);
			Collection<Command> heard = sense.getHearing();
			think(sense.getTime(), sense.getChangeSet(), heard);
		}
		long dtime = System.currentTimeMillis() - st_time;
		if (dtime > 600) {
			log.error("@" + sense.getTime() + " processSense time cost:" + dtime);
		} else if (dtime > 400) {
			log.warn("@" + sense.getTime() + " processSense time cost:" + dtime);
		}
	}

	@Override
	protected final void think(int time, ChangeSet changes, Collection<Command> heard) {
		AntCommand cc = this.ant.think();

		if (cc == null) {
			log.warn("Null AntCommand at time=" + time);
			cc = new AntRest(time);
		}

		this.context.getCommands().put(time, cc);

		if (cc instanceof AntExtinguish) {
			AntExtinguish tmp = ((AntExtinguish) cc);
			sendExtinguish(tmp.getTime(), tmp.getTarget(), tmp.getWater());
		} else if (cc instanceof AntLoad) {
			AntLoad tmp = ((AntLoad) cc);
			sendLoad(tmp.getTime(), tmp.getTarget());
		} else if (cc instanceof AntMove) {
			AntMove tmp = ((AntMove) cc);
			sendMove(tmp.getTime(), tmp.getPath());
		} else if (cc instanceof AntMoveXY) {
			AntMoveXY tmp = ((AntMoveXY) cc);
			sendMove(tmp.getTime(), tmp.getPath(), tmp.getDestX(), tmp.getDestY());
		} else if (cc instanceof AntRescue) {
			AntRescue tmp = ((AntRescue) cc);
			sendRescue(tmp.getTime(), tmp.getTarget());
		} else if (cc instanceof AntRest) {
			AntRest tmp = ((AntRest) cc);
			sendRest(tmp.getTime());
		} else if (cc instanceof AntSubscribe) {
			AntSubscribe tmp = ((AntSubscribe) cc);
			sendSubscribe(tmp.getTime(), tmp.getChannels());
		} else if (cc instanceof AntClear) {
			AntClear tmp = ((AntClear) cc);
			sendClear(tmp.getTime(), tmp.getTarget());
		} else if (cc instanceof AntUnload) {
			AntUnload tmp = ((AntUnload) cc);
			sendUnload(tmp.getTime());
		} else {
			// throw new RuntimeException();
			log.warn("Unknown AntCommand :" + cc + " at time=" + time);
		}

		handleMessages(time, cmu.pullSubscribe(), cmu.pullallSpeak());

	}

	private void handleMessages(int time, AntSubscribe sub, List<AntSpeak> speaks) {
		if (sub != null) {
			// TODO
			if (sub.getTime() != time) {
				log.warn("Subscribe time error, now=" + time + " sub:" + sub);
			} else {
				sendSubscribe(time, sub.getChannels());
			}
		}
		if (speaks == null) {
			log.warn("Null speaks");
		} else {
			for (AntSpeak speak : speaks) {
				if (speak.getTime() != time) {
					log.warn("Speak time error, now=" + time + " speak:" + speak);
				} else {
					sendSpeak(time, speak.getChannel(), speak.getData());
				}
			}
		}
	}

	public abstract StandardEntityURN getAgentType();

	public final E getHuman() {
		return me();
	}

	
	
	
	
	
	
	
	
	
	
	
	
	
	/**********************************************************/
	private static final int RANDOM_WALK_LENGTH = 50;
	
	protected SampleSearch search;
    protected List<EntityID> buildingIDs;
    protected List<EntityID> roadIDs;
    protected List<EntityID> refugeIDs;
    protected Map<EntityID, Set<EntityID>> neighbours;
	private Random rand;

    public static final class SampleSearch {
        private Map<EntityID, Set<EntityID>> graph;

        /**
           Construct a new SampleSearch.
           @param world The world model to construct the neighbourhood graph from.
        */
        public SampleSearch(StandardWorldModel world) {
            Map<EntityID, Set<EntityID>> neighbours = new LazyMap<EntityID, Set<EntityID>>() {
                @Override
                public Set<EntityID> createValue() {
                    return new HashSet<EntityID>();
                }
            };
            for (Entity next : world) {
                if (next instanceof Area) {
                    Collection<EntityID> areaNeighbours = ((Area)next).getNeighbours();
                    neighbours.get(next.getID()).addAll(areaNeighbours);
                }
            }
            setGraph(neighbours);
        }


        /**
           Construct a new ConnectionGraphSearch.
           @param graph The connection graph in the form of a map from EntityID to the set of neighbouring EntityIDs.
         */
        public SampleSearch(Map<EntityID, Set<EntityID>> graph) {
            setGraph(graph);
        }

        /**
           Set the neighbourhood graph.
           @param newGraph The new neighbourhood graph.
        */
        public void setGraph(Map<EntityID, Set<EntityID>> newGraph) {
            this.graph = newGraph;
        }

        /**
           Get the neighbourhood graph.
           @return The neighbourhood graph.
        */
        public Map<EntityID, Set<EntityID>> getGraph() {
            return graph;
        }

        /**
           Do a breadth first search from one location to the closest (in terms of number of nodes) of a set of goals.
           @param start The location we start at.
           @param goals The set of possible goals.
           @return The path from start to one of the goals, or null if no path can be found.
        */
        public List<EntityID> breadthFirstSearch(EntityID start, EntityID... goals) {
            return breadthFirstSearch(start, Arrays.asList(goals));
        }

        /**
           Do a breadth first search from one location to the closest (in terms of number of nodes) of a set of goals.
           @param start The location we start at.
           @param goals The set of possible goals.
           @return The path from start to one of the goals, or null if no path can be found.
        */
        public List<EntityID> breadthFirstSearch(EntityID start, Collection<EntityID> goals) {
            List<EntityID> open = new LinkedList<EntityID>();
            Map<EntityID, EntityID> ancestors = new HashMap<EntityID, EntityID>();
            open.add(start);
            EntityID next = null;
            boolean found = false;
            ancestors.put(start, start);
            do {
                next = open.remove(0);
                if (isGoal(next, goals)) {
                    found = true;
                    break;
                }
                Collection<EntityID> neighbours = graph.get(next);
                if (neighbours.isEmpty()) {
                    continue;
                }
                for (EntityID neighbour : neighbours) {
                    if (isGoal(neighbour, goals)) {
                        ancestors.put(neighbour, next);
                        next = neighbour;
                        found = true;
                        break;
                    }
                    else {
                        if (!ancestors.containsKey(neighbour)) {
                            open.add(neighbour);
                            ancestors.put(neighbour, next);
                        }
                    }
                }
            } while (!found && !open.isEmpty());
            if (!found) {
                // No path
                return null;
            }
            // Walk back from goal to start
            EntityID current = next;
            List<EntityID> path = new LinkedList<EntityID>();
            do {
                path.add(0, current);
                current = ancestors.get(current);
                if (current == null) {
                    throw new RuntimeException("Found a node with no ancestor! Something is broken.");
                }
            } while (current != start);
            return path;
        }

        private boolean isGoal(EntityID e, Collection<EntityID> test) {
            return test.contains(e);
        }
    }
    
    protected List<EntityID> randomWalk() {
        List<EntityID> result = new ArrayList<EntityID>(RANDOM_WALK_LENGTH);
        Set<EntityID> seen = new HashSet<EntityID>();
        EntityID current = ((Human) getHuman()).getPosition();
        for (int i = 0; i < RANDOM_WALK_LENGTH; ++i) {
            result.add(current);
            seen.add(current);
            List<EntityID> possible = new ArrayList<EntityID>(neighbours.get(current));
            Collections.shuffle(possible, rand);
            boolean found = false;
            for (EntityID next : possible) {
                if (seen.contains(next)) {
                    continue;
                }
                current = next;
                found = true;
                break;
            }
            if (!found) {
                // We reached a dead-end.
                break;
            }
        }
        return result;
    }
    
    public static class DistanceSorter implements Comparator<StandardEntity> {
        private StandardEntity reference;
        private StandardWorldModel world;

        /**
           Create a DistanceSorter.
           @param reference The reference point to measure distances from.
           @param world The world model.
        */
        public DistanceSorter(StandardEntity reference, StandardWorldModel world) {
            this.reference = reference;
            this.world = world;
        }

        @Override
        public int compare(StandardEntity a, StandardEntity b) {
            int d1 = world.getDistance(reference, a);
            int d2 = world.getDistance(reference, b);
            return d1 - d2;
        }
    }
    
    public static Set<EntityID> objectsToIDs(Collection<? extends Entity> entities) {
        Set<EntityID> result = new HashSet<EntityID>();
        for (Entity next : entities) {
            result.add(next.getID());
        }
        return result;
    }
}
