package ant.agent;

import java.util.ArrayList;
import java.util.Collection;
import java.util.EnumSet;
import java.util.List;

import rescuecore2.log.Logger;
import rescuecore2.messages.Command;
import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.PoliceForce;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.worldmodel.ChangeSet;
import rescuecore2.worldmodel.EntityID;
import ant.AbstractAntAgent;

/**
 * 
 * @author Ivan
 * 
 */
public final class CommonPoliceForce extends CommonAgent<PoliceForce> {
	private static final String DISTANCE_KEY = "clear.repair.distance";
	private int distance;

	public CommonPoliceForce(AbstractAntAgent<PoliceForce> ant) {
		super(ant);
	}

	@Override
	protected final EnumSet<StandardEntityURN> getRequestedEntityURNsEnum() {
		return EnumSet.of(StandardEntityURN.POLICE_FORCE);
	}

	@Override
	public final StandardEntityURN getAgentType() {
		return StandardEntityURN.POLICE_FORCE;
	}

	@Override
	protected void exceptionInitial() {
		super.exceptionInitial();
		distance = config.getIntValue(DISTANCE_KEY);
	}

	@Override
	protected void exceptionThink(int time, ChangeSet changed, Collection<Command> heard) {
		if (time == 3) {// config.getIntValue(kernel.KernelConstants.IGNORE_AGENT_COMMANDS_KEY))
			// {
			// Subscribe to channel 1
			sendSubscribe(time, 1);
		}
		// Am I near a blockade?
		Blockade target = getTargetBlockade();
		if (target != null) {
			sendClear(time, target.getID());
			return;
		}
		// Plan a path to a blocked area
		List<EntityID> path = search.breadthFirstSearch(me().getPosition(), getBlockedRoads());
		if (path != null) {
			Road r = (Road) model.getEntity(path.get(path.size() - 1));
			Blockade b = getTargetBlockade(r, -1);
			sendMove(time, path, b.getX(), b.getY());
			return;
		}
		sendMove(time, randomWalk());

	}
	
	private List<EntityID> getBlockedRoads() {
		Collection<StandardEntity> e = model
				.getEntitiesOfType(StandardEntityURN.ROAD);
		List<EntityID> result = new ArrayList<EntityID>();
		for (StandardEntity next : e) {
			Road r = (Road) next;
			if (r.isBlockadesDefined() && !r.getBlockades().isEmpty()) {
				result.add(r.getID());
			}
		}
		return result;
	}

	private Blockade getTargetBlockade() {
		Area location = (Area) location();
		Blockade result = getTargetBlockade(location, distance);
		if (result != null) {
			return result;
		}
		for (EntityID next : location.getNeighbours()) {
			location = (Area) model.getEntity(next);
			result = getTargetBlockade(location, distance);
			if (result != null) {
				return result;
			}
		}
		return null;
	}

	private Blockade getTargetBlockade(Area area, int maxDistance) {
		// Logger.debug("Looking for nearest blockade in " + area);
		if (area == null || !area.isBlockadesDefined()) {
			// Logger.debug("Blockades undefined");
			return null;
		}
		List<EntityID> ids = area.getBlockades();
		// Find the first blockade that is in range.
		int x = me().getX();
		int y = me().getY();
		for (EntityID next : ids) {
			Blockade b = (Blockade) model.getEntity(next);
			double d = findDistanceTo(b, x, y);
			// Logger.debug("Distance to " + b + " = " + d);
			if (maxDistance < 0 || d < maxDistance) {
				// Logger.debug("In range");
				return b;
			}
		}
		// Logger.debug("No blockades in range");
		return null;
	}

	private int findDistanceTo(Blockade b, int x, int y) {
		// Logger.debug("Finding distance to " + b + " from " + x + ", " + y);
		List<Line2D> lines = GeometryTools2D.pointsToLines(
				GeometryTools2D.vertexArrayToPoints(b.getApexes()), true);
		double best = Double.MAX_VALUE;
		Point2D origin = new Point2D(x, y);
		for (Line2D next : lines) {
			Point2D closest = GeometryTools2D.getClosestPointOnSegment(next,
					origin);
			double d = GeometryTools2D.getDistance(origin, closest);
			// Logger.debug("Next line: " + next + ", closest point: " + closest
			// + ", distance: " + d);
			if (d < best) {
				best = d;
				// Logger.debug("New best distance");
			}

		}
		return (int) best;
	}
	
}
