package ant.agent.job.fb;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.agent.behavior.FBBehavior;
import ant.agent.command.AntCommand;
import ant.agent.command.AntExtinguish;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.job.JobState;
import ant.component.fb.AbstractFireConvexHullComponent.Hull;
import ant.component.fb.EstimationFireConvexHullComponent;
import ant.component.fb.EstimationFireConvexHullComponent.HullForEstimation;
import ant.component.fb.ExtinguishRecordComponent;
import ant.component.fb.ExtinguishRecordComponent.BuildingExtinguishRecord;
import ant.component.fb.FireBuilding;
import ant.component.fb.FireBuildingComparatorWithFSC;
import ant.component.fb.FireConfirmComponent;
import ant.component.fb.FireConvexHullComponent;
import ant.component.fb.FireConvexHullPickComponent;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.liaison.channel.assign.ChannelAssignResult.BandwidthLevel;
import ant.liaison.message.MTargeBuildingUnreachable;
import ant.liaison.message.MVoiceConfirmBuilding;
import ant.liaison.message.MVoiceExtinguishBuilding;
import ant.map.component.AbstractRouteComponent;
import ant.map.objects.RouteResult;
import ant.utils.Pair;
import ant.utils.geo.Point;

/**
 * 
 * @author Oliver
 * 
 */
public class ExtinguishEntityJobWithFSC extends AbstractExtinguishJob {
	private static Log log = LogFactory.getLog(ExtinguishEntityJobWithFSC.class);

	private static final int MAX_CANDITATE_SIZE = 40;

	private static final int CONDITION_RECENT_LEVEL_MAX = 5;
	private static final double COEFFICIENT_EXTINGUISH_TIME = 0.5;
	private static final int COMMAND_MAX = 4;
	private static final double CONFIRM_ALONE_DISTANCE = 100000;

	private static final int MIN_HULL_RADIUS = 25000;
	private static final int MIN_BUILDING_SIZE_IN_HULL = 4;
	private static final double NEAR_EDGE_DISTANCE = 50;
	private static final int VERY_RECENT_TIME_LEVEL = 2;

	private static final double LITTLE_WATER_POWER_COEFFICIENT = 3.1;
	private static final double LITTLE_WATER_TOTAL_COEFFICIENT = 0.25;

	private AbstractRouteComponent arc;
	private ExtinguishRecordComponent erc;
	private FireBuildingComparatorWithFSC fbc;

	private Set<FireBuilding> ignoredBuildings;

	private ExtinguishPositionSearcher eps;
	private List<EntityID> path;
	private Point pos;
	private CurrentState currentState;
	private EntityID confirmTarget;

	private FireConfirmComponent fcc;
	private EstimationFireConvexHullComponent efc;
	private FireConvexHullPickComponent fcp;
	private FireConvexHullComponent fch;

	private enum CurrentState {
		ConfirmingFire, GoToExtinguishFire, ExtinguishingFire
	}

	public ExtinguishEntityJobWithFSC(AgentContext context) {
		super(context);
		this.arc = context.getAntComponentKit().getAbstractRouteComponent();
		this.erc = (ExtinguishRecordComponent) context.getAntComponentKit().getAntComponent(
				ExtinguishRecordComponent.class);
		this.fcc = (FireConfirmComponent) context.getAntComponentKit().getAntComponent(
				FireConfirmComponent.class);
		this.fch = (FireConvexHullComponent) context.getAntComponentKit().getAntComponent(
				FireConvexHullComponent.class);
		this.efc = (EstimationFireConvexHullComponent) context.getAntComponentKit()
				.getAntComponent(EstimationFireConvexHullComponent.class);
		this.fcp = (FireConvexHullPickComponent) context.getAntComponentKit().getAntComponent(
				FireConvexHullPickComponent.class);
		this.fbc = new FireBuildingComparatorWithFSC(context);
		this.ignoredBuildings = new HashSet<FireBuilding>();
	}

	@Override
	public void overtaken() {
		this.state = JobState.Inactive;
		currentState = null;
		path = null;
		pos = null;
		eps = null;
		confirmTarget = null;
	}

	@Override
	public void overleap() {
		super.overleap();
		overtaken();
	}

	@Override
	public boolean shouldExecute() {
		long sTime = System.currentTimeMillis();
		if (!canExtinguish()) {
			long dTime = System.currentTimeMillis() - sTime;
			if (dTime > 200) {
				log.error("ExtinguishEntityJobWithFSC cost:" + dTime);
			}
			return false;
		}

		List<FireBuilding> targets = searchNextTarget();
		if (targets.isEmpty()) {
			long dTime = System.currentTimeMillis() - sTime;
			if (dTime > 200) {
				log.error("ExtinguishEntityJobWithFSC cost:" + dTime + ", targets:"
						+ targets.size());
			}
			return false;
		}

		int count = 0;
		for (FireBuilding fireBuilding : targets) {
			count++;
			Building b = fireBuilding.getBuilding();
			if (fcc.shouldConfirmBuildingCondition(fireBuilding)) {
				Pair<Point, Road> confirmLocation = ((FBBehavior) behavior)
						.getPositionInRoadToMove(b);
				if (confirmLocation != null) {
					RouteResult route = arc.search(me.getPosition(), me.getX(), me.getY(),
							confirmLocation.second().getID());
					if (route.hasAcceptableRoute()) {
						this.path = route.getMoveRoute();
						this.pos = confirmLocation.first();
						this.confirmTarget = b.getID();
						currentState = CurrentState.ConfirmingFire;
						long dTime = System.currentTimeMillis() - sTime;
						if (dTime > 200) {
							log.error("ExtinguishEntityJobWithFSC cost:" + dTime + ", targets:"
									+ targets.size() + ", use:" + count);
						}
						return true;
					}
				}
				// confirmLocation == null, go to building directly
				RouteResult route = arc.search(me.getPosition(), me.getX(), me.getY(), b.getID());
				if (route.hasAcceptableRoute()) {
					this.path = route.getMoveRoute();
					this.pos = new Point(b.getX(), b.getY());
					this.confirmTarget = b.getID();
					currentState = CurrentState.ConfirmingFire;
					long dTime = System.currentTimeMillis() - sTime;
					if (dTime > 200) {
						log.error("ExtinguishEntityJobWithFSC cost:" + dTime + ", targets:"
								+ targets.size() + ", use:" + count);
					}
					return true;
				} else {
					// FIXME, this actually need to be confirmed
					ignoredBuildings.add(fireBuilding);
					continue;
				}
			} else {
				// no need to confirm fire
				if (!shouldExtinguishBuilding(fireBuilding)) {
					continue;
				}
				eps = new ExtinguishPositionSearcher(context, fireBuilding);
				if (eps.haveArrivedPosition()) {
					// no need to care about eps isValid
					currentState = CurrentState.ExtinguishingFire;
					long dTime = System.currentTimeMillis() - sTime;
					if (dTime > 200) {
						log.error("ExtinguishEntityJobWithFSC cost:" + dTime + ", targets:"
								+ targets.size() + ", use:" + count);
					}
					return true;
				} else if (!eps.isPositionValid()) {
					// TODO is it reasonable to ignore?
					ignoredBuildings.add(fireBuilding);
					continue;
				} else {
					// and eps isValid
					RouteResult route = arc.search(me.getPosition(), me.getX(), me.getY(),
							eps.getPositionID());
					if (route.hasAcceptableRoute()) {
						this.path = route.getMoveRoute();
						this.pos = eps.getExtinguishCoordinate();
						currentState = CurrentState.GoToExtinguishFire;
						long dTime = System.currentTimeMillis() - sTime;
						if (dTime > 200) {
							log.error("ExtinguishEntityJobWithFSC cost:" + dTime + ", targets:"
									+ targets.size() + ", use:" + count);
						}
						return true;
					} else {
						log.error("Position unreachable!");
						continue;
					}
				}
			}
		}

		// log.info("No position to extinguish!");
		// cmu.sendRadio(new
		// MTargeBuildingUnreachable(targets.get(0).getBuilding().getID()));
		addRadioMessage(new MTargeBuildingUnreachable(targets.get(0).getBuilding().getID()));
		addVoiceMessage(new MTargeBuildingUnreachable(targets.get(0).getBuilding().getID()));

		long dTime = System.currentTimeMillis() - sTime;
		if (dTime > 200) {
			log.error("ExtinguishEntityJobWithFSC cost:" + dTime + ", targets:" + targets.size()
					+ ", use:" + count);
		}
		return false;
	}

	private boolean shouldExtinguishBuilding(FireBuilding fireBuilding) {
		if (!fireBuilding.getCondition().isOnFire()) {
			return false;
		}

		int timeNow = context.getTimestamp();
		BuildingCondition bc = fireBuilding.getCondition();
		int buildingUpdateTime = bc.getTimestamp();

		if (timeNow - buildingUpdateTime <= 1) {
			// just updated
			return true;
		}
		if (timeNow - buildingUpdateTime > CONDITION_RECENT_LEVEL_MAX) {
			// long time no update
			return false;
		}

		BuildingExtinguishRecord record = erc.getExtinguishRecord(fireBuilding.getBuilding()
				.getID());
		if (record == null) {
			return true;
		}
		double needExtinguishCount = fireBuilding.getWaterQuantityToExtinguish() / maxPower;
		if (record.getExtinguishCountSinceLastUpdate(buildingUpdateTime) > Math.min(
				needExtinguishCount * COEFFICIENT_EXTINGUISH_TIME, COMMAND_MAX)) {
			// extinguish more time than it should be
			return false;
		}
		return true;
	}

	@Override
	public AntCommand execute() {
		if (currentState == CurrentState.ConfirmingFire) {
			// use voice to tell others I'm confirming
			// cmu.sendVoice(new MVoiceConfirmBuilding(confirmTarget,
			// me.getPosition()));
			addVoiceMessage(new MVoiceConfirmBuilding(confirmTarget, me.getPosition()));

			fcc.addSelfConfirmRecord(confirmTarget);
			return new AntMoveXY(context.getTimestamp(), path, pos.getX(), pos.getY());
		} else if (currentState == CurrentState.GoToExtinguishFire) {
			return new AntMoveXY(context.getTimestamp(), path, pos.getX(), pos.getY());
		} else if (currentState == CurrentState.ExtinguishingFire) {
			// use voice to tell others I'm extinguishing
			EntityID targetID = eps.getTargetFireBuilding().getBuilding().getID();
			// cmu.sendVoice(new MVoiceExtinguishBuilding(targetID));
			addVoiceMessage(new MVoiceExtinguishBuilding(targetID));

			return new AntExtinguish(context.getTimestamp(), targetID, maxPower);
		} else {
			log.error("CurrentState:" + currentState);
			return new AntMove(context.getTimestamp(), behavior.getRandomWalk());
		}
	}

	@Override
	public String getJobName() {
		return getClass().getName();
	}

	private boolean isGoingOutOfWater() {
		int w = agent.getWater();
		if (w < maxPower * LITTLE_WATER_POWER_COEFFICIENT
				|| w < maxWater * LITTLE_WATER_TOTAL_COEFFICIENT) {
			return true;
		}
		return false;
	}

	private List<FireBuilding> searchNextTarget() {
		List<FireBuilding> rv;
		BandwidthLevel bandwidthLevel = context.getAntCommunicationKit().getBandwidthLevel();
		HullForEstimation h = fcp.getHullForMyself();
		if (isGoingOutOfWater() || bandwidthLevel == BandwidthLevel.Unknown
				|| bandwidthLevel == BandwidthLevel.None || bandwidthLevel == BandwidthLevel.Poor
				|| !canAccessToHull(h)) {
			rv = getPoorBandwidthTargets();
		} else {
			rv = new ArrayList<FireBuilding>();
			if (h == null) {
				return rv;
			}
			Set<FireBuilding> worthyBuildings = h.getWorthyBuildings();
			for (FireBuilding fb : h.getBuildingsInHull()) {
				if (worthyBuildings.contains(fb)) {
					rv.add(fb);
				} else if (fb.getCurrentEstiFierynessEnum() == Fieryness.HEATING) {
					rv.add(fb);
				}
			}
		}
		Collections.sort(rv, fbc);
		// FIXME
		if (rv.size() > MAX_CANDITATE_SIZE) {
			rv = rv.subList(0, MAX_CANDITATE_SIZE);
		}
		return rv;
	}

	private boolean canAccessToHull(Hull hull) {
		if (hull == null) {
			return false;
		}
		for (FireBuilding fb : hull.getBuildingsOnHull()) {
			RouteResult route = arc.search(me.getPosition(), me.getX(), me.getY(), fb.getBuilding()
					.getID());
			if (route.hasAcceptableRoute()) {
				return true;
			}
		}
		return false;
	}

	private boolean shouldHullBeExtinguished(HullForEstimation hull) {
		if (hull.getWorthyBuildings().isEmpty()) {
			return false;
		}
		if (hull.getWaterQuantityToExtinguish() <= 0) {
			return false;
		}
		if (hull.isCoveredByOtherHull()) {
			return false;
		}
		return true;
	}

	private List<FireBuilding> getPoorBandwidthTargets() {
		List<FireBuilding> rv = new ArrayList<FireBuilding>();
		List<HullForEstimation> hullList = new ArrayList<HullForEstimation>();
		for (HullForEstimation hull : efc.getHulls()) {
			if (shouldHullBeExtinguished(hull) && !hull.isLittleOuterAffection()
					&& !hull.isDyingHull()) {
				hullList.add(hull);
			}
		}
		if (hullList.isEmpty()) {
			for (HullForEstimation hull : efc.getHulls()) {
				if (shouldHullBeExtinguished(hull)) {
					hullList.add(hull);
				}
			}
		}
		for (HullForEstimation hull : hullList) {
			Set<FireBuilding> worthyBuildings = hull.getWorthyBuildings();
			for (FireBuilding fireBuilding : hull.getBuildingsInHull()) {
				if (worthyBuildings.contains(fireBuilding)) {
					rv.add(fireBuilding);
				} else if (fireBuilding.getCurrentEstiFierynessEnum() == Fieryness.HEATING) {
					rv.add(fireBuilding);
				}
			}
		}
		return rv;
	}

	private boolean isBuildingQualified(FireBuilding fireBuilding, Hull hull) {
		BuidlingQualification qualification = fireBuilding.getQualification();
		if (qualification.isValid()) {
			qualification.clear();
		}

		BuildingCondition condition = fireBuilding.getCondition();
		if (!condition.isOnFire()) {
			// building should at least on fire
			return false;
		}

		int radius = hull.getPolygon().getRadius();
		if (radius < MIN_HULL_RADIUS
				|| hull.getBuildingsInHull().size() < MIN_BUILDING_SIZE_IN_HULL) {
			// building is in really small hull, no need to further
			// qualification
			qualification.setQualificationValue(radius, true);
			return true;
		}

		int x = fireBuilding.getBuilding().getX();
		int y = fireBuilding.getBuilding().getY();
		double distanceToNearestEdge = hull.getDistanceToNearestEdge(x, y);
		if (distanceToNearestEdge < radius * Hull.HULL_OUTER_RANGE
				|| distanceToNearestEdge <= NEAR_EDGE_DISTANCE) {
			// building in the outer range can be qualified
			// TODO better judgment on radius and distanceToNearestEdge
			qualification.setQualificationValue(radius, false, distanceToNearestEdge);
			return true;
		}

		// TODO building not in outer range but hasn't been badly burnt
		if (context.getTimestamp() - condition.getTimestamp() > VERY_RECENT_TIME_LEVEL) {
			if (condition.getFieryness() == 1) {
				qualification.setQualificationValue(radius, false, distanceToNearestEdge);
				return true;
			}
		}

		return false;
	}
}
