package ant.agent.job.fb;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.config.Config;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.ConfigKey;
import ant.component.fb.FireBuilding;
import ant.component.universal.BuildingConditionComponent;
import ant.map.component.AbstractRouteComponent;
import ant.map.objects.RouteResult;
import ant.utils.geo.Point;

/**
 * 
 * @author Oliver
 * 
 */
public class ExtinguishPositionSearcher {
	private static Log log = LogFactory.getLog(ExtinguishPositionSearcher.class);

	private AgentContext context;
	/**
	 * POSITION_XY_TO_CENTER is the POSITION_XY weight, and (1-
	 * POSITION_XY_TO_CENTER) is center weight
	 */
	private static final double POSITION_XY_TO_CENTER = 0.80;
	private static final int POSITION_SCANNER_EXTENDER = 5000;
	private static final double LONG_DISTANCE_TO_TARGET = 150000;
	private static final int MAX_ROUTE_SEARCH_COUNT = 20;
	private int maxDistance;
	private ArrayList<Area> roadPositions;
	private ArrayList<Area> buildingPositions;
	/**
	 * the area where agent stands to extinguish fire
	 */
	private Area position;
	/**
	 * the x Coordinate where agent stands to extinguish fire
	 */
	private int position_x;
	/**
	 * the y Coordinate where agent stands to extinguish fire
	 */
	private int position_y;

	private int maxExtinguishDistance;

	private EntityID targetID;
	private Building targetBuilding;
	private FireBuilding targetFireBuilding;

	private Human agent;
	private AbstractRouteComponent arc;
	private BuildingConditionComponent bcc;
	private DistanceComparatorForEPS dc;

	private boolean isPositionValid = false;
	private double distanceToTarget;

	public ExtinguishPositionSearcher(AgentContext context, FireBuilding targetFireBuilding) {
		this.context = context;
		Config config = context.getConfig();
		this.maxExtinguishDistance = ConfigKey.getFireMaxDistance(config);
		this.arc = context.getAntComponentKit().getAbstractRouteComponent();
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		this.agent = context.getHuman();

		this.roadPositions = new ArrayList<Area>();
		this.buildingPositions = new ArrayList<Area>();

		this.targetFireBuilding = targetFireBuilding;
		this.targetBuilding = targetFireBuilding.getBuilding();
		this.targetID = targetBuilding.getID();
		this.dc = new DistanceComparatorForEPS(context.getHuman(), context.getWorld(),
				targetBuilding);
		initial();
	}

	public void initial() {
		long sTime = System.currentTimeMillis();

		this.maxDistance = maxExtinguishDistance;
		this.distanceToTarget = distance(agent.getX(), agent.getY(), targetBuilding.getX(),
				targetBuilding.getY());

		for (StandardEntity e : context.getWorld().getObjectsInRange(targetID,
				maxExtinguishDistance + POSITION_SCANNER_EXTENDER)) {
			if (e instanceof Road) {
				searchPosition((Area) e, roadPositions);
			} else if (e instanceof Building && !e.getID().equals(targetID)) {
				searchPosition((Area) e, buildingPositions);
			}
		}

		if (setPosition()) {
			// success
			isPositionValid = true;
		} else {
			// failure
			// TODO
			isPositionValid = false;
		}

		long dTime = System.currentTimeMillis() - sTime;
		if (dTime > 50) {
			log.info("timeout:" + dTime + " roadPositions:" + roadPositions.size()
					+ " buildingPositions:" + buildingPositions.size());
		}
	}

	/**
	 * TODO better decision about the position (road or building) at which we
	 * can make less move to extinguish
	 * 
	 * @return true if a reachable position is found, false for otherwise
	 */
	private boolean setPosition() {
		position = null;
		double minDistance = Double.MAX_VALUE;
		EntityID here = agent.getPosition();
		int here_x = agent.getX();
		int here_y = agent.getY();
		int routeSearchCount = 0;
		if (roadPositions.size() != 0) {
			Collections.sort(roadPositions, dc);
			for (Area area : roadPositions) {
				if (routeSearchCount > MAX_ROUTE_SEARCH_COUNT) {
					break;
				} else {
					routeSearchCount++;
				}
				RouteResult route = arc.search(here, here_x, here_y, area.getID());
				if (route.hasAcceptableRoute() && route.getDistance() < minDistance) {
					minDistance = route.getDistance();
					position = area;
					if (distanceToTarget > LONG_DISTANCE_TO_TARGET) {
						break;
					}
				}
			}
		}

		if (position == null && buildingPositions.size() != 0) {
			Collections.sort(buildingPositions, dc);
			for (Area area : buildingPositions) {
				if (routeSearchCount > MAX_ROUTE_SEARCH_COUNT) {
					break;
				} else {
					routeSearchCount++;
				}
				RouteResult route = arc.search(here, here_x, here_y, area.getID());
				if (route.hasAcceptableRoute() && route.getDistance() < minDistance) {
					minDistance = route.getDistance();
					position = area;
					if (distanceToTarget > LONG_DISTANCE_TO_TARGET) {
						break;
					}
				}
			}
		}

		if (position == null) {
			// log.warn("All position is unreachable");
			return false;
		}

		getPositionCoordinate();
		return true;
	}

	private void searchPosition(Area area, ArrayList<Area> positions) {
		if (positions.contains(area))
			return;
		/**
		 * judge from center
		 */
		if (distance(area.getX(), area.getY(), targetBuilding.getX(), targetBuilding.getY()) <= maxDistance) {
			positions.add(area);
			return;
		}

		/**
		 * search apex
		 */
		for (int i = 0; i < area.getApexList().length - 1; i += 2) {
			if (distance((int) (area.getApexList()[i] * POSITION_XY_TO_CENTER + area.getX()
					* (1 - POSITION_XY_TO_CENTER)), (int) (area.getApexList()[i + 1]
					* POSITION_XY_TO_CENTER + area.getY() * (1 - POSITION_XY_TO_CENTER)),
					targetBuilding.getX(), targetBuilding.getY()) <= maxDistance) {
				positions.add(area);
				return;
			}
		}

		/**
		 * search edge midpoint
		 */
		for (Edge edge : area.getEdges()) {
			if (distance(
					(int) ((edge.getStartX() + edge.getEndX()) / 2 * POSITION_XY_TO_CENTER + area.getX()
							* (1 - POSITION_XY_TO_CENTER)),
					(int) ((edge.getStartY() + edge.getEndY()) / 2 * POSITION_XY_TO_CENTER + area
							.getY() * (1 - POSITION_XY_TO_CENTER)), targetBuilding.getX(),
					targetBuilding.getY()) <= maxDistance) {
				positions.add(area);
				return;
			}
		}
	}

	/**
	 * TODO what if position has block on it, and it is better if the coordinate
	 * is the closest of all
	 */
	private void getPositionCoordinate() {
		double minDistance = Double.MAX_VALUE;
		boolean hasCoordinate = false;
		double tmpDistance;
		int x, y;

		x = position.getX();
		y = position.getY();
		tmpDistance = distance(x, y, targetBuilding.getX(), targetBuilding.getY());
		if ((!hasCoordinate && tmpDistance <= maxDistance)
				|| (hasCoordinate && tmpDistance <= minDistance)) {
			position_x = x;
			position_y = y;
			minDistance = tmpDistance;
			hasCoordinate = true;
		}
		/**
		 * search apex
		 */
		for (int i = 0; i < position.getApexList().length - 1; i += 2) {
			x = (int) (position.getApexList()[i] * POSITION_XY_TO_CENTER + position.getX()
					* (1 - POSITION_XY_TO_CENTER));
			y = (int) (position.getApexList()[i + 1] * POSITION_XY_TO_CENTER + position.getY()
					* (1 - POSITION_XY_TO_CENTER));
			tmpDistance = distance(x, y, targetBuilding.getX(), targetBuilding.getY());
			if ((!hasCoordinate && tmpDistance <= maxDistance)
					|| (hasCoordinate && tmpDistance <= minDistance)) {
				position_x = x;
				position_y = y;
				minDistance = tmpDistance;
				hasCoordinate = true;
			}
		}

		/**
		 * search edge midpoint
		 */
		for (Edge edge : position.getEdges()) {
			x = (int) ((edge.getStartX() + edge.getEndX()) / 2 * POSITION_XY_TO_CENTER + position
					.getX() * (1 - POSITION_XY_TO_CENTER));
			y = (int) ((edge.getStartY() + edge.getEndY()) / 2 * POSITION_XY_TO_CENTER + position
					.getY() * (1 - POSITION_XY_TO_CENTER));
			tmpDistance = distance(x, y, targetBuilding.getX(), targetBuilding.getY());
			if ((!hasCoordinate && tmpDistance <= maxDistance)
					|| (hasCoordinate && tmpDistance <= minDistance)) {
				position_x = x;
				position_y = y;
				minDistance = tmpDistance;
				hasCoordinate = true;
			}
		}

		if (!hasCoordinate) {
			log.fatal("No coordinate found!");
		}
	}

	/**
	 * FIXME 站在着火建筑物内灭火怎么办？
	 * 
	 * @return
	 */
	public boolean haveArrivedPosition() {
		if (distance(agent.getX(), agent.getY(), targetBuilding.getX(), targetBuilding.getY()) <= maxDistance) {
			StandardEntity location = agent.getPosition(context.getWorld());
			if (location instanceof Building && ((Building) location).isOnFire()) {
				// 站在一个前往目标点路上的着火建筑内
				if (position == null) {
					return true;
				} else {
					return location.getID().equals(position.getID());
				}
			} else {
				return true;
			}
		} else {
			return false;
		}
	}

	private static double distance(int x1, int y1, int x2, int y2) {
		return Math.hypot(x1 - x2, y1 - y2);
	}

	public Area getPosition() {
		return position;
	}

	public EntityID getPositionID() {
		return position.getID();
	}

	public Point getExtinguishCoordinate() {
		return new Point(position_x, position_y);
	}

	public boolean isPositionValid() {
		return isPositionValid;
	}

	public FireBuilding getTargetFireBuilding() {
		return targetFireBuilding;
	}

	private class DistanceComparatorForEPS implements Comparator<Area> {
		private static final int ONE_STEP_MOVE_DISTANCE = 40000;
		private Human me;
		private StandardWorldModel world;
		private Building targetBuilding;

		public DistanceComparatorForEPS(Human me, StandardWorldModel world, Building targetBuilding) {
			this.targetBuilding = targetBuilding;
			this.me = me;
			this.world = world;
		}

		@Override
		public int compare(Area a, Area b) {
			int dMe2a = world.getDistance(me, a);
			int dMe2b = world.getDistance(me, b);

			if (dMe2a < ONE_STEP_MOVE_DISTANCE && dMe2b < ONE_STEP_MOVE_DISTANCE) {
				int dTarget2a = world.getDistance(targetBuilding, a);
				int dTarget2b = world.getDistance(targetBuilding, b);

				return dTarget2a - dTarget2b;
			} else {
				return dMe2a - dMe2b;
			}
		}
	}
}
