package ant.agent.job.fb;

import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.config.Config;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.ConfigKey;
import ant.agent.behavior.FBBehavior;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntRest;
import ant.agent.job.AntFireBrigadeJob;
import ant.agent.job.JobState;
import ant.liaison.message.MOutOfWater;
import ant.map.component.AbstractRouteComponent;
import ant.map.component.AntMapBasisComponent;
import ant.map.objects.RouteResult;

/**
 * 
 * @author Oliver
 * 
 */
public class GetWaterJob extends AntFireBrigadeJob {
	private static Log log = LogFactory.getLog(GetWaterJob.class);

	private static final float ENOUGH_WATER_LEVEL = 0.95f;
	private int maxWater;
	private AbstractRouteComponent arc;
	private Set<Refuge> refuges;
	private Refuge targeRefuge;

	public GetWaterJob(AgentContext context) {
		super(context);
		Config config = context.getConfig();
		this.maxWater = ConfigKey.getFireMaxWater(config);
		AntMapBasisComponent awc = context.getAntComponentKit().getAntMapBasisComponent();
		this.refuges = awc.getRefugeSet();
		this.arc = context.getAntComponentKit().getAbstractRouteComponent();
	}

	@Override
	public void overtaken() {
		this.state = JobState.Inactive;

		this.targeRefuge = null;
	}

	@Override
	public boolean shouldExecute() {
		if (!agent.isWaterDefined()) {
			this.state = JobState.Inactive;
			return false;
		}

		StandardEntity currentPosition = agent.getPosition(context.getWorld());
		if (((FBBehavior) behavior).isOutOfWater()) {
			if (currentPosition instanceof Refuge) {
				state = JobState.FillingWater;
				return true;
			}

			this.targeRefuge = getNearestRefuge();
			if (targeRefuge == null) {
				log.warn("No route to refuge at position " + currentPosition);
				return false;
			}

			MOutOfWater msg = new MOutOfWater();
			addRadioMessage(msg);
			addVoiceMessage(msg);

			state = JobState.GoingToRefugeToRefillWater;
			return true;
		}

		if (state == JobState.GoingToRefugeToRefillWater) {
			if (currentPosition instanceof Refuge) {
				state = JobState.FillingWater;
			}

			this.targeRefuge = getNearestRefuge();
			if (targeRefuge == null) {
				log.warn("No route to refuge at position " + currentPosition);
				return false;
			}

			return true;
		} else if (state == JobState.FillingWater) {
			if (currentPosition instanceof Refuge) {
				if (agent.getWater() > ENOUGH_WATER_LEVEL * maxWater) {
					this.state = JobState.Inactive;
					return false;
				} else {
					return true;
				}
			} else {
				// this not happen in most case
				log.warn("Last state is FillingWater, but not in refuge!!");
				state = JobState.GoingToRefugeToRefillWater;
				if (targeRefuge == null) {
					targeRefuge = getNearestRefuge();
					if (targeRefuge == null) {
						log.error("This should never happen!!!");
						return false;
					}
				}
				return true;
			}
		}

		this.state = JobState.Inactive;
		return false;
	}

	@Override
	public AntCommand execute() {
		if (state == JobState.GoingToRefugeToRefillWater) {
			RouteResult route = arc.search(agent.getPosition(), agent.getX(), agent.getY(),
					targeRefuge.getID());
			if (route.hasAcceptableRoute()) {
				return new AntMove(context.getTimestamp(), route.getMoveRoute());
			} else {
				log.error("No route to refuge, but state is GoingToRefugeToRefillWater!");
				return new AntRest(context.getTimestamp());
			}
		} else if (state == JobState.FillingWater) {
			return new AntRest(context.getTimestamp());
		} else {
			log.fatal("Unexpected state");
			return new AntRest(context.getTimestamp());
		}
	}

	@Override
	public String getJobName() {
		return getClass().getName();
	}

	private Refuge getNearestRefuge() {
		EntityID here = agent.getPosition();
		double minDistance = Double.MAX_VALUE;
		Refuge nearestRefuge = null;
		for (Refuge refuge : refuges) {
			RouteResult route = arc.search(here, agent.getX(), agent.getY(), refuge.getID());
			if (route.hasAcceptableRoute()) {
				if (route.getDistance() < minDistance) {
					minDistance = route.getDistance();
					nearestRefuge = refuge;
				}
			}
		}
		return nearestRefuge;
	}
}
