package ant.agent.job.pf;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Human;
import ant.AgentContext;
import ant.agent.command.AntCommand;
import ant.agent.command.AntRest;
import ant.agent.job.AntPoliceForceJob;
import ant.component.pf.ObjectiveJigsaw;
import ant.component.pf.ObjectiveJigsawAvenue;
import ant.component.pf.ObjectiveJigsawCross;
import ant.component.pf.ObjectiveJigsawDecisionComponent;
import ant.component.pf.ObjectiveJigsawMetaComponent;
import ant.component.pf.ObjectiveRaceComponent;
import ant.component.pf.Race;
import ant.liaison.message.MJigsawDoing;
import ant.map.partition.Partition;

/**
 * 
 * @author Ivan
 * 
 * @depend {@link ObjectiveRaceComponent}
 */
public class PFPartitionalObjectiveJob extends AntPoliceForceJob {
	private static final Log log = LogFactory.getLog(PFPartitionalObjectiveJob.class);
	
	private ObjectiveRaceComponent opc;
	private Race race;
	private PFWorkerPartitionalObjectiveJob workerjob;
	private PFSpritePartitionalObjectiveJob spritejob;

	public PFPartitionalObjectiveJob(AgentContext context) {
		super(context);
		this.opc = (ObjectiveRaceComponent) context.getAntComponentKit().getAntComponent(ObjectiveRaceComponent.class);
		this.race = opc.getMyRace();
		if (race == Race.Worker) {
			this.workerjob = new PFWorkerPartitionalObjectiveJob(context);
		} else {
			this.spritejob = new PFSpritePartitionalObjectiveJob(context);
		}
	}

	@Override
	public void overtaken() {
		if (race == Race.Worker)
			workerjob.overtaken();
		else
			spritejob.overtaken();
	}

	@Override
	public void overleap() {
		super.overleap();
		if (race == Race.Worker)
			workerjob.overleap();
		else
			spritejob.overleap();
	}

	@Override
	public boolean shouldExecute() {
		if (race == Race.Worker) {
			long _time_st = System.currentTimeMillis();
			boolean ret = workerjob.shouldExecute();
			long _time_ed = System.currentTimeMillis();
			if (_time_ed - _time_st > 50) {
				log.info("PFPartitionalObjectiveJob<Worker>.shouldExecute cost: " + (_time_ed - _time_st) + " milliseconds");
			}
			return ret;
		} else {
			long _time_st = System.currentTimeMillis();
			boolean ret = spritejob.shouldExecute();
			long _time_ed = System.currentTimeMillis();
			if (_time_ed - _time_st > 50) {
				log.info("PFPartitionalObjectiveJob<Sprite>.shouldExecute cost: " + (_time_ed - _time_st) + " milliseconds");
			}
			return ret;
		}
	}

	@Override
	public AntCommand execute() {
		if (race == Race.Worker) {
			return workerjob.execute();
		} else {
			return spritejob.execute();
		}
	}
	
	@Override
	public String getJobName() {
		return PFPartitionalObjectiveJob.class.getName();
	}

}

abstract class PFAbstractPartitionalObjectiveJob extends AntPoliceForceJob {

	protected ObjectiveJigsawMetaComponent omc;
	protected ObjectiveJigsawDecisionComponent odc;
	protected ObjectiveRaceComponent opc;
	protected Human human;

	public PFAbstractPartitionalObjectiveJob(AgentContext context) {
		super(context);
		this.omc = (ObjectiveJigsawMetaComponent) context.getAntComponentKit().getAntComponent(ObjectiveJigsawMetaComponent.class);
		this.odc = (ObjectiveJigsawDecisionComponent) context.getAntComponentKit().getAntComponent(ObjectiveJigsawDecisionComponent.class);
		this.opc = (ObjectiveRaceComponent) context.getAntComponentKit().getAntComponent(ObjectiveRaceComponent.class);
		this.human = context.getHuman();
	}
	
}

class PFWorkerPartitionalObjectiveJob extends PFAbstractPartitionalObjectiveJob {
	private static final Log log = LogFactory.getLog(PFWorkerPartitionalObjectiveJob.class);
	
	private AbstractOjExecutor executor = null;
	private Partition mypart;
	
	public PFWorkerPartitionalObjectiveJob(AgentContext context) {
		super(context);
		this.mypart = opc.getMyPartition();
	}

	@Override
	public void overtaken() {
	}

	@Override
	public void overleap() {
		super.overleap();
		executor = null;
	}
	
	@Override
	public boolean shouldExecute() {
		if (executor != null) {
			if (executor.isDone()) {
				executor = null;
			}
		}
		
		if (executor == null) {
			List<ObjectiveJigsaw> templist = omc.getSortedHighlySpotedCandicateObjectiveJigsaws();
			for (ObjectiveJigsaw oj : templist) {
				if (!oj.belongPartId(mypart.getID()))
					continue;
				
				if (oj.getDiscuss().isBestApplicant()) {
					if (oj instanceof ObjectiveJigsawCross) {
						executor = new ObjectiveJigsawCrossExecutor(context, behavior, (ObjectiveJigsawCross) oj);
						break;
					} else if (oj instanceof ObjectiveJigsawAvenue) {
						executor = new ObjectiveJigsawAvenueExecutor(context, behavior, (ObjectiveJigsawAvenue) oj);
						break;
					}
				} else {
					odc.discuss(oj);
					// Wait for next cycle
					break;
				}
			}
			
			if (executor == null) {
				return false;
			}
		}
		
		// GO~
		return true;
	}

	
	@Override
	public AntCommand execute() {
		if (executor != null) {
			AntCommand ret = executor.execute();
			if (ret != null) {
				addRadioMessage(new MJigsawDoing(executor.getObjectiveJigsaw().getObjectiveID()));
				return ret;
			} else {
				log.error("Bad state");
				return new AntRest(context.getTimestamp());
			}
		} else {
			log.error("Bad state");
			return new AntRest(context.getTimestamp());
		}
	}

	@Override
	public String getJobName() {
		return PFWorkerPartitionalObjectiveJob.class.getName();
	}
	
}

class PFSpritePartitionalObjectiveJob extends PFAbstractPartitionalObjectiveJob {
	private static final Log log = LogFactory.getLog(PFSpritePartitionalObjectiveJob.class);
	
	private AbstractOjExecutor executor = null;
	
	public PFSpritePartitionalObjectiveJob(AgentContext context) {
		super(context);
	}

	@Override
	public void overtaken() {
	}

	@Override
	public void overleap() {
		super.overleap();
		executor = null;
	}
	
	@Override
	public boolean shouldExecute() {
		if (executor != null) {
			if (executor.isDone()) {
				executor = null;
			}
		}
		
		if (executor == null) {
			List<ObjectiveJigsaw> templist = omc.getSortedHighlySpotedCandicateObjectiveJigsaws();
			for (ObjectiveJigsaw oj : templist) {
				if (oj.getDiscuss().isBestApplicant()) {
					if (oj instanceof ObjectiveJigsawCross) {
						executor = new ObjectiveJigsawCrossExecutor(context, behavior, (ObjectiveJigsawCross) oj);
						break;
					} else if (oj instanceof ObjectiveJigsawAvenue) {
						executor = new ObjectiveJigsawAvenueExecutor(context, behavior, (ObjectiveJigsawAvenue) oj);
						break;
					}
				} else {
					odc.discuss(oj);
					// Wait for next cycle
					break;
				}
			}
			
			if (executor == null) {
				return false;
			}
		}
		
		// GO~
		return true;
	}

	@Override
	public AntCommand execute() {
		if (executor != null) {
			AntCommand ret = executor.execute();
			if (ret != null) {
				addRadioMessage(new MJigsawDoing(executor.getObjectiveJigsaw().getObjectiveID()));
				return ret;
			} else {
				log.error("Bad state");
				return new AntRest(context.getTimestamp());
			}
		} else {
			log.error("Bad state");
			return new AntRest(context.getTimestamp());
		}
	}

	@Override
	public String getJobName() {
		return PFSpritePartitionalObjectiveJob.class.getName();
	}
	
}

