package ant.agent.job.pf;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Human;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.agent.behavior.PFBehavior.CleanLevel;
import ant.agent.command.AntClear;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntRest;
import ant.agent.job.AntPoliceForceJob;
import ant.component.pf.ObjectiveJigsaw;
import ant.component.pf.ObjectiveJigsawAvenue;
import ant.component.pf.ObjectiveJigsawCross;
import ant.component.pf.ObjectiveJigsawDecisionComponent;
import ant.component.pf.ObjectiveJigsawMetaComponent;
import ant.component.pf.ObjectiveProbeComponent;
import ant.component.pf.ObjectiveProbeSpriteComponent;
import ant.component.pf.ObjectiveProbeWorkerComponent;
import ant.component.pf.ObjectiveRaceComponent;
import ant.component.pf.Race;
import ant.component.universal.ChangeSetDetailComponent;
import ant.map.component.StaticCityGraphComponent;
import ant.map.partition.Partition;

/**
 * 
 * @author Ivan
 * 
 * @depend {@link ObjectiveRaceComponent}
 */
public class PFPartitionalProbeJob extends AntPoliceForceJob {
	private static final Log log = LogFactory.getLog(PFPartitionalProbeJob.class);
	
	private ObjectiveRaceComponent opc;
	private Race race;
	private PFWorkerPartitionalProbeJob workerjob;
	private PFSpritePartitionalProbeJob spritejob;

	public PFPartitionalProbeJob(AgentContext context) {
		super(context);
		this.opc = (ObjectiveRaceComponent) context.getAntComponentKit().getAntComponent(ObjectiveRaceComponent.class);
		this.race = opc.getMyRace();
		if (race == Race.Worker) {
			this.workerjob = new PFWorkerPartitionalProbeJob(context);
		} else {
			this.spritejob = new PFSpritePartitionalProbeJob(context);
		}
	}

	@Override
	public void overtaken() {
		if (race == Race.Worker)
			workerjob.overtaken();
		else
			spritejob.overtaken();
	}

	@Override
	public void overleap() {
		super.overleap();
		if (race == Race.Worker)
			workerjob.overleap();
		else
			spritejob.overleap();
	}

	@Override
	public boolean shouldExecute() {
		if (race == Race.Worker) {

			long _time_st = System.currentTimeMillis();
			boolean ret = workerjob.shouldExecute();
			long _time_ed = System.currentTimeMillis();
			if (_time_ed - _time_st > 50) {
				log.info("PFPartitionalProbeJob<Worker>.shouldExecute cost: "
						+ (_time_ed - _time_st) + " milliseconds");
			}
			return ret;
		} else {

			long _time_st = System.currentTimeMillis();
			boolean ret = spritejob.shouldExecute();
			long _time_ed = System.currentTimeMillis();
			if (_time_ed - _time_st > 50) {
				log.info("PFPartitionalProbeJob<Sprite>.shouldExecute cost: "
						+ (_time_ed - _time_st) + " milliseconds");
			}
			return ret;
		}
	}

	@Override
	public AntCommand execute() {
		if (race == Race.Worker) {

			long _time_st = System.currentTimeMillis();
			AntCommand ret = workerjob.execute();
			long _time_ed = System.currentTimeMillis();
			if (_time_ed - _time_st > 75) {
				log.info("PFPartitionalProbeJob<Worker>.execute cost: " + (_time_ed - _time_st)
						+ " milliseconds");
			}
			if (ret == null)
				log.error("WorkerJob return null");
			return ret;
		} else {

			long _time_st = System.currentTimeMillis();
			AntCommand ret = spritejob.execute();
			long _time_ed = System.currentTimeMillis();
			if (_time_ed - _time_st > 75) {
				log.info("PFPartitionalProbeJob<Sprite>.execute cost: " + (_time_ed - _time_st)
						+ " milliseconds");
			}
			if (ret == null)
				log.error("SpriteJob return null");
			return ret;
		}
	}
	
	@Override
	public String getJobName() {
		return PFPartitionalProbeJob.class.getName();
	}

}

/**************************************************************************************************************************/
/*********************************************** ABSTRACT PARTITIONAL PROBE ***********************************************/
/**************************************************************************************************************************/

abstract class PFAbstractPartitionalProbeJob extends AntPoliceForceJob {

	protected ObjectiveJigsawMetaComponent ojmc;
	protected ObjectiveJigsawDecisionComponent odc;
	protected ObjectiveRaceComponent opc;
	protected StaticCityGraphComponent city;
	protected ChangeSetDetailComponent csd;
	protected Human human;
	protected ObjectiveProbeComponent probc;

	public PFAbstractPartitionalProbeJob(AgentContext context) {
		super(context);
		this.ojmc = (ObjectiveJigsawMetaComponent) context.getAntComponentKit().getAntComponent(ObjectiveJigsawMetaComponent.class);
		this.odc = (ObjectiveJigsawDecisionComponent) context.getAntComponentKit().getAntComponent(ObjectiveJigsawDecisionComponent.class);
		this.opc = (ObjectiveRaceComponent) context.getAntComponentKit().getAntComponent(ObjectiveRaceComponent.class);
		this.city = context.getAntComponentKit().getStaticCityGraphComponent();
		this.csd = context.getAntComponentKit().getChangeSetDetailComponent();
		this.probc = (ObjectiveProbeComponent) context.getAntComponentKit().getAntComponent(ObjectiveProbeComponent.class);
		this.human = context.getHuman();
	}
	
}


/**************************************************************************************************************************/
/*********************************************** WORKER PARTITIONAL PROBE ***********************************************/
/**************************************************************************************************************************/

class PFWorkerPartitionalProbeJob extends PFAbstractPartitionalProbeJob {
	private static final Log log = LogFactory.getLog(PFWorkerPartitionalProbeJob.class);
	private enum WppStage {
		Inactive, MovingToPart, PartWorking
	}
	private Partition mypart;
	private ObjectiveProbeWorkerComponent pwc;
	private WppStage stage;
	
	private EntityID moveToPartTarget;
	private EntityID partWorkingTarget;
	private AbstractOjExecutor executor = null;
	
	public PFWorkerPartitionalProbeJob(AgentContext context) {
		super(context);
		this.mypart = opc.getMyPartition();
		this.pwc = probc.getProbeWorkerComponent();
		this.stage = WppStage.Inactive;
	}

	@Override
	public void overtaken() {
		stage = WppStage.Inactive;
		executor = null;
	}

	@Override
	public void overleap() {
		super.overleap();
		stage = WppStage.Inactive;
		executor = null;
	}
	
	@Override
	public boolean shouldExecute() {
		if (executor != null) {
			if (executor.isDone()) {
				executor = null;
			} else {
				return true;
			}
		}
		
		if (executor == null) {
			// here-executor
			EntityID loc = human.getPosition();
			ObjectiveJigsaw oj = ojmc.getExMappingObjectiveJigsaw(loc);
			if (oj.isSpoted()) {
				if (oj instanceof ObjectiveJigsawCross) {
					executor = new ObjectiveJigsawCrossExecutor(context, behavior, (ObjectiveJigsawCross) oj, false);
					if (executor.isDone()) {
						log.error("Add doneExecutor!");
					}
					return true;
				} else if (oj instanceof ObjectiveJigsawAvenue) {
					executor = new ObjectiveJigsawAvenueExecutor(context, behavior, (ObjectiveJigsawAvenue) oj, false);
					if (executor.isDone()) {
						log.error("Add doneExecutor!");
					}
					return true;
				}
			}
		}
		
		// Executor == null
		boolean isInMyPart = pwc.isInMyPartition();
		if (!isInMyPart) {
			// Move to myPartition
			
			// 如果上个状态是Inactive, 那么movetopart, return true
			if (stage == WppStage.Inactive) {
				stage = WppStage.MovingToPart;
				moveToPartTarget = pwc.getNearestRoadToPartition();
				return true;
			}
			
			if (stage == WppStage.MovingToPart) {
				if (moveToPartTarget == null) {
					moveToPartTarget = pwc.getNearestRoadToPartition();
				}
				return true;
			}
			
			if (stage == WppStage.PartWorking) {
				return true;
			}
			
			return true;
			
		} else {
			// In myPartition
			stage = WppStage.PartWorking;
			
			if (agent.getPosition().equals(partWorkingTarget)) {
				// 如果所在Jigsaw Totally clean
				partWorkingTarget = null;
			}
			
			if (partWorkingTarget == null) {
				return tryGetTask();
			} else {
				return true;
			}
		}
	}
	
	private boolean tryGetTask() {
		if (!pwc.hasTask()) {
			return false;
		} else {
			partWorkingTarget = pwc.getTask();
			return true;
		}
	}

	@Override
	public AntCommand execute() {
		int time = context.getTimestamp();
		
		if (executor != null) {
			AntCommand ret = executor.execute();
			if (ret == null) {
				log.error("Executor :" + executor + " ret null");
			}
			return ret;
		}
			
		if (stage == WppStage.MovingToPart) {
			List<EntityID> path = city.getStaticMoveRoute(agent.getPosition(), moveToPartTarget);
			EntityID id = behavior.cleverPathBlockadeSelect(path, CleanLevel.Normal);
			if (id == null) {
				return new AntMove(time, path);
			} else {
				return new AntClear(time, id);
			}
		} else {
			// PartWork
			if (agent.getPosition().equals(partWorkingTarget)) {
				EntityID id = behavior.cleverAroundBlockadeSelect(CleanLevel.Almost);
				if (id != null) {
					return new AntClear(time, id);
				} else {
					log.error("Bad Rest");
					return new AntRest(time);
				}
			} else {
				List<EntityID> path = pwc.getMidMoveRoute(agent.getPosition(), partWorkingTarget);
				EntityID id = behavior.cleverPathBlockadeSelect(path, CleanLevel.Normal);
				
				if (id != null) {
					return new AntClear(time, id);
				} else {
					return new AntMove(time, path);
				}
			}
		}
	}

	@Override
	public String getJobName() {
		return PFWorkerPartitionalProbeJob.class.getName();
	}
	
}



/**************************************************************************************************************************/
/*********************************************** SPRITE PARTITIONAL PROBE ***********************************************/
/**************************************************************************************************************************/


class PFSpritePartitionalProbeJob extends PFAbstractPartitionalProbeJob {
	private static final Log log = LogFactory.getLog(PFSpritePartitionalProbeJob.class);
	
	private AbstractOjExecutor executor = null;
	private ObjectiveJigsaw target = null;

	private ObjectiveProbeSpriteComponent psc;
	
	public PFSpritePartitionalProbeJob(AgentContext context) {
		super(context);
		this.psc = probc.getProbeSpriteComponent();
	}

	@Override
	public void overtaken() {
		executor = null;
		target = null;
	}

	@Override
	public void overleap() {
		super.overleap();
		executor = null;
		target = null;
	}
	
	@Override
	public boolean shouldExecute() {
		if (executor != null) {
			if (executor.isDone()) {
				executor = null;
			} else {
				return true;
			}
		}
		
		EntityID loc = human.getPosition();
		ObjectiveJigsaw oj = ojmc.getExMappingObjectiveJigsaw(loc);
		if (executor == null) {
			if (oj.isSpoted()) {
				if (oj instanceof ObjectiveJigsawCross) {
					executor = new ObjectiveJigsawCrossExecutor(context, behavior, (ObjectiveJigsawCross) oj, false);
					if (executor.isDone()) {
						log.error("Add doneExecutor!");
					}
					return true;
				} else if (oj instanceof ObjectiveJigsawAvenue) {
					executor = new ObjectiveJigsawAvenueExecutor(context, behavior, (ObjectiveJigsawAvenue) oj, false);
					if (executor.isDone()) {
						log.error("Add doneExecutor!");
					}
					return true;
				}
			}
		}
		
		if (target != null) {
			// not arrive
			if (oj.getObjectiveID() != target.getObjectiveID()) {
				return true;
			} else {
				target = null;
			}
		}
		
		if (target == null) {
			return tryGetTask();
		} else {
			return false;
		}
	}

	private boolean tryGetTask() {
		this.target = psc.getTask();
		if (target == null)
			return false;
		else
			return true;
	}

	@Override
	public AntCommand execute() {
		if (executor != null) {
			AntCommand ret = executor.execute();
			if (ret == null) {
				log.error("Executor ret null:" + executor);
			}
			return ret;
		}
		
		int time = context.getTimestamp();
		EntityID tw = target.getJigsawObject().getCenter().getID();
		if (agent.getPosition().equals(tw)) {
			EntityID id = behavior.cleverAroundBlockadeSelect(CleanLevel.Almost);
			if (id != null) {
				return new AntClear(time, id);
			} else {
				log.error("Bad Rest");
				return new AntRest(time);
			}
		} else {
			List<EntityID> path = psc.getMidMoveRoute(agent.getPosition(), tw);
			EntityID id = behavior.cleverPathBlockadeSelect(path, CleanLevel.Normal);
			if (id != null) {
				return new AntClear(time, id);
			} else {
				return new AntMove(time, path);
			}
		}
	}

	@Override
	public String getJobName() {
		return PFSpritePartitionalProbeJob.class.getName();
	}
	
}


