package ant.agent.job.pf;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.ConfigKey;
import ant.agent.command.AntClear;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.command.AntRest;
import ant.agent.job.AntPoliceForceJob;
import ant.agent.job.JobState;
import ant.component.universal.AgentLocationComponent;
import ant.component.universal.AgentLocationComponent.AgentLocationInfo;
import ant.utils.RandomSelector;

/**
 * 
 * @author Ivan
 * 
 */
public class PFStuckInBuildingJob extends AntPoliceForceJob {
	private static final Log log = LogFactory.getLog(PFStuckInBuildingJob.class);

	private enum StuckStep {
		Safe, CleanAround, MovingToEdge, MovingOut, EdgeClean
	}

	private static final int CHECK_STEP = 2;
	private static final int LEAST_MOVE_LEN = 2500;

	private EntityID stuckhere;
	private StuckStep stuckState;
	private int ignoreTimestamp;
	private StandardWorldModel world;
	private AgentLocationComponent alc;

	public PFStuckInBuildingJob(AgentContext context) {
		super(context);
		this.alc = (AgentLocationComponent) context.getAntComponentKit().getAntComponent(
				AgentLocationComponent.class);
		this.ignoreTimestamp = ConfigKey.getIgnoreTime(context.getConfig());
		this.world = context.getWorld();
	}

	@Override
	public void overtaken() {
		this.state = JobState.Inactive;
		this.stuckhere = null;
		this.stuckState = StuckStep.Safe;
	}

	@Override
	public void overleap() {
		super.overleap();
		this.state = JobState.Inactive;
		this.stuckhere = null;
		this.stuckState = StuckStep.Safe;
	}

	@Override
	public boolean shouldExecute() {
		int time = context.getTimestamp();
		if (time < ignoreTimestamp + CHECK_STEP)
			return false;

		if (stuckState == StuckStep.Safe) {
			stuckhere = context.getHuman().getPosition();

			if (!isAlwaysMoveOrRest())
				return false;
			if (!isAlwaysSamePlace())
				return false;
			if (!isNotMoveTooLong())
				return false;

			if (behavior.hasInsightInRangeBlockade()) {
				stuckState = StuckStep.CleanAround;
				return true;
			} else {
				stuckState = StuckStep.MovingToEdge;
				return true;
			}
		} else if (stuckState == StuckStep.CleanAround) {
			if (behavior.hasInsightInRangeBlockade()) {
				return true;
			} else {
				stuckState = StuckStep.MovingToEdge;
				return true;
			}
		} else if (stuckState == StuckStep.MovingToEdge || stuckState == StuckStep.EdgeClean) {
			if (behavior.hasInsightInRangeBlockade()) {
				stuckState = StuckStep.EdgeClean;
				return true;
			} else {
				stuckState = StuckStep.MovingOut;
				return true;
			}
		} else if (stuckState == StuckStep.MovingOut) {
			if (!context.getHuman().getPosition().equals(stuckhere)) {
				return false;
			}
			log.warn("PFStuck Moving out failed: " + stuckhere);
			return false;
		} else {
			log.info("Uexpected state!!" + stuckState);
			return false;
		}
	}

	@Override
	public AntCommand execute() {
		if (stuckState == StuckStep.CleanAround || stuckState == StuckStep.EdgeClean) {
			Blockade blk = behavior.getRandomOneOfInsightInRangeBlockade();
			if (blk != null) {
				return new AntClear(this.context.getTimestamp(), blk.getID());
			} else {
				log.warn("Bad stuck state, " + stuckState + " but not found any blockades");
				return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
			}
		} else if (stuckState == StuckStep.MovingToEdge) {
			Area here = (Area) world.getEntity(context.getHuman().getPosition());
			List<EntityID> path = new ArrayList<EntityID>();
			path.add(here.getID());
			Set<Edge> cad = new HashSet<Edge>();
			for (Edge e : here.getEdges()) {
				if (e.isPassable())
					cad.add(e);
			}
			if (cad.isEmpty()) {
				log.warn("Bad stuck state: no passable edge");
				return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
			} else {
				// FIXME don't random
				Edge pick = RandomSelector.randomSelect(cad, 1, behavior.getMyRandom()).iterator()
						.next();
				int x = (pick.getStartX() + pick.getEndX()) / 2;
				int y = (pick.getStartY() + pick.getEndY()) / 2;
				return new AntMoveXY(this.context.getTimestamp(), path, x, y);
			}
		} else if (stuckState == StuckStep.MovingOut) {
			Blockade blk = behavior.getRandomOneOfInsightInRangeBlockade();
			if (blk != null) {
				log.error("Bad clean for moving out");
				return new AntClear(this.context.getTimestamp(), blk.getID());
			} else {
				return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
			}
		}
		log.warn("Bad stuck state");
		return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
	}

	@Override
	public String getJobName() {
		return PFStuckInBuildingJob.class.getName();
	}

	private boolean isAlwaysSamePlace() {
		int time = context.getTimestamp();
		for (int i = 1; i < CHECK_STEP; i++) {
			AgentLocationInfo info1 = alc.getAgentLocationInfo(context.getAgentID(), time - i);
			AgentLocationInfo info2 = alc.getAgentLocationInfo(context.getAgentID(), time - i + 1);
			if (!info1.getLocation().equals(info2.getLocation()))
				return false;

		}
		return true;
	}

	private boolean isNotMoveTooLong() {
		int time = context.getTimestamp();
		int totals = 0;
		for (int i = 1; i < CHECK_STEP; i++) {
			AgentLocationInfo info1 = alc.getAgentLocationInfo(context.getAgentID(), time - i);
			AgentLocationInfo info2 = alc.getAgentLocationInfo(context.getAgentID(), time - i + 1);
			totals += alc.manhattanDistance(info1, info2);
		}
		if (totals > LEAST_MOVE_LEN) {
			return false;
		} else {
			return true;
		}
	}

	private boolean isAlwaysMoveOrRest() {
		int time = context.getTimestamp();
		for (int i = 1; i <= CHECK_STEP; i++) {
			AntCommand ac = this.context.getCommands().get(time - i);
			if ((ac instanceof AntMove) || (ac instanceof AntMoveXY) || (ac instanceof AntRest)) {
			} else {
				return false;
			}
		}
		return true;
	}

}
