package ant.agent.job.pf;

import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.ConfigKey;
import ant.agent.command.AntClear;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.command.AntRest;
import ant.agent.job.AntPoliceForceJob;
import ant.agent.job.JobState;
import ant.component.universal.AgentLocationComponent;
import ant.component.universal.AgentLocationComponent.AgentLocationInfo;
import ant.component.universal.ChangeSetDetailComponent;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Point;
import ant.utils.geo.Polygon;

/**
 * 
 * @author Ivan
 * 
 */
public class PFStuckInRoadJob extends AntPoliceForceJob {
	private static final Log log = LogFactory.getLog(PFStuckInRoadJob.class);

	private enum State {
		Safe, CleanNearest, TryMove, MoveEst, CleanAllAround, RandomMove
	}

	private static final int CHECK_STEP = 2;
	private static final int LEAST_MOVE_LEN = 2000;

	private State stuckState;
	private StandardWorldModel world;
	private AgentLocationComponent alc;
	private ChangeSetDetailComponent chs;

	private Road pickRoad = null;
	private EntityID pickBlockadeId = null;
	private EntityID stuckWhere;
	private int stuckX;
	private int stuckY;
	private List<EntityID> lastMoveCommandPath;

	public PFStuckInRoadJob(AgentContext context) {
		super(context);
		this.alc = (AgentLocationComponent) context.getAntComponentKit().getAntComponent(AgentLocationComponent.class);
		this.world = context.getWorld();
		this.chs = context.getAntComponentKit().getChangeSetDetailComponent();
	}

	@Override
	public void overtaken() {
		this.state = JobState.Inactive;
		this.stuckState = State.Safe;
		
		reset();
	}

	@Override
	public void overleap() {
		super.overleap();
		this.state = JobState.Inactive;
		this.stuckState = State.Safe;
		
		reset();
	}

	private void reset() {
		pickRoad = null;
		pickBlockadeId = null;
		stuckWhere = null;
		lastMoveCommandPath = null;
	}

	@Override
	public boolean shouldExecute() {
		int time = context.getTimestamp();

		if (time < ConfigKey.getIgnoreTime(context.getConfig()) + CHECK_STEP)
			return false;

		if (context.getHuman().isBuriednessDefined())
			if (context.getHuman().getBuriedness() > 0)
				return false;

		if (context.getHuman().getPosition(world) instanceof Building)
			return false;

		if (stuckState == State.Safe) {

			if (!isAlwaysMove())
				return false;
//			if (!isAlwaysSamePlace())
//				return false;
			if (!isNotMoveTooLong())
				return false;

			stuckX = context.getHuman().getX();
			stuckY = context.getHuman().getY();
			
			// stuckState = State.CleanNearest;
			stuckWhere = context.getHuman().getPosition();
			lastMoveCommandPath = null;
			
			AntCommand c = this.context.getCommands().get(time - 1);
			if (c instanceof AntMove) {
				lastMoveCommandPath = ((AntMove) c).getPath();
			} else if (c instanceof AntMoveXY) {
				lastMoveCommandPath = ((AntMoveXY) c).getPath();
			}
			
			pickBlockade();
			
			if (pickBlockadeId != null) {
				stuckState = State.CleanNearest;
				return true;
				
			} else {
				stuckState = State.RandomMove;
				log.warn("Out-of-Sight-Stuck");
				return true;
			}
		}
		
		if (stuckState == State.CleanNearest) {
			if (pickRoad != null && pickBlockadeId != null) {
				if (pickRoad.getBlockades().contains(pickBlockadeId)) {
					Blockade b = (Blockade) world.getEntity(pickBlockadeId);
					
					// log.info("B:" + b + " " + b.getRepairCost());
					if (b != null) {
						if (b.isRepairCostDefined()) {
							if (b.getRepairCost() > 0) {
								// Still Need clean
								return true;
							}
						}
					}
				}
				
				// Done
				pickRoad = null;
				pickBlockadeId = null;
				stuckState = State.TryMove;
			} else {
				log.warn("Bad stuck state");
				pickRoad = null;
				pickBlockadeId = null;
				stuckState = State.TryMove;
			}
		}
		
		if (stuckState == State.TryMove) {
			return true;
		}
		
		if (stuckState == State.MoveEst) {
			if (GeoUtils.dist(this.stuckX, this.stuckY, context.getHuman().getX(), context.getHuman().getY()) < LEAST_MOVE_LEN) {
				stuckX = context.getHuman().getX();
				stuckY = context.getHuman().getY();
				
				stuckState = State.CleanNearest;
				stuckWhere = context.getHuman().getPosition();
				lastMoveCommandPath = null;
				
				AntCommand c = this.context.getCommands().get(time - 1);
				if (c instanceof AntMove) {
					lastMoveCommandPath = ((AntMove) c).getPath();
				} else if (c instanceof AntMoveXY) {
					lastMoveCommandPath = ((AntMoveXY) c).getPath();
				}
				
				pickBlockade();
				
				if (pickBlockadeId != null) {
					stuckState = State.CleanNearest;
					return true;
					
				} else {
					// TODO should blind-clean
					// log.warn("Out-of-Sight-Stuck");
					return false;
				}
			} else {
				stuckState = State.Safe;
				return false;
			}
		}
		return false;
	}

	private void pickBlockade() {

		this.pickRoad = null;
		this.pickBlockadeId = null;

		int hx = context.getHuman().getX();
		int hy = context.getHuman().getY();
		
		double mindist = Double.MAX_VALUE;
		
		for (StandardEntity se : chs.getEntitiesInSight(StandardEntityURN.ROAD)) {
			Road r = (Road) se;
			if (!r.isBlockadesDefined())
				continue;

			for (EntityID bid : r.getBlockades()) {
				Blockade blk = (Blockade) world.getEntity(bid);
				if (!blk.isRepairCostDefined())
					continue;
				
				if (blk.getRepairCost() <= 0)
					continue;
				
				if (!behavior.isInClearRange(blk))
					continue;
				
				Point p = Polygon.getNearestEdgeOfPolygon(hx, hy, blk.getApexes());
				double temp = GeoUtils.dist(hx, hy, p.getX(), p.getY());
				if (temp < mindist) {
					mindist = temp;
					pickBlockadeId = blk.getID();
					pickRoad = r;
				}
			}
		}
	}

	@Override
	public AntCommand execute() {
		
		// log.debug("@" + context.getTimestamp() + " PF:" + context.getHuman().getID() + " Into StuckInBuildingState:" + stuckState);
		
		if (stuckState == State.Safe) {
			log.error("Bad StuckInRoad state");
			return new AntRest(context.getTimestamp());
		}
		
		if (stuckState == State.CleanNearest) {
			if (pickBlockadeId == null) {
				log.error("Bad StuckInRoad state");
				return new AntRest(context.getTimestamp());
			} else {
				return new AntClear(this.context.getTimestamp(), pickBlockadeId);
			}
		}
		
		if (stuckState == State.TryMove) {
			if (lastMoveCommandPath != null) {
				stuckState = State.MoveEst;
				return new AntMove(this.context.getTimestamp(), lastMoveCommandPath);
			} else {
				log.error("Bad StuckInRoad state");
				return new AntRest(context.getTimestamp());
			}
		}
		
		if (stuckState == State.RandomMove) {
			stuckState = State.Safe;
			return new AntMove(context.getTimestamp(), behavior.getRandomWalk());
		}
		
		log.error("Bad State for StuckInRoad");
		return new AntRest(context.getTimestamp());
	}
	
//	@Override
//	public AntCommand execute() {
//		if (stuckState == StuckStep.CleanAround) {
//			Blockade blk = behavior.getRandomOneOfInsightInRangeBlockade();
//			if (blk != null) {
//				return new AntClear(this.context.getTimestamp(), blk.getID());
//			} else {
//				log.warn("Bad stuck state");
//				stuckState = StuckStep.MovingOut;
//			}
//		}
//
//		if (stuckState == StuckStep.MovingOut) {
//			Blockade blk = behavior.getRandomOneOfInsightInRangeBlockade();
//			if (blk != null) {
//				return new AntClear(this.context.getTimestamp(), blk.getID());
//			} else {
//				return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
//			}
//		} else {
//			log.warn("Bad stuck state");
//			return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
//		}
//		return null;
//	}

	@Override
	public String getJobName() {
		return PFStuckInRoadJob.class.getName();
	}

	private boolean isAlwaysSamePlace() {
		int time = context.getTimestamp();
		for (int i = 1; i < CHECK_STEP; i++) {
			AgentLocationInfo info1 = alc.getAgentLocationInfo(context.getAgentID(), time - i);
			AgentLocationInfo info2 = alc.getAgentLocationInfo(context.getAgentID(), time - i + 1);
			if (!info1.getLocation().equals(info2.getLocation()))
				return false;

		}
		return true;
	}

	private boolean isNotMoveTooLong() {
		int time = context.getTimestamp();
		int totals = 0;
		for (int i = 1; i < CHECK_STEP; i++) {
			AgentLocationInfo info1 = alc.getAgentLocationInfo(context.getAgentID(), time - i);
			AgentLocationInfo info2 = alc.getAgentLocationInfo(context.getAgentID(), time - i + 1);
			totals += alc.manhattanDistance(info1, info2);
		}
		if (totals > LEAST_MOVE_LEN) {
			return false;
		} else {
			return true;
		}
	}

	private boolean isAlwaysMove() {
		int time = context.getTimestamp();
		for (int i = 1; i <= CHECK_STEP; i++) {
			AntCommand ac = this.context.getCommands().get(time - i);
			if ((ac instanceof AntMove) || (ac instanceof AntMoveXY)) {
			} else {
				return false;
			}
		}
		return true;
	}

}
