package ant.agent.job.universal;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AbstractAntCommunicationKit;
import ant.AgentContext;
import ant.ConfigKey;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.job.AbstractJob;
import ant.agent.job.JobState;
import ant.component.universal.AgentLocationComponent;
import ant.component.universal.AgentLocationComponent.AgentLocationInfo;
import ant.component.universal.RoadInfoCacheComponent;
import ant.utils.Pair;
import ant.utils.blockade.RoadUtils;
import ant.utils.blockade.WrappedRoad;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Line;
import ant.utils.geo.Point;
import ant.utils.geo.Polygon;
import ant.utils.geo.Segment;

/**
 * 
 * @author Oliver, Fun
 * 
 */
public abstract class AbstractStuckJob<E extends Human> extends AbstractJob<E> {
	private static Log log = LogFactory.getLog(AbstractStuckJob.class);

	private static final int LEAST_MOVE_LEN = 2500;
	private static final double TRANSLATION_COEFFICIENT = 1;

	protected Human me;
	protected StandardWorldModel world;
	protected int ignoreTimestamp;
	protected AbstractAntCommunicationKit cmu;

	private AgentLocationComponent alc;

	public AbstractStuckJob(AgentContext context) {
		super(context);
		this.me = context.getHuman();
		this.world = context.getWorld();
		this.ignoreTimestamp = ConfigKey.getIgnoreTime(context.getConfig());
		this.alc = context.getAntComponentKit().getAgentLocationComponent();
		this.cmu = context.getAntCommunicationKit();
	}

	@Override
	public void overtaken() {
		this.state = JobState.Inactive;
	}

	@Override
	public void overleap() {
		super.overleap();
		this.state = JobState.Inactive;
	}

	@Override
	public String getJobName() {
		return AbstractStuckJob.class.getName();
	}

	protected boolean isStuck(int checkStep) {
		if (!isAlwaysMove(checkStep)) {
			return false;
		}
		// FIXME 一些极端case
		if (!isAlwaysSamePlace(checkStep)) {
			return false;
		}
		if (!isNotMoveTooLong(checkStep)) {
			return false;
		}
		return true;
	}

	protected boolean isAlwaysMove(int checkStep) {
		int time = context.getTimestamp();
		for (int i = 1; i <= checkStep; i++) {
			AntCommand ac = this.context.getCommands().get(time - i);
			if ((ac instanceof AntMove) || (ac instanceof AntMoveXY)) {

			} else {
				return false;
			}
		}
		return true;
	}

	private boolean isAlwaysSamePlace(int checkStep) {
		int time = context.getTimestamp();
		for (int i = 1; i < checkStep; i++) {
			AgentLocationInfo info1 = alc.getAgentLocationInfo(context.getAgentID(), time - i);
			AgentLocationInfo info2 = alc.getAgentLocationInfo(context.getAgentID(), time - i + 1);
			if (!info1.getLocation().equals(info2.getLocation()))
				return false;

		}
		return true;
	}

	private boolean isNotMoveTooLong(int checkStep) {
		int time = context.getTimestamp();
		int totals = 0;
		for (int i = 1; i < checkStep; i++) {
			AgentLocationInfo info1 = alc.getAgentLocationInfo(context.getAgentID(), time - i);
			AgentLocationInfo info2 = alc.getAgentLocationInfo(context.getAgentID(), time - i + 1);
			totals += alc.manhattanDistance(info1, info2);
		}
		if (totals > LEAST_MOVE_LEN) {
			return false;
		} else {
			return true;
		}
	}

	/**
	 * 用Blockade截取相邻边，得到中点
	 * 
	 * @param context
	 * @param nextArea
	 * @return
	 */
	protected Pair<Integer, Integer> getTargetXY(AgentContext context, EntityID nextArea) {
		StandardWorldModel world = context.getWorld();
		RoadInfoCacheComponent ricc = context.getAntComponentKit().getRoadInfoCacheComponent();

		Area ahere = (Area) world.getEntity(context.getHuman().getPosition());
		Area anext = (Area) world.getEntity(nextArea);

		WrappedRoad where = null;
		WrappedRoad wnext = null;

		if (ahere instanceof Road) {
			where = ricc.getWrappedRoad(ahere.getID());
		}
		if (anext instanceof Road) {
			wnext = ricc.getWrappedRoad(anext.getID());
		}

		Edge cedge = ahere.getEdgeTo(nextArea);
		if (cedge == null) {
			log.error("Area " + ahere.getID() + " to " + anext.getID() + " edge is null!");
			return null;
		}
		Segment seg = new Segment(cedge.getStartX(), cedge.getStartY(), cedge.getEndX(),
				cedge.getEndY());

		if (where != null) {
			for (Polygon p : where.getExpBlockades()) {
				seg = GeoUtils.cutSegmentInsidePolygon(seg, p);
				if (seg == null) {
					return null;
				}
			}
		}

		if (wnext != null) {
			for (Polygon p : wnext.getExpBlockades()) {
				seg = GeoUtils.cutSegmentInsidePolygon(seg, p);
				if (seg == null) {
					return null;
				}
			}
		}

		return new Pair<Integer, Integer>(seg.getMidX(), seg.getMidY());
	}

	/**
	 * 离散化相邻边，取离Blockade最远点
	 * 
	 * @param context
	 * @param nextArea
	 * @return
	 */
	protected Pair<Integer, Integer> getTargetXY2(AgentContext context, EntityID nextArea) {
		StandardWorldModel world = context.getWorld();
		RoadInfoCacheComponent ricc = context.getAntComponentKit().getRoadInfoCacheComponent();

		Area ahere = (Area) world.getEntity(context.getHuman().getPosition());
		Area anext = (Area) world.getEntity(nextArea);

		WrappedRoad where = null;
		WrappedRoad wnext = null;

		if (ahere instanceof Road) {
			where = ricc.getWrappedRoad(ahere.getID());
		}
		if (anext instanceof Road) {
			wnext = ricc.getWrappedRoad(anext.getID());
		}

		Edge cedge = ahere.getEdgeTo(nextArea);
		if (cedge == null) {
			log.error("Area " + ahere.getID() + " to " + anext.getID() + " edge is null!");
			return null;
		}
		Segment seg = new Segment(cedge.getStartX(), cedge.getStartY(), cedge.getEndX(),
				cedge.getEndY());

		List<Segment> dseg = GeoUtils.discretizeSegment(seg,
				(RoadUtils.RESCUE_AGENT_RADIUS * 2 + 200));

		if (dseg.size() <= 0 || dseg.size() >= 25) { // 复杂度高
			return null;
		}

		Map<Point, Integer> pdmap = new HashMap<Point, Integer>();

		for (Segment next : dseg) {
			pdmap.put(next.getMidPoint(), Integer.MAX_VALUE);
		}

		if (where != null) {
			for (Blockade blockade : where.getBlockEntities()) {
				Point bcenter = new Point(blockade.getX(), blockade.getY());
				for (Entry<Point, Integer> entry : pdmap.entrySet()) {
					int dist = (int) GeoUtils.dist(entry.getKey(), bcenter);
					if (dist < entry.getValue()) {
						entry.setValue(dist);
					}
				}
			}
		}

		if (wnext != null) {
			for (Blockade blockade : wnext.getBlockEntities()) {
				Point bcenter = new Point(blockade.getX(), blockade.getY());
				for (Entry<Point, Integer> entry : pdmap.entrySet()) {
					int dist = (int) GeoUtils.dist(entry.getKey(), bcenter);
					if (dist < entry.getValue()) {
						entry.setValue(dist);
					}
				}
			}
		}

		int dist = 0;
		Point p = null;
		for (Entry<Point, Integer> entry : pdmap.entrySet()) {
			int tmp = entry.getValue();
			if (tmp > dist) {
				dist = tmp;
				p = entry.getKey();
			}
		}
		if (p == null) {
			return null;
		}
		return new Pair<Integer, Integer>(p.getX(), p.getY());
	}

	protected static Line getPassableLine(Line2D orignalLine, Collection<Blockade> blockades) {
		Set<Pair<Double, Double>> pieces = new HashSet<Pair<Double, Double>>();
		pieces.add(new Pair<Double, Double>(0d, 1d));
		for (Blockade b : blockades) {
			double minU = Double.MAX_VALUE;
			double maxU = Double.MIN_VALUE;
			int[] apex = b.getApexes();
			for (int i = 0; i < apex.length - 1; i += 2) {
				double u = getClosestPointRelativePostionOnSegment(orignalLine, apex[i],
						apex[i + 1]);
				if (minU > u) {
					minU = u;
				}
				if (maxU < u) {
					maxU = u;
				}
			}

			Set<Pair<Double, Double>> newPieces = new HashSet<Pair<Double, Double>>();
			for (Iterator<Pair<Double, Double>> it = pieces.iterator(); it.hasNext();) {
				Pair<Double, Double> piece = it.next();
				Double pieceMin = piece.first();
				Double pieceMax = piece.second();
				if (maxU <= pieceMin || minU >= pieceMax) {
					continue;
				}
				if (maxU <= pieceMax && minU >= pieceMin) {
					// split
					if (minU != pieceMin) {
						newPieces.add(new Pair<Double, Double>(pieceMin, minU));
					}
					if (maxU != pieceMax) {
						newPieces.add(new Pair<Double, Double>(maxU, pieceMax));
					}
				} else if (maxU >= pieceMax && minU <= pieceMin) {
					// cover
				} else {
					if (maxU < pieceMax) {
						// left shrink
						newPieces.add(new Pair<Double, Double>(maxU, pieceMax));
					} else if (minU > pieceMin) {
						// right shrink
						newPieces.add(new Pair<Double, Double>(pieceMin, minU));
					}
				}
				it.remove();
			}
			pieces.addAll(newPieces);
		}
		double maxD = Double.MIN_VALUE;
		Pair<Double, Double> thePiece = null;
		for (Pair<Double, Double> piece : pieces) {
			double d = piece.second() - piece.first();
			if (d > maxD) {
				maxD = d;
				thePiece = piece;
			}
		}

		if (thePiece != null) {
			double pieceLength = orignalLine.getDirection().getLength() * maxD;
			if (pieceLength > RoadUtils.RESCUE_AGENT_RADIUS * 2) {
				// translation away edgeLine center
				double dU = ((thePiece.first() + thePiece.second()) / 2 - 0.5)
						* TRANSLATION_COEFFICIENT;
				// log.info("Found the piece:" + thePiece + ", pieceLength:" +
				// pieceLength + " dU:"
				// + dU);
				Point2D p1 = orignalLine
						.getPoint(Math.min(Math.max(thePiece.first() + dU, 0), 0.99));
				Point2D p2 = orignalLine.getPoint(Math.max(Math.min(thePiece.second() + dU, 1),
						0.01));
				return new Line(new Point((int) p1.getX(), (int) p1.getY()), new Point(
						(int) p2.getX(), (int) p2.getY()));
			} else {
				// log.info("piece is not long enough! " + thePiece +
				// ", pieceLength:" + pieceLength);
			}
		} else {
			// log.info("Not found valid piece!");
		}
		return null;
	}

	private static double getClosestPointRelativePostionOnSegment(Line2D line, int pointX,
			int pointY) {
		Point2D p1 = line.getOrigin();
		Point2D p2 = line.getEndPoint();
		double u = (((pointX - p1.getX()) * (p2.getX() - p1.getX())) + ((pointY - p1.getY()) * (p2
				.getY() - p1.getY())))
				/ (line.getDirection().getLength() * line.getDirection().getLength());
		return u;
	}
}
