package ant.agent.job.universal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.job.JobState;
import ant.component.universal.LineOfSightPerceptionComponent;
import ant.component.universal.LineOfSightPerceptionComponent.LineInfo;
import ant.liaison.message.MStuckInBuilding;
import ant.liaison.message.MVoiceAgentStuckInBuilding;
import ant.map.component.StaticCityGraphComponent;
import ant.utils.Pair;
import ant.utils.blockade.RoadUtils;
import ant.utils.comparator.EdgeLengthComparator;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Line;
import ant.utils.geo.Point;

/**
 * 
 * @author Oliver
 * 
 */
public class AdvancedStuckInBuildingJob<E extends Human> extends AbstractStuckJob<E> {
	private static final Log log = LogFactory.getLog(AdvancedStuckInBuildingJob.class);

	private static final int CHECK_STEP = 2;
	private static final int DEFAULT_BUILDING_DISTANCE = 1500;
	private static final int DEFAULT_ROAD_DISTANCE = 20000;
	private static final int MIN_DISTANCE = 1000;
	private static final int REACH_POS_MIN_DISTANCE = 800;

	private enum StuckStep {
		NotStuck, MovingToPosInBuilding, MovingOut
	}

	private StuckStep stuckState;
	private StaticCityGraphComponent scg;

	private List<Edge> passableEdges;
	private Edge currentTryingEdge;
	private Set<Edge> triedEdge;
	private Point pos;
	private List<EntityID> pathToPos;
	private List<Building> tryBuildlings;
	private int buildingIndex;

	private LineOfSightPerceptionComponent los;

	public AdvancedStuckInBuildingJob(AgentContext context) {
		super(context);
		this.scg = context.getAntComponentKit().getStaticCityGraphComponent();
		this.passableEdges = new ArrayList<Edge>();
		this.triedEdge = new HashSet<Edge>();
		this.los = (LineOfSightPerceptionComponent) context.getAntComponentKit().getAntComponent(
				LineOfSightPerceptionComponent.class);
	}

	@Override
	public void overtaken() {
		super.overtaken();

		this.passableEdges.clear();
		this.triedEdge.clear();
		this.currentTryingEdge = null;
		this.pos = null;
		this.pathToPos = null;
		this.tryBuildlings = null;
		this.buildingIndex = -1;

		this.stuckState = StuckStep.NotStuck;
	}

	@Override
	public void overleap() {
		super.overleap();
		this.overtaken();
	}

	@Override
	public boolean shouldExecute() {
		
		// no need to check stuck at start
		if (context.getTimestamp() < ignoreTimestamp + CHECK_STEP)
			return false;

		StandardEntity here = me.getPosition(world);
		if (stuckState == StuckStep.NotStuck) {
			if (here instanceof Building && isStuck(CHECK_STEP)) {
				this.state = JobState.MovingOutOfStuck;

				this.tryBuildlings = getNeighbourBuildingsWithPassbaleEdgeToRoad((Building) here);
				// log.info("tryBuildlings:" + tryBuildlings);
				if (tryBuildlings.isEmpty()) {
					log.error("getNeighbourBuildingsWithPassbaleEdgeToRoad " + here + " isEmpty");
					return false;
				} else {
					this.buildingIndex = 0;
					passableEdges = getPassableEdgesToRoad(tryBuildlings.get(buildingIndex));
					setPosAndPathToBuilding();
					stuckState = StuckStep.MovingToPosInBuilding;
					return true;
				}
			} else {
				return false;
			}
		} else if (stuckState == StuckStep.MovingToPosInBuilding) {
			if (here instanceof Road) {
				log.warn("Get out of stuck when MovingToCenter!");
				return false;
			} else {
				if (!hasReachPosInBuilding()) {
					pathToPos = scg.getStaticMoveRoute(me.getPosition(),
							pathToPos.get(pathToPos.size() - 1));
					return true;
				}

				Pair<Point, Road> pair = getPositionInRoadToMove();
				if (pair != null) {
					pos = pair.first();
					pathToPos = scg.getStaticMoveRoute(me.getPosition(), pair.second().getID());
				} else {
					pos = null;
					pathToPos = behavior.getRandomWalk();
				}
				stuckState = StuckStep.MovingOut;
				return true;
			}
		} else if (stuckState == StuckStep.MovingOut) {
			if (here instanceof Road || !tryBuildlings.contains(here)) {
				// FIXME
				// log.info("Done moving out!");
				return false;
			} else {
				setPosAndPathToBuilding();
				stuckState = StuckStep.MovingToPosInBuilding;
				return true;
			}
		} else {
			log.warn("Unexpected state:" + stuckState);
			return false;
		}
	}

	@Override
	public AntCommand execute() {
		// log.info("buildingIndex:" + buildingIndex + ",pos:" + pos + ",myXY("
		// + me.getX() + ","
		// + me.getY() + "),stuckState:" + stuckState);
		if (stuckState == StuckStep.MovingToPosInBuilding) {
			return new AntMoveXY(this.context.getTimestamp(), pathToPos, pos.x(), pos.y());
		} else if (stuckState == StuckStep.MovingOut) {
			if (pos != null) {
				return new AntMoveXY(this.context.getTimestamp(), pathToPos, pos.x(), pos.y());
			} else {
				return new AntMove(this.context.getTimestamp(), pathToPos);
			}
		} else {
			log.fatal("error stuck state:" + stuckState);
			return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
		}
	}

	@Override
	public String getJobName() {
		return AdvancedStuckInBuildingJob.class.getName();
	}

	private void setPosAndPathToBuilding() {
		for (int i = 0; i < tryBuildlings.size(); i++) {
			Pair<Point, Building> pair = getPositionInBuildingToMove();
			if (pair != null) {
				// log.info("Found!" + tryBuildlings.get(buildingIndex));
				pos = pair.first();
				pathToPos = scg.getStaticMoveRoute(me.getPosition(), pair.second().getID());
				return;
			} else {
				buildingIndex++;
				if (buildingIndex == tryBuildlings.size()) {
					buildingIndex = 0;
					sendStuckMessage();
				}
				if (tryBuildlings.size() != 1) {
					passableEdges = getPassableEdgesToRoad(tryBuildlings.get(buildingIndex));
				}
			}
		}
		// log.warn("No pos and path found to move!");
		Building b = tryBuildlings.get(buildingIndex);
		pos = new Point(b.getX(), b.getY());
		pathToPos = scg.getStaticMoveRoute(me.getPosition(), b.getID());
	}

	private boolean hasReachPosInBuilding() {
		if (tryBuildlings.size() == 1) {
			return true;
		}
		// TODO
		if (Math.abs(me.getX() - pos.getX()) + Math.abs(me.getY() - pos.getY()) <= REACH_POS_MIN_DISTANCE) {
			return true;
		} else {
			return false;
		}
	}

	private void sendStuckMessage() {
		// FIXME sending policy needed
		if (buildingIndex < tryBuildlings.size()) {
			// cmu.sendRadio(new
			// MStuckInBuilding(tryBuildlings.get(buildingIndex).getID()));
			addRadioMessage(new MStuckInBuilding(tryBuildlings.get(buildingIndex).getID()));
			addVoiceMessage(new MVoiceAgentStuckInBuilding(context.getAgentID(), tryBuildlings.get(
					buildingIndex).getID()));
		} else {
			log.error("No stuckBuilding, why send stuck message?" + buildingIndex + " ,"
					+ tryBuildlings);
		}
	}

	private Pair<Point, Building> getPositionInBuildingToMove() {
		Point pos = null;
		boolean hasPos = false;
		Building taregtBuidling = null;
		Building stuckBuilding = tryBuildlings.get(buildingIndex);
		List<Building> relatedBuildings = getRelatedBuildings(stuckBuilding);
		Set<LineInfo> relatedBuildingsLines = los.getBuidlingsLines(relatedBuildings);
		relatedBuildingsLines.addAll(los.getBuidlingLines(stuckBuilding));
		for (Edge e : passableEdges) {
			if (triedEdge.contains(e)) {
				continue;
			}
			int distance = DEFAULT_BUILDING_DISTANCE;
			Line line = getLineOnEdge(e);
			Point ref = new Point(stuckBuilding.getX(), stuckBuilding.getY());
			while (true) {
				pos = GeoUtils.getPerpendicular(line, distance, ref, true);
				if (pos == null) {
					log.error("getPerpendicular failed!");
					break;
				}
				for (Building b : relatedBuildings) {
					if (GeoUtils.inPolygon(b.getApexList(), pos.getX(), pos.getY())
							&& los.canSeeBuildingXY(pos.getX(), pos.getY(), line.getMidPoint()
									.getX(), line.getMidPoint().getY(), relatedBuildingsLines)) {
						hasPos = true;
						taregtBuidling = b;
						break;
					}
				}
				if (hasPos) {
					currentTryingEdge = e;
					triedEdge.add(e);
					return new Pair<Point, Building>(pos, taregtBuidling);
				} else {
					distance /= 2;
					if (distance <= MIN_DISTANCE) {
						pos = null;
						break;
					}
				}
			}
		}

		triedEdge.clear();
		// log.info("No suitable perpendicular, use center instead!");
		return null;
	}

	private Line getLineOnEdge(Edge e) {
		Point2D origin = new Point2D(e.getStartX(), e.getStartY());
		Point2D end = new Point2D(e.getEndX(), e.getEndY());
		Line2D edgeLine = new Line2D(origin, end);
		Line line = getPassableLine(edgeLine, getRelatedRoadBlockades(e));
		if (line == null) {
			line = new Line(e.getStartX(), e.getStartY(), e.getEndX(), e.getEndY());
		}
		return line;
	}

	private List<Blockade> getRelatedRoadBlockades(Edge e) {
		List<Blockade> rv = new ArrayList<Blockade>();
		HashSet<Road> roads = new HashSet<Road>();
		EntityID neighbourID = e.getNeighbour();
		if (neighbourID == null) {
			log.error("no neighbour found!");
			return rv;
		}

		StandardEntity neighbour = world.getEntity(neighbourID);
		if (neighbour instanceof Road) {
			Road neighbourRoad = (Road) neighbour;
			roads.add(neighbourRoad);
			for (EntityID id : neighbourRoad.getNeighbours()) {
				StandardEntity entity = world.getEntity(id);
				if (entity instanceof Road) {
					Road r = (Road) entity;
					roads.add(r);
				}
			}
		} else {
			log.error(e + " getNeighbour, found road:" + neighbour);
		}

		for (Road r : roads) {
			if (!r.isBlockadesDefined()) {
				continue;
			}
			for (EntityID id : r.getBlockades()) {
				Blockade b = (Blockade) world.getEntity(id);
				if (b != null) {
					rv.add(b);
				}
			}
		}
		return rv;
	}

	private Pair<Point, Road> getPositionInRoadToMove() {
		if (currentTryingEdge == null) {
			// log.error(currentTryingEdge == null);
			return null;
		}
		List<Road> roads = getRelatedRoad(currentTryingEdge);
		Building stuckBuilding = tryBuildlings.get(buildingIndex);

		Point pos = null;
		boolean hasPos = false;
		Road targetRoad = null;
		int distance = DEFAULT_ROAD_DISTANCE;
		Line line = new Line(
				new Point(currentTryingEdge.getStartX(), currentTryingEdge.getStartY()), new Point(
						currentTryingEdge.getEndX(), currentTryingEdge.getEndY()));
		Point ref = new Point(stuckBuilding.getX(), stuckBuilding.getY());
		while (true) {
			pos = GeoUtils.getPerpendicular(line, distance, ref, false);
			if (pos == null) {
				log.error("getPerpendicular failed!");
				break;
			}
			for (Road r : roads) {
				if (GeoUtils.inPolygon(r.getApexList(), pos.getX(), pos.getY())) {
					hasPos = true;
					targetRoad = r;
					break;
				}
			}
			if (hasPos) {
				return new Pair<Point, Road>(pos, targetRoad);
			} else {
				distance /= 2;
				if (distance <= MIN_DISTANCE) {
					pos = null;
					break;
				}
			}
		}

		// log.warn("No such pos found to move on road! currentTryingEdge:"
		// + currentTryingEdge);
		return null;
	}

	private List<Road> getRelatedRoad(Edge e) {
		List<Road> rv = new ArrayList<Road>();
		HashSet<Road> roads = new HashSet<Road>();
		EntityID neighbourID = e.getNeighbour();
		if (neighbourID == null) {
			log.error("no neighbour found!");
			return rv;
		}

		StandardEntity neighbour = world.getEntity(neighbourID);
		if (neighbour instanceof Road) {
			Road neighbourRoad = (Road) neighbour;
			rv.add(neighbourRoad);
			roads.add(neighbourRoad);
			for (EntityID id : neighbourRoad.getNeighbours()) {
				StandardEntity entity = world.getEntity(id);
				if (entity instanceof Road) {
					Road r = (Road) entity;
					rv.add(r);
					roads.add(r);
				}
			}
			for (Road r : roads) {
				for (EntityID id : r.getNeighbours()) {
					StandardEntity entity = world.getEntity(id);
					if (entity instanceof Road && !roads.contains(entity)) {
						rv.add((Road) entity);
					}
				}
			}
		} else {
			log.error(e + " getNeighbour, found road:" + neighbour);
		}
		return rv;
	}

	private List<Building> getRelatedBuildings(Building building) {
		List<Building> rv = new ArrayList<Building>();
		rv.add(building);
		for (EntityID id : building.getNeighbours()) {
			StandardEntity entity = world.getEntity(id);
			if (entity instanceof Building) {
				Building b = (Building) entity;
				rv.add(b);
			}
		}
		return rv;
	}

	private List<Building> getNeighbourBuildingsWithPassbaleEdgeToRoad(Building building) {
		List<Building> rv = new ArrayList<Building>();

		HashSet<Building> nearBuildings = new HashSet<Building>();
		HashSet<Building> traversalBuildings = new HashSet<Building>();

		nearBuildings.add(building);
		traversalBuildings.add(building);
		while (true) {
			if (traversalBuildings.isEmpty()) {
				break;
			}
			Iterator<Building> it = traversalBuildings.iterator();
			Building b = it.next();
			it.remove();

			for (EntityID id : b.getNeighbours()) {
				StandardEntity entity = world.getEntity(id);
				if (entity instanceof Building) {
					Building nearBuilding = (Building) entity;
					if (!nearBuildings.contains(nearBuilding)) {
						nearBuildings.add(nearBuilding);
						traversalBuildings.add(nearBuilding);
					}
				} else if (entity instanceof Road) {
					rv.add(b);
				}
			}
		}
		return rv;
	}

	private List<Edge> getPassableEdgesToRoad(Building building) {
		List<Edge> rv = new ArrayList<Edge>();
		for (Edge e : building.getEdges()) {
			if (e.isPassable()
					&& EdgeLengthComparator.getLength(e) >= RoadUtils.RESCUE_AGENT_RADIUS * 2) {
				EntityID neighbour = e.getNeighbour();
				if (world.getEntity(neighbour) instanceof Road) {
					rv.add(e);
				}
			}
		}
		Collections.sort(rv, new EdgeLengthComparator());
		return rv;
	}
}
