package ant.agent.job.universal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import ant.AgentContext;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.job.AbstractJob;
import ant.agent.job.JobState;
import ant.component.universal.AgentPartitionAssignComponent;
import ant.component.universal.BuildingConditionComponent;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.component.universal.BuildingWithNoCivilianComponent;
import ant.component.universal.ChangeSetDetailComponent;
import ant.component.universal.LocationMapComponent;
import ant.component.universal.LocationMapComponent.AreaNode;
import ant.component.universal.LocationMapComponent.AreaNodeComparator;
import ant.map.component.AbstractRouteComponent;
import ant.map.component.StaticCityGraphComponent;
import ant.map.objects.RouteResult;
import ant.map.partition.Partition;

/**
 * 
 * @author Oliver
 * 
 * @param <E>
 *            Type of agent
 */
public class SearchJobBasedOnAgentLocation<E extends Human> extends AbstractJob<E> {
	private static Log log = LogFactory.getLog(SearchJobBasedOnAgentLocation.class);

	private static final int INITIAL_SEARCH_RANGE = 500000;
	private static final int MAX_SEARCH_RANGE = 1000000;
	private static final int SEARCH_RANGE_EXTENDER = 50000;
	private static final int RECENT_TIME_LEVEL = 4;
	private static final int MAX_CANDITATE_SIZE = 10;

	private ArrayList<Road> roads;
	private AbstractRouteComponent arc;
	private ChangeSetDetailComponent csd;
	private StaticCityGraphComponent scg;
	private LocationMapComponent lmc;
	private BuildingWithNoCivilianComponent bnc;
	private BuildingConditionComponent bcc;
	private int searchRange;
	private RouteResult route;
	private Area target;
	private Partition partition;

	private StandardWorldModel world;

	public SearchJobBasedOnAgentLocation(AgentContext context) {
		super(context);
		this.world = context.getWorld();
		this.arc = context.getAntComponentKit().getAbstractRouteComponent();
		this.scg = context.getAntComponentKit().getStaticCityGraphComponent();
		this.csd = context.getAntComponentKit().getChangeSetDetailComponent();
		this.lmc = (LocationMapComponent) context.getAntComponentKit().getAntComponent(
				LocationMapComponent.class);
		this.bnc = (BuildingWithNoCivilianComponent) context.getAntComponentKit().getAntComponent(
				BuildingWithNoCivilianComponent.class);
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		AgentPartitionAssignComponent apa = (AgentPartitionAssignComponent) context
				.getAntComponentKit().getAntComponent(AgentPartitionAssignComponent.class);
		this.partition = apa.getPartition(context.getAgentID());
		this.searchRange = INITIAL_SEARCH_RANGE;

		this.roads = new ArrayList<Road>();
		for (StandardEntity entity : context.getWorld().getEntitiesOfType(StandardEntityURN.ROAD)) {
			Road r = (Road) entity;
			roads.add(r);
		}
	}

	@Override
	public void overtaken() {
		state = JobState.Inactive;
		target = null;
	}

	@Override
	public boolean shouldExecute() {
		getTargetRoad();
		// log.info("target:" + target + " route:" + route.getMoveRoute()
		// + " route.hasAcceptableRoute():" + route.hasAcceptableRoute());
		return true;
	}

	@Override
	public AntCommand execute() {
		if (route.hasAcceptableRoute()) {
			state = JobState.Exploring;
			// log.info("Search:" + targetRoad);
			return new AntMove(context.getTimestamp(), route.getMoveRoute());
		} else if (target != null) {
			state = JobState.Waiting;
			return new AntMove(context.getTimestamp(), scg.getStaticMoveRoute(context.getHuman()
					.getPosition(), target.getID()));
		} else {
			state = JobState.ExploringRandom;
			return new AntMove(context.getTimestamp(), behavior.getRandomWalk());
		}
	}

	@Override
	public String getJobName() {
		return getClass().getName();
	}

	private void getTargetRoad() {
		Area here = (Area) context.getHuman().getPosition(context.getWorld());

		if (target != null && !lmc.hasAreaBeenRecentlyVisited(target.getID(), RECENT_TIME_LEVEL)
				&& !hasArrivedTargetRoad(here)) {
			BuildingCondition condition = bcc.getBuildingCondition(target.getID());
			if (condition != null && condition.isWorthyOfSearch()) {
				if (((Building) target).isOnFire()) {
					log.error("Target is on fire, but worthyOfSearch!!");
				}
				this.route = arc.search(here.getID(), here.getX(), here.getY(), target.getID());
				if (route.hasAcceptableRoute()) {
					return;
				} else {
					// log.info("!!!no accept route!!");
				}
			}
		}

		Set<Building> agentBuildingLocations = new HashSet<Building>();
		for (Human agent : csd.getAgentsInSight()) {
			StandardEntity location = agent.getPosition(world);
			if (location instanceof Building) {
				agentBuildingLocations.add((Building) location);
			}
		}
		List<AreaNode> areasToSearch = new ArrayList<LocationMapComponent.AreaNode>();
		for (StandardEntity entity : world.getObjectsInRange(here.getX(), here.getY(), searchRange)) {
			if (entity instanceof Building && !(entity instanceof Refuge) && !entity.equals(target)) {
				if (bnc.isBuildingWithNoCivilian(entity.getID())) {
					continue;
				}
				BuildingCondition condition = bcc.getBuildingCondition(entity.getID());
				if (condition == null) {
					log.error("Can't get building condition for " + entity);
					continue;
				} else if (!condition.isWorthyOfSearch()) {
					continue;
				}
				AreaNode node = lmc.getAreaNode(entity.getID());
				if (node != null) {
					if (!agentBuildingLocations.contains(node.getArea().getID())) {
						areasToSearch.add(node);
					}
				} else {
					log.error("node null!! " + entity);
				}
			}
		}
		Collections.sort(areasToSearch, new AreaNodeComparator(context, partition));

		// FIXME
		if (areasToSearch.size() > MAX_CANDITATE_SIZE) {
			areasToSearch = areasToSearch.subList(0, MAX_CANDITATE_SIZE);
		}

		// log.info("areasToSearch:" + areasToSearch);
		if (areasToSearch.size() != 0) {
			for (AreaNode node : areasToSearch) {
				if (!node.getArea().equals(target)) {
					route = arc.search(here.getID(), here.getX(), here.getY(), node.getArea()
							.getID());
					if (route.hasAcceptableRoute()) {
						target = node.getArea();
						return;
					} else {
						target = null;
						continue;
					}
				}
			}
			increaseSearchRange();
			if (target == null) {
				if (areasToSearch.size() != 0) {
					target = areasToSearch.get(0).getArea();
					// log.info("No suitable target, try to " + target);
					return;
				} else {
					log.error("Fail to find search target!!");
					return;
				}
			} else {
				log.error("Target:" + target + " areasToSearch:" + areasToSearch);
			}
		} else {
			increaseSearchRange();
			return;
		}
	}

	private void increaseSearchRange() {
		if (searchRange < MAX_SEARCH_RANGE) {
			searchRange += SEARCH_RANGE_EXTENDER;
			if (searchRange >= MAX_SEARCH_RANGE) {
				log.info("SEARCH_RANGE is already very far!!" + searchRange);
			}
		}
	}

	private boolean hasArrivedTargetRoad(Area here) {
		if (target == null)
			return false;
		if (here.equals(target)) {
			return true;
		}
		return false;
	}
}
