package ant.agent.job.universal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.job.AbstractJob;
import ant.agent.job.JobState;
import ant.component.universal.AgentPartitionAssignComponent;
import ant.component.universal.BuildingConditionComponent;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.component.universal.BuildingWithNoCivilianComponent;
import ant.component.universal.LocationMapComponent;
import ant.component.universal.LocationMapComponent.AreaNode;
import ant.component.universal.LocationMapComponent.AreaNodeComparator;
import ant.map.component.AbstractRouteComponent;
import ant.map.component.StaticCityGraphComponent;
import ant.map.objects.RouteResult;
import ant.map.partition.Partition;

/**
 * 
 * @author Fun
 * 
 * @param <E>
 */
public class SearchJobForATAndFB<E extends Human> extends AbstractJob<E> {
	private static Log log = LogFactory.getLog(SearchJobForATAndFB.class);

	private static final int INITIAL_SEARCH_RANGE = 500000;
	private static final int MAX_SEARCH_RANGE = 1000000;
	private static final int SEARCH_RANGE_EXTENDER = 50000;
	private static final int RECENT_TIME_LEVEL = 4;
	private static final int MAX_CANDITATE_SIZE = 10;
	private static final int MIN_CANDIDATE_SIZE = 3;

	private static final int UNREACHABLE_AREA_IGNORE_INTERVAL = 30;
	private static final int SHORTIME_IGNORE_INTERVAL = 8;

	private AbstractRouteComponent arc;
	private LocationMapComponent lmc;
	private BuildingWithNoCivilianComponent bnc;
	private BuildingConditionComponent bcc;
	private StaticCityGraphComponent scg;
	private int searchRange;
	private Area target;
	private List<EntityID> path;
	private Partition partition;
	private boolean moveFailedTry = false;

	private Map<EntityID, Integer> unreachableArea;
	private Map<EntityID, Integer> shortimeIgnoreArea;

	private StandardWorldModel world;

	public SearchJobForATAndFB(AgentContext context) {
		super(context);
		this.world = context.getWorld();
		this.arc = context.getAntComponentKit().getAbstractRouteComponent();
		this.lmc = (LocationMapComponent) context.getAntComponentKit().getAntComponent(LocationMapComponent.class);
		this.bnc = (BuildingWithNoCivilianComponent) context.getAntComponentKit().getAntComponent(
				BuildingWithNoCivilianComponent.class);
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		
		this.scg = context.getAntComponentKit().getStaticCityGraphComponent();
		AgentPartitionAssignComponent apa = (AgentPartitionAssignComponent) context.getAntComponentKit().getAntComponent(
				AgentPartitionAssignComponent.class);
		this.partition = apa.getPartition(context.getAgentID());
		this.searchRange = INITIAL_SEARCH_RANGE;

		this.unreachableArea = new HashMap<EntityID, Integer>();
		this.shortimeIgnoreArea = new HashMap<EntityID, Integer>();
	}

	@Override
	public boolean shouldExecute() {
		path = null;
		if (isKeepingTarget()) {
			state = JobState.Exploring;
			return true;
		}

		moveFailedTry = false;
		
		target = getNextTarget();

		if (target != null) {
			state = JobState.Exploring;
			return true;
		}

		state = JobState.ExploringRandom;

		return true;
	}

	private boolean isKeepingTarget() {
		Area here = (Area) context.getHuman().getPosition(context.getWorld());
		
		if (target != null && !lmc.hasAreaBeenRecentlyVisited(target.getID(), RECENT_TIME_LEVEL) && !isTargetReached(here)) {
			BuildingCondition condition = bcc.getBuildingCondition(target.getID());
			if (condition != null && condition.isWorthyOfSearch()) {
				if (((Building) target).isOnFire()) {
					log.error("Target is on fire, but worthyOfSearch!!");
				}
				RouteResult route = arc.search(here.getID(), here.getX(), here.getY(), target.getID());
				if (route.hasAcceptableRoute()) {
					path = route.getMoveRoute();
					return true;
				} else if(!moveFailedTry) {
					moveFailedTry = true;
					path = scg.getStaticMoveRoute(context.getHuman()
							.getPosition(), target.getID());
					return true;
				} else{
					addUnreachableArea(target);
					path = null;
					return false;
				}
			}
		}

		return false;
	}

	private Area getNextTarget() {
		Area here = (Area) context.getHuman().getPosition(context.getWorld());

		List<AreaNode> areasToSearch = new ArrayList<LocationMapComponent.AreaNode>();
		for (StandardEntity entity : world.getObjectsInRange(here.getX(), here.getY(), searchRange)) {
			if (entity instanceof Building && !(entity instanceof Refuge) && !entity.equals(target)) {
				if (bnc.isBuildingWithNoCivilian(entity.getID())) {
					continue;
				}

				if (isUnreachableArea((Building) entity)) {
					continue;
				}
				
				if (isShortTimeIgnore((Building) entity)){
					continue;
				}

				BuildingCondition condition = bcc.getBuildingCondition(entity.getID());
				
				if (condition == null) {
					log.error("Can't get building condition for " + entity);
					continue;
				} else if (!condition.isWorthyOfSearch()) {
					continue;
				}
				
				AreaNode node = lmc.getAreaNode(entity.getID());
				
				if (node == null) {
					log.error("node null!! " + entity);
				}else{
					areasToSearch.add(node);
				}
			}
		}

		Collections.sort(areasToSearch, new AreaNodeComparator(context, partition));

		if (areasToSearch.size() > MAX_CANDITATE_SIZE) {
			areasToSearch = areasToSearch.subList(0, MAX_CANDITATE_SIZE);
		}else if(areasToSearch.size() < MIN_CANDIDATE_SIZE){
			increaseSearchRange();
		}

		path = null;
		
		if (areasToSearch.size() != 0) {
			for (AreaNode node : areasToSearch) {
				RouteResult route = arc.search(here.getID(), here.getX(), here.getY(), node.getArea().getID());
				if (route.hasAcceptableRoute()) {
					path = route.getMoveRoute();
					return node.getArea();
				} else {
					addShortTimeIgnore(node.getArea());
					continue;
				}

			}
		}
		
		increaseSearchRange();
		
		return null;
	}
	
	private boolean isShortTimeIgnore(Area area){
		if(shortimeIgnoreArea.containsKey(area.getID())){
			int time = shortimeIgnoreArea.get(area.getID());
			if((context.getTimestamp() - time) < SHORTIME_IGNORE_INTERVAL){
				return true;
			}
		}
		return false;
	}

	private void addShortTimeIgnore(Area area){
		if(!shortimeIgnoreArea.containsKey(area.getID())){
			shortimeIgnoreArea.put(area.getID(), context.getTimestamp());
		}else {
			int time = shortimeIgnoreArea.get(area.getID());
			if(time < context.getTimestamp()){
				shortimeIgnoreArea.put(area.getID(), context.getTimestamp());
			}
		}
	}
	
	private boolean isUnreachableArea(Area area) {
		if (unreachableArea.containsKey(area.getID())) {
			int time = unreachableArea.get(area.getID());
			if ((context.getTimestamp() - time) < UNREACHABLE_AREA_IGNORE_INTERVAL) {
				return true;
			}
		}
		return false;
	}

	private void addUnreachableArea(Area area) {
		if (!unreachableArea.containsKey(area.getID())) {
			unreachableArea.put(area.getID(), context.getTimestamp());
		} else {
			int time = unreachableArea.get(area.getID());
			if (time < context.getTimestamp()) {
				unreachableArea.put(area.getID(), context.getTimestamp());
			}
		}
	}

	private void increaseSearchRange() {
		if (searchRange < MAX_SEARCH_RANGE) {
			searchRange += SEARCH_RANGE_EXTENDER;
			if (searchRange >= MAX_SEARCH_RANGE) {
				log.info("SEARCH_RANGE is already very far!!" + searchRange);
			}
		}
	}

	private boolean isTargetReached(Area here) {
		if (target == null)
			return false;
		if (here.equals(target)) {
			return true;
		}
		return false;
	}

	@Override
	public AntCommand execute() {
		if (state.equals(JobState.Exploring)) {
			if (path != null) {
				return new AntMove(context.getTimestamp(), path);
			} else {
				log.error("path is null!");
			}
		}
		return new AntMove(context.getTimestamp(), behavior.getRandomWalk());
	}

	@Override
	public void overtaken() {
		state = JobState.Inactive;
		target = null;
		path = null;
	}

	@Override
	public String getJobName() {
		return getClass().getName();
	}
}
