package ant.agent.job.universal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.job.JobState;
import ant.liaison.message.MStuckInBuilding;
import ant.liaison.message.MVoiceAgentStuckInBuilding;
import ant.map.component.StaticCityGraphComponent;
import ant.utils.Pair;
import ant.utils.comparator.EdgeLengthComparator;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Line;
import ant.utils.geo.Point;

/**
 * 
 * @author Oliver
 * 
 */
public class StuckInBuildingJob<E extends Human> extends AbstractStuckJob<E> {
	private static final Log log = LogFactory.getLog(StuckInBuildingJob.class);

	private static final int CHECK_STEP = 2;
	private static final int DEFAULT_DISTANCE = 20000;
	private static final int MIN_DISTANCE = 2000;

	private enum StuckStep {
		NotStuck, MovingToCenter, MovingOut
	}

	private StuckStep stuckState;

	private Building stuckBuilding;
	private StaticCityGraphComponent scg;
	private List<Edge> passableEdges;
	private Edge currentTryingEdge;
	private Set<Edge> triedEdge;
	private Point pos;
	private List<EntityID> pathToPos;

	public StuckInBuildingJob(AgentContext context) {
		super(context);

		this.scg = context.getAntComponentKit().getStaticCityGraphComponent();
		this.passableEdges = new ArrayList<Edge>();
		this.triedEdge = new HashSet<Edge>();
	}

	@Override
	public void overtaken() {
		super.overtaken();

		this.stuckBuilding = null;
		this.passableEdges.clear();
		this.stuckState = StuckStep.NotStuck;
	}

	@Override
	public void overleap() {
		super.overleap();

		this.stuckBuilding = null;
		this.passableEdges.clear();
		this.stuckState = StuckStep.NotStuck;
	}

	@Override
	public boolean shouldExecute() {
		// no need to check stuck at start
		if (context.getTimestamp() < ignoreTimestamp + CHECK_STEP)
			return false;

		if (stuckState == StuckStep.NotStuck) {
			StandardEntity here = context.getHuman().getPosition(world);
			if (here instanceof Building && isStuck(CHECK_STEP)) {
				this.state = JobState.MovingOutOfStuck;
				stuckBuilding = (Building) here;

				for (Edge e : stuckBuilding.getEdges()) {
					if (e.isPassable()) {
						passableEdges.add(e);
					}
				}
				Collections.sort(passableEdges, new EdgeLengthComparator());

				this.pos = getPositionInBuildingToMove();
				stuckState = StuckStep.MovingToCenter;
				return true;
			} else {
				return false;
			}
		} else if (stuckState == StuckStep.MovingToCenter) {
			if (!context.getHuman().getPosition().equals(stuckBuilding.getID())) {
				log.warn("Get out of stuck when MovingToCenter!");
				return false;
			} else {
				Pair<Point, Road> pair = getPositionInRoadToMove();
				if (pair != null) {
					pos = pair.first();
					pathToPos = scg.getStaticMoveRoute(context.getHuman().getPosition(), pair
							.second().getID());
				} else {
					pos = null;
					pathToPos = behavior.getRandomWalk();
				}
				stuckState = StuckStep.MovingOut;
				return true;
			}
		} else if (stuckState == StuckStep.MovingOut) {
			if (!context.getHuman().getPosition().equals(stuckBuilding.getID())) {
				// log.info("Done moving out!");
				return false;
			} else {
				log.debug("fail in MovingOut, try again!");
				sendStuckMessage();

				this.pos = getPositionInBuildingToMove();
				stuckState = StuckStep.MovingToCenter;
				return true;
			}
		} else {
			log.warn("Unexpected state:" + stuckState);
			return false;
		}
	}

	private void sendStuckMessage() {
		// FIXME sending policy needed
		if (stuckBuilding != null) {
			// cmu.sendRadio(new MStuckInBuilding(stuckBuilding.getID()));
			addRadioMessage(new MStuckInBuilding(stuckBuilding.getID()));
			addVoiceMessage(new MVoiceAgentStuckInBuilding(context.getAgentID(), stuckBuilding.getID()));
			
		} else {
			log.error("No stuckBuilding, why send stuck message?");
		}
	}

	@Override
	public AntCommand execute() {
		if (stuckState == StuckStep.MovingToCenter) {
			Area here = (Area) world.getEntity(context.getHuman().getPosition());
			List<EntityID> path = new ArrayList<EntityID>();
			path.add(here.getID());
			return new AntMoveXY(this.context.getTimestamp(), path, pos.x(), pos.y());
		} else if (stuckState == StuckStep.MovingOut) {
			if (pos != null) {
				return new AntMoveXY(this.context.getTimestamp(), pathToPos, pos.x(), pos.y());
			} else {
				return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
			}
		} else {
			log.fatal("error stuck state:" + stuckState);
			return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
		}
	}

	@Override
	public String getJobName() {
		return StuckInBuildingJob.class.getName();
	}

	private Point getPositionInBuildingToMove() {
		Point rv = null;
		boolean hasPos = false;
		for (Edge e : passableEdges) {
			if (triedEdge.contains(e)) {
				continue;
			}
			int distance = DEFAULT_DISTANCE;
			while (true) {
				rv = GeoUtils.getPerpendicular(new Line(new Point(e.getStartX(), e.getStartY()),
						new Point(e.getEndX(), e.getEndY())), distance,
						new Point(stuckBuilding.getX(), stuckBuilding.getY()), true);
				if (rv == null) {
					log.error("getPerpendicular failed!");
					break;
				}
				if (GeoUtils.inPolygon(stuckBuilding.getApexList(), rv.getX(), rv.getY())) {
					hasPos = true;
					break;
				} else {
					distance /= 2;
					if (distance <= MIN_DISTANCE) {
						rv = null;
						break;
					}
				}
			}
			if (hasPos) {
				currentTryingEdge = e;
				triedEdge.add(e);
				return rv;
			} else {
				continue;
			}
		}

		triedEdge.clear();
		// log.info("No suitable perpendicular, use center instead!");
		return new Point(stuckBuilding.getX(), stuckBuilding.getY());
	}

	private Pair<Point, Road> getPositionInRoadToMove() {
		Edge edge = null;
		List<Road> roads = new ArrayList<Road>();
		if (currentTryingEdge != null) {
			edge = currentTryingEdge;
			roads.addAll(getRelatedRoad(edge));
		} else {
			if (passableEdges.size() != 0) {
				edge = passableEdges.get(0);
				roads.addAll(getRelatedRoad(edge));
			} else {
				log.error("No passable edge found!");
				return null;
			}
		}

		Point pos = null;
		boolean hasPos = false;
		Road targetRoad = null;
		int distance = DEFAULT_DISTANCE;
		while (true) {
			pos = GeoUtils.getPerpendicular(new Line(new Point(edge.getStartX(), edge.getStartY()),
					new Point(edge.getEndX(), edge.getEndY())), distance,
					new Point(stuckBuilding.getX(), stuckBuilding.getY()), false);
			if (pos == null) {
				log.error("getPerpendicular failed!");
				break;
			}
			for (Road r : roads) {
				if (GeoUtils.inPolygon(r.getApexList(), pos.getX(), pos.getY())) {
					hasPos = true;
					targetRoad = r;
					break;
				}
			}
			if (hasPos) {
				break;
			} else {
				distance /= 2;
				if (distance <= MIN_DISTANCE) {
					pos = null;
					break;
				}
			}
		}
		if (hasPos) {
			// log.info("found road pos to move " + pos + " " + targetRoad);
			return new Pair<Point, Road>(pos, targetRoad);
		} else {
			// log.warn("No such pos found to move on road! currentTryingEdge:"
			// + currentTryingEdge);
			return null;
		}
	}

	private List<Road> getRelatedRoad(Edge e) {
		List<Road> rv = new ArrayList<Road>();
		EntityID neighbourID = e.getNeighbour();
		if (neighbourID == null) {
			log.error("no neighbour found!");
			return rv;
		}

		StandardEntity neighbour = world.getEntity(neighbourID);
		if (neighbour instanceof Road) {
			Road neighbourRoad = (Road) neighbour;
			rv.add(neighbourRoad);
			for (EntityID id : neighbourRoad.getNeighbours()) {
				StandardEntity entity = world.getEntity(id);
				if (entity instanceof Road) {
					Road r = (Road) entity;
					rv.add(r);
				}
			}
		}
		return rv;
	}
}
