package ant.agent.job.universal;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.agent.command.AntCommand;
import ant.agent.command.AntMove;
import ant.agent.command.AntMoveXY;
import ant.agent.job.JobState;
import ant.liaison.message.MStuckInRoad;
import ant.liaison.message.MVoiceAgentStuckInRoad;
import ant.map.component.StaticCityGraphComponent;
import ant.utils.Pair;
import ant.utils.geo.Converter;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Line;
import ant.utils.geo.Point;
import ant.utils.geo.Segment;

/**
 * 
 * @author Oliver
 * 
 */
public class StuckInRoadJob<E extends Human> extends AbstractStuckJob<E> {
	private static Log log = LogFactory.getLog(StuckInRoadJob.class);

	private static final int CHECK_STEP = 2;
	private static final int DEFAULT_DISTANCE_TO_POS = 1000;
	private static final int MIN_DISTANCE = 100;
	private static final int REACH_POS_MIN_DISTANCE = 800;
	private static final int DEFAULT_DISTANCE_TO_MOVE_OUT = 20000;
	private static final int WAIT_TIME_TO_SEND_STUCK_MESSAGE = 2;
	private static final int MAX_MOVE_TO_POS_TRY_TIME = 2;

	private StaticCityGraphComponent scg;

	private Road stuckRoad;
	private List<EntityID> stuckMovePath;
	private EntityID nextAreaID;
	private Area nextArea;
	private Edge edgeToNextArea;
	private boolean isLineInStuckRoad;
	private Line line;
	private Point ref;
	private Point pos;
	private List<EntityID> pathToPos;

	private StuckStep stuckState;

	private boolean needSendStuckMessageLater;
	private int currentWaitTimeToSendStuckMessage;
	private int moveToPosCount;

	private enum StuckStep {
		NotStuck, MovingToPos, MovingOut
	}

	public StuckInRoadJob(AgentContext context) {
		super(context);
		this.scg = context.getAntComponentKit().getStaticCityGraphComponent();
		this.moveToPosCount = 0;
	}

	@Override
	public void overtaken() {
		super.overtaken();
		this.stuckState = StuckStep.NotStuck;
		this.stuckRoad = null;
		this.stuckMovePath = null;
		this.nextAreaID = null;
		this.nextArea = null;
		this.edgeToNextArea = null;
		this.isLineInStuckRoad = false;
		this.line = null;
		this.ref = null;
		this.pos = null;
		this.pathToPos = null;
		this.needSendStuckMessageLater = false;
		this.moveToPosCount = 0;
	}

	@Override
	public void overleap() {
		super.overleap();
		if (this.needSendStuckMessageLater) {
			this.currentWaitTimeToSendStuckMessage++;
			if (currentWaitTimeToSendStuckMessage >= WAIT_TIME_TO_SEND_STUCK_MESSAGE) {
				sendStuckMessage();
				currentWaitTimeToSendStuckMessage = 0;
				needSendStuckMessageLater = false;
				this.stuckRoad = null;
			}
			// stuckRoad for message
		} else {
			this.stuckRoad = null;
		}
		this.stuckState = StuckStep.NotStuck;
		this.stuckMovePath = null;
		this.nextAreaID = null;
		this.nextArea = null;
		this.edgeToNextArea = null;
		this.isLineInStuckRoad = false;
		this.line = null;
		this.ref = null;
		this.pos = null;
		this.pathToPos = null;
		this.moveToPosCount = 0;
	}

	@Override
	public AntCommand execute() {
		if (stuckState == StuckStep.MovingToPos) {
			return new AntMoveXY(context.getTimestamp(), pathToPos, pos.getX(), pos.getY());
		} else if (stuckState == StuckStep.MovingOut) {
			if (pos != null) {
				return new AntMoveXY(this.context.getTimestamp(), pathToPos, pos.getX(), pos.getY());
			} else {
				return new AntMove(this.context.getTimestamp(), pathToPos);
			}
		} else {
			log.fatal("error stuck state:" + stuckState);
			return new AntMove(this.context.getTimestamp(), behavior.getRandomWalk());
		}
	}

	@Override
	public boolean shouldExecute() {
		// no need to check stuck at start
		if (context.getTimestamp() < ignoreTimestamp + CHECK_STEP)
			return false;

		StandardEntity here = context.getHuman().getPosition(world);
		if (stuckState == StuckStep.NotStuck) {
			if (!(here instanceof Road) || !isStuck(CHECK_STEP)) {
				return false;
			}
			stuckRoad = (Road) here;
			if (isAgentInBlockade(stuckRoad, context.getHuman())) {
				sendStuckMessage();
				return false;
			}

			this.stuckMovePath = getStuckMovePath();
			if (stuckMovePath == null) {
				log.error("stuckMovePath null!");
				return false;
			}

			for (EntityID next : stuckMovePath) {
				if (next.equals(stuckRoad.getID())) {
					continue;
				}
				if (stuckRoad.getNeighbours().contains(next)) {
					nextAreaID = next;
					break;
				}
			}
			if (nextAreaID == null) {
				if (stuckMovePath.size() > 1) {
					log.error("nextAreaID not found! " + stuckMovePath);
				} else {
					log.error("FIXME: short path problem!");
					sendStuckMessage();
				}
				return false;
			}
			nextArea = (Area) world.getEntity(nextAreaID);
			edgeToNextArea = stuckRoad.getEdgeTo(nextAreaID);

			if (!setPosToMove()) {
				this.needSendStuckMessageLater = true;
				return false;
			}
			this.state = JobState.MovingOutOfStuck;
			this.stuckState = StuckStep.MovingToPos;
			return true;
		} else if (stuckState == StuckStep.MovingToPos) {
			if (here instanceof Building) {
				log.warn("Get out of stuck when MovingToPos:" + pos);
				return false;
			} else {
				if (!hasReachPos()) {
					if (moveToPosCount == 0) {
//						log.info("hasReachPos first fail!");
					}
					moveToPosCount++;
					if (moveToPosCount >= MAX_MOVE_TO_POS_TRY_TIME) {
						this.needSendStuckMessageLater = true;
						return false;
					} else {
						pathToPos = scg.getStaticMoveRoute(me.getPosition(),
								pathToPos.get(pathToPos.size() - 1));
						return true;
					}
				}
				Pair<Point, Area> pair = getPositionToMoveOut();
				// log.info("getPositionToMoveOut:" + pair);
				if (pair != null) {
					pos = pair.first();
					pathToPos = scg.getStaticMoveRoute(me.getPosition(), pair.second().getID());
				} else {
					pos = null;
					pathToPos = behavior.getRandomWalk();
				}
				stuckState = StuckStep.MovingOut;
				return true;
			}
		} else if (stuckState == StuckStep.MovingOut) {
			if (!here.equals(stuckRoad)) {
				state = JobState.Inactive;
				// success
				return false;
			} else {
				sendStuckMessage();
				// failed!
//				log.info("Try failed in moving out stuck in road!");
				return false;
			}
		} else {
			log.error("Unexpected stuckstate:" + stuckState);
			return false;
		}
	}

	private boolean isAgentInBlockade(Road here, Human me) {
		if (here.isBlockadesDefined()) {
			return false;
		}
		for (EntityID id : here.getBlockades()) {
			Blockade b = (Blockade) world.getEntity(id);
			if (GeoUtils.inPolygon(b.getApexes(), me.getX(), me.getY())) {
				return true;
			}
		}
		return false;
	}

	private boolean setPosToMove() {
		this.line = null;
		this.ref = null;
		Segment segDiv = getDivSegment(stuckRoad, nextAreaID, me.getX(), me.getY());
		Set<Blockade> facedBlocakesOnStuckRoad = getFacedBlocakesOnStuckRoad(segDiv, stuckRoad,
				nextAreaID, world);
		if (!facedBlocakesOnStuckRoad.isEmpty()) {
			line = getPassableLine(new Line2D(new Point2D(segDiv.getX1(), segDiv.getY1()),
					new Point2D(segDiv.getX2(), segDiv.getY2())), facedBlocakesOnStuckRoad);
			if (line == null) {
				// log.error("Line is null");
				return false;
			}
			ref = new Point((edgeToNextArea.getStartX() + edgeToNextArea.getEndX()) / 2,
					(edgeToNextArea.getStartY() + edgeToNextArea.getEndY()) / 2);
			isLineInStuckRoad = true;
			// log.info("line:" + line + " ref" + ref + " isLineInStuckRoad:" +
			// isLineInStuckRoad);
		} else {
			if (nextArea instanceof Building) {
//				log.error("Never should agent stuck like this!");
				return false;
			}
			Road nextRoad = (Road) nextArea;
			List<Blockade> blockades = new ArrayList<Blockade>();
			if (nextRoad.isBlockadesDefined()) {
				for (EntityID id : nextRoad.getBlockades()) {
					blockades.add((Blockade) world.getEntity(id));
				}
			}
			line = getPassableLine(new Line2D(new Point2D(edgeToNextArea.getStartX(),
					edgeToNextArea.getStartY()), new Point2D(edgeToNextArea.getEndX(),
					edgeToNextArea.getEndY())), blockades);
			if (line == null) {
				// log.error("Line is null");
				return false;
			}
			ref = new Point(me.getX(), me.getY());
			isLineInStuckRoad = false;
		}
		int distance = DEFAULT_DISTANCE_TO_POS;
		while (true) {
			pos = GeoUtils.getPerpendicular(line, distance, ref, !isLineInStuckRoad);
			if (pos == null) {
				log.error("getPerpendicular failed!");
				break;
			}
			if (GeoUtils.inPolygon(stuckRoad.getApexList(), pos.getX(), pos.getY())) {
				break;
			} else {
				distance /= 2;
				if (distance <= MIN_DISTANCE) {
					pos = null;
					break;
				}
			}
		}
		if (pos == null) {
//			log.info("pos not found!");
			return false;
		}
		pathToPos = new ArrayList<EntityID>();
		pathToPos.add(stuckRoad.getID());
		return true;
	}

	private Pair<Point, Area> getPositionToMoveOut() {
		List<Area> areas = getRelatedArea();
		int distance = DEFAULT_DISTANCE_TO_MOVE_OUT;
		while (true) {
			pos = GeoUtils.getPerpendicular(line, distance, ref, isLineInStuckRoad);
			if (pos == null) {
				log.error("getPerpendicular failed!");
				break;
			}
			for (Area area : areas) {
				if (GeoUtils.inPolygon(area.getApexList(), pos.getX(), pos.getY())) {
					return new Pair<Point, Area>(pos, area);
				}
			}
			distance /= 2;
			if (distance <= MIN_DISTANCE) {
				pos = null;
				break;
			}
		}
		log.error("no pos to move!");
		return null;
	}

	private List<Area> getRelatedArea() {
		List<Area> rv = new ArrayList<Area>();
		if (isLineInStuckRoad) {
			rv.add(stuckRoad);
		}
		rv.add(nextArea);
		for (EntityID id : nextArea.getNeighbours()) {
			Area area = (Area) world.getEntity(id);
			rv.add(area);
		}
		return rv;
	}

	private List<EntityID> getStuckMovePath() {
		int time = context.getTimestamp();
		for (int i = 1; i <= CHECK_STEP; i++) {
			AntCommand ac = this.context.getCommands().get(time - i);
			if (ac instanceof AntMove) {
				return ((AntMove) ac).getPath();
			} else if (ac instanceof AntMoveXY) {
				return ((AntMoveXY) ac).getPath();
			}
		}
		return null;
	}

	private boolean hasReachPos() {
		if (Math.abs(me.getX() - pos.getX()) + Math.abs(me.getY() - pos.getY()) <= REACH_POS_MIN_DISTANCE) {
			return true;
		} else {
			return false;
		}
	}

	private void sendStuckMessage() {
		if (stuckRoad != null) {
			addRadioMessage(new MStuckInRoad(stuckRoad, context.getHuman().getX(), context
					.getHuman().getY()));
			addVoiceMessage(new MVoiceAgentStuckInRoad(context.getAgentID(), stuckRoad, context
					.getHuman().getX(), context.getHuman().getY()));
		} else {
			log.error("No stuckRoad, why send stuck message?");
		}
	}

	private Segment getDivSegment(Road stuckRoad, EntityID neighbour, int ax, int ay) {
		Edge edge = stuckRoad.getEdgeTo(neighbour);
		int lenX = edge.getEndX() - edge.getStartX();
		int lenY = edge.getEndY() - edge.getStartY();
		Segment seg1 = new Segment(ax - 10 * lenX, ay - 10 * lenY, ax, ay);
		Segment seg2 = new Segment(ax, ay, ax + 10 * lenX, ay + 10 * lenY);
		Point mid = new Point(ax, ay);

		List<Segment> segs = Converter.generateSegments(stuckRoad.getApexList());

		for (Segment next : segs) {
			Point intp = GeoUtils.getSegmentIntersectionPoint(seg1, next);
			if (intp != null) {
				seg1 = new Segment(intp, mid);
			}
			intp = GeoUtils.getSegmentIntersectionPoint(seg2, next);
			if (intp != null) {
				seg2 = new Segment(mid, intp);
			}
		}

		return new Segment(seg1.getOrigin(), seg2.getEnd());
	}

	private Set<Blockade> getFacedBlocakesOnStuckRoad(Segment divSeg, Road stuckRoad,
			EntityID neighbour, StandardWorldModel world) {
		Edge edge = stuckRoad.getEdgeTo(neighbour);
		Point flag = new Point((edge.getStartX() + edge.getEndX()) / 2,
				(edge.getStartY() + edge.getEndY()) / 2);

		Set<Blockade> ret = new HashSet<Blockade>();

		if (!stuckRoad.isBlockadesDefined()) {
			return ret;
		}
		for (EntityID bid : stuckRoad.getBlockades()) {
			Blockade blockade = (Blockade) world.getEntity(bid);
			if (GeoUtils.onSameSide(divSeg.getOrigin(), divSeg.getEnd(), flag,
					new Point(blockade.getX(), blockade.getY()))) {
				ret.add(blockade);
			}
		}
		return ret;
	}

}
