package ant.component.fb;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ant.AntComponent;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Polygon;
import ant.utils.geo.Segment;
import ant.utils.geo.convexhull.JarvisMarch;

/**
 * 
 * @author Oliver
 * 
 */
public abstract class AbstractFireConvexHullComponent<H extends AbstractFireConvexHullComponent.Hull> implements AntComponent {
	private static Log log = LogFactory.getLog(AbstractFireConvexHullComponent.class);
	protected List<H> hulls;

	protected abstract AbstractFireSpliteComponent getFireSpliteComponent();

	protected AbstractFireConvexHullComponent() {
		this.hulls = new ArrayList<H>();
	}

	protected void calHull() {
		long sTime = System.currentTimeMillis();

		hulls.clear();
		AbstractFireSpliteComponent afsc = getFireSpliteComponent();
		if (afsc == null) {
			return;
		}
		List<List<FireBuilding>> splitListResult = afsc.getSplitListResult();
		List<Set<FireBuilding>> splitSetResult = afsc.getSplitSetResult();
		for (int i = 0; i < splitListResult.size(); i++) {
			List<FireBuilding> list = splitListResult.get(i);
			Set<FireBuilding> set = splitSetResult.get(i);

			JarvisMarch<FireBuilding> jm = new JarvisMarch<FireBuilding>(list);
			List<FireBuilding> hull = jm.getHull();

			if (hull != null) {
				hulls.add(newHullInstance(hull, set, i));
			} else {
				String info = "";
				for (FireBuilding fb : list) {
					info += "\t" + fb.getBuilding().getID();
				}
				log.error("No hull found for " + info);
			}
		}

		long dTime = System.currentTimeMillis() - sTime;
		if (dTime > 50) {
			log.error("calHull cost:" + dTime);
		}
	}

	protected abstract H newHullInstance(List<FireBuilding> buildingsOnHull,
			Set<FireBuilding> buildingsInHull, int id);

	public H getHull(FireBuilding fireBuilding) {
		for (H hull : hulls) {
			if (hull.containsBuilding(fireBuilding)) {
				return hull;
			}
		}
		return null;
	}

	public List<H> getHulls() {
		return hulls;
	}

	public String printResult() {
		String info = "ConvexHull Result, got " + hulls.size() + " hulls";
		for (int i = 0; i < hulls.size(); i++) {
			H hull = hulls.get(i);
			info += "\nhull " + i + ":";
			for (FireBuilding fireBuilding : hull.getBuildingsInHull()) {
				info += "\t" + fireBuilding.getBuilding().getID();
			}
		}
		return info;
	}

	public static class Hull {
		public static final double HULL_OUTER_RANGE = 0.36d;
		protected List<FireBuilding> buildingsOnHull;
		protected Set<FireBuilding> buildingsInHull;
		protected Polygon polygon;
		protected int id;

		public Hull(List<FireBuilding> buildingsOnHull, Set<FireBuilding> buildingsInHull, int id) {
			this.buildingsOnHull = buildingsOnHull;
			this.buildingsInHull = buildingsInHull;
			this.id = id;
			int[] apexes = new int[buildingsOnHull.size() * 2];
			for (int i = 0; i < buildingsOnHull.size(); i++) {
				FireBuilding fb = buildingsOnHull.get(i);
				apexes[2 * i] = fb.x();
				apexes[2 * i + 1] = fb.y();
			}
			this.polygon = new Polygon(apexes);
		}

		public Polygon getPolygon() {
			return polygon;
		}

		public boolean containsBuilding(FireBuilding fireBuilding) {
			return buildingsInHull.contains(fireBuilding);
		}

		public List<FireBuilding> getBuildingsOnHull() {
			return buildingsOnHull;
		}

		public Set<FireBuilding> getBuildingsInHull() {
			return buildingsInHull;
		}

		public int getId() {
			return id;
		}

		public double getDistanceToNearestEdge(int x, int y) {
			Double minDistane = Double.MAX_VALUE;
			if (polygon.getEdges().isEmpty()) {
				minDistane = 0d;
			} else {
				for (Segment segment : polygon.getEdges()) {
					double dis = GeoUtils.getDistanceToLine(segment, x, y);
					if (dis < minDistane) {
						minDistane = dis;
					}
				}
			}
			return minDistane;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + id;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			Hull other = (Hull) obj;
			if (id != other.id) {
				return false;
			}
			return true;
		}

		@Override
		public String toString() {
			String info = "hull " + id + ":\nbuildingsInHull\t";
			for (FireBuilding fireBuilding : buildingsInHull) {
				info += " " + fireBuilding.getBuilding().getID();
			}
			info += "\nbuildingsOnHull\t";
			for (FireBuilding fireBuilding : buildingsOnHull) {
				info += " " + fireBuilding.getBuilding().getID();
			}
			info += "\npolygon=" + polygon;
			return info;
		}
	}
}
