package ant.component.fb;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.component.universal.HumanInfoComponent;
import ant.map.component.AntMapBasisComponent;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Point;

/**
 * 
 * @author Oliver
 * @depend {@link AntMapBasisComponent} {@link FireSplitComponent}
 */
public class EstimationFireConvexHullComponent extends
		AbstractFireConvexHullComponent<EstimationFireConvexHullComponent.HullForEstimation> implements AntComponent {
	private static Log log = LogFactory.getLog(EstimationFireConvexHullComponent.class);

	private EstimationFireSplitComponent fsc;
	private HumanInfoComponent hic;

	private Set<Refuge> refuges;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	protected HullForEstimation newHullInstance(List<FireBuilding> buildingsOnHull,
			Set<FireBuilding> buildingsInHull, int id) {
		return new HullForEstimation(buildingsOnHull, buildingsInHull, id);
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.refuges = context.getAntComponentKit().getAntMapBasisComponent().getRefugeSet();
		this.fsc = (EstimationFireSplitComponent) context.getAntComponentKit().getAntComponent(
				EstimationFireSplitComponent.class);
		this.hic = (HumanInfoComponent) context.getAntComponentKit().getAntComponent(
				HumanInfoComponent.class);
	}

	@Override
	protected AbstractFireSpliteComponent getFireSpliteComponent() {
		return fsc;
	}

	@Override
	public void update() {
		calHull();
		for (HullForEstimation hfe : hulls) {
			hfe.setCoveredByOtherHull(isCoveredByOtherHull(hfe));
			hfe.setDistanceToRefuge(calDistanceToRefuge(hfe));
		}
	}

	private boolean isCoveredByOtherHull(Hull hull) {
		for (HullForEstimation hfe : hulls) {
			if (hull.equals(hfe) || hfe.isSmallHull()) {
				continue;
			}
			boolean isInHull = true;
			for (FireBuilding fb : hull.getBuildingsOnHull()) {
				if (!GeoUtils.inPolygon(hfe.getPolygon().getApexes(), fb.x(), fb.y())) {
					isInHull = false;
					break;
				}
			}
			if (isInHull) {
				return true;
			}
		}
		return false;
	}

	private double calDistanceToRefuge(Hull hull) {
		Point center = hull.getPolygon().getCenter();
		double minDis = Double.MAX_VALUE;
		for (Refuge r : refuges) {
			if (GeoUtils.inPolygon(hull.getPolygon().getApexes(), r.getX(), r.getY())) {
				minDis = 0;
				break;
			} else {
				double dis = Math.hypot(r.getX() - center.getX(), r.getY() - center.getY());
				if (dis < minDis) {
					minDis = dis;
				}
			}
		}
		return minDis;
	}

	public class HullForEstimation extends ant.component.fb.AbstractFireConvexHullComponent.Hull {
		private static final double NEAR_IGNITION_TEMPERATURE = 8d;
		private static final int MIN_HULL_SIZE = 6;
		private static final double INNER_AFFECT_COEFFICIENT = 3;
		private static final double LITTLE_HULL_AFFECTION = 1000000;
		private static final double WORTHLESS_BUILDING_WATER_COEFFICIENT = 0.15;
		private static final double NEAR_WORTHY_BUILDING_WATER_COEFFICIENT = 0.45;
		private static final int PER_CIVILIAN_AFFECTION = 5000000;

		private double outerHullAffection;
		private double hullAffection;
		private Set<FireBuilding> worthlessBuildings;
		private Set<FireBuilding> worthyBuildings;
		private double waterQuantityToExtinguish;
		private double selfHullAffection;

		private boolean isCoveredByOtherHull;
		private double distanceToRefuge;
		private boolean isDyingHull;
		private int agentsMoveDistance;

		HullForEstimation(List<FireBuilding> buildingsOnHull, Set<FireBuilding> buildingsInHull,
				int id) {
			super(buildingsOnHull, buildingsInHull, id);
			this.hullAffection = -1;
			this.isCoveredByOtherHull = false;
			this.worthlessBuildings = new HashSet<FireBuilding>();
			this.worthyBuildings = new HashSet<FireBuilding>();
			calBuildingWorth();
			this.selfHullAffection = calSelfHullAffection();
			this.outerHullAffection = calOuterHullAffection();
			this.hullAffection = selfHullAffection + outerHullAffection;
			this.waterQuantityToExtinguish = calWaterQuantityToExtinguish();
			this.isDyingHull = calIsDyingHull();
		}

		private void calBuildingWorth() {
			for (FireBuilding fb : buildingsInHull) {
				boolean isWorthless = true;
				for (FireBuilding nearBuildings : fb.getNearFireBuildings().keySet()) {
					if (!nearBuildings.isCurrentEstiOnFire()
							&& nearBuildings.getCurrentEstiFierynessEnum() != Fieryness.BURNT_OUT) {
						isWorthless = false;
						break;
					}
				}
				if (isWorthless) {
					worthlessBuildings.add(fb);
				} else {
					worthyBuildings.add(fb);
				}
			}
		}

		private double calOuterHullAffection() {
			HashMap<FireBuilding, Double> buildingsAffectMap = new HashMap<FireBuilding, Double>();
			HashMap<FireBuilding, Double> innerBuildingsAffectMap = new HashMap<FireBuilding, Double>();
			for (FireBuilding fb : buildingsInHull) {
				for (Iterator<Entry<FireBuilding, Float>> it = fb.getNearFireBuildings().entrySet()
						.iterator(); it.hasNext();) {
					Entry<FireBuilding, Float> entry = it.next();
					FireBuilding nearFireBuilding = entry.getKey();
					if (buildingsInHull.contains(nearFireBuilding)
							|| nearFireBuilding.getCurrentEstiFierynessEnum() == Fieryness.BURNT_OUT
							|| nearFireBuilding.isCurrentEstiOnFire()) {
						continue;
					}
					Double affectCoefficient = buildingsAffectMap.get(nearFireBuilding);
					if (affectCoefficient == null) {
						affectCoefficient = new Double(entry.getValue());
					} else {
						affectCoefficient += entry.getValue();
					}
					buildingsAffectMap.put(nearFireBuilding, affectCoefficient);

					// inner
					for (Iterator<Entry<FireBuilding, Float>> innerIt = nearFireBuilding
							.getNearFireBuildings().entrySet().iterator(); innerIt.hasNext();) {
						Entry<FireBuilding, Float> innerEntry = innerIt.next();
						FireBuilding innerFireBuilding = innerEntry.getKey();
						if (buildingsInHull.contains(innerFireBuilding)
								|| buildingsAffectMap.containsKey(innerFireBuilding)
								|| innerFireBuilding.getCurrentEstiFierynessEnum() == Fieryness.BURNT_OUT
								|| innerFireBuilding.isCurrentEstiOnFire()) {
							continue;
						}
						Double innerAffectCoefficient = innerBuildingsAffectMap
								.get(innerFireBuilding);
						if (innerAffectCoefficient == null) {
							innerAffectCoefficient = new Double(innerEntry.getValue());
						} else {
							innerAffectCoefficient += entry.getValue() * innerEntry.getValue();
						}
						innerBuildingsAffectMap.put(innerFireBuilding, innerAffectCoefficient);
					}
				}
			}

			for (FireBuilding affectBuilding : buildingsAffectMap.keySet()) {
				innerBuildingsAffectMap.remove(affectBuilding);
			}

			double rv = 0;
			for (Iterator<Entry<FireBuilding, Double>> it = buildingsAffectMap.entrySet()
					.iterator(); it.hasNext();) {
				Entry<FireBuilding, Double> entry = it.next();
				double affection = getBuildingAffection(entry.getKey());
				rv += affection * entry.getValue();
			}

			// inner loop
			for (Iterator<Entry<FireBuilding, Double>> it = innerBuildingsAffectMap.entrySet()
					.iterator(); it.hasNext();) {
				Entry<FireBuilding, Double> entry = it.next();
				double affection = getBuildingAffection(entry.getKey());
				rv += affection * entry.getValue() * INNER_AFFECT_COEFFICIENT;
			}
			return rv;
		}

		private double calSelfHullAffection() {
			double rv = 0;
			for (FireBuilding fb : buildingsInHull) {
				rv += fb.getCurrentEstiFuel();
			}

			return rv;
		}

		private double getBuildingAffection(FireBuilding fb) {
			double rv = 0;

			// temperature
			double temper = fb.getCurrentEstiTemperature();
			double dt = fb.getIgnitionPoint() - temper;
			if (dt < NEAR_IGNITION_TEMPERATURE) {
				rv += Math.max(1, (NEAR_IGNITION_TEMPERATURE - dt) / 3) * fb.getCurrentEstiFuel();
			} else {
				rv += fb.getInitFuel();
			}

			// agent and civilian buried in building
			Set<EntityID> set = hic.getBuildingCivilianMap().get(fb.getBuilding().getID());
			if (set != null) {
				rv += set.size() * PER_CIVILIAN_AFFECTION;
			}

			return rv;
		}

		public double getHullAffection() {
			return hullAffection;
		}

		public double getSelfHullAffection() {
			return selfHullAffection;
		}

		public double getOuterHullAffection() {
			return outerHullAffection;
		}

		private double calWaterQuantityToExtinguish() {
			double rv = 0;
			if (isSmallHull()) {
				for (FireBuilding fb : buildingsInHull) {
					rv += fb.getWaterQuantityToExtinguish();
				}
			} else {
				for (FireBuilding fb : buildingsInHull) {
					if (worthyBuildings.contains(fb)) {
						rv += fb.getWaterQuantityToExtinguish();
					} else {
						boolean isNearWorthyFireBuilding = false;
						for (FireBuilding nearFireBuilding : fb.getNearFireBuildings().keySet()) {
							if (worthyBuildings.contains(nearFireBuilding)) {
								isNearWorthyFireBuilding = true;
							}
						}
						if (isNearWorthyFireBuilding) {
							rv += NEAR_WORTHY_BUILDING_WATER_COEFFICIENT
									* fb.getWaterQuantityToExtinguish();
						} else {
							rv += WORTHLESS_BUILDING_WATER_COEFFICIENT
									* fb.getWaterQuantityToExtinguish();
						}
					}
				}
			}
			return rv;
		}

		public boolean calIsDyingHull() {
			boolean rv = true;
			for (FireBuilding fb : worthyBuildings) {
				for (FireBuilding nearFireBuilding : fb.getNearFireBuildings().keySet()) {
					if (buildingsInHull.contains(nearFireBuilding)) {
						continue;
					}
					if (nearFireBuilding.getCurrentEstiFierynessEnum() == Fieryness.UNBURNT) {
						rv = false;
						break;
					}
				}
			}
			return rv;
		}

		public boolean isDyingHull() {
			return isDyingHull;
		}

		public double getWaterQuantityToExtinguish() {
			return waterQuantityToExtinguish;
		}

		private void setCoveredByOtherHull(boolean isCoveredByOtherHull) {
			this.isCoveredByOtherHull = isCoveredByOtherHull;
		}

		public boolean isCoveredByOtherHull() {
			return isCoveredByOtherHull;
		}

		public boolean isSmallHull() {
			return buildingsInHull.size() < MIN_HULL_SIZE;
		}

		public boolean isLittleOuterAffection() {
			return getOuterHullAffection() < LITTLE_HULL_AFFECTION;
		}

		public Set<FireBuilding> getWorthlessBuildings() {
			return worthlessBuildings;
		}

		public Set<FireBuilding> getWorthyBuildings() {
			return worthyBuildings;
		}

		private void setDistanceToRefuge(double distanceToRefuge) {
			this.distanceToRefuge = distanceToRefuge;
		}

		public double getDistanceToRefuge() {
			return distanceToRefuge;
		}

		public int getAgentsMoveDistance() {
			return agentsMoveDistance;
		}

		public void setAgentsMoveDistance(int agentsMoveDistance) {
			this.agentsMoveDistance = agentsMoveDistance;
		}
	}
}
