package ant.component.fb;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardWorldModel;
import ant.agent.job.fb.BuidlingQualification;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.utils.geo.convexhull.Planarizable;

/**
 * 
 * @author Oliver
 * 
 *         所有的esti-value将用在火区模拟上, 而currentEsti-value当前始终之后一个周期的esti-value的备份
 */
public class FireBuilding implements Planarizable {
	private static Log log = LogFactory.getLog(FireBuilding.class);

	private static final double STEFAN_BOLTZMANN_CONSTANT = 0.000000056704;
	private static final float GAMMA = 0.5f;
	private static final float WATER_COEFFICIENT = 20f;
	private static final int DES_TEMPERATURE = 45;
	private static final float COEFFICIENT_FIX = 1.0f;

	public final static float woodEnergie = 2400f;
	public final static float steelEnergie = 900f;
	public final static float concreteEnergie = 350f;
	public final static float woodCapacity = 1.1f;
	public final static float steelCapacity = 1.0f;
	public final static float concreteCapacity = 1.5f;
	public final static float woodIgnition = 47.0f;
	public final static float steelIgnition = 47.0f;
	public final static float concreteIgnition = 47.0f;
	public final static float refugeIgnition = Float.MAX_VALUE;

	/**
	 * assume that one building can only affect the other buildings in the
	 * affectRange, this arg will affect the complexity of all program
	 */
	protected static final int AFFECT_RANGE = 80000;

	/**
	 * the larger, the more the energy will radiated in one cycle
	 */
	static final double RADIATION_COEFFICENT = 0.011d;
	/**
	 * the larger, the bigger the fuel volume will be
	 */
	static final double FUEL_VOLUME = 1d;
	/**
	 * the larger, the faster the fuel will be consumed, FIXME need adjustment
	 */
	static final float CONSUM_COEFFICIENT = 0.9f;

	static final int FLOOR_HEIGHT = 3;

	static final float AVERAGE_TEMP_COEFFICIENT = 0.3f;

	/**
	 * values estimated for the last record
	 */
	private float fuel;
	private double energy;

	/**
	 * following values will not change after initial()
	 */
	private float volume;
	private float capacity;
	private float initFuel;
	private double initialEnergy;

	protected Collection<Wall> walls;
	protected double totalWallArea;
	private double airTemp;
	protected double nearBuildingArea;

	/**
	 * used to save the near by buildings and the hits
	 * 
	 * should never be used after initial()
	 */
	protected Map<FireBuilding, Float> nearFireBuildings;
	private FireSimulatorComponent fsc;
	protected Building building;

	/**
	 * the real ignition time obtain from model or hearing
	 */
	private int ignitionTime;

	/**
	 * the estimated ignition time of the building
	 */
	private int estiIgnitionTime;

	/**
	 * the estimated burnt out time of the building if it will be burned out
	 */
	private int estiBurntoutTime;

	private float estiFuel;
	private double estiEnergy;
	private double estiEnergyD;
	private int estiFieryness;

	private float currentEstiFuel;
	private double currentEstiEnergy;
	private int currentEstiFieryness;

	/**
	 * Flag for correction
	 */
	private boolean isCorrected;
	protected double energyCorrectionCoefficent;

	protected int totalHits;
	protected int totalRays;

	private boolean isUpdated;

	private StandardWorldModel world;
	protected BuildingCondition condition;
	private BuidlingQualification qualification;

	public FireBuilding() {
	}

	public FireBuilding(FireSimulatorComponent fsc, StandardWorldModel world, Building building) {
		this.fsc = fsc;
		this.world = world;
		this.building = building;
	}

	public void initial() {
		this.nearFireBuildings = new HashMap<FireBuilding, Float>();

		this.volume = building.getGroundArea() * building.getFloors() * FLOOR_HEIGHT;
		this.fuel = initFuel = (float) (getFuelDensity() * volume * FUEL_VOLUME);
		this.capacity = volume * getThermoCapacity();
		this.initialEnergy = FireSimulatorComponent.INITIAL_TEMP * capacity;
		this.energy = initialEnergy;

		this.airTemp = FireSimulatorComponent.INITIAL_TEMP;

		this.ignitionTime = building.isIgnitionDefined() ? (building.getIgnition() ? 0 : -1) : -1;

		this.condition = fsc.getBuildingCondition(building);
		this.isUpdated = false;
		this.isCorrected = false;
		initialValues();

		initWalls();
		initialNearFireBuildings();
	}

	public void initialEnergyCorrectionCoefficent() {
		energyCorrectionCoefficent = 0;
		nearBuildingArea = building.getGroundArea();

		for (Iterator<Entry<FireBuilding, Float>> iter = nearFireBuildings.entrySet().iterator(); iter
				.hasNext();) {
			Entry<FireBuilding, Float> entry = iter.next();
			FireBuilding fireBuilding = entry.getKey();
			float connectedValue = entry.getValue();
			nearBuildingArea += fireBuilding.getBuilding().getGroundArea();
			Float backConnectionValue = fireBuilding.getNearFireBuildings().get(this);
			if (backConnectionValue != null) {
				energyCorrectionCoefficent += connectedValue * backConnectionValue;
				// log.info(building.getID() + " "
				// + fireBuilding.getBuilding().getID()
				// + " connectedValue " + connectedValue
				// + " backConnectionValue " + backConnectionValue);
			}
		}
	}

	private void initialValues() {
		estiIgnitionTime = ignitionTime;
		estiBurntoutTime = -1;
		estiFuel = initFuel;
		setEstiEnergy(initialEnergy);
		estiEnergyD = 0;
		estiFieryness = building.isFierynessDefined() ? building.getFieryness() : 0;
		saveState();
	}

	/**
	 * find the fireBuildings in the affectRange
	 */
	protected void initialNearFireBuildings() {
		nearFireBuildings = new HashMap<FireBuilding, Float>();
		int affectRange = AFFECT_RANGE + getDistanceToWalls();
		for (StandardEntity entity : world.getObjectsInRange(building, affectRange)) {
			if (entity instanceof Building) {
				FireBuilding fireBuilding = fsc.getFireBuilding(entity.getID());
				if (fireBuilding.equals(this)) {
					continue;
				} else {
					nearFireBuildings.put(fireBuilding, new Float(-1));
				}
			}
		}
	}

	protected int getDistanceToWalls() {
		int rv = 0;
		int[] apexs = building.getApexList();
		int x = building.getX();
		int y = building.getY();
		for (int i = 0; i < apexs.length; i += 2) {
			int v = Math.abs(x - apexs[i]) + Math.abs(y - apexs[i + 1]);
			if (v > rv) {
				rv = v;
			}
		}
		return rv;
	}

	/**
	 * Initialize the walls of building and calculate the totalWallArea of the
	 * building at the same time
	 */
	protected void initWalls() {
		if (walls != null)
			return;
		totalWallArea = 0;
		walls = new LinkedList<Wall>();
		int[] apexes = building.getApexList();
		int fx = building.getApexList()[0];
		int fy = building.getApexList()[1];
		int lx = fx;
		int ly = fy;
		for (int n = 2; n < apexes.length; n++) {
			int tx = apexes[n];
			int ty = apexes[++n];
			Wall w = new Wall(lx, ly, tx, ty, this);
			if (w.validate()) {
				walls.add(w);
				totalWallArea += FLOOR_HEIGHT * 1000 * w.length;
			} else {
				// FIXME
				// log.warn("warning: ignoring odd wall at building " +
				// building.getID());
			}
			lx = tx;
			ly = ty;
		}
		Wall w = new Wall(lx, ly, fx, fy, this);
		walls.add(w);
		totalWallArea = totalWallArea / 1000000d;
	}

	/**
	 * better way than initialConnectedValues();
	 * 
	 * initial connectedBuildings and connectedValues
	 * 
	 */
	void initWallValues() {
		totalHits = 0;
		totalRays = 0;
		int selfHits = 0;
		for (Iterator<Wall> w = walls.iterator(); w.hasNext();) {
			Wall wall = w.next();
			// wall.findHits(sfe);
			wall.findHitsInNearFireBuildings();
			totalHits += wall.hits;
			selfHits += wall.selfHits;
			totalRays += wall.rays;
		}
		Iterator<FireBuilding> it = nearFireBuildings.keySet().iterator();
		while (it.hasNext()) {
			FireBuilding fireBuilding = it.next();
			if (nearFireBuildings.get(fireBuilding) < 0) {
				it.remove();
			}
		}
		float base = totalRays;
		for (Iterator<Entry<FireBuilding, Float>> iter = nearFireBuildings.entrySet().iterator(); iter
				.hasNext();) {
			Entry<FireBuilding, Float> entry = iter.next();
			FireBuilding fireBuilding = entry.getKey();
			Float value = entry.getValue();
			nearFireBuildings.put(fireBuilding, value / base);
		}
		// log.info(building.getID() + "{" + (((float) totalHits) * 100 /
		// ((float) totalRays)) + "%,"
		// + totalRays + "," + totalHits + "," + selfHits + "}");
	}

	/**
	 * Based on the relationship between fieryness and fuel,
	 * 
	 * TODO better estimation, use Records
	 * 
	 * @param fieryness
	 * @return the estimated value of fuel
	 */
	private float estiFuel(int fieryness) {
		float returnFuel;
		switch (fieryness) {
		case 0:
		case 4:
			returnFuel = initFuel;
			break;
		case 1:
		case 5:
			returnFuel = (float) (initFuel * 0.66);
			break;
		case 2:
		case 6:
			returnFuel = (float) (initFuel * 0.33);
			break;
		case 3:
		case 7:
			returnFuel = (float) (initFuel * 0.10);
			break;
		case 8:
			returnFuel = 0;
			break;
		default:
			returnFuel = initFuel;
		}

		return returnFuel;
	}

	void correctValues() {
		setEstiEnergy(estiEnergy + estiEnergyD);
		estiEnergyD = 0d;
		estiFuel = getEstiFuel();
		estiFieryness = adjustEstiFieryness();
		if (estiBurntoutTime < 0 && estiFieryness == 8) {
			estiBurntoutTime = fsc.getCurrentTime() + fsc.getSimulateTime();
		} else if (estiBurntoutTime > 0 && estiFieryness != 8) {
			estiBurntoutTime = -1;
		}

		if (estiIgnitionTime < 0 && isEstiOnFire()) {
			estiIgnitionTime = fsc.getCurrentTime() + fsc.getSimulateTime();
		} else if (estiIgnitionTime > 0 && !isEstiOnFire()) {
			estiIgnitionTime = -1;
		}
	}

	boolean isEstiOnFire() {
		return estiFieryness == 1 || estiFieryness == 2 || estiFieryness == 3;
	}

	/**
	 * @return the estiFieryness value according to estiFuel and estiTemperature
	 */
	private int adjustEstiFieryness() {
		if (getEstiTemperature() >= getIgnitionPoint()) {
			if (estiFuel >= initFuel * 0.66)
				return 1; // burning, slightly damaged
			if (estiFuel >= initFuel * 0.33)
				return 2; // burning, more damaged
			if (estiFuel > 0)
				return 3; // burning, severly damaged
		}
		if (estiFuel == initFuel)
			return 0;
		if (estiFuel >= initFuel * 0.66)
			return 5; // extinguished, slightly damaged
		if (estiFuel >= initFuel * 0.33)
			return 6; // extinguished, more damaged
		if (estiFuel > 0)
			return 7; // extinguished, severely damaged
		return 8; // completely burnt down
	}

	/**
	 * from currentEsti_value to esti_value and initial necessary esti-values
	 */
	void readState() {
		estiFuel = currentEstiFuel;
		setEstiEnergy(currentEstiEnergy);
		estiFieryness = currentEstiFieryness;
		estiEnergyD = 0;
	}

	/**
	 * from esti_value to currentEsti_value
	 */
	void saveState() {
		currentEstiFuel = estiFuel;
		currentEstiEnergy = estiEnergy;
		currentEstiFieryness = estiFieryness;
	}

	void updateFromCondition() {
		if (condition.isOnFire() && ignitionTime < 0)
			ignitionTime = fsc.getCurrentTime();

		energy = condition.getTemperature() != BuildingCondition.INVALID_VALUE ? condition
				.getTemperature() * capacity : estiEnergy;

		fuel = estiFuel(condition.getFieryness() != BuildingCondition.INVALID_VALUE ? condition
				.getFieryness() : 0);
		isUpdated = true;
		isCorrected = true;
	}

	public void restoreCondition() {
		currentEstiFuel = estiFuel(condition.getFieryness() != BuildingCondition.INVALID_VALUE ? condition
				.getFieryness() : 0);
		currentEstiEnergy = condition.getTemperature() != BuildingCondition.INVALID_VALUE ? condition
				.getTemperature() * capacity
				: estiEnergy;
		currentEstiFieryness = condition.getFieryness() == -1 ? 0 : condition.getFieryness();
	}

	/**
	 * save current real values to esti-values;
	 * 
	 * fireBuildings that have real changes should call this function after
	 * correction to nearFireBuildings are made
	 */
	public void updateEstiValues() {
		estiFuel = fuel;
		setEstiEnergy(energy);
		estiFieryness = condition.getFieryness();
		if (estiBurntoutTime < 0 && condition.getFieryness() == 8) {
			estiBurntoutTime = fsc.getCurrentTime();
		} else if (estiBurntoutTime > 0 && condition.getFieryness() != 8) {
			estiBurntoutTime = -1;
		}

		if (estiIgnitionTime < 0 && isEstiOnFire()) {
			estiIgnitionTime = fsc.getCurrentTime() + fsc.getSimulateTime();
		} else if (estiIgnitionTime > 0 && !isEstiOnFire()) {
			estiIgnitionTime = -1;
		}
	}

	public float getInitFuel() {
		return initFuel;
	}

	float getEstiConsum() {
		if (estiFuel == 0) {
			return 0;
		}
		float tf = (float) (getEstiTemperature() / 1000f);
		float lf = estiFuel / initFuel;
		float f = tf * lf * 0.2f * CONSUM_COEFFICIENT;
		if (f < 0.005f)
			f = 0.005f;
		return (float) (initFuel * f);
	}

	/**
	 * TODO better control.
	 * 
	 * Radiation energy is proportional to t^4 * wallArea;
	 * 
	 * @return
	 */
	double getEstiRadiationEnergy() {
		/**
		 * make sure that building's energy won't be less than initial energy
		 */
		if (getEstiTemperature() <= FireSimulatorComponent.INITIAL_TEMP) {
			return 0;
		}
		double t = getEstiTemperature() + FireSimulatorComponent.ABSOLUTE_TEMPERATURE + 20;
		double radEn = (t * t * t * t) * RADIATION_COEFFICENT * totalWallArea
				* STEFAN_BOLTZMANN_CONSTANT;
		if (radEn == Double.NaN || radEn == Double.POSITIVE_INFINITY
				|| radEn == Double.NEGATIVE_INFINITY)
			radEn = Double.MAX_VALUE * 0.75;
		return radEn;
	}

	private float getFuelDensity() {
		switch (building.getBuildingCode()) {
		case 0:
			return woodEnergie;
		case 1:
			return steelEnergie;
		default:
			return concreteEnergie;
		}
	}

	private float getThermoCapacity() {
		switch (building.getBuildingCode()) {
		case 0:
			return woodCapacity;
		case 1:
			return steelCapacity;
		default:
			return concreteCapacity;
		}
	}

	public float getIgnitionPoint() {
		if (building instanceof Refuge) {
			return refugeIgnition;
		}
		switch (building.getBuildingCode()) {
		case 0:
			return woodIgnition;
		case 1:
			return steelIgnition;
		default:
			return concreteIgnition;
		}
	}

	public double getEnergyCorrectionCoefficent() {
		return energyCorrectionCoefficent;
	}

	public double getHitRate() {
		return totalHits * 1.0 / totalRays;
	}

	public double getAirTemp() {
		return airTemp;
	}

	public void setAirTemp(double airTemp) {
		this.airTemp = airTemp;
	}

	public double getAverageTemp() {
		return getAverageAirTemp() * AVERAGE_TEMP_COEFFICIENT + getEstiTemperature()
				* (1 - AVERAGE_TEMP_COEFFICIENT);
	}

	public double getAverageAirTemp() {
		double rv = airTemp;
		for (Iterator<Entry<FireBuilding, Float>> iter = nearFireBuildings.entrySet().iterator(); iter
				.hasNext();) {
			Entry<FireBuilding, Float> entry = iter.next();
			FireBuilding fireBuilding = entry.getKey();
			rv += fireBuilding.getAirTemp() * fireBuilding.getBuilding().getGroundArea();
		}
		return (rv + airTemp * building.getGroundArea()) / nearBuildingArea;
	}

	public Collection<Wall> getWalls() {
		if (walls == null) {
			log.warn("walls is null!,reinitial");
			initWalls();
		}
		return walls;
	}

	public void setWalls(Collection<Wall> walls) {
		this.walls = walls;
	}

	public float getCapacity() {
		return capacity;
	}

	public double getEnergy() {
		return energy;
	}

	public int getIgnitionTime() {
		return ignitionTime;
	}

	// ////////////////////////////////////////////////////////
	public int getEstiIgnitionTime() {
		return estiIgnitionTime;
	}

	public void setEstiIgnitionTime(int estiIgnitionTime) {
		this.estiIgnitionTime = estiIgnitionTime;
	}

	public int getEstiBurntoutTime() {
		return estiBurntoutTime;
	}

	public void setEstiBurntoutTime(int estiBurntoutTime) {
		this.estiBurntoutTime = estiBurntoutTime;
	}

	public Building getBuilding() {
		return building;
	}

	public boolean isCorrected() {
		return isCorrected;
	}

	public void setCorrected(boolean isCorrected) {
		this.isCorrected = isCorrected;
	}

	public boolean isUpdated() {
		return isUpdated;
	}

	public void setUpdated(boolean isUpdated) {
		this.isUpdated = isUpdated;
	}

	public Map<FireBuilding, Float> getNearFireBuildings() {
		return nearFireBuildings;
	}

	public double getEstiTemperature() {
		double rv = estiEnergy / capacity;
		if (rv == Double.NaN || rv == Double.POSITIVE_INFINITY || rv == Double.NEGATIVE_INFINITY)
			rv = Double.MAX_VALUE * 0.75;
		return rv;
	}

	public double getCurrentEstiTemperature() {
		double rv = currentEstiEnergy / capacity;
		if (rv == Double.NaN || rv == Double.POSITIVE_INFINITY || rv == Double.NEGATIVE_INFINITY)
			rv = Double.MAX_VALUE * 0.75;
		return rv;
	}

	public float getEstiFuel() {
		return estiFuel;
	}

	void setEstiFuel(float estiFuel) {
		this.estiFuel = estiFuel;
	}

	public float getCurrentEstiFuel() {
		return currentEstiFuel;
	}

	public float getFuelRate() {
		return currentEstiFuel / initFuel;
	}

	public double getEstiEnergy() {
		return estiEnergy;
	}

	public double getCurrentEstiEnergy() {
		return currentEstiEnergy;
	}

	public void setEstiEnergy(double estiEnergy) {
		if (estiEnergy < initialEnergy)
			this.estiEnergy = initialEnergy;
		else
			this.estiEnergy = estiEnergy;
	}

	public double getEstiEnergyD() {
		return estiEnergyD;
	}

	public void setEstiEnergyD(double estiEnergyD) {
		this.estiEnergyD = estiEnergyD;
	}

	public int getEstiFieryness() {
		return estiFieryness;
	}

	public Fieryness getEstiFierynessEnum() {
		return StandardEntityConstants.Fieryness.values()[estiFieryness];
	}

	public int getCurrentEstiFieryness() {
		return currentEstiFieryness;
	}

	public Fieryness getCurrentEstiFierynessEnum() {
		return StandardEntityConstants.Fieryness.values()[currentEstiFieryness];
	}

	public boolean isCurrentEstiOnFire() {
		return currentEstiFieryness == 1 || currentEstiFieryness == 2 || currentEstiFieryness == 3;
	}

	public BuildingCondition getCondition() {
		return condition;
	}

	public BuidlingQualification getQualification() {
		if (qualification == null) {
			qualification = new BuidlingQualification();
		}
		return qualification;
	}

	int waterCool(int waterQuantity) {
		if (waterQuantity <= 0 || currentEstiEnergy <= 0) {
			log.warn("waterQuantity:" + waterQuantity + " currentEstiEnergy:" + currentEstiEnergy);
			return 0;
		}
		double lWATER_COEFFICIENT = (currentEstiFieryness > 0 && currentEstiFieryness < 4 ? WATER_COEFFICIENT
				: WATER_COEFFICIENT * GAMMA);
		double effect = waterQuantity * lWATER_COEFFICIENT;
		int consumed = waterQuantity;
		if (effect > currentEstiEnergy) {
			double pc = currentEstiEnergy / effect;
			effect = currentEstiEnergy;
			consumed *= pc;
		}
		currentEstiEnergy -= effect;
		return waterQuantity - consumed;
	}

	/**
	 * 
	 * @return the water quantity needed to make the fire to temperature ==
	 *         DES_TEMPERATURE
	 */
	public double getWaterQuantityToExtinguish() {
		double lWATER_COEFFICIENT = COEFFICIENT_FIX
				* (currentEstiFieryness > 0 && currentEstiFieryness < 4 ? WATER_COEFFICIENT
						: WATER_COEFFICIENT * GAMMA);
		double dE = currentEstiEnergy - DES_TEMPERATURE * capacity;
		if (dE < 0) {
			return 0;
		} else {
			return dE / lWATER_COEFFICIENT;
		}
	}

	double getWaterQuantityToExtinguishWhenTemperatureIs(int temperature) {
		double lWATER_COEFFICIENT = COEFFICIENT_FIX * WATER_COEFFICIENT;
		double dE = (temperature - DES_TEMPERATURE) * capacity;
		if (dE < 0) {
			return 0;
		} else {
			return dE / lWATER_COEFFICIENT;
		}
	}

	@Override
	public String toString() {
		String info = "" + nearFireBuildings.size();
		for (Iterator<Entry<FireBuilding, Float>> it = nearFireBuildings.entrySet().iterator(); it
				.hasNext();) {
			Entry<FireBuilding, Float> entry = it.next();
			info += "\n(" + entry.getKey().getBuilding().getID() + "," + entry.getValue() + ")";
		}

		return "FireBuilding:" + building.getID() + " [fuel=" + String.format("%.0f", fuel)
				+ ", energy=" + String.format("%.0f", energy) + ", initFuel=" + initFuel
				+ ", volume=" + volume + ", capacity=" + capacity + ", airTemp=" + airTemp
				+ ", ignitionTime=" + ignitionTime + ", estiBurntoutTime=" + estiBurntoutTime
				+ ", estiTemperature=" + String.format("%.0f", getEstiTemperature())
				+ ", estiFuel=" + String.format("%.0f", estiFuel) + ", estiEnergy="
				+ String.format("%.0f", estiEnergy) + ", estiFieryness=" + estiFieryness
				+ ", isCorrected=" + isCorrected + "]";// + info;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((building == null) ? 0 : building.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		FireBuilding other = (FireBuilding) obj;
		if (building == null) {
			if (other.building != null) {
				return false;
			}
		} else if (!building.equals(other.building)) {
			return false;
		}
		return true;
	}

	@Override
	public int x() {
		return getBuilding().getX();
	}

	@Override
	public int y() {
		return getBuilding().getY();
	}
}
