package ant.component.fb;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AbstractAntCommunicationKit;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.component.fb.ExtinguishRecordComponent.BuildingExtinguishRecord;
import ant.component.universal.BuildingConditionComponent;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.liaison.message.IMessage;
import ant.liaison.message.MVoiceConfirmBuilding;
import ant.map.component.StaticCityGraphComponent;
import ant.map.objects.RouteResult;

/**
 * 
 * @author Oliver
 * @depend {@link BuildingConditionComponent}
 * @depend {@link StaticCityGraphComponent}
 */
public class FireConfirmComponent implements AntComponent {
	private static Log log = LogFactory.getLog(FireConfirmComponent.class);

	private static final double AGENT_MAX_MOVE_PER_STEP = 40000;
	private static final double CONFIRM_TIME_COEFFICIENT = 2d;
	private static final int MAX_NEED_CONFIRM_TIME = 15;
	private static final int RECENT_UPDATE_TIME = 2;

	private static final int CONDITION_RECENT_LEVEL_MAX = 5;
	private static final double COEFFICIENT_EXTINGUISH_TIME = 0.5;
	private static final int COMMAND_MAX = 4;
	private static final double CONFIRM_ALONE_DISTANCE = 100000;

	private AgentContext context;
	private int maxPower;
	private Human me;
	private AbstractAntCommunicationKit cmu;
	private BuildingConditionComponent bcc;
	private ExtinguishRecordComponent erc;
	private StaticCityGraphComponent scg;

	/********* buildingID****agentID****confirmRecord ****************/
	private Map<EntityID, Map<EntityID, ConfirmRecord>> confirmRecordMap;

	public enum ConfirmState {
		Confirming, Confirmed, TimeOut
	}

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.maxPower = ConfigKey.getFireMaxPower(context.getConfig());
		this.me = context.getHuman();
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		this.erc = (ExtinguishRecordComponent) context.getAntComponentKit().getAntComponent(
				ExtinguishRecordComponent.class);
		this.scg = context.getAntComponentKit().getStaticCityGraphComponent();
		this.cmu = context.getAntCommunicationKit();

		this.confirmRecordMap = new HashMap<EntityID, Map<EntityID, ConfirmRecord>>();
	}

	@Override
	public void update() {
		// check whether records are still valid
		for (Iterator<Entry<EntityID, Map<EntityID, ConfirmRecord>>> it = confirmRecordMap
				.entrySet().iterator(); it.hasNext();) {
			Entry<EntityID, Map<EntityID, ConfirmRecord>> entry = it.next();
			Map<EntityID, ConfirmRecord> map = entry.getValue();
			boolean shouldClear = map.isEmpty();
			for (ConfirmRecord cr : map.values()) {
				BuildingCondition bc = cr.getConfirmTarget();
				int dt = context.getTimestamp() - bc.getTimestamp();
				if (dt <= RECENT_UPDATE_TIME) {
					shouldClear = true;
					break;
				}
			}
			if (shouldClear) {
				map.clear();
				it.remove();
			}
		}

		// voice msg of others' confirm
		for (IMessage msg : cmu.getVoiceReceiveBox()) {
			if (msg instanceof MVoiceConfirmBuilding
					&& !msg.getSender().equals(context.getAgentID())) {
				addConfirmRecord((MVoiceConfirmBuilding) msg);
			}
		}

		// update exists records
		for (Map<EntityID, ConfirmRecord> map : confirmRecordMap.values()) {
			for (ConfirmRecord cr : map.values()) {
				cr.update(context.getTimestamp());
			}
		}

		// print();
	}

	public boolean shouldConfirmBuildingCondition(FireBuilding fireBuilding) {
		Building b = fireBuilding.getBuilding();
		EntityID bid = b.getID();
		if (
		// others were confirming
		!amIConfirming(bid) && isBuildingInConfirmingState(bid)
		// the building is near, no need to confirm together
				&& Math.hypot(b.getX() - me.getX(), b.getY() - me.getY()) < CONFIRM_ALONE_DISTANCE
		//
		) {
			// log.info("others were confirming:" + bid);
			return false;
		}

		int timeNow = context.getTimestamp();
		BuildingCondition bc = fireBuilding.getCondition();
		int buildingUpdateTime = bc.getTimestamp();

		if (timeNow - buildingUpdateTime <= 1) {
			// just updated
			return false;
		}
		if (timeNow - buildingUpdateTime > CONDITION_RECENT_LEVEL_MAX) {
			// long time no update
			return true;
		}

		BuildingExtinguishRecord record = erc.getExtinguishRecord(bid);
		if (record == null) {
			return false;
		}
		double needExtinguishCount = fireBuilding.getWaterQuantityToExtinguish() / maxPower;
		if (record.getExtinguishCountSinceLastUpdate(buildingUpdateTime) > Math.min(
				needExtinguishCount * COEFFICIENT_EXTINGUISH_TIME, COMMAND_MAX)) {
			// extinguish more time than it should be
			return true;
		}
		return false;
	}

	private void print() {
		// print
		String info = "confirmRecordMap:" + confirmRecordMap.size();
		for (Iterator<Entry<EntityID, Map<EntityID, ConfirmRecord>>> it = confirmRecordMap
				.entrySet().iterator(); it.hasNext();) {
			Entry<EntityID, Map<EntityID, ConfirmRecord>> entry = it.next();
			Map<EntityID, ConfirmRecord> map = entry.getValue();
			info += "\n" + entry.getKey() + " " + map.size() + " ";
			int timeOutRecordCount = 0;
			for (ConfirmRecord cr : map.values()) {
				if (cr.getState() == ConfirmState.TimeOut) {
					timeOutRecordCount++;
				}
				info += "\n\t" + cr;
			}
			info += "\ntimeOutRecordCount:" + timeOutRecordCount;
		}
		if (!confirmRecordMap.isEmpty())
			log.info(info);
	}

	private void addConfirmRecord(MVoiceConfirmBuilding msg) {
		EntityID executor = msg.getSender();
		EntityID bid = msg.getBuilding();
		BuildingCondition bc = bcc.getBuildingCondition(bid);
		Map<EntityID, ConfirmRecord> records = getBuildingConfirmRecords(bid);
		int needConfirmTime = calNeedConfirmTime(msg.getPosition(), bid);
		ConfirmRecord record = new ConfirmRecord(executor, msg.getTimestamp(), bc, needConfirmTime);
		// if (records.containsKey(executor)) {
		// log.warn("Record is already there!");
		// }
		records.put(executor, record);
	}

	public void addSelfConfirmRecord(EntityID bid) {
		EntityID myID = context.getAgentID();
		BuildingCondition bc = bcc.getBuildingCondition(bid);
		Map<EntityID, ConfirmRecord> records = getBuildingConfirmRecords(bid);
		int needConfirmTime = calNeedConfirmTime(context.getHuman().getPosition(context.getWorld())
				.getID(), bid);
		ConfirmRecord record = new ConfirmRecord(myID, context.getTimestamp(), bc, needConfirmTime);
		// if (records.containsKey(myID)) {
		// log.warn("Record is already there!");
		// }
		records.put(myID, record);
	}

	private Map<EntityID, ConfirmRecord> getBuildingConfirmRecords(EntityID buildingID) {
		Map<EntityID, ConfirmRecord> rv = confirmRecordMap.get(buildingID);
		if (rv == null) {
			rv = new HashMap<EntityID, FireConfirmComponent.ConfirmRecord>();
			confirmRecordMap.put(buildingID, rv);
		}
		return rv;
	}

	private int calNeedConfirmTime(EntityID from, EntityID target) {
		RouteResult route = scg.getStaticMoveRouteExt(from, target);
		if (!route.hasAcceptableRoute()) {
			log.error("route == null!");
			return 0;
		}
		double maxCost = CONFIRM_TIME_COEFFICIENT * route.getDistance() / AGENT_MAX_MOVE_PER_STEP;
		if (maxCost < 2) {
			return 1;
		} else if (maxCost > MAX_NEED_CONFIRM_TIME) {
			// log.warn("NeedConfirmTime large:" + maxCost);
			return MAX_NEED_CONFIRM_TIME;
		}
		return (int) maxCost;
	}

	public boolean amIConfirming(EntityID buildingID) {
		Map<EntityID, ConfirmRecord> map = confirmRecordMap.get(buildingID);
		if (map == null) {
			return false;
		}

		ConfirmRecord record = map.get(context.getAgentID());
		if (record != null && record.getState() == ConfirmState.Confirming) {
			return true;
		}
		return false;
	}

	public boolean isBuildingInConfirmingState(EntityID buildingID) {
		Map<EntityID, ConfirmRecord> map = confirmRecordMap.get(buildingID);
		if (map == null) {
			return false;
		}
		for (ConfirmRecord cr : map.values()) {
			if (cr.getState() == ConfirmState.Confirming) {
				return true;
			}
		}
		return false;
	}

	public static class ConfirmRecord {
		private EntityID executor;
		private int startConfirmTime;
		private BuildingCondition confirmTarget;
		private ConfirmState state;
		private int needConfirmTime;

		public ConfirmRecord(EntityID executor, int startConfirmTime,
				BuildingCondition confirmTarget, int needConfirmTime) {
			this.executor = executor;
			this.startConfirmTime = startConfirmTime;
			this.confirmTarget = confirmTarget;
			this.needConfirmTime = needConfirmTime;
			this.state = ConfirmState.Confirming;
		}

		public void update(int now) {
			if (now > startConfirmTime + needConfirmTime) {
				state = ConfirmState.TimeOut;
			}
		}

		public ConfirmState getState() {
			return state;
		}

		public BuildingCondition getConfirmTarget() {
			return confirmTarget;
		}

		public EntityID getExecutor() {
			return executor;
		}

		public int getStartConfirmTime() {
			return startConfirmTime;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((confirmTarget == null) ? 0 : confirmTarget.hashCode());
			result = prime * result + ((executor == null) ? 0 : executor.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			ConfirmRecord other = (ConfirmRecord) obj;
			if (confirmTarget == null) {
				if (other.confirmTarget != null) {
					return false;
				}
			} else if (!confirmTarget.equals(other.confirmTarget)) {
				return false;
			}
			if (executor == null) {
				if (other.executor != null) {
					return false;
				}
			} else if (!executor.equals(other.executor)) {
				return false;
			}
			return true;
		}

		@Override
		public String toString() {
			return "ConfirmRecord [state=" + state + ", executor=" + executor + ", confirmTarget="
					+ confirmTarget.getBuilding() + ", startConfirmTime=" + startConfirmTime
					+ ", needConfirmTime=" + needConfirmTime + "]";
		}
	}

}
