package ant.component.fb;

import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.component.fb.EstimationFireConvexHullComponent.HullForEstimation;
import ant.component.universal.AgentLocationComponent;
import ant.component.universal.AgentLocationComponent.AgentLocationInfo;
import ant.utils.Pair;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Point;
import ant.utils.geo.Polygon;

/**
 * 
 * @author Oliver
 * @depend {@link EstimationFireConvexHullComponent}
 *         {@link AgentLocationComponent} {@link ActiveFireBrigadeComponent}
 */
public class FireConvexHullPickComponent implements AntComponent {
	private static Log log = LogFactory.getLog(FireConvexHullPickComponent.class);

	private static final int DEFAULT_LEAST_AGENTS_NEEDED = 3;
	private static final int DYING_HULL_LEAST_AGENTS_NEEDED = 1;
	private static final int LITTLE_OUTER_AFFECTION_HULL_MOST_AGENTS_NEEDED = 2;

	private static final double MORE_AGENT_NEEDED_COEFFICIENT = 2d;
	private static final float AGENT_NEEDED_MAX_POWER_COFFICIENT = 4f;
	private static final float AGENT_NEEDED_MAX_WATER_COEFFICIENT = 0.6f;
	private static final double MIN_AGENT_NEEDED_FOR_LARGE_HULL = 8;

	private static final int AGENT_MOVE_SPEED = 40000;
	private static final double AGENTS_MOVE_COEFFICIENT = 1;
	private static final double MOVE_DISTANCE_COEFFICIENT = 2.0;
	private static final double FIT_HULL_DELTA_DISTANCE = AGENT_MOVE_SPEED * 5;

	private static final double DYING_HULL_AGENT_NEED_COEFFICIENT = 2.5;

	private AgentContext context;
	private EstimationFireConvexHullComponent efc;
	private AgentLocationComponent alc;
	private ActiveFireBrigadeComponent afb;

	private int maxWater;
	private int maxPower;
	private float agentAverageWater;

	private HullForEstimationComparator hullForEstimationComparator;

	private List<EntityID> fireBrigades;
	private HashMap<EntityID, HullForEstimation> agentHullMap;
	private HashMap<HullForEstimation, List<EntityID>> hullAgentsMap;
	private HashMap<EntityID, List<HullForEstimation>> agentHullsMap;
	private HashMap<EntityID, List<HullForEstimation>> preLiminaryAgentHullsMap;
	private List<HullForEstimation> allHulls;
	private List<EntityID> activeAgents;
	private HashSet<EntityID> freeAgents;
	private HashMap<EntityID, Map<HullForEstimation, Double>> agentHullDistanceMap;

	private HashMap<EntityID, Pair<HullForEstimation, Double>> agentNearestHullMap;

	private HashMap<HullForEstimation, Integer> hullAgentNeedMap;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.efc = (EstimationFireConvexHullComponent) context.getAntComponentKit()
				.getAntComponent(EstimationFireConvexHullComponent.class);
		this.alc = (AgentLocationComponent) context.getAntComponentKit().getAntComponent(
				AgentLocationComponent.class);
		this.afb = (ActiveFireBrigadeComponent) context.getAntComponentKit().getAntComponent(
				ActiveFireBrigadeComponent.class);
		this.maxWater = ConfigKey.getFireMaxWater(context.getConfig());
		this.maxPower = ConfigKey.getFireMaxPower(context.getConfig());
		this.agentAverageWater = Math.min(maxWater * AGENT_NEEDED_MAX_WATER_COEFFICIENT, maxPower
				* AGENT_NEEDED_MAX_POWER_COFFICIENT);
		this.hullForEstimationComparator = new HullForEstimationComparator();

		this.fireBrigades = new ArrayList<EntityID>();
		for (Human h : context.getAntComponentKit().getAntMapBasisComponent().getAgentSet()) {
			if (h instanceof FireBrigade) {
				fireBrigades.add(h.getID());
			}
		}
		this.agentHullMap = new HashMap<EntityID, HullForEstimation>();
		this.agentHullsMap = new HashMap<EntityID, List<HullForEstimation>>();
		for (EntityID id : fireBrigades) {
			agentHullsMap.put(id, new ArrayList<HullForEstimation>());
		}
		this.hullAgentsMap = new HashMap<HullForEstimation, List<EntityID>>();
		this.hullAgentNeedMap = new HashMap<HullForEstimation, Integer>();
		this.agentNearestHullMap = new HashMap<EntityID, Pair<HullForEstimation, Double>>();
		this.allHulls = new ArrayList<HullForEstimation>();
		this.activeAgents = new ArrayList<EntityID>();
		this.preLiminaryAgentHullsMap = new HashMap<EntityID, List<HullForEstimation>>();
		this.freeAgents = new HashSet<EntityID>();
		this.agentHullDistanceMap = new HashMap<EntityID, Map<HullForEstimation, Double>>();
		for (EntityID id : fireBrigades) {
			agentHullDistanceMap.put(id, new HashMap<HullForEstimation, Double>());
		}
	}

	@Override
	public void update() {
		getReady();
		if (allHulls.isEmpty()) {
			return;
		}
		preliminaryAssign();
		advancedAssign();
		// adjustAssign();
	}

	private void getReady() {
		allHulls.clear();
		agentHullMap.clear();
		for (List<HullForEstimation> list : agentHullsMap.values()) {
			list.clear();
		}
		hullAgentsMap.clear();
		hullAgentNeedMap.clear();
		activeAgents.clear();
		agentNearestHullMap.clear();

		preLiminaryAgentHullsMap.clear();
		for (Map<HullForEstimation, Double> map : agentHullDistanceMap.values()) {
			map.clear();
		}

		freeAgents.clear();

		// getReady for allHulls
		for (HullForEstimation h : efc.getHulls()) {
			if (h.getWorthyBuildings().isEmpty() || h.getWaterQuantityToExtinguish() <= 0
					|| h.isCoveredByOtherHull()) {
				continue;
			}
			allHulls.add(h);
		}

		// getReady for activeAgents，activeAgentsLocations
		activeAgents.addAll(fireBrigades);
		for (Iterator<EntityID> it = activeAgents.iterator(); it.hasNext();) {
			EntityID id = it.next();
			if (!afb.isAgentActive(id)) {
				it.remove();
				continue;
			}
		}
	}

	private void preliminaryAssign() {
		// preliminary assign
		freeAgents.addAll(activeAgents);
		for (HullForEstimation h : allHulls) {
			int agentsNeeded = (int) (MORE_AGENT_NEEDED_COEFFICIENT * getHullAgentsNeeded(h));
			Collections.sort(activeAgents, new ActiveAgentSorter(h));
			ArrayList<EntityID> agentsForHull = new ArrayList<EntityID>();
			int moveDistance = 0;
			for (int i = 0; i < Math.min(agentsNeeded, activeAgents.size()); i++) {
				EntityID agentID = activeAgents.get(i);
				agentsForHull.add(agentID);
				List<HullForEstimation> hullsForAgent = agentHullsMap.get(agentID);
				hullsForAgent.add(h);
				// maintain freeAgents
				freeAgents.remove(agentID);
				moveDistance += getAgentHullDistance(agentID, h);
			}
			h.setAgentsMoveDistance(moveDistance);
			hullAgentsMap.put(h, agentsForHull);
		}

		// copy agentHullsMap to preLiminaryAgentHullsMap
		for (Iterator<Entry<EntityID, List<HullForEstimation>>> it = agentHullsMap.entrySet()
				.iterator(); it.hasNext();) {
			Entry<EntityID, List<HullForEstimation>> entry = it.next();
			ArrayList<HullForEstimation> list = new ArrayList<HullForEstimation>(entry.getValue());
			preLiminaryAgentHullsMap.put(entry.getKey(), list);
		}

		for (EntityID id : activeAgents) {
			double minDis = Double.MAX_VALUE;
			HullForEstimation nearestHull = null;
			for (HullForEstimation h : allHulls) {
				double dis = getAgentHullDistance(id, h);
				if (dis < minDis) {
					minDis = dis;
					nearestHull = h;
				}
			}
			agentNearestHullMap.put(id, new Pair<HullForEstimation, Double>(nearestHull, minDis));
		}
	}

	private void advancedAssign() {
		// sort hulls
		Collections.sort(allHulls, hullForEstimationComparator);

		Set<HullForEstimation> assignedHulls = new HashSet<HullForEstimation>();

		// first assign
		for (HullForEstimation h : allHulls) {
			List<EntityID> agents = hullAgentsMap.get(h);
			int agentsNeeded = getHullAgentsNeeded(h);
			Collections.sort(agents, new HullAgentComparator(h));
			int hitCount = 0;
			for (Iterator<EntityID> agentIt = agents.iterator(); agentIt.hasNext();) {
				EntityID id = agentIt.next();
				Pair<HullForEstimation, Double> nearestHullPair = agentNearestHullMap.get(id);
				if (hitCount < agentsNeeded
						&& (nearestHullPair.first().equals(h) || (getAgentHullDistance(id, h) - nearestHullPair
								.second()) < FIT_HULL_DELTA_DISTANCE)) {
					// hit
					hitCount++;
					agentHullMap.put(id, h);
					for (Iterator<HullForEstimation> hullIt = agentHullsMap.get(id).iterator(); hullIt
							.hasNext();) {
						HullForEstimation hull = hullIt.next();
						if (!hull.equals(h)) {
							hullAgentsMap.get(hull).remove(id);
							hullIt.remove();
						}
					}

					if (hitCount == agentsNeeded) {
						assignedHulls.add(h);
					}
				} else {
					List<HullForEstimation> hulls = agentHullsMap.get(id);
					if (hulls.size() == 1) {
						if (!hulls.get(0).equals(h)) {
							log.error("Hull unfit!");
						}
						freeAgents.add(id);
						agentHullsMap.get(id).clear();
					} else {
						for (Iterator<HullForEstimation> hullIt = hulls.iterator(); hullIt
								.hasNext();) {
							HullForEstimation hull = hullIt.next();
							if (hull.equals(h)) {
								hullIt.remove();
								break;
							}
						}
					}
					agentIt.remove();
				}
			}
			hullAgentsMap.put(h, agents);
		}

		// second assign
		List<Entry<EntityID, Pair<HullForEstimation, Double>>> entryList = new ArrayList<Entry<EntityID, Pair<HullForEstimation, Double>>>(
				agentNearestHullMap.entrySet());
		Collections.sort(entryList,
				new Comparator<Entry<EntityID, Pair<HullForEstimation, Double>>>() {
					@Override
					public int compare(Entry<EntityID, Pair<HullForEstimation, Double>> o1,
							Entry<EntityID, Pair<HullForEstimation, Double>> o2) {
						double d = o1.getValue().second() - o2.getValue().second();
						if (d < 0) {
							return -1;
						} else if (d > 0) {
							return 1;
						} else {
							return 0;
						}
					}
				});
		for (Entry<EntityID, Pair<HullForEstimation, Double>> entry : entryList) {
			EntityID id = entry.getKey();
			Pair<HullForEstimation, Double> hullPair = entry.getValue();
			HullForEstimation h = hullPair.first();
			if (agentHullMap.containsKey(id)) {
				continue;
			} else if (assignedHulls.contains(h)) {
				continue;
			}
			List<EntityID> agents = hullAgentsMap.get(h);
			int agentsNeeded = getHullAgentsNeeded(h);
			if (agents.size() < agentsNeeded) {
				agentHullMap.put(id, h);
				agents.add(id);
				for (Iterator<HullForEstimation> hullIt = agentHullsMap.get(id).iterator(); hullIt
						.hasNext();) {
					HullForEstimation hull = hullIt.next();
					if (!hull.equals(h)) {
						hullAgentsMap.get(hull).remove(id);
						hullIt.remove();
					}
				}
				if (agents.size() == agentsNeeded) {
					assignedHulls.add(h);
				}
			} else {
				freeAgents.add(id);
			}
		}

		// third assign
		List<EntityID> freeAgentsList = new ArrayList<EntityID>(freeAgents);
		for (HullForEstimation h : allHulls) {
			if (assignedHulls.contains(h)) {
				continue;
			} else if (freeAgentsList.isEmpty()) {
				break;
			}
			int agentsNeeded = getHullAgentsNeeded(h);
			List<EntityID> agents = hullAgentsMap.get(h);

			Collections.sort(freeAgentsList, new HullAgentComparator(h));
			int freeAgentIndex = Math.min(agentsNeeded - agents.size(), freeAgentsList.size());
			List<EntityID> noLongerFreeAgents = freeAgentsList.subList(0, freeAgentIndex);
			agents.addAll(noLongerFreeAgents);
			freeAgentsList = freeAgentsList.subList(freeAgentIndex, freeAgentsList.size());
			for (EntityID id : noLongerFreeAgents) {
				agentHullMap.put(id, h);
				List<HullForEstimation> hulls = agentHullsMap.get(id);
				if (hulls != null) {
					for (Iterator<HullForEstimation> it = hulls.iterator(); it.hasNext();) {
						HullForEstimation hull = it.next();
						if (!hull.equals(h)) {
							hullAgentsMap.get(hull).remove(id);
							it.remove();
						}
					}
				} else {
					log.warn("hulls should not be null!");
					hulls = new ArrayList<HullForEstimation>();
					hulls.add(h);
				}
			}
			hullAgentsMap.put(h, agents);
		}
	}

	private void adjustAssign() {
		// sort hulls
		Collections.sort(allHulls, hullForEstimationComparator);

		// Adjust assign
		boolean hasLargeHull = false;
		List<EntityID> freeAgentsList = new ArrayList<EntityID>(freeAgents);
		for (Iterator<HullForEstimation> iterator = allHulls.iterator(); iterator.hasNext();) {
			HullForEstimation h = iterator.next();
			List<EntityID> agents = hullAgentsMap.get(h);
			int agentsNeeded = getHullAgentsNeeded(h);
			if (agentsNeeded > agents.size() + freeAgentsList.size()) {
				if (isLargeHull(h)) {
					hasLargeHull = true;
					Set<EntityID> assignedAgents = agentHullMap.keySet();
					agents.clear();
					for (EntityID id : activeAgents) {
						if (assignedAgents.contains(id)) {
							continue;
						}
						agents.add(id);
					}
				} else {
					agents.addAll(freeAgentsList);
					freeAgentsList.clear();
				}
			} else if (agentsNeeded > agents.size()) {
				Collections.sort(freeAgentsList, new HullAgentComparator(h));
				List<EntityID> noLongerFreeAgents = freeAgentsList.subList(0,
						agentsNeeded - agents.size());
				agents.addAll(noLongerFreeAgents);
				freeAgentsList = freeAgentsList.subList(agentsNeeded - agents.size(),
						freeAgentsList.size());
			} else {
				Collections.sort(agents, new HullAgentComparator(h));
				for (EntityID id : agents.subList(agentsNeeded, agents.size())) {
					agentHullsMap.get(id).remove(h);
				}
				agents = agents.subList(0, agentsNeeded);
			}

			for (EntityID id : agents) {
				agentHullMap.put(id, h);
				List<HullForEstimation> hulls = agentHullsMap.get(id);
				if (hulls != null) {
					for (Iterator<HullForEstimation> it = hulls.iterator(); it.hasNext();) {
						HullForEstimation hull = it.next();
						if (!hull.equals(h)) {
							hullAgentsMap.get(hull).remove(id);
							it.remove();
						}
					}
				}
			}
			hullAgentsMap.put(h, agents);
			if (hasLargeHull) {
				break;
			}
		}
	}

	private boolean isLargeHull(HullForEstimation h) {
		if (h.getWaterQuantityToExtinguish() / agentAverageWater > MIN_AGENT_NEEDED_FOR_LARGE_HULL) {
			return true;
		}
		return false;
	}

	private int getHullAgentsNeeded(HullForEstimation h) {
		Integer rv = hullAgentNeedMap.get(h);
		if (rv == null) {
			if (h == null) {
				log.error("h:" + h);
			}
			rv = calAgentsNeeded(h);
			hullAgentNeedMap.put(h, rv);
		}
		return rv;
	}

	private int calAgentsNeeded(HullForEstimation h) {
		if (h == null) {
			log.error("h:" + h);
			return 0;
		}
		if (h.isLittleOuterAffection()) {
			return Math.min((int) (h.getWaterQuantityToExtinguish() / agentAverageWater + 1),
					LITTLE_OUTER_AFFECTION_HULL_MOST_AGENTS_NEEDED);
		} else if (h.isDyingHull()) {
			int minNeeded = DYING_HULL_LEAST_AGENTS_NEEDED;
			// return minNeeded;
			int rv = (int) (h.getWaterQuantityToExtinguish() / agentAverageWater + 1 / DYING_HULL_AGENT_NEED_COEFFICIENT);
			rv = Math.max(minNeeded, rv);
			return rv;
		}
		int minNeeded = DEFAULT_LEAST_AGENTS_NEEDED;
		int rv = Math.max((int) (h.getWaterQuantityToExtinguish() / agentAverageWater + 1),
				minNeeded);
		// test for it
		return rv;
	}

	private double getAgentHullDistance(EntityID agentID, HullForEstimation h) {
		Map<HullForEstimation, Double> map = agentHullDistanceMap.get(agentID);
		if (map == null) {
			log.error("Distance map null for " + agentID);
			map = new HashMap<HullForEstimation, Double>();
			agentHullDistanceMap.put(agentID, map);
		}

		Double dis = map.get(h);
		if (dis == null) {
			AgentLocationInfo info = alc.getAgentLastLocationInfo(agentID);
			dis = calMoveDistance(h.getPolygon(), info);
		}
		return dis;
	}

	private double calMoveDistance(Polygon p, AgentLocationInfo info) {
		Point center = p.getCenter();
		int locationX, locationY;
		if (info.isCoordinateDefined()) {
			locationX = info.getX();
			locationY = info.getY();
		} else {
			Area location = (Area) context.getWorld().getEntity(info.getLocation());
			locationX = location.getX();
			locationY = location.getY();
		}

		if (GeoUtils.inPolygon(p.getApexes(), locationX, locationY)) {
			// TODO
			return 0;
		} else {
			return Math.hypot(center.getX() - locationX, center.getY() - locationY);
		}
	}

	public HashMap<EntityID, HullForEstimation> getAgentHullMap() {
		return agentHullMap;
	}

	public HullForEstimation getHullForAgent(EntityID id) {
		return agentHullMap.get(id);
	}

	public HullForEstimation getHullForMyself() {
		return agentHullMap.get(context.getAgentID());
	}

	public HashMap<EntityID, List<HullForEstimation>> getPreliminaryAgentHullsMap() {
		return preLiminaryAgentHullsMap;
	}

	private class AgentHullComparator implements Comparator<HullForEstimation> {
		private EntityID agentID;

		public AgentHullComparator(EntityID agentID) {
			this.agentID = agentID;
		}

		@Override
		public int compare(HullForEstimation o1, HullForEstimation o2) {
			return getHullAgentValue(o1) - getHullAgentValue(o2);
		}

		private int getHullAgentValue(HullForEstimation h) {
			int rv = 0;

			double moveTime = getAgentHullDistance(agentID, h) * MOVE_DISTANCE_COEFFICIENT
					/ AGENT_MOVE_SPEED;
			// TODO may consider other coefficient

			return rv;
		}
	}

	private class HullAgentComparator implements Comparator<EntityID> {
		private HullForEstimation hull;

		public HullAgentComparator(HullForEstimation hull) {
			this.hull = hull;
		}

		@Override
		public int compare(EntityID o1, EntityID o2) {
			List<HullForEstimation> hulls1 = agentHullsMap.get(o1);
			List<HullForEstimation> hulls2 = agentHullsMap.get(o2);

			if (hulls1.size() != hulls2.size()) {
				return hulls1.size() - hulls2.size();
			} else {
				double d1 = getAgentHullDistance(o1, hull);
				double d2 = getAgentHullDistance(o2, hull);
				if (d1 < d2) {
					return -1;
				} else if (d1 > d2) {
					return 1;
				} else {
					return o1.getValue() - o2.getValue();
				}
			}
		}
	}

	private class ActiveAgentSorter implements Comparator<EntityID> {
		private HullForEstimation hull;

		public ActiveAgentSorter(HullForEstimation hull) {
			this.hull = hull;
		}

		@Override
		public int compare(EntityID o1, EntityID o2) {
			double dd = getAgentHullDistance(o1, hull) - getAgentHullDistance(o2, hull);
			if (dd < 0) {
				return -1;
			} else if (dd > 0) {
				return 1;
			} else {
				return o1.getValue() - o2.getValue();
			}
		}
	}

	private class HullForEstimationComparator implements Comparator<HullForEstimation> {
		private static final int NO_LITTEL_AFFECTION_COEFFICIENT = 400;
		private static final int MAX_POWER_COEFFICIENT = 10;
		private static final int MAX_WATER_COEFFICIENT = 3;
		private static final double EASY_EXTINGUISH_COEFFICIENT = 50;
		private static final int HULL_VALUE_MARGIN = 50;
		private static final double HULL_WATER_VALUE_MARGIN_COEFFICIENT = 3;
		private static final double HULL_REFUGE_DISTANCE_MARGIN = 100000;
		private static final int HULL_RADIUS_MARGIN = 50000;
		private static final int NEED_MANY_AGENT_COEFFICIENT = 100;
		private static final double WATER_NEED_PUNISHMENT = 5;

		@Override
		public int compare(HullForEstimation o1, HullForEstimation o2) {
			int dValue = getValue(o1) - getValue(o2);
			if (Math.abs(dValue) > HULL_VALUE_MARGIN) {
				return -dValue;
			}

			double dWater = o1.getWaterQuantityToExtinguish() - o2.getWaterQuantityToExtinguish();
			if (Math.abs(dWater) > agentAverageWater * HULL_WATER_VALUE_MARGIN_COEFFICIENT) {
				return dWater < 0 ? -1 : 1;
			}

			double dRefugeDistance = o1.getDistanceToRefuge() - o2.getDistanceToRefuge();
			if (Math.abs(dRefugeDistance) > HULL_REFUGE_DISTANCE_MARGIN) {
				// log.info("Use dRefugeDistance");
				return dRefugeDistance < 0 ? -1 : 1;
			}

			int dRadius = o1.getPolygon().getRadius() - o2.getPolygon().getRadius();
			if (Math.abs(dRadius) > HULL_RADIUS_MARGIN) {
				// log.info("Use dRadius");
				return dRadius < 0 ? -1 : 1;
			}

			return getDistanceToMapCenter(o1) - getDistanceToMapCenter(o2);
		}

		private int getDistanceToMapCenter(HullForEstimation h) {
			Rectangle2D bounds = context.getWorld().getBounds();
			Point hCenter = h.getPolygon().getCenter();
			return (int) Math.hypot(bounds.getCenterX() - hCenter.getX(), bounds.getCenterY()
					- hCenter.getY());
		}

		private int getValue(HullForEstimation h) {
			int rv = 0;
			double affection = h.getHullAffection();
			double water = h.getWaterQuantityToExtinguish();
			double moveTime = h.getAgentsMoveDistance() * MOVE_DISTANCE_COEFFICIENT
					/ AGENT_MOVE_SPEED;

			if (!h.isLittleOuterAffection()) {
				rv += NO_LITTEL_AFFECTION_COEFFICIENT;
			}
			if (water != 0) {
				double rate = affection / water;
				rv += rate;
			} else {
				log.error("water zero for hull:" + h);
			}

			rv -= moveTime * AGENTS_MOVE_COEFFICIENT;

			if (water < maxWater * MAX_WATER_COEFFICIENT
					|| water < maxPower * MAX_POWER_COEFFICIENT) {
				rv += EASY_EXTINGUISH_COEFFICIENT;
			}

			if (water / agentAverageWater < MIN_AGENT_NEEDED_FOR_LARGE_HULL) {
				rv += NEED_MANY_AGENT_COEFFICIENT;
			} else {
				rv -= (water / agentAverageWater) * WATER_NEED_PUNISHMENT;
			}

			return rv;
		}
	}

}
