package ant.component.fb;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.component.fb.ExtinguishRecordComponent.WaterQuantity;
import ant.component.universal.BuildingConditionComponent;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.map.component.AntMapBasisComponent;

/**
 * @author Oliver
 * @depend {@link AntMapBasisComponent}
 * @depend {@link BuildingConditionComponent}
 * @depend {@link ExtinguishRecordComponent}
 */
public class FireSimulatorComponent implements AntComponent {
	private static Log log = LogFactory.getLog(FireSimulatorComponent.class);

	/**
	 * assume this is the global initial temperature of world
	 */
	public static final float INITIAL_TEMP = 0;

	public static final float RADIATION_COEFFICENT = 0.011f;
	public static final float AIR_TO_AIR_COEFFICIENT = 1.0f;
	public static final float AIR_TO_BUILDING_COEFFICIENT = 0.0015f;
	/**
	 * One cycle in the simulation corresponds to one minute in the disaster
	 * space
	 */
	public static final float TIME_STEP_LENGTH = 60f;
	public static final int MAX_SIMULATE_TIME = 60;
	public static final float ENERGY_LOSS = 0.86f;

	public static final int ABSOLUTE_TEMPERATURE = 273;
	private static final int AVERAGE_ON_FIRE_TEMPERATURE = 150;

	private static final int MIN_NEIGHBOUR_DISTANCE_FOR_NORMAL_MAP = 15000;
	private static final int MAX_NEIGHBOUR_DISTANCE_FOR_NORMAL_MAP = 24000;
	private static final double BASE_NEIGHBOUR_DISTANCE = 20000;
	private static final double BASE_EXTINGUISH_TIME = 0.7;
	private static final double NEIGHBOUR_DISTANCE_COEFFICIENT_WEIGHT = 0.8;

	private static final double MIN_WORLD_FIRE_BUILDING_SITUATION = 0.5;
	private static final double MAX_WORLD_FIRE_BUILDING_SITUATION = 3.5;

	private Map<EntityID, FireBuilding> fireBuildings;

	/**
	 * record the current estimation time step
	 */
	private int simulateTime;

	private AgentContext context;
	private StandardWorldModel world;

	private BuildingConditionComponent bcc;
	private ExtinguishRecordComponent erc;

	private double worldFireBuildingSituation;

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.world = world;
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		this.erc = (ExtinguishRecordComponent) context.getAntComponentKit().getAntComponent(
				ExtinguishRecordComponent.class);

		this.fireBuildings = new HashMap<EntityID, FireBuilding>();
		for (Building b : context.getAntComponentKit().getAntMapBasisComponent().getBuildingSet()) {
			fireBuildings.put(b.getID(), new FireBuilding(this, world, b));
		}
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.initial();
		}
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.initWallValues();
		}

		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.initialEnergyCorrectionCoefficent();
		}

		this.worldFireBuildingSituation = calWorldFireBuildingSituation();

		// to save memory, test and found it useless
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.getWalls().clear();
			fireBuilding.setWalls(null);
		}

	}

	private double calWorldFireBuildingSituation() {
		double averageWaterToExtinguish = 0;
		double averageWallLength = 0;
		double averageBuildingSize = 0;
		double averageNeighbourDistance = 0;
		int neighbourCount = 0;
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			averageWaterToExtinguish += fireBuilding
					.getWaterQuantityToExtinguishWhenTemperatureIs(AVERAGE_ON_FIRE_TEMPERATURE);
			double wallLength = 0;
			for (Wall w : fireBuilding.getWalls()) {
				wallLength += w.length;
			}
			wallLength /= fireBuilding.getWalls().size();
			averageWallLength += wallLength;

			Building b = fireBuilding.getBuilding();
			int x = b.getX();
			int y = b.getY();
			int[] apex = b.getApexList();
			double buildingSize = 0;
			for (int i = 0; i < apex.length - 1; i += 2) {
				buildingSize += Math.hypot(x - apex[i], y - apex[i + 1]);
			}
			buildingSize = buildingSize / apex.length * 2;
			averageBuildingSize += buildingSize;

			for (FireBuilding neighbour : fireBuilding.getNearFireBuildings().keySet()) {
				averageNeighbourDistance += Math.hypot(x - neighbour.getBuilding().getX(), y
						- neighbour.getBuilding().getY());
				neighbourCount++;
			}
		}
		averageWaterToExtinguish /= fireBuildings.size();
		averageBuildingSize /= fireBuildings.size();
		averageWallLength /= fireBuildings.size();
		averageNeighbourDistance /= neighbourCount;

		// TODO remove unused part

		int maxWater = ConfigKey.getFireMaxWater(context.getConfig());
		int maxPower = ConfigKey.getFireMaxPower(context.getConfig());
		log.info("averageWaterToExtinguish:" + String.format("%.2f", averageWaterToExtinguish)
				+ " maxWater:" + maxWater + " maxPower:" + maxPower + " agentToExtingusih:"
				+ String.format("%.3f", averageWaterToExtinguish / maxWater) + " timeToExtinguish:"
				+ String.format("%.3f", averageWaterToExtinguish / maxPower)
				+ " averageWallLength:" + String.format("%.3f", averageWallLength)
				+ " averageBuildingSize:" + String.format("%.3f", averageBuildingSize)
				+ " averageNeighbourDistance:" + String.format("%.3f", averageNeighbourDistance));
		double neighbourDistanceCoefficient;
		if (averageNeighbourDistance > MIN_NEIGHBOUR_DISTANCE_FOR_NORMAL_MAP
				&& averageNeighbourDistance < MAX_NEIGHBOUR_DISTANCE_FOR_NORMAL_MAP) {
			neighbourDistanceCoefficient = 1.0;
		} else {
			neighbourDistanceCoefficient = averageNeighbourDistance / BASE_NEIGHBOUR_DISTANCE;
		}
		double rv = neighbourDistanceCoefficient * NEIGHBOUR_DISTANCE_COEFFICIENT_WEIGHT
				+ averageWaterToExtinguish / maxPower / BASE_EXTINGUISH_TIME
				* (1 - NEIGHBOUR_DISTANCE_COEFFICIENT_WEIGHT);
		if (rv < MIN_WORLD_FIRE_BUILDING_SITUATION) {
			log.warn("calWorldFireBuildingSituation low:" + rv);
			rv = MIN_WORLD_FIRE_BUILDING_SITUATION;
		} else if (rv > MAX_WORLD_FIRE_BUILDING_SITUATION) {
			log.warn("calWorldFireBuildingSituation high:" + rv);
			rv = MAX_WORLD_FIRE_BUILDING_SITUATION;
		}
		log.info("WorldFireBuildingSituation:" + rv);
		return rv;

		// IO-berlin1
		// averageWaterToExtinguish:12940.65 maxWater:60000 maxPower:4000
		// agentToExtingusih:0.216 timeToExtinguish:3.235
		// averageWallLength:16603.564 averageBuildingSize:17929.238
		// averageNeighbourDistance:47509.613

		// IO-berlin2
		// averageWaterToExtinguish:12940.65 maxWater:40000 maxPower:4000
		// agentToExtingusih:0.324 timeToExtinguish:3.235
		// averageWallLength:16603.564 averageBuildingSize:17929.238
		// averageNeighbourDistance:47509.613

		// IO-berlin3
		// averageWaterToExtinguish:12940.65 maxWater:150000 maxPower:4000
		// agentToExtingusih:0.086 timeToExtinguish:3.235
		// averageWallLength:16603.564 averageBuildingSize:17929.238
		// averageNeighbourDistance:47509.613

		// berlin5
		// averageWaterToExtinguish:12940.65 maxWater:150000 maxPower:10000
		// agentToExtingusih:0.086 timeToExtinguish:1.294
		// averageWallLength:16603.564 averageBuildingSize:17929.238
		// averageNeighbourDistance:47509.613

		// IO-Istanbul1
		// averageWaterToExtinguish:7041.62 maxWater:60000 maxPower:4000
		// agentToExtingusih:0.117 timeToExtinguish:1.760
		// averageWallLength:8991.310 averageBuildingSize:11296.290
		// averageNeighbourDistance:32982.017

		// IO-Istanbul2
		// averageWaterToExtinguish:7041.62 maxWater:60000 maxPower:4000
		// agentToExtingusih:0.117 timeToExtinguish:1.760
		// averageWallLength:8991.310 averageBuildingSize:11296.290
		// averageNeighbourDistance:32982.017

		// IO-Istanbul3
		// averageWaterToExtinguish:7041.62 maxWater:50000 maxPower:5000
		// agentToExtingusih:0.141 timeToExtinguish:1.408
		// averageWallLength:8991.310 averageBuildingSize:11296.290
		// averageNeighbourDistance:32982.017

		// Kobe2
		// averageWaterToExtinguish:1526.88 maxWater:7500 maxPower:500
		// agentToExtingusih:0.204 timeToExtinguish:3.054
		// averageWallLength:5588.734 averageBuildingSize:6119.645
		// averageNeighbourDistance:19006.719

		// IO-Kobe2
		// averageWaterToExtinguish:1526.88 maxWater:30000 maxPower:3000
		// agentToExtingusih:0.051 timeToExtinguish:0.509
		// averageWallLength:5588.734 averageBuildingSize:6119.645
		// averageNeighbourDistance:19006.719

		// Paris2
		// averageWaterToExtinguish:2929.21 maxWater:60000 maxPower:4000
		// agentToExtingusih:0.049 timeToExtinguish:0.732
		// averageWallLength:5175.351 averageBuildingSize:7449.943
		// averageNeighbourDistance:23153.023

		// Paris4
		// averageWaterToExtinguish:2929.21 maxWater:60000 maxPower:4000
		// agentToExtingusih:0.049 timeToExtinguish:0.732
		// averageWallLength:5175.351 averageBuildingSize:7449.943
		// averageNeighbourDistance:23153.023

		// VC2
		// averageWaterToExtinguish:1349.05 maxWater:20000 maxPower:2000
		// agentToExtingusih:0.067 timeToExtinguish:0.675
		// averageWallLength:4933.170 averageBuildingSize:5705.489
		// averageNeighbourDistance:15930.768

		// VC4
		// averageWaterToExtinguish:1349.05 maxWater:30000 maxPower:3000
		// agentToExtingusih:0.045 timeToExtinguish:0.450
		// averageWallLength:4933.170 averageBuildingSize:5705.489
		// averageNeighbourDistance:15930.768
	}

	@Override
	public void update() {
		simulateTime = 0;
		readState();
		cool();
		updateChangedBuildings();
		makeCorrection();
		simulateOneStep();
		saveState();
	}

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	public void simulateOneStep() {
		burn();
		// airTempTransfer();
		// airEnergyTransfer();
		exchangeBuilding();
		finishStep();
	}

	private void debugOut() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			// if (fireBuilding.getEstiEnergy() == Double.NaN)
			log.info("debug @ " + context.getTimestamp() + " " + fireBuilding);
		}
	}

	/**
	 * 
	 * @param buildngID
	 * @return 0 for invalid buildingID; -1 for buildings that won't be burnt
	 *         out; other value for the the time when the building will burnt
	 *         out
	 */
	public int getBurntOutTime(EntityID buildngID) {
		FireBuilding fireBuilding = fireBuildings.get(buildngID);
		if (fireBuilding == null) {
			log.error("Invalid buildingID");
			return 0;
		}
		while ((fireBuilding.getEstiBurntoutTime() == -1) && (simulateTime < MAX_SIMULATE_TIME)) {
			simulateOneStep();
		}

		return fireBuilding.getEstiBurntoutTime();
	}

	/**
	 * 
	 * @param buildngID
	 * @return 0 for invalid buildingID; -1 for buildings that won't be on fire;
	 *         other value for the the time when the building will catch fire
	 */
	public int getIgnitionTime(EntityID buildngID) {
		FireBuilding fireBuilding = fireBuildings.get(buildngID);
		if (fireBuilding == null) {
			log.error("Invalid buildingID");
			return 0;
		}
		// if (fireBuilding.getIgnitionTime() != -1)
		// return fireBuilding.getIgnitionTime();
		while ((fireBuilding.getEstiIgnitionTime() == -1) && (simulateTime < MAX_SIMULATE_TIME)) {
			simulateOneStep();
		}

		return fireBuilding.getEstiIgnitionTime();
	}

	private void readState() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.readState();
		}
	}

	/**
	 * Changes will reflect on nonEsti-Values.
	 * 
	 * as long as the building was changed, new Record of the building's state
	 * will be added to Records
	 * 
	 * @param changeSet
	 * @param container
	 */
	private void updateChangedBuildings() {
		for (BuildingCondition bc : bcc.getCurrentUpdatedConditions()) {
			fireBuildings.get(bc.getBuilding().getID()).updateFromCondition();
		}
	}

	/**
	 * 对与观察值不符的建筑进行修正，使得火区模拟能够不背离实际情况
	 */
	public void makeCorrection() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			/**
			 * condition of correction making
			 */
			if (fireBuilding.isUpdated()
					&& fireBuilding.getCondition().getFieryness() != fireBuilding
							.getEstiFieryness()) {
				/**
				 * for facts that don't fit with the estimation, better
				 * correction should be made.
				 */
				double dE = fireBuilding.getEnergy() - fireBuilding.getEstiEnergy();
				double pureEnergyD = dE / (1.0 + fireBuilding.getEnergyCorrectionCoefficent());
				for (Iterator<Entry<FireBuilding, Float>> iter = fireBuilding
						.getNearFireBuildings().entrySet().iterator(); iter.hasNext();) {
					Entry<FireBuilding, Float> entry = iter.next();
					FireBuilding b = entry.getKey();
					Float connectedValue = entry.getValue();
					/**
					 * only correct the unchanged
					 */
					if (!b.isUpdated()) {
						// return energy
						b.setEstiEnergyD(b.getEstiEnergyD() + pureEnergyD * connectedValue);
						b.setCorrected(true);
					}
				}
				/**
				 * For every observed fireBuilding
				 */
				fireBuilding.updateEstiValues();
			}
		}

		// make the returned energy real
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			if (fireBuilding.isCorrected() && !fireBuilding.isUpdated())
				fireBuilding.correctValues();
		}

		SecondCorrection();
	}

	private void SecondCorrection() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			/**
			 * condition of correction making
			 */
			if (fireBuilding.isUpdated() && fireBuilding.isCorrected()) {
				for (FireBuilding nearBuilding : fireBuilding.getNearFireBuildings().keySet()) {
					/**
					 * only those corrected and unchanged and vary in fieryness,
					 * will need second correction
					 */
					if (nearBuilding.isCorrected()
							&& !nearBuilding.isUpdated()
							&& nearBuilding.getEstiFieryness() != nearBuilding
									.getCurrentEstiFieryness()) {
						double dE = nearBuilding.getEnergy() - nearBuilding.getEstiEnergy();
						double pureEnergyD = dE
								/ (1.0 + nearBuilding.getEnergyCorrectionCoefficent());
						for (Iterator<Entry<FireBuilding, Float>> iter = nearBuilding
								.getNearFireBuildings().entrySet().iterator(); iter.hasNext();) {
							Entry<FireBuilding, Float> entry = iter.next();
							FireBuilding b = entry.getKey();
							Float connectedValue = entry.getValue();
							/**
							 * only correct the uncorrected
							 */
							if (!b.isCorrected()) {
								// return energy
								b.setEstiEnergyD(b.getEstiEnergyD() + pureEnergyD * connectedValue);
							}
						}
					}// end of innner if
				}// end of inner for
			}// end of outter if
		}// end of outter for

		// make the returned energy real
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			if (fireBuilding.getEstiEnergyD() != 0.0d)
				fireBuilding.correctValues();
		}

		// finish correction
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.setUpdated(false);
			fireBuilding.setCorrected(false);
		}
	}

	private void saveState() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			fireBuilding.saveState();
		}
	}

	private void cool() {
		for (Iterator<Entry<EntityID, WaterQuantity>> it = erc.getBuildingWaterQuanatityMap()
				.entrySet().iterator(); it.hasNext();) {
			Entry<EntityID, WaterQuantity> entry = it.next();
			WaterQuantity wq = entry.getValue();
			int waterleft = fireBuildings.get(entry.getKey()).waterCool(wq.getQuantity());
			if (waterleft <= 0) {
				it.remove();
			} else {
				wq.setQuantity(waterleft);
			}
		}
	}

	/**
	 * Use estimated values
	 */
	private void burn() {
		for (FireBuilding b : fireBuildings.values()) {
			if (b.getEstiTemperature() >= b.getIgnitionPoint() && b.getEstiFuel() > 0) {
				float consumed = b.getEstiConsum();
				if (consumed > b.getEstiFuel()) {
					consumed = b.getEstiFuel();
					b.setEstiFuel(0f);
					b.setEstiEnergy(b.getEstiEnergy() + consumed);
				} else {
					b.setEstiFuel(b.getEstiFuel() - consumed);
					b.setEstiEnergy(b.getEstiEnergy() + consumed);
				}
			}
		}
	}

	/**
	 * adjust air temperature based on the building temperature and nearby air
	 * temperature; energy loss is considered
	 */
	private void airTempTransfer() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			double oldTemp = fireBuilding.getAirTemp();
			double newTemp = oldTemp + (fireBuilding.getAverageTemp() - oldTemp)
					* AIR_TO_AIR_COEFFICIENT;
			fireBuilding.setAirTemp(newTemp);
			// if (fireBuilding.getBuilding().getID().getValue() <= 951
			// && fireBuilding.getBuilding().getID().getValue() >= 949)
			log.info("1NewTemp for " + fireBuilding.getBuilding().getID() + " is " + newTemp);
		}
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			double deltaT = fireBuilding.getAirTemp() - INITIAL_TEMP;
			double newTemp = INITIAL_TEMP + relTemp(deltaT);
			fireBuilding.setAirTemp(newTemp);
			// if (fireBuilding.getBuilding().getID().getValue() <= 951
			// && fireBuilding.getBuilding().getID().getValue() >= 949)
			log.info("2NewTemp for " + fireBuilding.getBuilding().getID() + " is " + newTemp);
		}
	}

	/**
	 * do energy transfer based on the difference in temperature between
	 * building temperature and nearby average temperature
	 */
	private void airEnergyTransfer() {
		log.info("In airEnergyTransfer");
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			double T = fireBuilding.getEstiTemperature();
			double dT = fireBuilding.getAverageTemp() - T;
			float coefficient = (AIR_TO_BUILDING_COEFFICIENT * TIME_STEP_LENGTH)
					/ fireBuilding.getCapacity();
			T += dT * coefficient;
			fireBuilding.setEstiEnergy(T * fireBuilding.getCapacity());
			// if (fireBuilding.getBuilding().getID().getValue() <= 951
			// && fireBuilding.getBuilding().getID().getValue() >= 949)
			log.info("New temp for " + fireBuilding.getBuilding().getID() + " is " + T);
		}
	}

	/**
	 * do energy loss
	 * 
	 * @param deltaT
	 * @return
	 */
	private double relTemp(double deltaT) {
		// return deltaT * ENERGY_LOSS * TIME_STEP_LENGTH;
		return deltaT * ENERGY_LOSS;
	}

	/**
	 * Use estimated values
	 * 
	 * 与燃烧产生的热量有关 基本关系： 散发热量 正比于 (温度+273)^4 * 墙面积 总热量 正比于 温度
	 */
	private void exchangeBuilding() {
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			double radEn = fireBuilding.getEstiRadiationEnergy();
			// log.info(fireBuilding.getBuilding().getID() + " radEn:" + radEn);
			for (Iterator<Entry<FireBuilding, Float>> it = fireBuilding.getNearFireBuildings()
					.entrySet().iterator(); it.hasNext();) {
				Entry<FireBuilding, Float> entry = it.next();
				FireBuilding b = entry.getKey();
				double oldEnergy = b.getEstiEnergyD();
				double connectionValue = entry.getValue().doubleValue();
				double a = radEn * connectionValue * ENERGY_LOSS;
				double sum = oldEnergy + a;
				b.setEstiEnergyD(sum);
				// log.info("oldEnergy " + oldEnergy + " connectionValue=" +
				// connectionValue + " a="
				// + a);
			}
			fireBuilding.setEstiEnergy(fireBuilding.getEstiEnergy() - radEn);
		}

		for (FireBuilding b : fireBuildings.values()) {
			b.setEstiEnergy(b.getEstiEnergy() + b.getEstiEnergyD());
			b.setEstiEnergyD(0.0);
		}
	}

	private void finishStep() {
		for (Iterator<Entry<EntityID, FireBuilding>> i = fireBuildings.entrySet().iterator(); i
				.hasNext();) {
			FireBuilding fireBuilding = i.next().getValue();
			/**
			 * correct esti-values so that next simulation can continue
			 */
			fireBuilding.correctValues();
			/**
			 * set burnt out time, iff fieryness==8 & it hasn't been set
			 */
			if (fireBuilding.getEstiFieryness() == 8 && fireBuilding.getEstiBurntoutTime() < 0)
				fireBuilding.setEstiBurntoutTime(context.getTimestamp() + simulateTime);

			/**
			 * set ignition time, iff isEstiOnFire() & it hasn't been set
			 */
			if (fireBuilding.isEstiOnFire() && fireBuilding.getEstiIgnitionTime() < 0)
				fireBuilding.setEstiIgnitionTime(context.getTimestamp() + simulateTime);
		}
		simulateTime++;
	}

	public FireBuilding getFireBuilding(EntityID id) {
		return fireBuildings.get(id);
	}

	/**
	 * 
	 * @param currentLocation
	 *            the center which we calculate the distance
	 * @param range
	 *            the limitation of distance from the target building to
	 *            currentLocation
	 * @param currentLocation
	 * @param recentLevel
	 * @return List of fireBuilding which is on fire and in the range limitation
	 */
	public List<FireBuilding> getBurningBuildingsInRange(EntityID currentLocation, int range,
			int recentLevel) {
		ArrayList<FireBuilding> burningBuildings = new ArrayList<FireBuilding>();
		for (StandardEntity entity : world.getObjectsInRange(currentLocation, range)) {
			if (entity instanceof Building) {
				if (entity instanceof Refuge)
					continue;
				FireBuilding fireBuilding = getFireBuilding(entity.getID());
				if (fireBuilding == null) {
					log.warn("FireBuilding not found:" + entity);
				} else {
					if (fireBuilding.getCondition().isOnFire()
							&& (context.getTimestamp() - fireBuilding.getCondition().getTimestamp()) <= recentLevel) {
						burningBuildings.add(fireBuilding);
					}
				}
			}
		}
		return burningBuildings;
	}

	/**
	 * 
	 * @param recentLevel
	 * @return list of fireBuilding whose fieryness = 1
	 *         <p>
	 *         and now() - lastRecordTime() <= recentLevel
	 */
	public List<FireBuilding> getRecentlyBurningBuildings(int recentLevel) {
		ArrayList<FireBuilding> burningBuildings = new ArrayList<FireBuilding>();
		for (FireBuilding fireBuilding : fireBuildings.values()) {
			BuildingCondition condition = fireBuilding.getCondition();
			if (condition == null)
				continue;
			else if (condition.getFieryness() == 1
					&& (context.getTimestamp() - fireBuilding.getCondition().getTimestamp()) <= recentLevel) {
				burningBuildings.add(fireBuilding);
			}
		}
		return burningBuildings;
	}

	public Collection<FireBuilding> getAllFireBuildings() {
		return fireBuildings.values();
	}

	int getSimulateTime() {
		return simulateTime;
	}

	int getCurrentTime() {
		return context.getTimestamp();
	}

	BuildingCondition getBuildingCondition(Building building) {
		if (building != null) {
			return bcc.getBuildingCondition(building.getID());
		} else {
			log.error("getBuildingCondition for null building!");
			return null;
		}
	}

	public double getWorldFireBuildingSituation() {
		return worldFireBuildingSituation;
	}
}
