package ant.component.fb;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.component.universal.BuildingConditionComponent;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.component.universal.ChangeSetDetailComponent;
import ant.map.component.AntMapBasisComponent;

/**
 * 
 * @author Oliver
 * @depend {@link ChangeSetDetailComponent}<br>
 *         {@link AntMapBasisComponent}<br>
 *         {@link BuildingConditionComponent}<br>
 *         {@link FireSimulatorComponent}<br>
 *         {@link RealTimeFireTraceBuildingManagerComponent}<br>
 */
public class FireTraceComponent implements AntComponent {
	private static Log log = LogFactory.getLog(FireTraceComponent.class);

	private static final int TRACE_MAX_TIME = 5;
	private static final double INITIAL_D_TRACE_VALUE = 1d;
	private static final double NO_VISIT_TRACE_VALUE = 3d;
	private static final int CACHE_SIZE = 5;
	private static final double MIN_TRACE_VALUE = 0.1;

	private AgentContext context;
	private ChangeSetDetailComponent csd;
	private FireSimulatorComponent fsc;
	private BuildingConditionComponent bcc;
	private RealTimeFireTraceBuildingManagerComponent rtftc;

	private Map<EntityID, TraceInfo> traceMap;
	private List<Map<TraceInfo, Double>> currentTraceListCache;
	private List<TraceInfo> currentTraceList;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.csd = context.getAntComponentKit().getChangeSetDetailComponent();
		this.fsc = (FireSimulatorComponent) context.getAntComponentKit().getAntComponent(
				FireSimulatorComponent.class);
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		this.rtftc = (RealTimeFireTraceBuildingManagerComponent) context.getAntComponentKit().getAntComponent(
				RealTimeFireTraceBuildingManagerComponent.class);
		this.traceMap = new HashMap<EntityID, TraceInfo>();
		for (Building b : context.getAntComponentKit().getAntMapBasisComponent().getBuildingSet()) {
			traceMap.put(b.getID(), new TraceInfo(bcc.getBuildingCondition(b.getID())));
		}

		this.currentTraceListCache = new LinkedList<Map<TraceInfo, Double>>();
		this.currentTraceList = new ArrayList<TraceInfo>();
	}

	@Override
	public void update() {
		HashSet<TraceInfo> toClearTraceInfo = new HashSet<TraceInfo>();
		for (BuildingCondition bc : bcc.getCurrentUpdatedConditions()) {
			// 直接删除getCurrentUpdatedConditions中包含的trace，也就是说看到的Building的traceInfo应该清零
			traceMap.get(bc.getBuilding().getID()).clearTraceValue();

			if (bc.isEverOnFire() || (bc.isNoZeroTemperatureEver() && bc.getTemperature() == 0)) {
				// 曾经着火，或者曾经温度非零，但是当前温度降为0了
				clearTraceInfo(bc, toClearTraceInfo);
			}
		}

		ArrayList<BuildingCondition> traceBuildings = new ArrayList<BuildingCondition>();
		for (StandardEntity entity : csd.getEntitiesInSight(StandardEntityURN.BUILDING,
				StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.FIRE_STATION,
				StandardEntityURN.POLICE_OFFICE, StandardEntityURN.REFUGE)) {
			Building b = (Building) entity;
			if (b.isTemperatureDefined() && b.getTemperature() > 0 && !b.isOnFire()) {
				BuildingCondition bc = bcc.getBuildingCondition(b.getID());
				traceBuildings.add(bc);
			}
		}

		Collections.sort(traceBuildings, new Comparator<BuildingCondition>() {
			@Override
			public int compare(BuildingCondition o1, BuildingCondition o2) {
				return o1.getTemperature() - o2.getTemperature();
			}
		});

		HashMap<TraceInfo, Double> traceSources = new HashMap<TraceInfo, Double>();
		while (!traceBuildings.isEmpty()) {
			Iterator<BuildingCondition> it = traceBuildings.iterator();
			BuildingCondition bc = it.next();
			it.remove();

			traceBuilding(bc, traceBuildings, traceSources, toClearTraceInfo);
		}

		for (Iterator<Entry<TraceInfo, Double>> it = traceSources.entrySet().iterator(); it
				.hasNext();) {
			Entry<TraceInfo, Double> entry = it.next();
			TraceInfo ti = entry.getKey();

			if (toClearTraceInfo.contains(ti)) {
				toClearTraceInfo.remove(ti);
			}
			Double dTraceValue = entry.getValue();
			ti.updateTraceInfo(dTraceValue, context.getTimestamp());
		}

		for (TraceInfo ti : toClearTraceInfo) {
			ti.clearTraceValue();
		}

		updateCurrentTraceList(traceSources);
	}

	/**
	 * 为了维护最近几次更新的traceSources，为了方便job使用
	 * 
	 * @param traceSources
	 */
	private void updateCurrentTraceList(HashMap<TraceInfo, Double> traceSources) {
		if (currentTraceListCache.size() > CACHE_SIZE) {
			currentTraceListCache.remove(0);
		}
		currentTraceListCache.add(traceSources);

		currentTraceList.clear();
		for (Map<TraceInfo, Double> m : currentTraceListCache) {
			for (TraceInfo ti : m.keySet()) {
				if (ti.getTraceValue() > MIN_TRACE_VALUE) {
					if (bcc.getCurrentUpdatedConditions().contains(ti.getBuildingCondition())) {
						log.error("bcc.getCurrentUpdatedConditions().contains(ti.getBuildingCondition())");
					} else {
						currentTraceList.add(ti);
					}
				}
			}
		}
		Collections.sort(currentTraceList, new TraceInfoComparator());
	}

	private void traceBuilding(BuildingCondition bc, ArrayList<BuildingCondition> traceBuildings,
			HashMap<TraceInfo, Double> traceSources, HashSet<TraceInfo> toClearTraceInfo) {
		ArrayList<Entry<FireBuilding, Float>> currentSource = new ArrayList<Entry<FireBuilding, Float>>();
		boolean hasFoundSource = false;
		for (Iterator<Entry<FireBuilding, Float>> it = getNearBuildings(bc).entrySet().iterator(); it
				.hasNext();) {
			Entry<FireBuilding, Float> entry = it.next();
			BuildingCondition condition = entry.getKey().getCondition();
			TraceInfo traceInfo = traceMap.get(condition.getBuilding().getID());

			if (condition.getTimestamp() != 0
					&& condition.getTimestamp() + TRACE_MAX_TIME >= context.getTimestamp()) {
				// 该building在最近有更新
				if (condition.isEverOnFire()) {
					hasFoundSource = true;
					continue;
				} else if (condition.getTemperature() > bc.getTemperature()) {
					hasFoundSource = true;
					if (traceSources.containsKey(traceInfo)) {
						// no need to trace the building again
						continue;
					} else {
						// 找到了与当前building有连接关系的building，应该trace该building
						// ?
						traceBuildings.add(condition);
						continue;
					}
				} else {
					// 这个condition不是source
					continue;
				}
			}

			// the condition hasn't been updated for a long time
			currentSource.add(entry);
		}

		if (hasFoundSource) {
			for (Entry<FireBuilding, Float> entry : currentSource) {
				BuildingCondition condition = entry.getKey().getCondition();
				TraceInfo traceInfo = traceMap.get(condition.getBuilding().getID());
				toClearTraceInfo.add(traceInfo);
			}
		} else {
			for (Entry<FireBuilding, Float> entry : currentSource) {
				Float connectValue = entry.getValue();
				BuildingCondition condition = entry.getKey().getCondition();
				TraceInfo traceInfo = traceMap.get(condition.getBuilding().getID());

				double dTraceValue = INITIAL_D_TRACE_VALUE;
				if (traceInfo.getTraceTime() != context.getTimestamp()
						&& condition.getTimestamp() == 0) {
					// 若之前没有更新过这个building的情况，则trace这个building有更大的收益
					dTraceValue += NO_VISIT_TRACE_VALUE;
				}
				dTraceValue *= connectValue;

				Double v = traceSources.get(traceInfo);
				// 更新traceValue
				if (v != null) {
					traceSources.put(traceInfo, v + dTraceValue);
				} else {
					traceSources.put(traceInfo, dTraceValue);
				}
			}
		}
	}

	private void clearTraceInfo(BuildingCondition bc, HashSet<TraceInfo> toClearTraceInfo) {
		for (FireBuilding fb : getNearBuildings(bc).keySet()) {
			TraceInfo ti = traceMap.get(fb.getBuilding().getID());
			toClearTraceInfo.add(ti);
		}
	}

	private Map<FireBuilding, Float> getNearBuildings(BuildingCondition bc) {
		if (fsc != null) {
			return fsc.getFireBuilding(bc.getBuilding().getID()).getNearFireBuildings();
		} else if (rtftc != null) {
			return rtftc.getTracedBuilding(bc.getBuilding().getID()).getNearFireBuildings();
		} else {
			// TODO
			return null;
		}
	}

	public TraceInfo getTraceInfo(EntityID id) {
		return traceMap.get(id);
	}

	public List<TraceInfo> getCurrentTraceList() {
		return currentTraceList;
	}

	public static class TraceInfo {
		private static final double OLD_TRACE_VALUE_COEFFICIENT = 0.2d;
		private int traceTime;
		private double traceValue;
		private BuildingCondition buildingCondition;

		public TraceInfo(BuildingCondition buildingCondition) {
			this.buildingCondition = buildingCondition;
			this.traceValue = 0;
			this.traceTime = -1;
		}

		public double getTraceValue() {
			return traceValue;
		}

		public int getTraceTime() {
			return traceTime;
		}

		public void updateTraceInfo(double dTraceValue, int traceTime) {
			if (this.traceTime != traceTime) {
				traceValue = traceValue * OLD_TRACE_VALUE_COEFFICIENT
						/ (traceTime - this.traceTime);
			}
			this.traceValue += dTraceValue;
			this.traceTime = traceTime;
		}

		public void clearTraceValue() {
			this.traceValue = 0;
			this.traceTime = -1;
		}

		public BuildingCondition getBuildingCondition() {
			return buildingCondition;
		}

		@Override
		public String toString() {
			return "TraceInfo [traceTime=" + traceTime + ", traceValue=" + traceValue
					+ ", buildingCondition=" + buildingCondition + "]";
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result
					+ ((buildingCondition == null) ? 0 : buildingCondition.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			TraceInfo other = (TraceInfo) obj;
			if (buildingCondition == null) {
				if (other.buildingCondition != null) {
					return false;
				}
			} else if (!buildingCondition.equals(other.buildingCondition)) {
				return false;
			}
			return true;
		}
	}

	public static class TraceInfoComparator implements Comparator<TraceInfo> {
		@Override
		public int compare(TraceInfo o1, TraceInfo o2) {
			double dTraceValue = o1.getTraceValue() - o2.getTraceValue();
			if (dTraceValue > 0) {
				return -1;
			} else if (dTraceValue < 0) {
				return 1;
			} else {
				return 0;
			}
		}
	}
}
