package ant.component.fb;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.Iterator;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.gui.ScreenTransform;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Line;
import ant.utils.geo.Point;

/**
 * refer to the wall in simulator
 * 
 * @author Oliver
 * 
 */
public class Wall {
	private static Log log = LogFactory.getLog(Wall.class);

	FireBuilding owner;
	/**
	 * will be used to calculate the rays for the owner FireBuilding
	 * 
	 * rays = (int) Math.ceil(length * RAY_RATE);
	 */
	int rays;
	double length;
	Point a;
	Point b;
	int selfHits;
	int hits;

	public static final int MAX_SAMPLE_DISTANCE = 50000;
	public static final float RAY_RATE = 0.002f;
	public static final int LENGTH_DIVIDER = 10000;

	public Wall(int x1, int y1, int x2, int y2, FireBuilding owner) {
		a = new Point(x1, y1);
		b = new Point(x2, y2);
		length = a.distance(b);
		rays = (int) Math.ceil(length * RAY_RATE);
		this.owner = owner;
	}

	public int getActualOutRay() {
		return rays / 2;
	}

	/**
	 * the wall is invalid when the two points of the edge is at the same
	 * position
	 * 
	 * @return true for valid wall, false for invalid wall
	 */
	public boolean validate() {
		return !(a.x() == b.x() && a.y() == b.y());
	}

	public String toString() {
		return "wall (" + a.x() + "," + a.y() + ")-(" + b.x() + "," + b.y() + "), length="
				+ String.format("%.0f", length) + "mm, rays=" + rays;
	}

	public void findHitsInNearFireBuildings() {
		selfHits = 0;
		int startPointCount = (int) Math.ceil(length / LENGTH_DIVIDER);
		int dx = (b.x() - a.x()) / (startPointCount + 1);
		int dy = (b.y() - a.y()) / (startPointCount + 1);
		for (int i = 0; i < startPointCount; i++) {
			// creating ray
			Point start = new Point(a.x() + dx * (i + 1), a.y() + dy * (i + 1));
			double delta = startPointCount * 2d * Math.PI / rays;
			double sin = Math.sin(delta);
			double cos = Math.cos(delta);
			double x, y, _x, _y;

			//使第一条线是向外发射的
			if (GeoUtils.cross(owner.getBuilding().getX(), owner.getBuilding().getY(), a.x(),
					a.y(), b.x(), b.y()) > 0) {
				_x = b.x() - start.x();
				_y = b.y() - start.y();
				double temp = Math.sqrt(_x * _x + _y * _y);
				x = MAX_SAMPLE_DISTANCE * _x / temp;
				y = MAX_SAMPLE_DISTANCE * _y / temp;
			} else {
				_x = a.x() - start.x();
				_y = a.y() - start.y();
				double temp = Math.sqrt(_x * _x + _y * _y);
				x = MAX_SAMPLE_DISTANCE * _x / temp;
				y = MAX_SAMPLE_DISTANCE * _y / temp;
			}
			// x = 0;
			// y = MAX_SAMPLE_DISTANCE;
			int actualOutRay = rays / (2 * startPointCount);
			for (int emitted = 0; emitted < actualOutRay; emitted++) {
				_x = x * cos - y * sin;
				_y = y * cos + x * sin;
				x = _x;
				y = _y;
				Point end = new Point((int) (x + start.x()), (int) (y + start.y()));
				// intersect
				Wall closest = null;
				double minDist = Double.MAX_VALUE;
				double tempDist = 0;

				for (Iterator<Entry<FireBuilding, Float>> iter = owner.getNearFireBuildings()
						.entrySet().iterator(); iter.hasNext();) {
					Entry<FireBuilding, Float> entry = iter.next();
					FireBuilding fireBuilding = entry.getKey();
					for (Iterator<Wall> it = fireBuilding.getWalls().iterator(); it.hasNext();) {
						Wall other = it.next();
						if (other == this)
							continue;
						tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(),
								end.x(), end.y(), other.a.x(), other.a.y(), other.b.x(),
								other.b.y());
						if (tempDist >= 0 && tempDist < minDist) {
							minDist = tempDist;
							closest = other;
						}
					}
				}
				for (Iterator<Wall> it = owner.getWalls().iterator(); it.hasNext();) {
					Wall other = it.next();
					if (other == this)
						continue;
					tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(), end.x(),
							end.y(), other.a.x(), other.a.y(), other.b.x(), other.b.y());
					if (tempDist >= 0 && tempDist < minDist) {
						minDist = tempDist;
						closest = other;
					}
				}

				if (closest == null) {
					// Nothing was hit
					continue;
				}
				if (closest.owner == this.owner) {
					// The source building was hit
					selfHits++;
				}
				if (closest != this && closest != null && closest.owner != owner) {
					hits++;
					Float value = owner.getNearFireBuildings().get(closest.owner);
					float temp = 0f;
					if (value != null && value > 0) {
						temp = value;
					}
					temp += 1f;
					owner.getNearFireBuildings().put(closest.owner, new Float(temp));
				}
			}
		}
	}
	
	public void drawRays(FireSimulatorComponent fsc, Graphics2D g, ScreenTransform t, int clickCount) {
		int rayCount = 0;
		int startPointCount = (int) Math.ceil(length / LENGTH_DIVIDER);
		int dx = (b.x() - a.x()) / (startPointCount + 1);
		int dy = (b.y() - a.y()) / (startPointCount + 1);
		for (int i = 0; i < startPointCount; i++) {
			// creating ray
			Point start = new Point(a.x() + dx * (i + 1), a.y() + dy * (i + 1));
			double delta = startPointCount * 2d * Math.PI / rays;
			Point cross = null;
			double sin = Math.sin(delta);
			double cos = Math.cos(delta);
			double x, y, _x, _y;

			if (GeoUtils.cross(owner.getBuilding().getX(), owner.getBuilding().getY(), a.x(),
					a.y(), b.x(), b.y()) > 0) {
				_x = b.x() - start.x();
				_y = b.y() - start.y();
				double temp = Math.sqrt(_x * _x + _y * _y);
				x = MAX_SAMPLE_DISTANCE * _x / temp;
				y = MAX_SAMPLE_DISTANCE * _y / temp;
			} else {
				_x = a.x() - start.x();
				_y = a.y() - start.y();
				double temp = Math.sqrt(_x * _x + _y * _y);
				x = MAX_SAMPLE_DISTANCE * _x / temp;
				y = MAX_SAMPLE_DISTANCE * _y / temp;
			}
			// x = 0;
			// y = MAX_SAMPLE_DISTANCE;
			int actualOutRay = rays / (2 * startPointCount);
			for (int emitted = 0; emitted < actualOutRay; emitted++, rayCount++) {
				_x = x * cos - y * sin;
				_y = y * cos + x * sin;
				x = _x;
				y = _y;
				Point end = new Point((int) (x + start.x()), (int) (y + start.y()));

				if (rayCount == clickCount) {
					log.info("Ray:" + start + "->" + end);
				}
				g.setColor(Color.WHITE);
				g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()), t.xToScreen(end.x()),
						t.yToScreen(end.y()));

				// intersect
				Wall closest = null;
				double minDist = Double.MAX_VALUE;
				double tempDist = 0;

				for (Iterator<Entry<FireBuilding, Float>> iter = owner.getNearFireBuildings()
						.entrySet().iterator(); iter.hasNext();) {
					Entry<FireBuilding, Float> entry = iter.next();
					FireBuilding fireBuilding = entry.getKey();
					for (Iterator<Wall> it = fireBuilding.getWalls().iterator(); it.hasNext();) {
						Wall other = it.next();
						if (other == this)
							continue;
						tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(),
								end.x(), end.y(), other.a.x(), other.a.y(), other.b.x(),
								other.b.y());
						if (tempDist >= 0 && tempDist < minDist) {
							minDist = tempDist;
							closest = other;
							cross = GeoUtils.getIntersectionPoint(new Line(start, end), new Line(
									other.a, other.b));
						}
					}
				}
				for (Iterator<Wall> it = owner.getWalls().iterator(); it.hasNext();) {
					Wall other = it.next();
					if (other == this)
						continue;
					tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(), end.x(),
							end.y(), other.a.x(), other.a.y(), other.b.x(), other.b.y());
					if (tempDist >= 0 && tempDist < minDist) {
						minDist = tempDist;
						closest = other;
						cross = GeoUtils.getIntersectionPoint(new Line(start, end), new Line(
								other.a, other.b));
					}
				}

				if (closest == null) {
					// Nothing was hit
					if (rayCount == clickCount) {
						g.setColor(Color.BLACK);
					} else {
						g.setColor(Color.CYAN);
					}
					g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()),
							t.xToScreen(end.x()), t.yToScreen(end.y()));
					continue;
				}
				if (closest.owner == this.owner) {
					// The source building was hit
					if (rayCount == clickCount) {
						g.setColor(Color.BLACK);
					} else {
						g.setColor(Color.RED);
					}
					g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()),
							t.xToScreen(cross.x()), t.yToScreen(cross.y()));
				}
				if (closest != this && closest != null && closest.owner != owner) {
					if (rayCount == clickCount) {
						g.setColor(Color.BLACK);
					} else {
						g.setColor(Color.YELLOW);
					}
					g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()),
							t.xToScreen(cross.x()), t.yToScreen(cross.y()));
				}
			}
		}
	}

	public void _findHitsInNearFireBuildings() {
		selfHits = 0;
		// creating ray
		Point start = new Point((a.x() + b.x()) / 2, (a.y() + b.y()) / 2);
		double delta = 2d * Math.PI / rays;
		double sin = Math.sin(delta);
		double cos = Math.cos(delta);
		double x, y, _x, _y;

		if (GeoUtils.cross(owner.getBuilding().getX(), owner.getBuilding().getY(), a.x(), a.y(),
				b.x(), b.y()) > 0) {
			_x = b.x() - start.x();
			_y = b.y() - start.y();
			double temp = Math.sqrt(_x * _x + _y * _y);
			x = MAX_SAMPLE_DISTANCE * _x / temp;
			y = MAX_SAMPLE_DISTANCE * _y / temp;
		} else {
			_x = a.x() - start.x();
			_y = a.y() - start.y();
			double temp = Math.sqrt(_x * _x + _y * _y);
			x = MAX_SAMPLE_DISTANCE * _x / temp;
			y = MAX_SAMPLE_DISTANCE * _y / temp;
		}
		// x = 0;
		// y = MAX_SAMPLE_DISTANCE;
		int actualOutRay = rays / 2;
		for (int emitted = 0; emitted < actualOutRay; emitted++) {
			_x = x * cos - y * sin;
			_y = y * cos + x * sin;
			x = _x;
			y = _y;
			Point end = new Point((int) (x + start.x()), (int) (y + start.y()));
			// intersect
			Wall closest = null;
			double minDist = Double.MAX_VALUE;
			double tempDist = 0;

			for (Iterator<Entry<FireBuilding, Float>> iter = owner.getNearFireBuildings()
					.entrySet().iterator(); iter.hasNext();) {
				Entry<FireBuilding, Float> entry = iter.next();
				FireBuilding fireBuilding = entry.getKey();
				for (Iterator<Wall> it = fireBuilding.getWalls().iterator(); it.hasNext();) {
					Wall other = it.next();
					if (other == this)
						continue;
					tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(), end.x(),
							end.y(), other.a.x(), other.a.y(), other.b.x(), other.b.y());
					if (tempDist >= 0 && tempDist < minDist) {
						minDist = tempDist;
						closest = other;
					}
				}
			}
			for (Iterator<Wall> it = owner.getWalls().iterator(); it.hasNext();) {
				Wall other = it.next();
				if (other == this)
					continue;
				tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(), end.x(),
						end.y(), other.a.x(), other.a.y(), other.b.x(), other.b.y());
				if (tempDist >= 0 && tempDist < minDist) {
					minDist = tempDist;
					closest = other;
				}
			}

			if (closest == null) {
				// Nothing was hit
				continue;
			}
			if (closest.owner == this.owner) {
				// The source building was hit
				selfHits++;
			}
			if (closest != this && closest != null && closest.owner != owner) {
				hits++;
				Float value = owner.getNearFireBuildings().get(closest.owner);
				float temp = 0f;
				if (value != null && value > 0) {
					temp = value;
				}
				temp += 1f;
				owner.getNearFireBuildings().put(closest.owner, new Float(temp));
			}
		}
	}

	public void _drawRays(FireSimulatorComponent fsc, Graphics2D g, ScreenTransform t,
			int clickCount) {
		int rayCount = 0;
		// creating ray
		Point start = new Point((a.x() + b.x()) / 2, (a.y() + b.y()) / 2);
		Point cross = null;
		double delta = 2d * Math.PI / rays;
		double sin = Math.sin(delta);
		double cos = Math.cos(delta);
		double x, y, _x, _y;

		if (GeoUtils.cross(owner.getBuilding().getX(), owner.getBuilding().getY(), a.x(), a.y(),
				b.x(), b.y()) > 0) {
			_x = b.x() - start.x();
			_y = b.y() - start.y();
			double temp = Math.sqrt(_x * _x + _y * _y);
			x = MAX_SAMPLE_DISTANCE * _x / temp;
			y = MAX_SAMPLE_DISTANCE * _y / temp;
		} else {
			_x = a.x() - start.x();
			_y = a.y() - start.y();
			double temp = Math.sqrt(_x * _x + _y * _y);
			x = MAX_SAMPLE_DISTANCE * _x / temp;
			y = MAX_SAMPLE_DISTANCE * _y / temp;
		}
		// x = 0;
		// y = MAX_SAMPLE_DISTANCE;
		int actualOutRay = rays / 2;
		for (int emitted = 0; emitted < actualOutRay; emitted++, rayCount++) {
			_x = x * cos - y * sin;
			_y = y * cos + x * sin;
			x = _x;
			y = _y;
			Point end = new Point((int) (x + start.x()), (int) (y + start.y()));

			if (rayCount == clickCount) {
				log.info("Ray:" + start + "->" + end);
			}
			g.setColor(Color.WHITE);
			g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()), t.xToScreen(end.x()),
					t.yToScreen(end.y()));

			// intersect
			Wall closest = null;
			double minDist = Double.MAX_VALUE;
			double tempDist = 0;

			for (Iterator<Entry<FireBuilding, Float>> iter = owner.getNearFireBuildings()
					.entrySet().iterator(); iter.hasNext();) {
				Entry<FireBuilding, Float> entry = iter.next();
				FireBuilding fireBuilding = entry.getKey();
				for (Iterator<Wall> it = fireBuilding.getWalls().iterator(); it.hasNext();) {
					Wall other = it.next();
					if (other == this)
						continue;
					tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(), end.x(),
							end.y(), other.a.x(), other.a.y(), other.b.x(), other.b.y());
					if (tempDist >= 0 && tempDist < minDist) {
						minDist = tempDist;
						closest = other;
						cross = GeoUtils.getIntersectionPoint(new Line(start, end), new Line(
								other.a, other.b));
					}
				}
			}
			for (Iterator<Wall> it = owner.getWalls().iterator(); it.hasNext();) {
				Wall other = it.next();
				if (other == this)
					continue;
				tempDist = GeoUtils.getP1toIntersectDistSquare(start.x(), start.y(), end.x(),
						end.y(), other.a.x(), other.a.y(), other.b.x(), other.b.y());
				if (tempDist >= 0 && tempDist < minDist) {
					minDist = tempDist;
					closest = other;
					cross = GeoUtils.getIntersectionPoint(new Line(start, end), new Line(other.a,
							other.b));
				}
			}

			if (closest == null) {
				// Nothing was hit
				if (rayCount == clickCount) {
					g.setColor(Color.BLACK);
				} else {
					g.setColor(Color.CYAN);
				}
				g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()), t.xToScreen(end.x()),
						t.yToScreen(end.y()));
				continue;
			}
			if (closest.owner == this.owner) {
				// The source building was hit
				if (rayCount == clickCount) {
					g.setColor(Color.BLACK);
				} else {
					g.setColor(Color.RED);
				}
				g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()), t.xToScreen(cross.x()),
						t.yToScreen(cross.y()));
			}
			if (closest != this && closest != null && closest.owner != owner) {
				if (rayCount == clickCount) {
					g.setColor(Color.BLACK);
				} else {
					g.setColor(Color.YELLOW);
				}
				g.drawLine(t.xToScreen(start.x()), t.yToScreen(start.y()), t.xToScreen(cross.x()),
						t.yToScreen(cross.y()));
			}
		}
	}
}
