package ant.component.pf;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.component.universal.AgentLocationComponent;
import ant.component.universal.BeliefClearComponent;
import ant.component.universal.BeliefClearListener;
import ant.map.component.AntMapBasisComponent;
import ant.map.component.StaticCityGraphComponent;
import ant.map.objects.LabeledArea;
import ant.map.objects.LabeledGraph;
import ant.map.objects.Route;
import ant.map.objects.StandardGraphEdge;
import ant.map.search.StandardLabeledSPFASearch;
import ant.utils.RandomSelector;
import ant.utils.geo.GeoUtils;

/**
 * repulsive-vector algorithm
 * 
 * @author Ivan
 * 
 * @depend {@link StaticCityGraphComponent} {@link AntMapBasisComponent}
 *         {@link BeliefClearComponent} {@link AgentLocationComponent}
 */
public class ObjectiveProbeSpriteComponent implements AntComponent {
	private static final int IGNORE_TIMESTAMP = 3;
	private static final int TARGET_CLEAR_PENALTY = 4;
	
	private AgentContext context;
	private StandardWorldModel world;
	
	private ObjectiveRaceComponent opc;
	private StaticCityGraphComponent scg;
	private BeliefClearComponent bcc;
	private ObjectiveJigsawMetaComponent ojmc;

	private LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>> graph;
	private StandardLabeledSPFASearch algorithm;
	private Set<EntityID> unclearRoadSet;
	private Random rand;

	@Override
	public String getComponentName() {
		return ObjectiveProbeSpriteComponent.class.getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.world = world;
		this.opc = (ObjectiveRaceComponent) context.getAntComponentKit().getAntComponent(ObjectiveRaceComponent.class);
		this.ojmc = (ObjectiveJigsawMetaComponent) context.getAntComponentKit().getAntComponent(ObjectiveJigsawMetaComponent.class);
				
		this.bcc = context.getAntComponentKit().getBeliefClearComponent();
		this.scg = this.context.getAntComponentKit().getStaticCityGraphComponent();
		this.graph = this.scg.createGraph();
		this.algorithm = new StandardLabeledSPFASearch(this.graph);
		this.bcc.addBeleifClearListener(new BeliefClearListener() {
			@Override
			public void beliefClear(Road r) {
				graphvisit(r);
			}
		});
		unclearRoadSet = new HashSet<EntityID>();
		for (Road r : context.getAntComponentKit().getAntMapBasisComponent().getRoadSet()) {
			unclearRoadSet.add(r.getID());
		}
		this.rand = new Random(context.getAgentID().getValue() + 1);
	}

	private void graphvisit(Road se) {
		if (unclearRoadSet.contains(se.getID())) {
			unclearRoadSet.remove(se.getID());
			LabeledArea to = this.scg.getLabeledAreaMap().get(se.getID());
			for (StandardGraphEdge<LabeledArea> edge : this.graph.edgesOf(to)) {
				edge.setWeight(edge.getWeight() * TARGET_CLEAR_PENALTY);
			}
		}
	}

	@Override
	public void update() {
	}

	public List<EntityID> getMidMoveRoute(EntityID from, EntityID to) {
		LabeledArea labelFrom = this.scg.getLabeledAreaMap().get(from);
		LabeledArea labelTo = this.scg.getLabeledAreaMap().get(to);
		Route<LabeledArea> data = this.algorithm.spSearch(labelFrom, labelTo);
		if (data == null) {
			return null;
		} else {
			List<EntityID> r = new ArrayList<EntityID>();
			if (data.getRoute().size() > 1) {
				for (int i = 1; i < data.getRoute().size(); i++) {
					r.add(data.getRoute().get(i).getID());
				}
			} else {
				for (int i = 0; i < data.getRoute().size(); i++) {
					r.add(data.getRoute().get(i).getID());
				}
			}
			return r;
		}
	}

	public ObjectiveJigsaw getTask() {
		int x = context.getHuman().getX();
		int y = context.getHuman().getY();
		
		ObjectiveJigsaw pick = null;
		double mindist = Double.MAX_VALUE;
		for (ObjectiveJigsaw oj : ojmc.getSpotedJigsaws()) {
			double temp = GeoUtils.manhattanDistance(x, y, oj.getX(), oj.getY());
			if (temp < mindist) {
				mindist = temp;
				pick = oj;
			}
		}
		
		if (pick != null) {
			return pick;
		}
		
		HashSet<ObjectiveJigsaw> alist = new HashSet<ObjectiveJigsaw>(ojmc.getObjectiveJigsaws());
		return RandomSelector.randomSelect(alist, 1, rand).iterator().next();
	}
	
//	/**
//	 * near -> far
//	 */
//	private EntityID multiscaleSelect(Set<EntityID> unvisitedRoad) {
//		int x = this.context.getHuman().getX();
//		int y = this.context.getHuman().getY();
//		int range = 200000;
//		Set<Area> temp = new HashSet<Area>();
//		for (EntityID id : unvisitedRoad) {
//			temp.add((Area) world.getEntity(id));
//		}
//		Set<Area> candidate = new HashSet<Area>();
//		while (true) {
//			candidate.clear();
//			range *= 2;
//			for (Area a : temp) {
//				if (GeoUtils.manhattanDistance(x, y, a.getX(), a.getY()) < range) {
//					candidate.add(a);
//				}
//			}
//			if (!candidate.isEmpty()) {
//				Area pick = RandomSelector.randomSelect(candidate, 1, rand).iterator().next();
//				return pick.getID();
//			}
//		}
//	}

	
}
