package ant.component.pf;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.component.universal.BeliefClearComponent;
import ant.component.universal.BeliefClearListener;
import ant.map.component.AntMapBasisComponent;
import ant.map.component.StaticCityGraphComponent;
import ant.map.objects.LabeledArea;
import ant.map.objects.LabeledGraph;
import ant.map.objects.Route;
import ant.map.objects.StandardGraphEdge;
import ant.map.partition.Partition;
import ant.map.search.StandardLabeledSPFASearch;
import ant.utils.geo.GeoUtils;

/**
 * 
 * @author Ivan
 *
 * @depend {@link StaticCityGraphComponent} {@link AntMapBasisComponent} {@link BeliefClearComponent}
 */
public class ObjectiveProbeWorkerComponent implements AntComponent {
	private static final double OUT_OF_PARTITION_PENALTY = 2;
	private static final double TARGET_CLEAR_PENALTY = 3;
	
	private AgentContext context;
	private StandardWorldModel world;
	private ObjectiveRaceComponent opc;
	private Partition part;

	private StaticCityGraphComponent scg;
	private LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>> graph;
	private StandardLabeledSPFASearch algorithm;
	private BeliefClearComponent bcc;
	private Partition mypart;
	private Set<EntityID> unclearRoadSet;
	private ObjectiveJigsawMetaComponent ojmc;
	
	@Override
	public String getComponentName() {
		return ObjectiveProbeWorkerComponent.class.getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.world = world;
		this.ojmc = (ObjectiveJigsawMetaComponent) context.getAntComponentKit().getAntComponent(ObjectiveJigsawMetaComponent.class);
		this.opc = (ObjectiveRaceComponent) context.getAntComponentKit().getAntComponent(ObjectiveRaceComponent.class);
		this.bcc = context.getAntComponentKit().getBeliefClearComponent();
		this.scg = this.context.getAntComponentKit().getStaticCityGraphComponent();
		
		this.part = opc.getMyPartition();
		this.graph = this.scg.createGraph();
		this.algorithm = new StandardLabeledSPFASearch(this.graph);
		
		this.bcc.addBeleifClearListener(new BeliefClearListener() {
			@Override
			public void beliefClear(Road r) {
				graphvisit(r);
			}
		});
		
		mypart = opc.getMyPartition();
		
		for (LabeledArea v : this.graph.getVertexSet()) {
			if (!mypart.getAreaSet().contains(v.getArea())) {
				for (StandardGraphEdge<LabeledArea> edge : graph.edgesOf(v)) {
					edge.setWeight(edge.getWeight() * OUT_OF_PARTITION_PENALTY);
				}
			}
		}
		
		unclearRoadSet = new HashSet<EntityID>();
		for (EntityID id : this.mypart.getAreaIDSet()) {
			if (world.getEntity(id) instanceof Road)
				unclearRoadSet.add(id);
		}
	}
	
	private void graphvisit(Road se) {
		if (unclearRoadSet.contains(se.getID())) {
			unclearRoadSet.remove(se.getID());
			LabeledArea to = this.scg.getLabeledAreaMap().get(se.getID());
			for (StandardGraphEdge<LabeledArea> edge : this.graph.edgesOf(to)) {
				edge.setWeight(edge.getWeight() * TARGET_CLEAR_PENALTY);
			}
		}
	}
	
	@Override
	public void update() {

	}
	
	public boolean isInMyPartition() {
		ObjectiveJigsaw oj = ojmc.getExMappingObjectiveJigsaw(context.getHuman().getPosition());
		return (oj.belongPartId(mypart.getID()));
		// return part.getAreaIDSet().contains(context.getHuman().getPosition());
	}
	
	/**
	 * @return null, if in my partition <br>
	 */
	public EntityID getNearestRoadToPartition() {
		if (isInMyPartition())
			return null;
		EntityID target = null;
		double dist = Double.MAX_VALUE;
		double tmp;
		Area here = (Area) context.getHuman().getPosition(world);
		for (Area a : part.getAreaSet()) {
			if (a instanceof Road) {
				tmp = GeoUtils.manhattanDistance(here, a);
				if (tmp < dist) {
					dist = tmp;
					target = a.getID();
				}
			}
		}
		return target;
	}
	
	public List<EntityID> getMidMoveRoute(EntityID from, EntityID to) {
		LabeledArea labelFrom = this.scg.getLabeledAreaMap().get(from);
		LabeledArea labelTo = this.scg.getLabeledAreaMap().get(to);
		Route<LabeledArea> data = this.algorithm.spSearch(labelFrom, labelTo);
		if (data == null) {
			return null;
		} else {
			List<EntityID> r = new ArrayList<EntityID>();
			if (data.getRoute().size() > 1) {
				for (int i = 1; i < data.getRoute().size(); i++) {
					r.add(data.getRoute().get(i).getID());
				}
			} else {
				for (int i = 0; i < data.getRoute().size(); i++) {
					r.add(data.getRoute().get(i).getID());
				}
			}
			return r;
		}
	}
	
	public boolean hasTask() {
		return (!unclearRoadSet.isEmpty());
	}
	
	/**
	 * 最远一个的UnClearRoad
	 */
	public EntityID getTask() {
		if (unclearRoadSet.isEmpty())
			return null;

		int x = context.getHuman().getX();
		int y = context.getHuman().getY();
		double dist;
		double max = 0;
		EntityID target = null;
		for (EntityID id : unclearRoadSet) {
			dist = GeoUtils.manhattanDistance(((Road) world.getEntity(id)), x, y);
			if (dist >= max) {
				max = dist;
				target = id;
			}
		}
		return target;
	}
	
}
