package ant.component.universal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AbstractAntCommunicationKit;
import ant.AbstractMessageProducerComponent;
import ant.AgentContext;
import ant.liaison.message.IAgentLocationMessage;
import ant.liaison.message.IMessage;
import ant.liaison.message.MAgentRoadCorLocation;
import ant.liaison.message.MStuckInRoad;
import ant.liaison.message.MVoiceAgentStuckInRoad;
import ant.map.component.AntMapBasisComponent;

/**
 * 
 * @author Oliver
 * @depend {@link ChangeSetDetailComponent} {@link AntMapBasisComponent}
 */
public class AgentLocationComponent extends AbstractMessageProducerComponent {
	private static final int VOICE_LOCATION_MSG_MAX_SIZE = 5;

	private static Log log = LogFactory.getLog(AgentLocationComponent.class);

	private AgentContext context;
	private AbstractAntCommunicationKit ack;
	private ChangeSetDetailComponent csd;
	private Set<Human> agents;

	private HashMap<EntityID, LinkedList<AgentLocationInfo>> agentLocations;
	private Set<AgentLocationInfo> updatedAgentLocationInfoSet;
	private HashMap<EntityID, LinkedList<AgentLocationInfo>> updatedAgentLocations;

	private LocationMsgComparator cmp;

	private HashSet<AgentLocationInfo> updatedAgentCurrentLocationInfoSet;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.ack = context.getAntCommunicationKit();
		this.csd = context.getAntComponentKit().getChangeSetDetailComponent();
		this.agents = context.getAntComponentKit().getAntMapBasisComponent().getAgentSet();

		this.cmp = new LocationMsgComparator(context);

		this.agentLocations = new HashMap<EntityID, LinkedList<AgentLocationInfo>>();
		this.updatedAgentLocations = new HashMap<EntityID, LinkedList<AgentLocationInfo>>();
		for (Human agent : agents) {
			LinkedList<AgentLocationInfo> list = new LinkedList<AgentLocationInfo>();
			list.add(new AgentLocationInfo(agent, 0));
			agentLocations.put(agent.getID(), list);

			LinkedList<AgentLocationInfo> updateList = new LinkedList<AgentLocationInfo>();
			updatedAgentLocations.put(agent.getID(), updateList);
		}
		this.updatedAgentLocationInfoSet = new HashSet<AgentLocationInfo>();
		this.updatedAgentCurrentLocationInfoSet = new HashSet<AgentLocationInfo>();
	}

	@Override
	public void update() {
		updatedAgentLocationInfoSet.clear();
		updatedAgentCurrentLocationInfoSet.clear();
		for (List<AgentLocationInfo> list : updatedAgentLocations.values()) {
			list.clear();
		}

		updateFromMessage();
		updateFromChangeSet();

		sendLocationMessages();
	}

	private void sendLocationMessages() {
		if (context.getAntCommunicationKit().isNoRadio()) {
			List<MAgentRoadCorLocation> list = new ArrayList<MAgentRoadCorLocation>();
			for (Human h : agents) {
				AgentLocationInfo info = getAgentLastLocationInfo(h.getID());
				if (info != null && info.isCoordinateDefined() && info.isFireSeenAroundMsgDefined) {
					MAgentRoadCorLocation msg = info.getMsg();
					if (!msg.isFireSeenAround()) {
						list.add(msg);
					}
				}
			}
			if (list.size() >= VOICE_LOCATION_MSG_MAX_SIZE) {
				Collections.sort(list, cmp);
				list = list.subList(0, VOICE_LOCATION_MSG_MAX_SIZE);
			}
			for (MAgentRoadCorLocation msg : list) {
				addVoiceMessage(msg);
			}
		}
	}

	private void updateFromMessage() {
		for (IMessage msg : ack.getRadioReceiveBox()) {
			if (msg instanceof IAgentLocationMessage) {
				AgentLocationInfo newInfo = new AgentLocationInfo((IAgentLocationMessage) msg);
				updateInfo(msg.getSender(), newInfo);
			}
		}
		for (IMessage msg : ack.getVoiceReceiveBox()) {
			if (msg instanceof IAgentLocationMessage) {
				AgentLocationInfo newInfo = new AgentLocationInfo((IAgentLocationMessage) msg);
				updateInfo(msg.getSender(), newInfo);
			}
		}
	}

	private void updateFromChangeSet() {
		for (Human h : csd.getAgentsInSight()) {
			AgentLocationInfo newInfo = new AgentLocationInfo(h, context.getTimestamp());
			updateInfo(h.getID(), newInfo);
		}
	}

	private void updateInfo(EntityID agentID, AgentLocationInfo newInfo) {
		LinkedList<AgentLocationInfo> infos = agentLocations.get(agentID);
		LinkedList<AgentLocationInfo> updatedInfos = updatedAgentLocations.get(agentID);
		if (infos == null) {
			log.info("agentID:" + agentID);
		}
		if (infos.size() != 0) {
			for (ListIterator<AgentLocationInfo> it = infos.listIterator(infos.size()); it
					.hasPrevious();) {
				AgentLocationInfo info = it.previous();
				int dt = newInfo.getTimestamp() - info.getTimestamp();
				if (dt > 0) {
					// newInfo is newer
					it.next();
					it.add(newInfo);
					updatedAgentLocationInfoSet.add(newInfo);
					updatedAgentCurrentLocationInfoSet.add(newInfo);
					addNewInfoToUpdatedLocationInfos(updatedInfos, newInfo);
					break;
				} else if (dt == 0) {
					// newInfo is the same
					if (info.shouldReplace(newInfo)) {
						it.remove();
						it.add(newInfo);
						it.previous();
						// FIXME should this added into
						// updatedAgentLocationInfos?
					}
					break;
				} else if (!it.hasPrevious()) {
					// newInfo is newer
					it.add(newInfo);
					updatedAgentLocationInfoSet.add(newInfo);
					addNewInfoToUpdatedLocationInfos(updatedInfos, newInfo);
					break;
				}
			}
		} else {
			infos.add(newInfo);
			updatedAgentLocationInfoSet.add(newInfo);
			addNewInfoToUpdatedLocationInfos(updatedInfos, newInfo);
		}
	}

	private void addNewInfoToUpdatedLocationInfos(LinkedList<AgentLocationInfo> updatedInfos,
			AgentLocationInfo newInfo) {
		if (updatedInfos.size() != 0) {
			for (ListIterator<AgentLocationInfo> it = updatedInfos
					.listIterator(updatedInfos.size()); it.hasPrevious();) {
				AgentLocationInfo info = it.previous();
				int dt = newInfo.getTimestamp() - info.getTimestamp();
				if (dt > 0) {
					// newInfo is newer
					it.next();
					it.add(newInfo);
					updatedInfos.add(newInfo);
					break;
				} else if (dt == 0) {
					// newInfo is the same
					log.error("dt == 0!");
					break;
				} else if (!it.hasPrevious()) {
					// newInfo is newer
					it.add(newInfo);
					updatedInfos.add(newInfo);
					break;
				}
			}
		} else {
			updatedInfos.add(newInfo);
		}
	}

	/**
	 * 
	 * @param agentID
	 * @return List of AgntLocationInfo in time order
	 */
	public List<AgentLocationInfo> getAgentLocationInfos(EntityID agentID) {
		return agentLocations.get(agentID);
	}

	/**
	 * 
	 * @param agentID
	 * @return List of updatedAgntLocationInfo in time order
	 */
	public List<AgentLocationInfo> getUpdatedAgentLocationInfos(EntityID agentID) {
		return updatedAgentLocations.get(agentID);
	}

	public AgentLocationInfo getAgentLastLocationInfo(EntityID agentID) {
		LinkedList<AgentLocationInfo> infos = agentLocations.get(agentID);
		return infos.getLast();
	}

	public EntityID getAgentLastLocation(EntityID agentID) {
		return getAgentLastLocationInfo(agentID).getLocation();
	}

	public EntityID getAgentInitialLocation(EntityID agentID) {
		LinkedList<AgentLocationInfo> infos = agentLocations.get(agentID);
		return infos.get(0).getLocation();
	}

	/**
	 * @return Map<AgentID, AreaID>
	 */
	public Map<EntityID, EntityID> getAgentsLastLocation() {
		Map<EntityID, EntityID> ret = new HashMap<EntityID, EntityID>();
		for (Human agent : agents) {
			ret.put(agent.getID(), getAgentLastLocationInfo(agent.getID()).getLocation());
		}
		return ret;
	}

	/**
	 * 有可能返回值中包含插入的agent以前位置的信息
	 * 
	 * @return
	 */
	public Set<AgentLocationInfo> getUpdatedAgentLocationInfoSet() {
		return updatedAgentLocationInfoSet;
	}

	/**
	 * 只返回更新的信息
	 * 
	 * @return
	 */
	public Set<AgentLocationInfo> getUpdatedAgentCurrentLocationInfoSet() {
		return updatedAgentCurrentLocationInfoSet;
	}

	/**
	 * 从最近一个记录往前遍历
	 * 
	 * @param agentID
	 * @param timestamp
	 * @return agentLocationInfo according to the agentID and timestamp
	 */
	public AgentLocationInfo getAgentLocationInfo(EntityID agentID, int timestamp) {
		LinkedList<AgentLocationInfo> infos = agentLocations.get(agentID);
		if (infos != null) {
			for (int i = infos.size() - 1; i >= 0; i--) {
				AgentLocationInfo info = infos.get(i);
				if (info.getTimestamp() == timestamp) {
					return info;
				}
			}
		}
		return null;
	}

	public EntityID getAgentLocation(EntityID agentID, int timestamp) {
		AgentLocationInfo info = getAgentLocationInfo(agentID, timestamp);
		if (info != null) {
			return info.getLocation();
		}
		return null;
	}

	public double manhattanDistance(AgentLocationInfo o1, AgentLocationInfo o2) {
		int x1, x2, y1, y2;
		if (o1.isCoordinateDefined()) {
			x1 = o1.getX();
			y1 = o1.getY();
		} else {
			Area location = (Area) context.getWorld().getEntity(o1.getLocation());
			x1 = location.getX();
			y1 = location.getY();
		}
		if (o2.isCoordinateDefined()) {
			x2 = o2.getX();
			y2 = o2.getY();
		} else {
			Area location = (Area) context.getWorld().getEntity(o2.getLocation());
			x2 = location.getX();
			y2 = location.getY();
		}
		return Math.abs(x1 - x2) + Math.abs(y1 - y2);
	}

	private static class LocationMsgComparator implements Comparator<MAgentRoadCorLocation> {

		private static final int TIME_COEFFICIENT = 20;
		private static final double DIS_COEFFICIENT = 50000d;
		private AgentContext context;
		private int lastUseTime;
		private Area here;
		private StandardWorldModel world;

		public LocationMsgComparator(AgentContext context) {
			this.context = context;
			this.world = context.getWorld();
			this.lastUseTime = -1;
		}

		@Override
		public int compare(MAgentRoadCorLocation o1, MAgentRoadCorLocation o2) {
			return getValue(o1) - getValue(o2);
		}

		private int getValue(MAgentRoadCorLocation msg) {
			int rv = 0;
			int dt = context.getTimestamp() - msg.getTimestamp();
			rv += dt * TIME_COEFFICIENT;

			double dis = getDistance(msg.getRoadCoord().getX(), msg.getRoadCoord().getY());
			rv += dis / DIS_COEFFICIENT;
			return rv;
		}

		private double getDistance(int x, int y) {
			return Math.hypot(x - getAgentCurrentPosition().getX(), y
					- getAgentCurrentPosition().getY());
		}

		private Area getAgentCurrentPosition() {
			if (this.lastUseTime != context.getTimestamp()) {
				this.here = (Area) context.getHuman().getPosition(world);
			}
			return here;
		}
	}

	public static class AgentLocationInfo {
		private static final int INVALID_VALUE = -1;
		private int timestamp;
		private EntityID location;
		private UpdateSource updateSource;
		private boolean isCoordinateDefined;

		private int x;
		private int y;
		private boolean isFireSeenAroundMsgDefined;
		private MAgentRoadCorLocation msg;
		private EntityID agentID;

		public AgentLocationInfo(IAgentLocationMessage msg) {
			this.agentID = msg.getSender();
			this.location = msg.getLocation();
			this.timestamp = msg.getTimestamp();
			this.updateSource = UpdateSource.Message;

			if (msg instanceof MAgentRoadCorLocation) {
				MAgentRoadCorLocation arc = (MAgentRoadCorLocation) msg;
				this.x = arc.getX();
				this.y = arc.getY();
				this.isCoordinateDefined = true;
				this.isFireSeenAroundMsgDefined = true;
				this.msg = arc;
			} else if (msg instanceof MStuckInRoad) {
				MStuckInRoad sr = (MStuckInRoad) msg;
				this.x = sr.getRoadCoord().getX();
				this.y = sr.getRoadCoord().getY();
				this.isCoordinateDefined = true;
			} else if (msg instanceof MVoiceAgentStuckInRoad) {
				MVoiceAgentStuckInRoad vsr = (MVoiceAgentStuckInRoad) msg;
				this.x = vsr.getRoadCoord().getX();
				this.y = vsr.getRoadCoord().getY();
				this.isCoordinateDefined = true;
			} else {
				this.x = INVALID_VALUE;
				this.y = INVALID_VALUE;
				this.isCoordinateDefined = false;
			}
		}

		public AgentLocationInfo(Human agent, int time) {
			this.agentID = agent.getID();
			this.location = agent.getPosition();
			this.x = agent.getX();
			this.y = agent.getY();
			this.timestamp = time;
			this.updateSource = UpdateSource.ChangeSet;
			this.isCoordinateDefined = true;
		}

		public EntityID getLocation() {
			return location;
		}

		public boolean isFireSeenAroundMsgDefined() {
			return isFireSeenAroundMsgDefined;
		}

		public MAgentRoadCorLocation getMsg() {
			if (isFireSeenAroundMsgDefined) {
				return msg;
			} else {
				throw new IllegalAccessError();
			}
		}

		public boolean isCoordinateDefined() {
			return isCoordinateDefined;
		}

		/**
		 * @return Coordinate x, if isCoordinateDefined == true;
		 * @throws IllegalAccessError
		 */
		public int getX() {
			if (isCoordinateDefined) {
				return x;
			} else {
				throw new IllegalAccessError();
			}
		}

		/**
		 * @return Coordinate y, if isCoordinateDefined == true;
		 * @throws IllegalAccessError
		 */
		public int getY() {
			if (isCoordinateDefined) {
				return y;
			} else {
				throw new IllegalAccessError();
			}
		}

		public UpdateSource getUpdateSource() {
			return updateSource;
		}

		public int getTimestamp() {
			return timestamp;
		}

		public EntityID getAgentID() {
			return agentID;
		}

		/**
		 * 
		 * @param otherInfo
		 * @return this.timestamp < other.timestamp
		 */
		public boolean shouldReplace(AgentLocationInfo other) {
			if (this.timestamp != other.timestamp) {
				log.error("this.timestamp != other.timestamp!");
				return false;
			}
			if (this.getUpdateSource() == UpdateSource.ChangeSet) {
				if (other.getUpdateSource() == UpdateSource.ChangeSet) {
					log.error("this.getUpdateSource() == UpdateSource.ChangeSet,other.getUpdateSource() == UpdateSource.ChangeSet!");
				}
				return false;
			} else if (this.getUpdateSource() == UpdateSource.Message) {
				if (other.getUpdateSource() == UpdateSource.ChangeSet) {
					return true;
				} else {
					if (this.isCoordinateDefined) {
						return false;
					} else {
						return other.isCoordinateDefined;
					}
				}
			}
			return false;
		}

		@Override
		public String toString() {
			if (updateSource == UpdateSource.ChangeSet)
				return "C:" + location + "@" + timestamp;
			else
				return "M:" + location + "@" + timestamp;
		}
	}
}
