package ant.component.universal;

import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityConstants;
import rescuecore2.standard.entities.StandardEntityConstants.Fieryness;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.liaison.message.IBuildingFire;
import ant.liaison.message.IMessage;
import ant.liaison.message.MAgentRoadCorLocation;
import ant.liaison.message.MBuildingBurning;
import ant.map.component.AntMapBasisComponent;

/**
 * 
 * @author Oliver
 * @depend {@link AntMapBasisComponent}
 * @depend {@link ChangeSetDetailComponent}
 * @depend {@link LineOfSightPerceptionComponent}
 */
public class BuildingConditionComponent implements AntComponent {
	private static Log log = LogFactory.getLog(BuildingConditionComponent.class);

	private AgentContext context;
	private ChangeSetDetailComponent csd;
	private LineOfSightPerceptionComponent los;
	private int ignoreTime;

	private Map<EntityID, BuildingCondition> buildings;
	private Set<BuildingCondition> currentUpdatedConditions;

	private boolean needPrintLog;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.csd = context.getAntComponentKit().getChangeSetDetailComponent();
		this.los = (LineOfSightPerceptionComponent) context.getAntComponentKit().getAntComponent(
				LineOfSightPerceptionComponent.class);
		this.ignoreTime = ConfigKey.getIgnoreTime(context.getConfig());

		this.buildings = new HashMap<EntityID, BuildingConditionComponent.BuildingCondition>();
		for (Building b : context.getAntComponentKit().getAntMapBasisComponent().getBuildingSet()) {
			buildings.put(b.getID(), new BuildingCondition(b));
		}
		this.currentUpdatedConditions = new HashSet<BuildingCondition>();
	}

	@Override
	public void update() {
		long sTime = System.currentTimeMillis();
		currentUpdatedConditions.clear();

		// update from message
		updateFromMessage(context.getAntCommunicationKit().getRadioReceiveBox());
		updateFromMessage(context.getAntCommunicationKit().getVoiceReceiveBox());

		// update from changeSet
		for (StandardEntity entity : csd.getEntitiesInSight(StandardEntityURN.BUILDING,
				StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.FIRE_STATION,
				StandardEntityURN.POLICE_OFFICE, StandardEntityURN.REFUGE)) {
			Building b = (Building) entity;
			BuildingCondition condition = buildings.get(b.getID());
			if (condition == null) {
				log.error("Can't get condition for " + b);
				condition = new BuildingCondition(b);
				buildings.put(b.getID(), condition);
			}
			if (condition.update(b, context.getTimestamp())) {
				currentUpdatedConditions.add(condition);
			}
		}
		long dTime = System.currentTimeMillis() - sTime;
		if (dTime > 200) {
			this.needPrintLog = true;
		} else {
			this.needPrintLog = false;
		}
	}

	public boolean needPrintLog() {
		return needPrintLog;
	}

	private void updateFromMessage(Set<IMessage> messages) {
		for (IMessage msg : messages) {
			if (msg instanceof MBuildingBurning) {
				MBuildingBurning bb = (MBuildingBurning) msg;
				BuildingCondition condition = buildings.get(bb.getBuilding());
				if (condition == null) {
					log.error("Can't get condition for " + bb);
					StandardEntity entity = context.getWorld().getEntity(bb.getBuilding());
					if (entity != null && entity instanceof Building) {
						condition = new BuildingCondition((Building) entity);
						buildings.put(bb.getBuilding(), condition);
					} else {
						log.error(bb.getBuilding() + " is not id of building!");
						continue;
					}
				}
				if (condition.update(bb)) {
					currentUpdatedConditions.add(condition);
				}
			} else if (msg instanceof IBuildingFire) {
				IBuildingFire mf = (IBuildingFire) msg;
				EntityID b = mf.getLocation();
				BuildingCondition condition = buildings.get(b);
				if (condition == null) {
					log.error("Can't get condition for " + b);
					condition = new BuildingCondition((Building) context.getWorld().getEntity(b));
					buildings.put(b, condition);
				}
				if (condition.updateWhetherOnFire(mf.getTimestamp(), mf.isBuildingOnFire())) {
					currentUpdatedConditions.add(condition);
				}
			}
			// /**
			else if (msg instanceof MAgentRoadCorLocation
			// FIXME 目前为了减小计算压力，只在FB接收
					&& context.getAgentType() != StandardEntityURN.AMBULANCE_TEAM) {
				if (msg.getSender().equals(context.getAgentID())
				// || context.getTimestamp() == ignoreTime + 3
				) {
					continue;
				}
				MAgentRoadCorLocation arc = (MAgentRoadCorLocation) msg;
				if (!arc.isFireSeenAround()) {
					Map<Building, Integer> seenBuildings = los.findVisibleBuildings(arc.getX(),
							arc.getY(), false);
					for (Iterator<Entry<Building, Integer>> it = seenBuildings.entrySet()
							.iterator(); it.hasNext();) {
						Entry<Building, Integer> entry = it.next();
						if (entry.getValue() >= LineOfSightPerceptionComponent.MIN_LOS_BELIEVE_RAY_COUNT) {
							Building b = entry.getKey();
							BuildingCondition condition = buildings.get(b.getID());
							if (condition == null) {
								log.error("Can't get condition for " + b);
								condition = new BuildingCondition(b);
								buildings.put(b.getID(), condition);
							}
							if (condition.updateWhetherOnFire(arc.getTimestamp(), false)) {
								currentUpdatedConditions.add(condition);
							}
						}
					}
				}
			}
			// **/
			// TODO other msg that may affect the condition
		}
	}

	public BuildingCondition getBuildingCondition(EntityID id) {
		return buildings.get(id);
	}

	public Set<BuildingCondition> getCurrentUpdatedConditions() {
		return currentUpdatedConditions;
	}

	public static class BuildingCondition {
		public static final int INVALID_VALUE = -1;
		private static final EnumSet<StandardEntityConstants.Fieryness> SAFE_TO_ENTER = EnumSet.of(
				StandardEntityConstants.Fieryness.UNBURNT,
				StandardEntityConstants.Fieryness.MINOR_DAMAGE,
				StandardEntityConstants.Fieryness.MODERATE_DAMAGE,
				StandardEntityConstants.Fieryness.SEVERE_DAMAGE,
				StandardEntityConstants.Fieryness.BURNT_OUT);
		
		// TODO need adjustment
		private static final EnumSet<StandardEntityConstants.Fieryness> WORTH_SEARCH = EnumSet.of(
				StandardEntityConstants.Fieryness.UNBURNT,
				StandardEntityConstants.Fieryness.MINOR_DAMAGE,
				StandardEntityConstants.Fieryness.MODERATE_DAMAGE);

		private static final EnumSet<StandardEntityConstants.Fieryness> EVER_ON_FIRE = EnumSet.of(
				StandardEntityConstants.Fieryness.HEATING,
				StandardEntityConstants.Fieryness.BURNING,
				StandardEntityConstants.Fieryness.INFERNO,
				StandardEntityConstants.Fieryness.MINOR_DAMAGE,
				StandardEntityConstants.Fieryness.MODERATE_DAMAGE,
				StandardEntityConstants.Fieryness.SEVERE_DAMAGE,
				StandardEntityConstants.Fieryness.BURNT_OUT);

		private static final EnumSet<StandardEntityConstants.Fieryness> EXTINGUISHED = EnumSet.of(
				StandardEntityConstants.Fieryness.MINOR_DAMAGE,
				StandardEntityConstants.Fieryness.MODERATE_DAMAGE,
				StandardEntityConstants.Fieryness.SEVERE_DAMAGE);
		private static final int DEFAULT_FIRESEENAROUND_TEMPERATURE = 0;
		private static final int DEFAULT_HEATING_TEMPERTURE = 100;
		private static final int DEFAULT_BURNING_TEMPERTURE = 200;
		private static final int DEFAULT_INFERNO_TEMPERTURE = 150;
		private static final int DEFAULT_ON_FIRE_TEMPERATURE = 120;

		private int timestamp;
		private int dataUpdateTime;
		private int ignitionTime;
		private UpdateSource updateSource;
		private int temperature;
		private int fieryness;
		private Building building;

		private boolean isNoZeroTemperatureEver;

		public BuildingCondition(Building b) {
			this.building = b;
			this.timestamp = INVALID_VALUE;
			this.dataUpdateTime = INVALID_VALUE;
			this.ignitionTime = INVALID_VALUE;
			this.temperature = INVALID_VALUE;
			this.fieryness = INVALID_VALUE;
			this.updateSource = null;
			this.isNoZeroTemperatureEver = false;
		}

		private boolean update(Building b, int time) {
			if (!building.equals(b)) {
				log.error("Invalid Building Condition Update! (" + b + " != " + building + ")");
				return false;
			}
			updateSource = UpdateSource.ChangeSet;
			timestamp = time;
			dataUpdateTime = time;
			temperature = b.isTemperatureDefined() ? b.getTemperature() : INVALID_VALUE;
			fieryness = b.isFierynessDefined() ? b.getFieryness() : INVALID_VALUE;

			if (isEverOnFire() && ignitionTime == INVALID_VALUE) {
				ignitionTime = timestamp;
			}
			if (temperature > 0) {
				isNoZeroTemperatureEver = true;
			}
			return true;
		}

		private boolean update(MBuildingBurning msg) {
			if (!building.getID().equals(msg.getBuilding())) {
				log.error("Invalid Building Condition Update! (" + msg.getBuilding() + " != "
						+ building.getID() + ")");
				return false;
			}

			int msgTime = msg.getTimestamp();
			if (msgTime > timestamp) {
				timestamp = msgTime;
				dataUpdateTime = msgTime;
				updateSource = UpdateSource.Message;
				temperature = msg.getTemperature() >= 0 ? msg.getTemperature() : INVALID_VALUE;
				fieryness = msg.getFieryness() >= 0 ? msg.getFieryness() : INVALID_VALUE;

				if (isEverOnFire() && ignitionTime == INVALID_VALUE) {
					ignitionTime = timestamp;
				}
				if (temperature > 0) {
					isNoZeroTemperatureEver = true;
				}
				return true;
			}
			if (msgTime > dataUpdateTime) {
				// 数据没有更新，但是timestamp更新了
				updateSource = UpdateSource.Message;
				dataUpdateTime = msgTime;
				int msgTemperature = msg.getTemperature();
				int msgFieryness = msg.getFieryness();
				Fieryness msgFierynessEnum = StandardEntityConstants.Fieryness.values()[msgFieryness];
				if (isOnFire()) {
					if (!Building.BURNING.contains(msgFierynessEnum)) {
						// msg not on fire, conflict
						switch (msgFierynessEnum) {
						case UNBURNT:
						case WATER_DAMAGE:
						case MINOR_DAMAGE:
							msgFieryness = StandardEntityConstants.Fieryness.HEATING.ordinal();
							msgTemperature = DEFAULT_HEATING_TEMPERTURE;
							break;
						case MODERATE_DAMAGE:
							msgFieryness = StandardEntityConstants.Fieryness.BURNING.ordinal();
							msgTemperature = DEFAULT_BURNING_TEMPERTURE;
							break;
						case SEVERE_DAMAGE:
							msgFieryness = StandardEntityConstants.Fieryness.INFERNO.ordinal();
							msgTemperature = DEFAULT_INFERNO_TEMPERTURE;
							break;
						case BURNT_OUT:
						default:
							log.error("conflict building state!");
							break;
						}
					} else {
						temperature = msgTemperature >= 0 ? msgTemperature : INVALID_VALUE;
					}
				} else {
					if (Building.BURNING.contains(msgFierynessEnum)) {
						// msg on fire, conflict
						switch (msgFierynessEnum) {
						case HEATING:
							msgFieryness = StandardEntityConstants.Fieryness.MINOR_DAMAGE.ordinal();
							break;
						case BURNING:
							msgFieryness = StandardEntityConstants.Fieryness.MODERATE_DAMAGE
									.ordinal();
							break;
						case INFERNO:
							msgFieryness = StandardEntityConstants.Fieryness.SEVERE_DAMAGE
									.ordinal();
							break;
						default:
							break;
						}
					} else {
						temperature = msgTemperature >= 0 ? msgTemperature : INVALID_VALUE;
					}
				}
				if (fieryness != msgFieryness) {
					fieryness = msgFieryness >= 0 ? msgFieryness : INVALID_VALUE;
				}
				if (isEverOnFire() && ignitionTime == INVALID_VALUE) {
					ignitionTime = timestamp;
				}
				if (temperature > 0) {
					isNoZeroTemperatureEver = true;
				}
				return true;
			} else {
				return false;
			}
		}

		private boolean updateWhetherOnFire(int msgTime, boolean isOnFire) {
			if (msgTime > timestamp) {
				timestamp = msgTime;
			} else {
				return false;
			}
			updateSource = UpdateSource.Message;

			if (isOnFire) {
				if (!isOnFire()) {
					// conflict
					Fieryness fierynessEnum = getFierynessEnum();
					if (fierynessEnum == null) {
						fieryness = StandardEntityConstants.Fieryness.HEATING.ordinal();
					} else {
						switch (fierynessEnum) {
						case MINOR_DAMAGE:
							fieryness = StandardEntityConstants.Fieryness.HEATING.ordinal();
							break;
						case MODERATE_DAMAGE:
							fieryness = StandardEntityConstants.Fieryness.BURNING.ordinal();
							break;
						case SEVERE_DAMAGE:
							fieryness = StandardEntityConstants.Fieryness.INFERNO.ordinal();
							break;
						default:
							break;
						}
					}
					temperature = DEFAULT_ON_FIRE_TEMPERATURE;
				}
			} else {
				if (isOnFire()) {
					// conflict
					switch (getFierynessEnum()) {
					case HEATING:
						fieryness = StandardEntityConstants.Fieryness.MINOR_DAMAGE.ordinal();
						break;
					case BURNING:
						fieryness = StandardEntityConstants.Fieryness.MODERATE_DAMAGE.ordinal();
						break;
					case INFERNO:
						fieryness = StandardEntityConstants.Fieryness.SEVERE_DAMAGE.ordinal();
						break;
					default:
						break;
					}
					temperature = DEFAULT_FIRESEENAROUND_TEMPERATURE;
				} else if (temperature == INVALID_VALUE) {
					temperature = 0;
				}
			}
			return true;
		}

		public int getFieryness() {
			return fieryness;
		}

		public StandardEntityConstants.Fieryness getFierynessEnum() {
			if (fieryness == INVALID_VALUE) {
				return null;
			}
			return StandardEntityConstants.Fieryness.values()[fieryness];
		}

		/**
		 * Find out if this building is on fire.
		 * 
		 * @return True iff this buildings fieryness indicates that it is
		 *         burning.
		 */
		public boolean isOnFire() {
			if (fieryness == INVALID_VALUE) {
				return false;
			}
			return Building.BURNING.contains(getFierynessEnum());
		}

		public boolean isBurntOut() {
			if (fieryness == INVALID_VALUE) {
				return false;
			}
			return getFierynessEnum() == StandardEntityConstants.Fieryness.BURNT_OUT;
		}

		public boolean isExtinguished() {
			if (fieryness == INVALID_VALUE) {
				return false;
			}
			return EXTINGUISHED.contains(getFierynessEnum());
		}

		/**
		 * Find out if this building is safe to search.
		 * 
		 * @return True iff this buildings fieryness indicates that it is
		 *         unburnt, burn out or put off.
		 */
		public boolean isSafeToEnter() {
			if (fieryness == INVALID_VALUE) {
				return true;
			}
			return SAFE_TO_ENTER.contains(getFierynessEnum());
		}

		public boolean isWorthyOfSearch() {
			if (fieryness == INVALID_VALUE) {
				return true;
			}
			return WORTH_SEARCH.contains(getFierynessEnum());
		}

		public boolean isEverOnFire() {
			if (fieryness == INVALID_VALUE) {
				return false;
			}
			return EVER_ON_FIRE.contains(getFierynessEnum());
		}

		public boolean isNoZeroTemperatureEver() {
			return isNoZeroTemperatureEver;
		}

		public int getTemperature() {
			return temperature;
		}

		public int getTimestamp() {
			return timestamp;
		}

		public UpdateSource getUpdateSource() {
			return updateSource;
		}

		public Building getBuilding() {
			return building;
		}

		public int getIgnitionTime() {
			return ignitionTime;
		}

		public boolean hasBeenUpdated() {
			return timestamp != INVALID_VALUE;
		}

		@Override
		public String toString() {
			return "BuildingCondition [timestamp=" + timestamp + ", ignitionTime=" + ignitionTime
					+ ", updateSource=" + updateSource + ", temperature=" + temperature
					+ ", fieryness=" + fieryness + ", building=" + building
					+ ", isNoZeroTemperatureEver=" + isNoZeroTemperatureEver + "]";
		}

		@Override
		public int hashCode() {
			return (building == null) ? 0 : building.hashCode();
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			BuildingCondition other = (BuildingCondition) obj;
			if (building == null) {
				if (other.building != null) {
					return false;
				}
			} else if (!building.equals(other.building)) {
				return false;
			}
			return true;
		}
	}

}
