package ant.component.universal;

import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.Pair;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.misc.geometry.Vector2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;

/**
 * Line of sight perception.
 * 
 * @author Oliver
 */
public class LineOfSightPerceptionComponent implements AntComponent {
	private static Log log = LogFactory.getLog(LineOfSightPerceptionComponent.class);

	public static final int MIN_LOS_BELIEVE_RAY_COUNT = 2;
	private static final int DEFAULT_RAY_COUNT = 72;
	private static final int MAX_USE_TIME_PER_STEP = 11;
	private static final boolean USE_LOS_CACHE = true;

	private AgentContext context;
	private StandardWorldModel world;
	private int viewDistance;
	private int rayCount;
	private HashMap<EntityID, Collection<LineInfo>> linesMap;
	private Vector2D[] vectors;

	/**
	 * cacheHitCount:10883 obCoordinateCacheSize:3620 Kobe2 <br>
	 * cacheHitCount:4193 obCoordinateCacheSize:3232 Paris4 <br>
	 * cacheHitCount:10143 obCoordinateCacheSize:3637 Berlin5 <br>
	 * cacheHitCount:10089 obCoordinateCacheSize:3232 VC2 <br>
	 * cacheHitCount:8837 obCoordinateCacheSize:4148 Istanbul5<br>
	 */
	private int cacheHitCount;
	private int overUseStepCount;
	private HashMap<Pair<Integer, Integer>, Map<Building, Integer>> losCache;
	private int lastUseTime;
	private int missCount;

	private int overflowCount;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.world = world;
		this.viewDistance = ConfigKey.getPerceptionMaxDistance(context.getConfig());
		this.linesMap = new HashMap<EntityID, Collection<LineInfo>>();

		this.rayCount = DEFAULT_RAY_COUNT;
		this.vectors = new Vector2D[rayCount];
		double dAngle = Math.PI * 2 / rayCount;
		for (int i = 0; i < rayCount; ++i) {
			double angle = i * dAngle;
			vectors[i] = new Vector2D(Math.sin(angle), Math.cos(angle)).scale(viewDistance);
		}

		this.cacheHitCount = 0;
		this.losCache = new HashMap<Pair<Integer, Integer>, Map<Building, Integer>>();
		this.lastUseTime = -1;
		this.overUseStepCount = 0;
		this.overflowCount = 0;
	}

	@Override
	public void update() {
		// if (context.getTimestamp() != 0 && context.getTimestamp() % 50 == 0
		// && USE_LOS_CACHE) {
		// log.info("@" + context.getTimestamp() + " cacheHitCount:" +
		// cacheHitCount
		// + " losCacheSize:" + losCache.size() + " overflowCount:" +
		// overflowCount
		// + " overUseStepCount:" + overUseStepCount);
		// }
	}

	public Map<Building, Integer> findVisibleBuildings(int x, int y, boolean mustReturnResult) {
		if (!USE_LOS_CACHE) {
			return calVisibleBuildings(x, y);
		}

		if (lastUseTime != context.getTimestamp()) {
			if (missCount > MAX_USE_TIME_PER_STEP) {
				overUseStepCount++;
			}
			this.missCount = 0;
			lastUseTime = context.getTimestamp();
		}

		Pair<Integer, Integer> location = new Pair<Integer, Integer>(x, y);
		Map<Building, Integer> rv = losCache.get(location);
		if (rv == null) {
			if (!mustReturnResult && missCount > MAX_USE_TIME_PER_STEP) {
				this.overflowCount++;
				return new HashMap<Building, Integer>();
			}
			missCount++;
			rv = calVisibleBuildings(x, y);
			losCache.put(location, rv);
		} else {
			cacheHitCount++;
		}

		return rv;
	}

	private Map<Building, Integer> calVisibleBuildings(int x, int y) {
		Map<Building, Integer> rv = new HashMap<Building, Integer>();
		// getLines
		Set<LineInfo> lines = getAllBuidlingLines(x, y);

		// Cast rays
		Point2D location = new Point2D(x, y);
		for (int i = 0; i < rayCount; ++i) {
			Ray ray = new Ray(new Line2D(location, vectors[i]), lines);
			for (LineInfo hit : ray.getLinesHit()) {
				Building b = hit.getBuilding();
				Integer count = rv.get(b);
				if (count == null) {
					rv.put(b, 1);
				} else {
					rv.put(b, count + 1);
				}
			}
		}
		return rv;
	}

	public boolean canSee(int obX, int obY, int targetX, int targetY) {
		if (!inViewDistance(obX, obY, targetX, targetY)) {
			return false;
		}

		Set<LineInfo> lines = getAllBuidlingLines(obX, obY);

		Point2D ob = new Point2D(obX, obY);
		Point2D target = new Point2D(targetX, targetY);
		Ray ray = new Ray(new Line2D(ob, target), lines);
		if (ray.getVisibleLength() >= 1) {
			return true;
		}
		return false;
	}

	public boolean canSeeBuildingXY(int obX, int obY, int targetX, int targetY,
			Collection<LineInfo> lines) {
		if (!inViewDistance(obX, obY, targetX, targetY)) {
			return false;
		}
		Point2D ob = new Point2D(obX, obY);
		Point2D target = new Point2D(targetX, targetY);
		Ray ray = new Ray(new Line2D(ob, target), lines);
		if (ray.getVisibleLength() >= 1) {
			return true;
		}
		return false;
	}

	public boolean canSeeInBuilding(int obX, int obY, int targetX, int targetY, Building building) {
		if (!inViewDistance(obX, obY, targetX, targetY)) {
			return false;
		}
		Point2D ob = new Point2D(obX, obY);
		Point2D target = new Point2D(targetX, targetY);
		Ray ray = new Ray(new Line2D(ob, target), getBuidlingLines(building));
		if (ray.getVisibleLength() >= 1) {
			return true;
		}
		return false;
	}

	private Set<LineInfo> getAllBuidlingLines(int x, int y) {
		Set<LineInfo> result = new HashSet<LineInfo>();
		Collection<StandardEntity> entities = world.getObjectsInRange(x, y, viewDistance);
		for (StandardEntity next : entities) {
			if (next instanceof Building) {
				result.addAll(getBuidlingLines((Building) next));
			}
		}
		return result;
	}

	public Collection<LineInfo> getBuidlingLines(Building b) {
		Collection<LineInfo> lines = linesMap.get(b.getID());
		if (lines == null) {
			lines = new ArrayList<LineInfo>();
			for (Edge edge : b.getEdges()) {
				lines.add(new LineInfo(edge.getLine(), b, !edge.isPassable()));
			}
			linesMap.put(b.getID(), lines);
		}
		return lines;
	}

	public Set<LineInfo> getBuidlingsLines(Collection<Building> buildings) {
		Set<LineInfo> result = new HashSet<LineInfo>();
		for (Building next : buildings) {
			result.addAll(getBuidlingLines((Building) next));
		}
		return result;
	}

	public boolean inViewDistance(int obX, int obY, int targetX, int targetY) {
		return Math.hypot(obX - targetX, obY - targetY) < viewDistance;
	}

	public Pair<Integer, Integer> getEncodedCoordinate(Area originalPosition, int originalX,
			int originalY) {
		Pair<Integer, Integer> ij = encode(originalPosition, originalX, originalY);
		Pair<Integer, Integer> xy = decode(originalPosition, ij.first(), ij.second());
		return xy;
	}

	private Pair<Integer, Integer> encode(Area a, int x, int y) {
		Rectangle rect = a.getShape().getBounds();
		double minx = rect.getMinX();
		double miny = rect.getMinY();
		double maxx = rect.getMaxX();
		double maxy = rect.getMaxY();
		double dx = (maxx - minx) / 4;
		double dy = (maxy - miny) / 4;

		boolean b = false;
		int i = 0, j = 0;
		for (i = 0; i < 4; i++) { // x
			for (j = 0; j < 4; j++) { // y
				double t_minx = minx + i * dx;
				double t_miny = miny + j * dy;
				double t_maxx = t_minx + dx;
				double t_maxy = t_miny + dy;
				if (x >= t_minx && x <= t_maxx)
					if (y >= t_miny && y <= t_maxy) {
						b = true;
						break;
					}
			}
			if (b)
				break;
		}

		return new Pair<Integer, Integer>(i, j);
	}

	private Pair<Integer, Integer> decode(Area a, int i, int j) {
		Rectangle rect = a.getShape().getBounds();
		double minx = rect.getMinX();
		double miny = rect.getMinY();
		double maxx = rect.getMaxX();
		double maxy = rect.getMaxY();
		double dx = (maxx - minx) / 4;
		double dy = (maxy - miny) / 4;
		int x = (int) (minx + i * dx + dx / 2);
		int y = (int) (miny + j * dy + dy / 2);
		return new Pair<Integer, Integer>(x, y);
	}

	public int getMissCount() {
		return missCount;
	}

	private static class Ray {
		private double length;
		private List<LineInfo> hit;

		public Ray(Line2D ray, Collection<LineInfo> otherLines) {
			List<Pair<LineInfo, Double>> intersections = new ArrayList<Pair<LineInfo, Double>>();
			// Find intersections with other lines
			for (LineInfo other : otherLines) {
				double d1 = ray.getIntersection(other.getLine());
				double d2 = other.getLine().getIntersection(ray);
				if (d2 >= 0 && d2 <= 1 && d1 > 0 && d1 <= 1) {
					intersections.add(new Pair<LineInfo, Double>(other, d1));
				}
			}
			Collections.sort(intersections, new Comparator<Pair<LineInfo, Double>>() {
				@Override
				public int compare(Pair<LineInfo, Double> o1, Pair<LineInfo, Double> o2) {
					double d1 = o1.second();
					double d2 = o2.second();
					if (d1 < d2) {
						return -1;
					}
					if (d1 > d2) {
						return 1;
					}
					return 0;
				}
			});
			hit = new ArrayList<LineInfo>();
			length = 1;
			for (Pair<LineInfo, Double> next : intersections) {
				LineInfo l = next.first();
				hit.add(l);
				if (l.isBlocking()) {
					length = next.second();
					break;
				}
			}
		}

		/**
		 * if rv >=1, no intersections with walls
		 * 
		 * @return The visible length of the ray.
		 */
		public double getVisibleLength() {
			return length;
		}

		/** List of lines hit in order. */
		public List<LineInfo> getLinesHit() {
			return hit;
		}
	}

	public static class LineInfo {
		private Line2D line;
		private Building building;
		private boolean blocking;

		public LineInfo(Line2D line, Building building, boolean blocking) {
			this.line = line;
			this.building = building;
			this.blocking = blocking;
		}

		public Line2D getLine() {
			return line;
		}

		public Building getBuilding() {
			return building;
		}

		public boolean isBlocking() {
			return blocking;
		}
	}
}
