package ant.component.universal;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.component.universal.AgentLocationComponent.AgentLocationInfo;
import ant.map.component.AntMapBasisComponent;
import ant.map.partition.Partition;

/**
 * 
 * @author Oliver
 * @depend {@link AgentLocationComponent} {@link AntMapBasisComponent}
 */
public class LocationMapComponent implements AntComponent {
	private static Log log = LogFactory.getLog(LocationMapComponent.class);

	private AgentLocationComponent alc;
	private HashMap<EntityID, AreaNode> locationMap;

	private AgentContext context;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.alc = (AgentLocationComponent) context.getAntComponentKit().getAntComponent(
				AgentLocationComponent.class);
		this.locationMap = new HashMap<EntityID, AreaNode>();
		for (Area a : context.getAntComponentKit().getAntMapBasisComponent().getAreaSet()) {
			locationMap.put(a.getID(), new AreaNode(a));
		}
	}

	@Override
	public void update() {
		for (AgentLocationInfo info : alc.getUpdatedAgentLocationInfoSet()) {
			AreaNode node = locationMap.get(info.getLocation());
			node.update(info);
		}
	}

	public AreaNode getAreaNode(EntityID areaID) {
		return locationMap.get(areaID);
	}

	/**
	 * 
	 * @param areaID
	 * @param recentTime
	 * @return true if this area is visited in recentTime
	 */
	public boolean hasAreaBeenRecentlyVisited(EntityID areaID, int recentTime) {
		if (recentTime < 0) {
			log.error("recentTime < 0!, set it to default value 0");
			recentTime = 0;
		}
		AreaNode node = getAreaNode(areaID);
		if (node == null) {
			log.error("Are u sure that " + areaID + " is an areaID??");
			return false;
		}
		int lastVisitTime = node.getLastVisitTime();
		int currentTime = context.getTimestamp();
		int dTime = currentTime - lastVisitTime;
		return dTime < recentTime;
	}

	public static class AreaNode {
		private Area area;
		private int lastVisitTime;
		private int visitCount;

		// FIXME not used yet
		private ArrayList<AgentLocationInfo> agentLocationInfos;

		public AreaNode(Area area) {
			this.area = area;
			this.lastVisitTime = 0;
			this.visitCount = 0;
			this.agentLocationInfos = new ArrayList<AgentLocationInfo>();
		}

		public Area getArea() {
			return area;
		}

		public int getNodeValue(int currentTime) {
			// TODO
			int dTime = currentTime - lastVisitTime;

			return Math.max(0, 20 - dTime) + Math.min(visitCount, 10);
		}

		public int getLastVisitTime() {
			return lastVisitTime;
		}

		public int getVisitCount() {
			return visitCount;
		}

		void update(AgentLocationInfo newInfo) {
			// agentLocationInfos.add(newInfo);
			int newInfoTimestamp = newInfo.getTimestamp();
			if (newInfoTimestamp >= lastVisitTime) {
				lastVisitTime = newInfoTimestamp;
				visitCount++;
			} else if (newInfoTimestamp == lastVisitTime - 1) {
				visitCount++;
			} else {
				// log.error("newInfo is too old for update! newInfo:" + newInfo
				// + " lastVisitTime:" + lastVisitTime);
			}
		}

		@Override
		public String toString() {
			return "AreaNode [area=" + area + ", lastVisitTime=" + lastVisitTime + ", visitCount="
					+ visitCount + "]";
		}
	}

	public static class AreaNodeComparator implements Comparator<AreaNode> {
		private StandardEntity here;
		private StandardWorldModel world;
		private Set<Area> partitionAreas;

		public AreaNodeComparator(AgentContext context, Partition partition) {
			this.world = context.getWorld();
			this.here = context.getHuman().getPosition(context.getWorld());
			this.partitionAreas = partition.getAreaSet();
		}

		@Override
		public int compare(AreaNode o1, AreaNode o2) {
			if (partitionAreas.contains(o1.getArea()) && !partitionAreas.contains(o2.getArea())) {
				return -1;
			} else if (!partitionAreas.contains(o1.getArea())
					&& partitionAreas.contains(o2.getArea())) {
				return 1;
			} else if (o1.getLastVisitTime() != o2.getLastVisitTime()) {
				return o1.getLastVisitTime() - o2.getLastVisitTime();
			} else if (o1.getVisitCount() != o2.getVisitCount()) {
				return o1.getVisitCount() - o2.getVisitCount();
			} else {
				return world.getDistance(o1.getArea(), here)
						- world.getDistance(o2.getArea(), here);
			}
		}
	}
}
