package ant.component.universal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.MessageProducer;
import ant.component.universal.BuildingConditionComponent.BuildingCondition;
import ant.liaison.message.IMessage;
import ant.liaison.message.MAgentBuildingLocation;
import ant.liaison.message.MAgentBuried;
import ant.liaison.message.MAgentLocation;
import ant.liaison.message.MAgentRoadCorLocation;
import ant.liaison.message.MBuildingBurning;
import ant.liaison.message.MFindCivilianInBuilding;
import ant.liaison.message.MFindCivilianOnRoad;
import ant.liaison.message.MVoiceAgentBuried;
import ant.utils.blockade.PassableJudgeType;
import ant.utils.blockade.RoadUtils;

/**
 * 
 * @author Oliver
 * @depend {@link ChangeSetDetailComponent}
 *         {@link LineOfSightPerceptionComponent}
 *         {@link BuildingConditionComponent} {@link RoadInfoCacheComponent}
 */
public class RoutineMessageComponent implements AntComponent, MessageProducer {
	private static Log log = LogFactory.getLog(RoutineMessageComponent.class);
	private static final int INVALID_VALUE = -1;
	private static final int ALWAYS_SEND_TIME = 3;
	private static final int BURIED_MSG_SEND_RATE = 3;
	private static final int CACHE_SIZE_FOR_BUILDINGS = 50;
	private static final int MAX_BUILDINGS_SENT_ON_VOICE = 15;
	private static final int BELIEVE_BUILDING_CONDITION_RECENT_TIME = 3;

	private AgentContext context;
	private ChangeSetDetailComponent csd;
	private BuildingConditionComponent bcc;
	private Human me;

	private int buridMsgFirstSentTime;
	private Set<IMessage> routineRadioMessages;
	private Set<IMessage> routineVoiceMessages;
	private Set<BuildingCondition> buildingConditionsForVoiceMessage;
	private BuildingConditionComparatorForVoiceMessage cmp;
	private LineOfSightPerceptionComponent los;
	private RoadInfoCacheComponent ric;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.csd = context.getAntComponentKit().getChangeSetDetailComponent();
		this.bcc = (BuildingConditionComponent) context.getAntComponentKit().getAntComponent(
				BuildingConditionComponent.class);
		this.los = (LineOfSightPerceptionComponent) context.getAntComponentKit().getAntComponent(
				LineOfSightPerceptionComponent.class);

		this.ric = (RoadInfoCacheComponent) context.getAntComponentKit()
				.getRoadInfoCacheComponent();

		this.buildingConditionsForVoiceMessage = new HashSet<BuildingCondition>();
		this.routineRadioMessages = new HashSet<IMessage>();
		this.routineVoiceMessages = new HashSet<IMessage>();
		this.buridMsgFirstSentTime = INVALID_VALUE;
		this.me = context.getHuman();
		this.cmp = new BuildingConditionComparatorForVoiceMessage();
	}

	@Override
	public void update() {
		if (context.getHuman().getID().getValue() < 0) {
			// in case viewer launch component
			return;
		}
		if (context.getTimestamp() < ConfigKey.getIgnoreTime(context.getConfig())) {
			return;
		}
		routineRadioMessages.clear();
		routineVoiceMessages.clear();

		updateFromChangeSet();
		updateBuildingConditionsForVoiceMessage();
	}

	private void updateFromChangeSet() {
		sendAgentMessages();
		sendBuildingMessages();
		sendCivilianMessages();
	}

	private void sendAgentMessages() {
		StandardEntity here = me.getPosition(context.getWorld());
		// buried?
		if (shouldSendAgentBuriedMessage(me)) {
			Building b = (Building) here;
			boolean seeCivilianInBuilding = context.getBehavior().seeCivilianIn(b);
			routineRadioMessages.add(new MAgentBuried(me.getHP(), me.getDamage(), me
					.getBuriedness(), here.getID(), seeCivilianInBuilding,
					b.isFierynessDefined() ? b.isOnFire() : false));
			routineVoiceMessages.add(new MVoiceAgentBuried(me.getHP(), me.getDamage(), me
					.getBuriedness(), me.getID(), here.getID(), seeCivilianInBuilding, b
					.isFierynessDefined() ? b.isOnFire() : false));
		} else {
			// if buried, no need to send location message
			// location info
			if (here instanceof Building) {
				if (here instanceof Refuge) {
					routineRadioMessages.add(new MAgentLocation(here.getID()));
					routineVoiceMessages.add(new MAgentLocation(here.getID()));
				} else {
					Building b = (Building) here;
					boolean seeCivilianInBuilding = context.getBehavior().seeCivilianIn(b);
					routineRadioMessages.add(new MAgentBuildingLocation(here.getID(),
							seeCivilianInBuilding, b.isFierynessDefined() ? b.isOnFire() : false));
					routineVoiceMessages.add(new MAgentBuildingLocation(here.getID(),
							seeCivilianInBuilding, b.isFierynessDefined() ? b.isOnFire() : false));
				}
			} else if (here instanceof Road) {
				Road road = (Road) here;
				boolean isBlockadeSeenOnTheRoad = isBlockadeSeenOnTheRoad(road);
				boolean isFireSeenAround = isFireSeenAround();
				// TODO IF 信道条件比较差不要发XY了
				routineRadioMessages.add(new MAgentRoadCorLocation((Road) here, context.getHuman()
						.getX(), context.getHuman().getY(), isFireSeenAround,
						isBlockadeSeenOnTheRoad, isNoCivilianInSight()));
				routineVoiceMessages.add(new MAgentRoadCorLocation((Road) here, context.getHuman()
						.getX(), context.getHuman().getY(), isFireSeenAround,
						isBlockadeSeenOnTheRoad, isNoCivilianInSight()));
			} else {
				log.error("Unexpexted current location!!!");
			}
		}
	}

	private void sendBuildingMessages() {
		// send changed buildings in sight to radio channel
		for (StandardEntity entity : csd.getChangedEntities(StandardEntityURN.BUILDING).values()) {
			Building building = (Building) entity;
			if (shouldSendBuildingBurningMessage(building)) {
				routineRadioMessages.add(new MBuildingBurning(entity.getID(), building
						.getFieryness(), building.getTemperature()));
			}
		}
		// send every building in sight to voice channel
		for (StandardEntity entity : csd.getEntitiesInSight(StandardEntityURN.BUILDING,
				StandardEntityURN.FIRE_STATION, StandardEntityURN.POLICE_OFFICE,
				StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.REFUGE)) {
			Building building = (Building) entity;
			if (shouldSendBuildingBurningMessage(building)) {
				routineVoiceMessages.add(new MBuildingBurning(entity.getID(), building
						.getFieryness(), building.getTemperature()));
			}
		}
	}

	private void sendCivilianMessages() {
		for (StandardEntity entity : csd.getChangedEntities(StandardEntityURN.CIVILIAN).values()) {
			Civilian civilian = (Civilian) entity;
			if (shouldSendFindCivilianMessage(civilian)) {
				StandardEntity pos = civilian.getPosition(context.getWorld());
				if (pos instanceof Building) {
					routineRadioMessages.add(new MFindCivilianInBuilding(entity.getID(), civilian
							.getPosition(), civilian.getHP(), civilian.getDamage(), civilian
							.getBuriedness()));
					routineVoiceMessages.add(new MFindCivilianInBuilding(entity.getID(), civilian
							.getPosition(), civilian.getHP(), civilian.getDamage(), civilian
							.getBuriedness()));
				} else if (pos instanceof Road) {
					routineRadioMessages.add(new MFindCivilianOnRoad(entity.getID(), civilian
							.getPosition(), civilian.getHP(), civilian.getDamage(), civilian
							.getBuriedness()));
					routineVoiceMessages.add(new MFindCivilianOnRoad(entity.getID(), civilian
							.getPosition(), civilian.getHP(), civilian.getDamage(), civilian
							.getBuriedness()));
				}
			}
		}
	}

	private void updateBuildingConditionsForVoiceMessage() {
		for (BuildingCondition bc : bcc.getCurrentUpdatedConditions()) {
			if (bc.isEverOnFire()) {
				buildingConditionsForVoiceMessage.add(bc);
			} else if (bc.getTemperature() > 0) {
				buildingConditionsForVoiceMessage.add(bc);
			}
		}

		ArrayList<BuildingCondition> list = new ArrayList<BuildingCondition>(
				buildingConditionsForVoiceMessage);
		if (buildingConditionsForVoiceMessage.size() > CACHE_SIZE_FOR_BUILDINGS) {
			Collections.sort(list, cmp);
			buildingConditionsForVoiceMessage.clear();
			buildingConditionsForVoiceMessage.addAll(list.subList(0, CACHE_SIZE_FOR_BUILDINGS));
		}

		for (int i = 0, sentCount = 0; i < list.size(); i++) {
			if (sentCount >= MAX_BUILDINGS_SENT_ON_VOICE) {
				break;
			}
			BuildingCondition bc = list.get(i);
			if (bc.getTimestamp() == context.getTimestamp()
					&& bc.getUpdateSource() == UpdateSource.ChangeSet) {
				// building in sight
				continue;
			}
			MBuildingBurning msg = new MBuildingBurning(bc.getBuilding().getID(),
					bc.getFieryness(), bc.getTemperature());
			msg.setTimestamp(bc.getTimestamp());
			routineVoiceMessages.add(msg);
			sentCount++;
		}
	}

	private boolean isBlockadeSeenOnTheRoad(Road road) {
		if (!road.isBlockadesDefined())
			return true;

		if (road.getBlockades().size() == 0)
			return false;

		for (Edge e1 : road.getEdges()) {
			if (!e1.isPassable())
				continue;
			for (Edge e2 : road.getEdges()) {
				if (!e2.isPassable())
					continue;
				if (e1.equals(e2))
					continue;

				if (RoadUtils.isPassable(ric, e1, road, e2, PassableJudgeType.Default) == false) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * FIXME
	 * 
	 * @param b
	 * @return
	 */
	private boolean shouldSendBuildingBurningMessage(Building b) {
		return !(b instanceof Refuge) && b.isFierynessDefined() && b.isTemperatureDefined()
				&& b.getFieryness() > 0;
	}

	private boolean shouldSendFindCivilianMessage(Civilian civilian) {
		if (!civilian.isHPDefined() || civilian.getHP() == 0) {
			return false;
		}
		if (!civilian.isBuriednessDefined() || civilian.getBuriedness() == 0) {
			return false;
		}
		if (!civilian.isDamageDefined()
		// || civilian.getDamage() == 0
		) {
			return false;
		}
		if (civilian.isPositionDefined()) {
			StandardEntity pos = civilian.getPosition(context.getWorld());
			if (pos instanceof Refuge || !(pos instanceof Area)) {
				return false;
			}
		} else {
			return false;
		}
		return true;
	}

	private boolean shouldSendAgentBuriedMessage(Human me) {
		if (me.isBuriednessDefined() && me.getBuriedness() > 0 && me.isHPDefined()
				&& me.getHP() > 0 && me.isDamageDefined()) {
			if (buridMsgFirstSentTime == INVALID_VALUE) {
				this.buridMsgFirstSentTime = context.getTimestamp();
				return true;
			} else {
				if (context.getTimestamp() - buridMsgFirstSentTime < ALWAYS_SEND_TIME) {
					return true;
				} else {
					if ((context.getTimestamp() - buridMsgFirstSentTime - ALWAYS_SEND_TIME)
							% BURIED_MSG_SEND_RATE == 0) {
						return true;
					} else {
						return false;
					}
				}
			}
		}
		return false;
	}

	private boolean isNoCivilianInSight() {
		List<StandardEntity> civilians = csd.getEntitiesInSight(StandardEntityURN.CIVILIAN);
		for (StandardEntity se : civilians) {
			Civilian c = (Civilian) se;
			if (c.isHPDefined()) {
				if (c.getHP() > 0) {
					return false;
				}
			} else {
				log.error("civilian " + c + "hp undefined");
			}
		}
		return true;
	}

	private boolean isFireSeenAround() {
		Human me = context.getHuman();
		Area here = (Area) me.getPosition(context.getWorld());
		Pair<Integer, Integer> encodedCoordinate = los.getEncodedCoordinate(here, me.getX(), me
				.getY());
		int encodeX = encodedCoordinate.first();
		int encodeY = encodedCoordinate.second();
		Map<Building, Integer> losBuildingMapForEncodeXY = los.findVisibleBuildings(encodeX,
				encodeY, true);
		for (Iterator<Entry<Building, Integer>> it = losBuildingMapForEncodeXY.entrySet()
				.iterator(); it.hasNext();) {
			Entry<Building, Integer> entry = it.next();
			if (entry.getValue() >= LineOfSightPerceptionComponent.MIN_LOS_BELIEVE_RAY_COUNT) {
				Building building = entry.getKey();
				BuildingCondition bc = bcc.getBuildingCondition(building.getID());
				int dt = context.getTimestamp() - bc.getTimestamp();
				if (bc.getTimestamp() > 0 && dt <= BELIEVE_BUILDING_CONDITION_RECENT_TIME) {
					// 若最近几周期内有更新
					if (bc.isOnFire()) {
						return true;
					}
					// 若最近几周期内有更新，而且也没有着火，其实可以认为没着火。
				} else {
					// 否则，还是认为着火，这样不至于消息收到后被误解
					// log.error(building + " hasn't been seen recently!");
					return true;
				}
			}
		}
		return false;
	}

	@Override
	public Set<IMessage> pollVoiceMessages() {
		return routineVoiceMessages;
	}

	@Override
	public Set<IMessage> pollRadioMessages() {
		return routineRadioMessages;
	}

	static class BuildingConditionComparatorForVoiceMessage implements
			Comparator<BuildingCondition> {
		private static final int ON_FIRE_VALUE = 40;
		private static final int ON_LONGER_ON_FIRE_VALUE = 20;
		private static final int NO_FIRE_VALUE = 10;
		private static final int TEMPERATURE_VALUE_COEFFICIENT = 10;

		@Override
		public int compare(BuildingCondition o1, BuildingCondition o2) {
			int dv = getBuildingConditionValueForVoiceMessage(o2)
					- getBuildingConditionValueForVoiceMessage(o1);

			if (dv != 0) {
				return dv;
			} else {
				return o1.getTimestamp() - o2.getTimestamp();
			}
		}

		private int getBuildingConditionValueForVoiceMessage(BuildingCondition bc) {
			int rv = 0;
			if (bc.isOnFire()) {
				rv += ON_FIRE_VALUE * (4 - bc.getFieryness());
			} else if (bc.isBurntOut() || bc.isExtinguished()) {
				rv += ON_LONGER_ON_FIRE_VALUE;
			} else {
				rv += NO_FIRE_VALUE;
				if (bc.getTemperature() > 0) {
					rv += bc.getTemperature() / TEMPERATURE_VALUE_COEFFICIENT;
				}
			}
			return rv;
		}

	}

}
