package ant.liaison;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ant.liaison.exception.DecodingException;
import ant.liaison.exception.EncodingException;
import ant.liaison.exception.MessagePropertyDescriptionException;
import ant.liaison.message.IMessage;

/**
 * 
 * @author Ivan
 * 
 * @depend {@link MessagePropertyCodecComponent}
 */
public class SingleMessageCodec {
	private static final Log log = LogFactory.getLog(SingleMessageCodec.class);

	private Map<Class<? extends IMessage>, MessagePropertiesOrder> mmap;
	private MessagePropertyCodecComponent comp;
	private boolean noVoice;

	/** NoVoice **/
	public SingleMessageCodec(MessagePropertyCodecComponent comp) throws MessagePropertyDescriptionException {
		this.comp = comp;
		this.mmap = MessagePropertiesOrder.getMessagePropertiesOrderMap(comp);
		this.noVoice = true;
	}

	/** All **/
	SingleMessageCodec(MessagePropertyCodecComponent comp, boolean noVoice) throws MessagePropertyDescriptionException {
		this.comp = comp;
		this.mmap = MessagePropertiesOrder.getMessagePropertiesOrderMap(comp);
		this.noVoice = noVoice;
	}
	
	public BitArray encode(IMessage message) throws EncodingException {
		if (noVoice)
			return _encode(message, MessageType.getNoVoiceTypeBitLength());
		else
			return _encode(message, MessageType.getAllTypeBitLength());
	}

	public IMessage decode(BitArray data) throws DecodingException {
		if (noVoice)
			return _decode(data, MessageType.getNoVoiceTypeBitLength(), MessageType.getNoVoiceOnlyMessageCount());
		else
			return _decode(data, MessageType.getAllTypeBitLength(), MessageType.values().length);
	}

	private BitArray _encode(IMessage message, int mtlength) throws EncodingException {
		Class<? extends IMessage> clazz = message.getMessageType().getBindClazz();
		if (!mmap.containsKey(clazz))
			throw new EncodingException("Not found message Class:" + clazz);

		BitArray ret;
		int val;
		int nbits;

		try {
			ret = new BitArray(0);
			ret.append(BitCodecUtils.encode(message.getMessageType().ordinal(), mtlength));
			MessagePropertiesOrder po = mmap.get(clazz);
			for (int i = 0; i < po.size(); i++) {
				MessageProperty p = po.getBitProperty(i);
				nbits = comp.getPropertyBitLength(p);
				Method get = po.getGetMethod(i);
				Object data = get.invoke(message, (Object[]) null);
				val = comp.encodeProperty(p, data);
				ret.append(BitCodecUtils.encode(val, nbits));
			}
		} catch (IllegalArgumentException e) {
			throw new EncodingException(e);
		} catch (IllegalAccessException e) {
			throw new EncodingException(e);
		} catch (InvocationTargetException e) {
			throw new EncodingException(e);
		}

		return ret;
	}

	/**
	 * data truncation
	 */
	public IMessage _decode(BitArray data, int mtlength, int mcount) throws DecodingException {
		// BitArray arr = data.truncation(MessageType.getTypeBitLength());
		// int val = BitCodecUtils.decode(arr);
		int val = data.truncationAndDecode(mtlength);

		if (val < 0 || val >= mcount) {
			throw new DecodingException("MessageType Invalid:" + val);
		}
		
		MessageType type = MessageType.values()[val];
		
		if (comp.ignoreMessage(type.getAcceptor())) {
			// log.info("Ignore decoding message: " + type);
			return null;
		}
		
		Class<? extends IMessage> clazz = type.getBindClazz();
		if (!mmap.containsKey(clazz))
			throw new DecodingException("Not found message Class:" + clazz);

		// data.startDecode();
		IMessage msg = null;
		try {
			msg = (IMessage) clazz.getConstructor((Class<?>[]) null).newInstance((Object[]) null);
			MessagePropertiesOrder po = mmap.get(clazz);
			for (int i = 0; i < po.size(); i++) {
				Method set = po.getSetMethod(i);
				MessageProperty prop = po.getBitProperty(i);
				int propBitLen = comp.getPropertyBitLength(prop);
				if (data.length() < propBitLen) {
					throw new DecodingException("Not enough length for " + clazz);
				}
				// arr = data.truncation(propBitLen);
				// val = BitCodecUtils.decode(arr);

				// val = data.iterDecode(propBitLen);
				val = data.truncationAndDecode(propBitLen);

				set.invoke(msg, comp.decodeProperty(prop, val));
			}
		} catch (IllegalArgumentException e) {
			throw new DecodingException(e);
		} catch (SecurityException e) {
			throw new DecodingException(e);
		} catch (InstantiationException e) {
			throw new DecodingException(e);
		} catch (IllegalAccessException e) {
			throw new DecodingException(e);
		} catch (InvocationTargetException e) {
			throw new DecodingException(e);
		} catch (NoSuchMethodException e) {
			throw new DecodingException(e);
		}

		return msg;
	}

}
