package ant.liaison.channel.detect;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.config.Config;
import rescuecore2.messages.Command;
import rescuecore2.misc.Pair;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.agent.command.AntSpeak;
import ant.liaison.channel.ChannelCondition;
import ant.liaison.channel.ChannelMannager;
import ant.liaison.channel.RadioChannel;

/**
 * 
 * @author Oliver
 * 
 */
public class ChannelConditionDetector implements AntComponent {
	private static Log log = LogFactory.getLog(ChannelConditionDetector.class);

	public final static String DETECT_MSG = "*";
	// FIXME if there is too many msg in one cycle for a channel, will lead to
	// kernel panic
	private final static int MAX_MSG_COUNT = 500;
	private final static int DETECT_RESULT_MSG_LENGTH = 3;
	private static final double DEFAULT_SUCCESS_RATE = 0.90;

	private EntityID agentID;
	private int ignoreTime;
	private int agentCount;
	private int subscribeMax;
	private ArrayList<AntSpeak> messages;
	private ArrayList<RadioChannel> subscribeChannels;
	private Map<Integer, RadioChannel> radioChannels;
	private DetectorAssigner detectorAssigner;

	private AgentContext context;

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.agentID = context.getAgentID();
		Config config = context.getConfig();
		this.ignoreTime = ConfigKey.getIgnoreTime(config);
		this.agentCount = ConfigKey.getTotalAgents(config);
		this.subscribeMax = ConfigKey.getChannelPlatoonMax(config);
		this.messages = new ArrayList<AntSpeak>();
		ChannelMannager channelManager = (ChannelMannager) context.getAntCommunicationKit()
				.getAntComponent(ChannelMannager.class);
		this.radioChannels = channelManager.getRadioChannels();
		for (RadioChannel rc : radioChannels.values()) {
			ChannelCondition condition = new ChannelCondition(rc.getId(), Math.min(
					rc.getBandWidth() / DETECT_MSG.length(), MAX_MSG_COUNT)
					/ agentCount);
			rc.setCondition(condition);
		}
		ArrayList<EntityID> agents = new ArrayList<EntityID>();
		for (StandardEntity entity : world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM,
				StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.POLICE_FORCE)) {
			agents.add(entity.getID());
		}
		this.detectorAssigner = new DetectorAssigner(agents, radioChannels, subscribeMax);
		// detectorAssigner.print();
		this.subscribeChannels = detectorAssigner.getAgentToChannelsMap().get(agentID);
	}

	@Override
	public void update() {
		if (subscribeMax == 0) {
			return;
		}
		if (context.getTimestamp() == ignoreTime) {
			sendDetectMessage(context);
		} else if (context.getTimestamp() == ignoreTime + 1) {
			detectCondition(context);
			sendDetectResult(context);
			// printChannelConditions();
		} else if (context.getTimestamp() == ignoreTime + 2) {
			concludeCondition(context);
			// printChannelConditions();
		}
	}

	private void sendDetectMessage(AgentContext context) {
		messages.clear();
		int time = context.getTimestamp();
		byte[] bytes = DETECT_MSG.getBytes();
		for (RadioChannel rc : radioChannels.values()) {
			ChannelCondition condition = rc.getCondition();
			AntSpeak antSpeak = new AntSpeak(time, rc.getId(), bytes);
			for (int i = 0; i < condition.getBaseMsgSentNum(); i++) {
				messages.add(antSpeak);
			}
		}
	}

	private void detectCondition(AgentContext context) {
		Collection<Command> hearing = context.getNaiveHearing();
		for (Command c : hearing) {
			if (c instanceof AKSpeak) {
				AKSpeak akSpeak = (AKSpeak) c;
				String content = new String(akSpeak.getContent());
				RadioChannel rc = radioChannels.get(akSpeak.getChannel());
				if (rc == null) {
					// this should be civilian says ouch or help
					// if (content.length() != 0 && !content.equals("Ouch") &&
					// !content.equals("Help")) {
					// log.error("Unknown msg with content:" + content);
					// }
					continue;
				} else {
					rc.getCondition().receiveNewMsg(content);
				}
			}
		}
	}

	private void sendDetectResult(AgentContext context) {
		messages.clear();
		ArrayList<RadioChannel> sortedRadioChannels = detectorAssigner.getSortedRadioChannels();
		for (RadioChannel rc : subscribeChannels) {
			rc.getCondition().ajustMsgSentCount();

			byte[] msg = encodeDetectResultMsg(rc.getId(), rc.getCondition().getMsgReceivedCount());
			int index = sortedRadioChannels.indexOf(subscribeChannels.get(0));
			for (int i = 0; i < sortedRadioChannels.size(); i += subscribeMax) {
				RadioChannel ch = sortedRadioChannels.get((i + index) % sortedRadioChannels.size());
				ChannelCondition condition = ch.getCondition();
				if (condition.getBaseMsgSentNum() / DETECT_RESULT_MSG_LENGTH < 1) {
					log.warn("condition.getBaseMsgSentNum() / DETECT_RESULT_MSG_LENGTH < 1, with condition:"
							+ condition);
					// FIXME should think of better way if this really happened
					messages.add(new AntSpeak(context.getTimestamp(), ch.getId(), msg));
				} else {
					for (int j = 0; j < condition.getBaseMsgSentNum() / DETECT_RESULT_MSG_LENGTH; j++) {
						messages.add(new AntSpeak(context.getTimestamp(), ch.getId(), msg));
					}
				}
			}
		}
	}

	private void concludeCondition(AgentContext context) {
		Collection<Command> hearing = context.getNaiveHearing();
		HashMap<Pair<EntityID, Integer>, Integer> detectResults = new HashMap<Pair<EntityID, Integer>, Integer>();
		for (Command c : hearing) {
			if (c instanceof AKSpeak) {
				AKSpeak akSpeak = (AKSpeak) c;
				byte[] content = akSpeak.getContent();
				if (content.length == 0) {
					continue;
				}
				if (radioChannels.get(akSpeak.getChannel()) == null) {
					// this should be civilian says ouch or help
					// String str = new String(content);
					// if (!str.equals("Ouch") && !str.equals("Help")) {
					// log.error("Unknown msg with content:" + str);
					// }
					continue;
				} else {
					if (!akSpeak.getAgentID().equals(agentID)) {
						Pair<Integer, Integer> result = decodeDetectResultMsg(content);
						detectResults.put(
								new Pair<EntityID, Integer>(akSpeak.getAgentID(), result.first()),
								result.second());
					}
				}
			}
		}

		for (Iterator<Entry<Pair<EntityID, Integer>, Integer>> it = detectResults.entrySet()
				.iterator(); it.hasNext();) {
			Entry<Pair<EntityID, Integer>, Integer> entry = it.next();
			ChannelCondition condition = radioChannels.get(entry.getKey().second()).getCondition();
			condition.updateCondition(entry.getValue());
		}

		for (RadioChannel rc : radioChannels.values()) {
			ChannelCondition condition = rc.getCondition();
			if (condition.getMsgSentCount() == 0) {
				log.error("Get no infomation about channel " + rc.getId()
						+ ", ChannelConditionDetect Failed!");
				rc.setSuccessRate(DEFAULT_SUCCESS_RATE);
			} else {
				if (condition.getMsgReceivedCount() == 0) {
					rc.setSuccessRate(DEFAULT_SUCCESS_RATE);
					// ////////////////////////////////////////////
					log.error(condition + " " + hearing.size());
					for (Command c : hearing) {
						if (c instanceof AKSpeak) {
							if (((AKSpeak) c).getChannel() != 0)
								log.info(c);
						}
					}
					// ////////////////////////////////////////////
				} else {
					rc.setSuccessRate(condition.getMsgReceivedCount() * 1.0d
							/ (condition.getMsgSentCount() * agentCount));
				}
			}
		}
	}

	/**
	 * @param channelID
	 * @param value
	 * @return the detect result msg, byte[0] for channelID, byte[1,2] for value
	 */
	private byte[] encodeDetectResultMsg(int channelID, int value) {
		if (channelID >= 128 || channelID < 0 || value >= 65536 || value < 0) {
			log.error("transferIntToBytes overflow!");
			return null;
		}
		byte[] rv = new byte[DETECT_RESULT_MSG_LENGTH];
		rv[0] = (byte) channelID;
		rv[1] = (byte) (value % 256 - 128);
		rv[2] = (byte) (value / 256 - 128);
		return rv;
	}

	/**
	 * @param bytes
	 * @return Pair of (channelID, value)
	 */
	private Pair<Integer, Integer> decodeDetectResultMsg(byte[] bytes) {
		if (bytes.length != DETECT_RESULT_MSG_LENGTH) {
			log.error("Invalid bytes of length != DETECT_RESULT_MSG_LENGTH");
			return null;
		}
		int value = 0;
		value += bytes[2] + 128;
		value *= 256;
		value += bytes[1] + 128;

		Pair<Integer, Integer> rv = new Pair<Integer, Integer>((int) bytes[0], value);
		return rv;
	}

	/**
	 * Get the messages to be sent in this cycle
	 * 
	 * @return the messages of type AntSpeak
	 */
	public ArrayList<AntSpeak> getMessages() {
		return messages;
	}

	/**
	 * Get the channels to be subscribed in this cycle
	 * 
	 * @return the channel ids
	 */
	public int[] getSubscribeChannels() {
		int[] rv = new int[subscribeChannels.size()];
		for (int i = 0; i < subscribeChannels.size(); i++) {
			rv[i] = subscribeChannels.get(i).getId();
		}
		return rv;
	}

	@Override
	public String getComponentName() {
		return this.getClass().getName();
	}

	/**
	 * For debug only, print the channel conditions
	 */
	public void printChannelConditions() {
		String str = "";
		for (Iterator<Entry<Integer, RadioChannel>> it = radioChannels.entrySet().iterator(); it
				.hasNext();) {
			Entry<Integer, RadioChannel> entry = it.next();
			ChannelCondition condition = entry.getValue().getCondition();
			if (condition.getBaseMsgSentNum() > 0)
				str += "\n" + condition + " SuccessRate:"
						+ String.format("%.2f", entry.getValue().getSuccessRate());
		}
		log.info(str);
	}
}