package ant.liaison.channel.io;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.messages.Command;
import rescuecore2.standard.entities.Civilian;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.standard.messages.AKSpeak;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.ConfigKey;
import ant.liaison.BitCodecUtils;
import ant.liaison.MultiMessageCodec;
import ant.liaison.SingleMessageCodec;
import ant.liaison.channel.AbstractChannel;
import ant.liaison.channel.ChannelMannager;
import ant.liaison.channel.RadioChannel;
import ant.liaison.channel.VoiceChannel;
import ant.liaison.exception.DecodingException;
import ant.liaison.message.AbstractCivilianMessage;
import ant.liaison.message.AbstractCivilianMessage.MCivilianSayHelp;
import ant.liaison.message.AbstractCivilianMessage.MCivilianSayOuch;
import ant.liaison.message.IMessage;

/**
 * 
 * @author Oliver
 * 
 */
public class MessageReceiver implements AntComponent {
	private static Log log = LogFactory.getLog(MessageReceiver.class);

	private AgentContext context;
	private int ignoreTime;
	private Set<IMessage> radioMessages;
	private Set<IMessage> radioMessagesWithSelf;
	private HashSet<IMessage> voiceMessages;
	private HashSet<AbstractCivilianMessage> civilianMessages;
	private SingleMessageCodec smc;
	private MultiMessageCodec mmc;
	private ChannelMannager channelMannager;

	@Override
	public String getComponentName() {
		return getClass().getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.ignoreTime = ConfigKey.getIgnoreTime(context.getConfig());
		this.radioMessages = new HashSet<IMessage>();
		this.radioMessagesWithSelf = new HashSet<IMessage>();
		this.voiceMessages = new HashSet<IMessage>();
		this.civilianMessages = new HashSet<AbstractCivilianMessage>();
		if (smc == null) {
			smc = context.getAntCommunicationKit().getSingleMessageCodec();
		}
		this.channelMannager = (ChannelMannager) context.getAntCommunicationKit().getAntComponent(
				ChannelMannager.class);
	}

	@Override
	public void update() {
		if (context.getTimestamp() <= ignoreTime) {
			// won't receive any message
			return;
		}
		clearMessageBoxes();

		Collection<Command> cc = context.getNaiveHearing();
		long sTime = System.currentTimeMillis();
		for (Command c : cc) {
			if (c instanceof AKSpeak) {
				AKSpeak message = (AKSpeak) c;
				if (message.getContent().length == 0) {
					continue;
				}
				AbstractChannel ch = channelMannager.getChannel(message.getChannel());
				if (ch instanceof VoiceChannel) {
					if (addCivilianMessages(message)) {
						continue;
					} else if (addVoiceMessages(message)) {
						continue;
					}
				}
				if (ch instanceof RadioChannel) {
					if (!shouldListenFromRadio()) {
						continue;
					}
					if (addRadioMessages(message)) {
						continue;
					}
				}
				log.error("Unrecognized msg:" + message);
			}
		}

		long dTime = System.currentTimeMillis() - sTime;
		if (dTime > 500) {
			log.error("@" + context.getTimestamp() + " MessageReceiver update cost:" + dTime
					+ " with receiveMessages " + radioMessagesWithSelf.size());
		} else if (dTime > 100) {
			log.warn("@" + context.getTimestamp() + " MessageReceiver update cost:" + dTime
					+ " with receiveMessages " + radioMessagesWithSelf.size());
		}
	}

	/**
	 * channelDetect and channelAssign takes time
	 * 
	 * @return true if radio should be listened
	 */
	private boolean shouldListenFromRadio() {
		return context.getTimestamp() > ignoreTime + 2;
	}

	private void clearMessageBoxes() {
		civilianMessages.clear();
		voiceMessages.clear();
		radioMessages.clear();
		radioMessagesWithSelf.clear();
	}

	private boolean addVoiceMessages(AKSpeak message) {
		byte[] content = message.getContent();
		if (content.length == 0) {
			return true;
		}
		List<IMessage> list = mmc.decode(content);
		if (list.isEmpty()) {
			log.error("Receive Voice Message with size 0," + message + " content:"
					+ new String(content));
			return false;
		}
		EntityID id = message.getAgentID();
		for (IMessage msg : list) {
			msg.setSender(id);
		}
		voiceMessages.addAll(list);
		return true;
	}

	private boolean addCivilianMessages(AKSpeak message) {
		EntityID sender = message.getAgentID();
		StandardEntity entity = context.getWorld().getEntity(sender);
		if (entity == null || entity instanceof Civilian) {
			String content = new String(message.getContent());
			if (content.length() == 0) {
				return true;
			} else if (content.equals("Help")) {
				MCivilianSayHelp msg = new AbstractCivilianMessage.MCivilianSayHelp(sender);
				msg.setTimestamp(context.getTimestamp() - 1);
				civilianMessages.add(msg);
				return true;
			} else if (content.equals("Ouch")) {
				MCivilianSayOuch msg = new AbstractCivilianMessage.MCivilianSayOuch(sender);
				msg.setTimestamp(context.getTimestamp() - 1);
				civilianMessages.add(msg);
				return true;
			} else {
				log.error("UnExpected msg received!!" + message);
				return true;
			}
		}
		return false;
	}

	private boolean addRadioMessages(AKSpeak message) {
		try {
			byte[] bytes = message.getContent();
			IMessage msg = smc.decode(BitCodecUtils.translateToBitArray(bytes));
			
			if (msg == null) {
				// 会收到无法解码的单条信息
				return true;
			}
			
			msg.setSender(message.getAgentID());
			msg.setTimestamp(context.getTimestamp() - 1);

			if (!context.getAgentID().equals(message.getAgentID())) {
				radioMessages.add(msg);
			}
			radioMessagesWithSelf.add(msg);
			return true;
		} catch (DecodingException e) {
			log.error(e);
			return false;
		}
	}

	public void setSingleMessageCodec(SingleMessageCodec singleMessageCodec) {
		this.smc = singleMessageCodec;
	}

	public void setMultiMessageCodec(MultiMessageCodec multiMessageCodec) {
		this.mmc = multiMessageCodec;
	}

	public Set<IMessage> getRadioMessages() {
		return radioMessages;
	}

	public Set<IMessage> getRadioMessagesWithSelf() {
		return radioMessagesWithSelf;
	}

	public HashSet<IMessage> getVoiceMessages() {
		return voiceMessages;
	}

	public Set<AbstractCivilianMessage> getMessagesFromCivilian() {
		return civilianMessages;
	}
}
