package ant.map.component;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.FireBrigade;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardEntityURN;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.utils.comparator.EntityComparatorById;
import ant.utils.comparator.EntityIDComparator;

/**
 * @provide entity set <br>
 *          entity label <br>
 * 
 * @author Ivan
 * 
 * @depend null
 */
public class AntMapBasisComponent implements AntComponent {
	private static final Log log = LogFactory.getLog(AntMapBasisComponent.class);

	private AgentContext context;
	private StandardWorldModel world;

	// Unmodifiable
	private Set<Human> agentSet;
	private Set<Human> pfSet;
	private Set<Human> atSet;
	private Set<FireBrigade> fbSet;

	private Set<Area> areaSet;
	private Set<Road> roadSet;
	private Set<Refuge> refugeSet;
	private Set<Building> buildingSet;
	private List<EntityID> areaLabelList;
	private List<EntityID> agentLabelList;

	private Map<EntityID, Integer> areaLabelMap;
	private Map<EntityID, Integer> agentLabelMap;

	@Override
	public String getComponentName() {
		return AntMapBasisComponent.class.getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.world = world;

		initialEntitySet();
		initialLabelAreaID();
		initialLabelAgentID();
		log.info("World contains: " + areaSet.size() + " areas, " + roadSet.size() + " roads, "
				+ buildingSet.size() + " buildings");
	}

	@Override
	public void update() {
	}

	private void initialLabelAgentID() {
		agentLabelList = new ArrayList<EntityID>();
		agentLabelMap = new HashMap<EntityID, Integer>();

		for (Human human : this.agentSet) {
			agentLabelList.add(human.getID());
		}
		Collections.sort(agentLabelList, new EntityIDComparator());
		agentLabelList = Collections.unmodifiableList(agentLabelList);

		for (int i = 0; i < agentLabelList.size(); i++) {
			agentLabelMap.put(agentLabelList.get(i), i);
		}
	}

	private void initialLabelAreaID() {
		areaLabelList = new ArrayList<EntityID>();
		areaLabelMap = new HashMap<EntityID, Integer>();

		for (Area area : this.areaSet) {
			areaLabelList.add(area.getID());
		}
		Collections.sort(areaLabelList, new EntityIDComparator());
		areaLabelList = Collections.unmodifiableList(areaLabelList);

		for (int i = 0; i < areaLabelList.size(); i++) {
			areaLabelMap.put(areaLabelList.get(i), i);
		}

	}

	private void initialEntitySet() {
		// Initial agent set
		this.agentSet = new TreeSet<Human>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE,
				StandardEntityURN.FIRE_BRIGADE, StandardEntityURN.AMBULANCE_TEAM)) {

			this.agentSet.add((Human) e);
		}
		this.agentSet = Collections.unmodifiableSet(this.agentSet);

		this.pfSet = new TreeSet<Human>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.POLICE_FORCE)) {
			this.pfSet.add((Human) e);
		}
		this.pfSet = Collections.unmodifiableSet(this.pfSet);

		this.atSet = new TreeSet<Human>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.AMBULANCE_TEAM)) {
			this.atSet.add((Human) e);
		}
		this.atSet = Collections.unmodifiableSet(this.atSet);

		this.fbSet = new TreeSet<FireBrigade>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.FIRE_BRIGADE)) {
			this.fbSet.add((FireBrigade) e);
		}
		this.fbSet = Collections.unmodifiableSet(this.fbSet);

		// Initial area set
		this.areaSet = new TreeSet<Area>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.BUILDING,
				StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.FIRE_STATION,
				StandardEntityURN.POLICE_OFFICE, StandardEntityURN.REFUGE, StandardEntityURN.ROAD)) {

			this.areaSet.add((Area) e);
		}
		this.areaSet = Collections.unmodifiableSet(this.areaSet);

		this.roadSet = new TreeSet<Road>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.ROAD)) {
			this.roadSet.add((Road) e);
		}
		this.roadSet = Collections.unmodifiableSet(this.roadSet);

		this.refugeSet = new TreeSet<Refuge>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.REFUGE)) {
			this.refugeSet.add((Refuge) e);
		}
		this.refugeSet = Collections.unmodifiableSet(this.refugeSet);

		this.buildingSet = new TreeSet<Building>(new EntityComparatorById());
		for (final StandardEntity e : world.getEntitiesOfType(StandardEntityURN.BUILDING,
				StandardEntityURN.AMBULANCE_CENTRE, StandardEntityURN.FIRE_STATION,
				StandardEntityURN.POLICE_OFFICE, StandardEntityURN.REFUGE)) {

			this.buildingSet.add((Building) e);
		}
		this.buildingSet = Collections.unmodifiableSet(this.buildingSet);
	}

	public Set<Human> getAgentSet() {
		return agentSet;
	}

	public Set<Human> getPfSet() {
		return pfSet;
	}

	public Set<Human> getAtSet() {
		return atSet;
	}

	public Set<FireBrigade> getFbSet() {
		return fbSet;
	}

	public Set<Area> getAreaSet() {
		return areaSet;
	}

	public int getAreaCount() {
		return areaSet.size();
	}

	public Set<Road> getRoadSet() {
		return roadSet;
	}

	public int getRoadCount() {
		return roadSet.size();
	}

	public Set<Building> getBuildingSet() {
		return buildingSet;
	}

	public int getBuildingCount() {
		return buildingSet.size();
	}

	public Set<Refuge> getRefugeSet() {
		return refugeSet;
	}

	public int getAreaLabel(EntityID areaId) {
		return areaLabelMap.get(areaId);
	}

	public int getAgentLabel(EntityID agentId) {
		return agentLabelMap.get(agentId);
	}

	public List<EntityID> getAreaLabelList() {
		return this.areaLabelList;
	}

	public List<EntityID> getAgentLabelList() {
		return this.agentLabelList;
	}

	public int getAreaSize() {
		return areaSet.size();
	}

	public int getAgentSize() {
		return agentSet.size();
	}
}
