package ant.map.component;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Refuge;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.map.objects.AreaGraph;
import ant.map.objects.IGraph;
import ant.map.objects.LabeledArea;
import ant.map.objects.LabeledGraph;
import ant.map.objects.Route;
import ant.map.objects.RouteResult;
import ant.map.objects.RouteResult.RouteResultFactory;
import ant.map.objects.StandardGraphEdge;
import ant.map.search.AbstractSSSPTree;
import ant.map.search.StandardLabeledSPFASearch;
import ant.utils.geo.GeoUtils;

/**
 * provide labeled area graph, distance and sssptree to refuge, static route from a to b<br>
 * node distance : manhattanDistance
 * 
 * @author Ivan
 * 
 * @depend {@link AntMapBasisComponent}
 */
public class StaticCityGraphComponent implements IGraphComponent<LabeledArea, StandardGraphEdge<LabeledArea>> {
	private static final Log log = LogFactory.getLog(StaticCityGraphComponent.class);

	private LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>> graph = null;
	private Map<EntityID, LabeledArea> mmap;
	private StandardLabeledSPFASearch algorighm;
	private AntMapBasisComponent mapbasis;
	private Set<LabeledArea> labeledAreas;
	private AreaGraph staticAreaGraph;
	private StandardWorldModel world;
	
	private EntityID cachedSSSPCenter = null;
	private AbstractSSSPTree<LabeledArea> cachedSSSPTree = null;
	
	// private Set<Refuge> refugeSet;
	// private Map<EntityID, AbstractSSSPTree<LabeledArea>> refugeSSSPMap;
	// private AgentContext context;

	public LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>> getLabeledGraph() {
		return graph;
	}

	@Override
	public String getComponentName() {
		return StaticCityGraphComponent.class.getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		// this.refugeSet = context.getAntComponentKit().getAntMapBasisComponent().getRefugeSet();
		// this.context = context;
		
		this.world = world;
		this.mmap = new HashMap<EntityID, LabeledArea>();
		this.mapbasis = context.getAntComponentKit().getAntMapBasisComponent();
		int size = mapbasis.getAreaSize();
		this.graph = new LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>>(size);
		this.labeledAreas = new HashSet<LabeledArea>();
		for (int i = 0; i < size; i++) {
			EntityID id = mapbasis.getAreaLabelList().get(i);
			LabeledArea v = new LabeledArea((Area) world.getEntity(id), i);
			this.graph.addVertex(v);
			mmap.put(id, v);
			this.labeledAreas.add(v);
		}

		for (LabeledArea a : graph.getVertexSet()) {
			for (EntityID id : a.getArea().getNeighbours()) {
				LabeledArea b = mmap.get(id);
				double dist = GeoUtils.manhattanDistance(a, b);
				this.graph.setEdge(a, b, new StandardGraphEdge<LabeledArea>(a, b, dist));
			}
		}

		this.mmap = Collections.unmodifiableMap(mmap);
		this.algorighm = new StandardLabeledSPFASearch(this.graph);
		
		intiialAreaGraph();
	}

	@Override
	public void update() {
	}

	public LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>> createGraph() {
		int size = mapbasis.getAreaSize();
		LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>> retgraph = new LabeledGraph<LabeledArea, StandardGraphEdge<LabeledArea>>(
				size);
		// this.refugeSet = context.getAntComponentKit().getAntMapBasisComponent().getRefugeSet();
		// this.labeledAreas = new HashSet<LabeledArea>();

		for (int i = 0; i < size; i++) {
			EntityID id = mapbasis.getAreaLabelList().get(i);
			LabeledArea v = new LabeledArea((Area) world.getEntity(id), i);
			retgraph.addVertex(v);
			// mmap.put(id, v);
			// this.labeledAreas.add(v);
		}

		for (LabeledArea a : graph.getVertexSet()) {
			for (EntityID id : a.getArea().getNeighbours()) {
				LabeledArea b = mmap.get(id);
				double dist = GeoUtils.manhattanDistance(a, b);
				retgraph.setEdge(a, b, new StandardGraphEdge<LabeledArea>(a, b, dist));
			}
		}
		return retgraph;
	}

	private void intiialAreaGraph() {
		staticAreaGraph = new AreaGraph();
		for (Area a : mapbasis.getAreaSet()) {
			staticAreaGraph.addVertex(a);
		}

		for (Area from : mapbasis.getAreaSet()) {
			for (EntityID id : from.getNeighbours()) {
				Area to = (Area) world.getEntity(id);

				double dist = GeoUtils.manhattanDistance(from.getX(), from.getY(), to.getX(), to.getY());
				staticAreaGraph.setEdge(from, to, new StandardGraphEdge<Area>(from, to, dist));
				staticAreaGraph.setEdge(to, from, new StandardGraphEdge<Area>(to, from, dist));
			}

		}
	}

//	private void initialSSSP() {
//		
//		this.refugeSSSPMap = new HashMap<EntityID, AbstractSSSPTree<LabeledArea>>();
//
//		if (refugeSet.isEmpty())
//			return;
//
//		for (Refuge refuge : refugeSet) {
//			LabeledArea labelArea = mmap.get(refuge.getID());
//			AbstractSSSPTree<LabeledArea> tree = algorighm.ssspSearch(labelArea);
//			refugeSSSPMap.put(refuge.getID(), tree);
//		}
//
//		this.refugeSSSPMap = Collections.unmodifiableMap(refugeSSSPMap);
//	}

//	public double getDistanceToRefuge(EntityID start, EntityID refugeId) {
//		if (!(world.getEntity(start) instanceof Area)) {
//			throw new IllegalArgumentException();
//		}
//
//		if (!(world.getEntity(refugeId) instanceof Refuge)) {
//			throw new IllegalArgumentException();
//		}
//
//		LabeledArea la = mmap.get(start);
//		return refugeSSSPMap.get(refugeId).getDistance(la);
//	}

//	public AbstractSSSPTree<LabeledArea> getToRefugeAbstractSSSPTree(EntityID refugeId) {
//		return refugeSSSPMap.get(refugeId);
//	}

	public AbstractSSSPTree<LabeledArea> getAbstractSSSPTree(EntityID areaId) {
		if (!areaId.equals(cachedSSSPCenter)) {
			cachedSSSPCenter = areaId;
			LabeledArea labelArea = mmap.get(areaId);
			cachedSSSPTree = this.algorighm.ssspSearch(labelArea);
		}
		
		return cachedSSSPTree;
		// LabeledArea labelArea = mmap.get(areaId);
		// return this.algorighm.ssspSearch(labelArea);
	}
	
	public double getStaticDistanceToTreeRoot(EntityID from, EntityID root) {
		AbstractSSSPTree<LabeledArea> tree = getAbstractSSSPTree(root);
		LabeledArea la = mmap.get(from);
		return tree.getDistance(la);
	}

	private Route<LabeledArea> getStaticRawRoute(EntityID from, EntityID to) {
		LabeledArea labelFrom = mmap.get(from);
		LabeledArea labelTo = mmap.get(to);
		// log.debug("Search from: " + from + " to:" + to + ", " + labelFrom +
		// " " + labelTo);
		return this.algorighm.spSearch(labelFrom, labelTo);
	}

	public List<EntityID> getStaticMoveRoute(EntityID from, EntityID to) {
		Route<LabeledArea> data = getStaticRawRoute(from, to);
		if (data == null) {
			return null;
		} else {
			List<EntityID> r = new ArrayList<EntityID>();
			if (data.getRoute().size() > 1) {
				for (int i = 1; i < data.getRoute().size(); i++) {
					r.add(data.getRoute().get(i).getID());
				}
			} else {
				for (int i = 0; i < data.getRoute().size(); i++) {
					r.add(data.getRoute().get(i).getID());
				}
			}
			return r;
		}
	}

	public RouteResult getStaticMoveRouteExt(EntityID from, EntityID to) {
		Route<LabeledArea> data = getStaticRawRoute(from, to);
		if (data == null) {
			new RouteResultFactory(false).newInstance();
		}
		List<EntityID> r = new ArrayList<EntityID>();
		if (data.getRoute().size() > 1) {
			for (int i = 1; i < data.getRoute().size(); i++) {
				r.add(data.getRoute().get(i).getID());
			}
		} else {
			for (int i = 0; i < data.getRoute().size(); i++) {
				r.add(data.getRoute().get(i).getID());
			}
		}

		RouteResultFactory factory = new RouteResultFactory(true);
		factory.setDistance(data.getDistance());
		factory.setRawRoute(r);
		factory.setFrom(from);
		factory.setTo(to);
		LabeledArea labelTo = mmap.get(to);
		factory.addSuitableDestPoint(labelTo.getX(), labelTo.getY());
		return factory.newInstance();
	}

	public Map<EntityID, LabeledArea> getLabeledAreaMap() {
		return mmap;
	}

	@Override
	public IGraph<LabeledArea, StandardGraphEdge<LabeledArea>> getGraph() {
		return this.graph;
	}

//	public Set<LabeledArea> getLabeledAreas() {
//		return labeledAreas;
//	}

	
	public AreaGraph getStaticAreaGraph() {
		return this.staticAreaGraph;
	}

	public double getDistanceToRefuge(Area from, Refuge refuge) {
		return GeoUtils.manhattanDistance(from, refuge);
	}
}
