package ant.map.jigsaw;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.AntComponent;
import ant.map.component.AntMapBasisComponent;
import ant.map.jigsaw.JigsawMarkerComponent.RoadMark;

/**
 * FIXME 如果有不连通的块怎么办
 * 
 * @author Ivan
 * 
 * @depend {@link JigsawMarkerComponent} {@link AntMapBasisComponent}
 */
public class JigsawMainComponent implements AntComponent {
	private static final Log log = LogFactory.getLog(JigsawMainComponent.class);

	private JigsawMarkerComponent marker;
	private StandardWorldModel world;
	private AgentContext context;
	private ArrayList<JigsawObject> items;

	private Set<JigsawAttach> attaches;
	private Set<JigsawAvenue> avenues;
	private Set<JigsawCross> crosses;

	private Map<EntityID, Integer> revmap;
	private int attachCount = 0;
	private int avenueCount = 0;
	private int crossCount = 0;

	@Override
	public String getComponentName() {
		return JigsawMainComponent.class.getName();
	}

	@Override
	public void initial(AgentContext context, StandardWorldModel world) {
		this.context = context;
		this.world = world;
		this.marker = (JigsawMarkerComponent) context.getAntComponentKit().getAntComponent(
				JigsawMarkerComponent.class);
		this.items = new ArrayList<JigsawObject>();
		this.revmap = new HashMap<EntityID, Integer>();
		this.attaches = new HashSet<JigsawAttach>();
		this.avenues = new HashSet<JigsawAvenue>();
		this.crosses = new HashSet<JigsawCross>();

		try {
			buildJigsaws();
			checkJigsaws();
		} catch (Exception e) {
			log.error(e);
		}
		log.info("<Jigsaw> AvenueCount:" + avenueCount + ", CrossCount:" + crossCount
				+ ", AttachCount:" + attachCount);
	}

	private void checkJigsaws() {
		for (int i = 0; i < items.size(); i++) {
			if (items.get(i).getJigsawID() != i) {
				log.error("Not same id! :" + items.get(i));
			}
		}
		for (Area a : context.getAntComponentKit().getAntMapBasisComponent().getAreaSet()) {
			if (!revmap.containsKey(a.getID())) {
				log.error("Not contains id! :" + a);
			}
		}
	}

	// 1. Cross
	// 2. Avenue
	// 3. Attach
	private void buildJigsaws() {
		int id = 0;
		Set<EntityID> visited = new HashSet<EntityID>();

		// Cross
		for (Road r : context.getAntComponentKit().getAntMapBasisComponent().getRoadSet()) {
			if (!visited.contains(r.getID())) {
				if (marker.getMark(r.getID()) == RoadMark.Cross) {
					Set<Area> tmp = spreadRoad(r, RoadMark.Cross, visited);
					for (Area a : tmp)
						revmap.put(a.getID(), id);
					JigsawCross c = new JigsawCross(id, tmp);
					items.add(c);
					crosses.add(c);
					id++;
					crossCount++;
				}
			}
		}

		// Avenue
		for (Road r : context.getAntComponentKit().getAntMapBasisComponent().getRoadSet()) {
			if (!visited.contains(r.getID())) {
				if (marker.getMark(r.getID()) == RoadMark.Avenue) {
					Set<Area> tmp = spreadRoad(r, RoadMark.Avenue, visited);
					for (Area a : tmp)
						revmap.put(a.getID(), id);
					JigsawAvenue c = new JigsawAvenue(id, tmp);
					items.add(c);
					avenues.add(c);
					id++;
					avenueCount++;
				}
			}
		}

		// Attach
		for (Road r : context.getAntComponentKit().getAntMapBasisComponent().getRoadSet()) {
			if (!visited.contains(r.getID())) {
				if (marker.getMark(r.getID()) == RoadMark.Entrance) {
					Set<Area> tmp = spreadAttach(r, visited);
					for (Area a : tmp)
						revmap.put(a.getID(), id);
					JigsawAttach c = new JigsawAttach(id, tmp);
					items.add(c);
					attaches.add(c);
					id++;
					attachCount++;
				}
			}
		}

		for (Building b : context.getAntComponentKit().getAntMapBasisComponent().getBuildingSet()) {
			if (!visited.contains(b.getID())) {
				Set<Area> tmp = spreadAttach(b, visited);
				for (Area a : tmp)
					revmap.put(a.getID(), id);
				JigsawAttach c = new JigsawAttach(id, tmp);
				items.add(c);
				attaches.add(c);
				id++;
				attachCount++;
			}
		}

		// Attaching
		for (JigsawAttach ja : attaches) {
			for (Area a : ja.getObjects()) {
				for (EntityID neid : a.getNeighbours()) {
					if (!ja.contains(neid)) {
						JigsawObject obj = getJigsaw(neid);
						if (obj.getType() == JigsawType.Attach) {
							log.error("Bad jigsaw attaching!");
						} else {
							obj.addAttach(ja);
							ja.addEnclosed(obj);
						}
					}
				}
			}
		}

		for (JigsawAvenue ja : avenues) {
			for (Area a : ja.getObjects()) {
				for (EntityID neid : a.getNeighbours()) {
					if (!ja.contains(neid)) {
						JigsawObject obj = getJigsaw(neid);
						if (obj.getType() == JigsawType.Cross) {
							JigsawCross jc = (JigsawCross) obj;
							ja.addCross(jc);
							jc.addAvenue(ja);
						} else if (obj.getType() == JigsawType.Avenue) {
							log.error("Bad jigsaw avenue!");
						}
					}
				}
			}
		}

	}

	private Set<Area> spreadAttach(Area a, Set<EntityID> visited) {
		LinkedList<Area> open = new LinkedList<Area>();
		Set<Area> ret = new HashSet<Area>();
		open.add(a);
		visited.add(a.getID());
		ret.add(a);
		while (!open.isEmpty()) {
			Area current = open.pop();
			ret.add(current);
			for (EntityID id : current.getNeighbours()) {
				if (!visited.contains(id)) {
					Area next = (Area) world.getEntity(id);
					if (next instanceof Road) {
						RoadMark other = marker.getMark(id);
						if (other == RoadMark.Entrance) {
							visited.add(id);
							open.add(next);
							ret.add(a);
						}
					} else {
						visited.add(id);
						open.add(next);
						ret.add(a);
					}
				}
			}
		}
		return ret;
	}

	private Set<Area> spreadRoad(Road r, RoadMark mark, Set<EntityID> visited) {
		if (marker.getMark(r.getID()) != mark) {
			throw new IllegalArgumentException();
		}
		LinkedList<Area> open = new LinkedList<Area>();
		Set<Area> ret = new HashSet<Area>();
		open.add(r);
		visited.add(r.getID());
		ret.add(r);
		while (!open.isEmpty()) {
			Area current = open.pop();
			ret.add(current);
			for (EntityID id : current.getNeighbours()) {
				if (!visited.contains(id)) {
					Area next = (Area) world.getEntity(id);
					if (next instanceof Road) {
						RoadMark other = marker.getMark(id);
						if (other == mark) {
							visited.add(id);
							open.add(next);
							ret.add(r);
						}
					}
				}
			}
		}
		return ret;
	}

	@Override
	public void update() {
	}

	public int getJigsawID(EntityID id) {
		return this.revmap.get(id);
	}

	public JigsawObject getJigsaw(EntityID id) {
		return items.get(revmap.get(id));
	}

	public ArrayList<JigsawObject> getJigsawObjects() {
		return this.items;
	}

	public int getTotalJigsaws() {
		return attachCount + avenueCount + crossCount;
	}
}
