package ant.map.objects;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import ant.utils.LabeledObject;

/**
 * continuous label from 0 - size-1
 * 
 * @author Ivan
 * 
 */
public class LabeledGraph<V extends LabeledObject, E extends IGraphEdge<V>> implements
		IGraph<V, E> {

	private int size;
	
	// index (= labeled) 
	private ArrayList<EdgeList> forward;
	private Set<V> vSet;
	
	// reverse table
	private HashMap<Integer, V> vMap;
	
	/**
	 * @param size : max vertex size
	 */
	public LabeledGraph(int size) {
		this.size = size;
		this.forward = new ArrayList<EdgeList>(size);
		for (int i = 0; i < size; i++) {
			this.forward.add(i, null);
		}
		vSet = new HashSet<V>();
		vMap = new HashMap<Integer, V>();
	}
	
	public int getVertexSize() {
		return this.vSet.size(); 
	}
	
	public V getVertex(int label) {
		return vMap.get(label);
	}
	
	@Override
	public void addVertex(V v) {
		if (v.getLabel() >= size || v.getLabel() < 0) {
			throw new IllegalArgumentException();
		}

		if (forward.get(v.getLabel()) == null) {
			forward.set(v.getLabel(), new EdgeList());
			vSet.add(v);
			vMap.put(v.getLabel(), v);
		}
	}

	@Override
	public boolean containsEdge(V source, V target) {
		if (source.getLabel() >= size || source.getLabel() < 0) {
			throw new IllegalArgumentException();
		}

		ArrayList<E> tmp = forward.get(source.getLabel()).list;

		for (E item : tmp) {
			if (item.getTarget().getLabel() == target.getLabel()) {
				return true;
			}
		}
		return false;
	}

	@Override
	public boolean containsVertex(V v) {
		if (v.getLabel() >= size || v.getLabel() < 0) {
			throw new IllegalArgumentException();
		}

		return (forward.get(v.getLabel()) != null);
	}
	
	/**
	 * Neighbors of v
	 */
	@Override
	public Collection<E> edgesOf(V v) {
		if (v.getLabel() >= size || v.getLabel() < 0) {
			throw new IllegalArgumentException();
		}
		return forward.get(v.getLabel()).list;
	}

	/**
	 * return null if not found
	 */
	@Override
	public E getEdge(V source, V target) {
		if (source.getLabel() >= size || source.getLabel() < 0) {
			throw new IllegalArgumentException();
		}

		ArrayList<E> tmp = forward.get(source.getLabel()).list;

		for (E item : tmp) {
			if (item.getTarget().getLabel() == target.getLabel()) {
				return item;
			}
		}

		return null;
	}

	@Override
	public Set<V> getNeighbours(V v) {
		throw new UnsupportedOperationException();
	}

	@Override
	public Set<V> getVertexSet() {
		return vSet;
	}

	@Override
	public void removeEdge(V source, V target) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void setEdge(V source, V target, E edge) {
		if (source.getLabel() >= size || source.getLabel() < 0) {
			throw new IllegalArgumentException();
		}
		if (target.getLabel() >= size || target.getLabel() < 0) {
			throw new IllegalArgumentException();
		}
		if (source.getLabel() != edge.getSource().getLabel()
				|| target.getLabel() != edge.getTarget().getLabel()) {
			throw new IllegalArgumentException();
		}

		ArrayList<E> tmp = forward.get(source.getLabel()).list;

		for (int i = 0; i < tmp.size(); i++) {
			if (tmp.get(i).getTarget().getLabel() == target.getLabel()) {
				tmp.set(i, edge);
			}
		}

		tmp.add(edge);
	}
	
	class EdgeList {
		ArrayList<E> list;

		public EdgeList() {
			this.list = new ArrayList<E>();
		}

		public void addEdge(E edge) {
			this.list.add(edge);
		}
	}
}
