package ant.map.partition;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.utils.LazyMap;
import ant.utils.geo.GeoUtils;

/**
 * 
 * @author Ivan
 * 
 */
public class CellPartitionAlgorithm extends AbstractPartitionAlgorithm {
	private static final Log log = LogFactory.getLog(CellPartitionAlgorithm.class);
	private static final int CELL_SCALE = 2;
	private int count;
	private StandardWorldModel world;
	private ArrayList<Partition> cellParts;
	private int splitIndex = 0;

	public CellPartitionAlgorithm(AgentContext context, int partitionCount) {
		super(context);
		this.count = partitionCount;
		this.world = context.getWorld();
	}

	@Override
	public int getPartitionCount() {
		return cellParts.size();
	}

	@Override
	public ArrayList<Partition> getPartitions() {
		return cellParts;
	}

	@Override
	public void partition() {
		long _time_st = System.currentTimeMillis();
		partitionInternal();
		long _time_ed = System.currentTimeMillis();
		log.debug("CellPartition.partSize:" + getPartitionCount());
		log.debug("CellPartition.partition cost: " + (_time_ed - _time_st) + " milliseconds");
	}

	private void partitionInternal() {
		HashMap<EntityID, Partition> mmap = new HashMap<EntityID, Partition>();
		// TODO CELL_SCALE
		int temp = this.count * CELL_SCALE;

		long _time_st = System.currentTimeMillis();
		// SimplePartitionAlgorithm simple = new SimplePartitionAlgorithm(context, temp);
		// simple.partition();
		BasicAreaPartitionAlgorithm simple = new BasicAreaPartitionAlgorithm(context, temp);
		log.debug("Starting simple");
		simple.partition();
		long _time_ed = System.currentTimeMillis();
		log.debug("CellPartitionAlg.base Partition cost: " + (_time_ed - _time_st) + " milliseconds");

		ArrayList<Partition> rawParts = simple.getPartitions();
		cellParts = new ArrayList<Partition>();

		for (Partition p : rawParts) {
			cellParts.addAll(split(p));
		}

		for (Partition p : cellParts) {
			for (Area a : p.getAreaSet()) {
				mmap.put(a.getID(), p);
			}
			p.convCalc();
		}

		// merge
		// TODO optimize
		while (cellParts.size() > this.count) {
			int pick = -1;
			int min = Integer.MAX_VALUE;
			for (int i = 0; i < cellParts.size(); i++) {
				if (cellParts.get(i).size() < min) {
					min = cellParts.get(i).size();
					pick = i;
				}
			}
			Partition target = cellParts.get(pick);
			cellParts.remove(pick);
			Partition neighbor = connectivity(target, mmap);
			for (Area a : target.getAreaSet()) {
				neighbor.addArea(a);
				mmap.put(a.getID(), neighbor);
			}
			neighbor.convCalc();
		}
		mmap = null;
	}

	private Partition connectivity(Partition target, HashMap<EntityID, Partition> mmap) {
		LazyMap<Partition, Integer> counter = new LazyMap<Partition, Integer>() {
			@Override
			public Integer createValue() {
				return 0;
			}
		};
		Partition tmp;
		for (Area a : target.getAreaSet()) {
			for (EntityID neid : a.getNeighbours()) {
				tmp = mmap.get(neid);
				if (tmp != target) {
					counter.put(tmp, counter.get(tmp) + 1);
					// counter.put(tmp, 1);
				}
			}
		}

		for (Entry<Partition, Integer> entry : counter.entrySet()) {
			int data = (int) (entry.getValue() * 1000 / Math.sqrt(entry.getKey().size()));
			entry.setValue(data);
		}

		int max = 0;
		tmp = null;
		for (Entry<Partition, Integer> entry : counter.entrySet()) {
			if (entry.getValue() > max) {
				max = entry.getValue();
				tmp = entry.getKey();
			} else if (entry.getValue() == max && max != 0 && tmp != null) {
				// double dist1 = GeoUtils.dist(target.getCx(), target.getCy(),
				// tmp.getCx(), tmp.getCy());
				// double dist2 = GeoUtils.dist(target.getCx(), target.getCy(),
				// entry.getKey().getCx(), entry.getKey().getCy());
				double dist1 = GeoUtils.manhattanDistance(target, tmp);
				double dist2 = GeoUtils.manhattanDistance(target, entry.getKey());
				if (dist2 < dist1) {
					max = entry.getValue();
					tmp = entry.getKey();
				}
			}
		}
		return tmp;
	}

	private Collection<Partition> split(Partition partition) {
		ArrayList<Partition> ret = new ArrayList<Partition>();
		HashSet<Area> marked = new HashSet<Area>();

		for (Area a : partition.getAreaSet()) {
			if (!marked.contains(a)) {
				Collection<Area> part = spread(a, partition.getAreaSet(), marked);
				// log.debug("Spread Size:" + part.size());
				Partition sub = new Partition(splitIndex);
				for (Area tmp : part)
					sub.addArea(tmp);
				sub.convCalc();
				splitIndex++;
				ret.add(sub);
			}
		}
		return ret;
	}

	/**
	 * 
	 * @param a
	 *            seed
	 * @param areaSet
	 *            limited set
	 * @param marked
	 * @return sub parts set
	 */
	private Collection<Area> spread(Area a, Set<Area> areaSet, HashSet<Area> marked) {
		ArrayList<Area> ret = new ArrayList<Area>();
		LinkedList<Area> queue = new LinkedList<Area>();
		queue.add(a);
		marked.add(a);
		while (!queue.isEmpty()) {
			Area target = queue.pollFirst();
			ret.add(target); // visited
			for (EntityID neid : target.getNeighbours()) {
				Area nei = (Area) world.getEntity(neid);
				if (areaSet.contains(nei)) {
					if (!marked.contains(nei)) { // not visited
						queue.add(nei);
						marked.add(nei);
					}
				}
			}
		}
		return ret;
	}

}
