package ant.map.partition;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.map.component.AntMapBasisComponent;
import ant.utils.CounterSet;
import ant.utils.RandomSelector;
import ant.utils.geo.GeoUtils;

/**
 * 
 * @author Ivan
 *
 */
public class FastCellPartitionAlgorithm extends AbstractPartitionAlgorithm {
	private static final Log log = LogFactory.getLog(FastCellPartitionAlgorithm.class);
	private static final int CELL_SCALE = 3;
	private static final double CONV_DISTANCE = 5;
	
	private int count;
	private StandardWorldModel world;
	private AntMapBasisComponent ambc;
	private Set<Area> areas;
	private ArrayList<Partition> parts;
	private Map<EntityID, Integer> reverse;
	
	public FastCellPartitionAlgorithm(AgentContext context, int partitionCount) {
		super(context);
		this.count = partitionCount;
		this.world = context.getWorld();
		this.ambc = context.getAntComponentKit().getAntMapBasisComponent();
		this.areas = ambc.getAreaSet();
		// log.debug(partitionCount + "," + areas.size());
	}

	@Override
	public void partition() {
		this.reverse = new HashMap<EntityID, Integer>(this.areas.size(), 0.8f);
		Map<Integer, Set<Area>> mmap = cellpartition();
		this.parts = new ArrayList<Partition>();
		int id = 0;
		for (Entry<Integer, Set<Area>> entry : mmap.entrySet()) {
			Partition p = new Partition(id);
			Set<Area> tmp = entry.getValue();
			p.addAll(tmp);
			p.convCalc();
			parts.add(p);
			for (Area a : tmp) {
				reverse.put(a.getID(), id);
			}
			id++;
		}
		if (parts.size() != this.count) {
			log.warn("Partition Numbers do not agree, parts.size=" + parts.size() + " expcount:" + count);
		}
		
		// debug();
	}

	private void debug() {
		StringBuffer sb = new StringBuffer();
		for (int i = 0; i < parts.size(); i++) {
			Partition pp = parts.get(i);
			long cc = 0;
			for (EntityID id : pp.getAreaIDSet()) {
				cc += id.getValue();
			}
			sb.append(cc);
			sb.append(",");
		}
		// System.out.println(sb);
		// log.info(sb);
	}

	private Map<Integer, Set<Area>> cellpartition() {
		Map<Integer, Set<Area>> mmap = kmpartition();
		for (Entry<Integer, Set<Area>> item : mmap.entrySet()) {
			Integer k = item.getKey();
			for (Area a : item.getValue()) {
				reverse.put(a.getID(), k);
			}
		}
		// System.out.println(mmap.keySet().size());
		// System.out.println(mmap.size());
		while (mmap.keySet().size() > count) {
			int min = Integer.MAX_VALUE;
			int source = -1;
			for (Entry<Integer, Set<Area>> item : mmap.entrySet()) {
				if (item.getValue().size() < min) {
					min = item.getValue().size();
					source = item.getKey();
				}
			}
			
			CounterSet<Integer> connValue = new CounterSet<Integer>();
			Set<Area> sourceAreaSet = mmap.get(source);
			double centerX = 0, centerY = 0;
			
			for (Area a : sourceAreaSet) {
				centerX += a.getX();
				centerY += a.getY();
				for (EntityID neid : a.getNeighbours()) {
					Area nei = (Area) world.getEntity(neid);
					if (!sourceAreaSet.contains(nei)) {
						connValue.inc(reverse.get(neid));
					}
				}
			}
			
			if (sourceAreaSet.size() != 0) {
				centerX = centerX / sourceAreaSet.size();
				centerY = centerY / sourceAreaSet.size();
			}
			
			double max = 0;
			int target = -1;
			for (Entry<Integer, Integer> entry : connValue.getMap().entrySet()) {
				int size = mmap.get(entry.getKey()).size();
				double grade = entry.getValue() / Math.sqrt(size);
				if (grade > max) {
					max = grade;
					target = entry.getKey();
				}
			}
			// merge
			// System.out.println(connValue.getMap().entrySet().size()+ "," + target+","+count);
			// System.out.println(sourceAreaSet);
			// 不连通情况
			if (target == -1) {
				log.warn("Not connecting!!");
				double mindist = Double.MAX_VALUE;
				for (Entry<Integer, Set<Area>> item : mmap.entrySet()) {
					if (item.getKey() != source) {
						double localmindist = Double.MAX_VALUE;
						for (Area tempa : item.getValue()) {
							double tempdist = GeoUtils.manhattanDistance(tempa, centerX, centerY);
							if (tempdist < localmindist) {
								localmindist = tempdist;
							}
						}
						if (localmindist < mindist) {
							mindist = localmindist;
							target = item.getKey();
						}
					}
				}
			}
			
			Set<Area> targetAreaSet = mmap.get(target);
			for (Area a : sourceAreaSet) {
				reverse.put(a.getID(), target);
			}
			targetAreaSet.addAll(sourceAreaSet);
			//System.out.println("pre:" + connValue.getMap().entrySet().size());
			mmap.remove(source);
			//System.out.println("post:" + connValue.getMap().entrySet().size());
		}
		return mmap;
	}
	
	private Map<Integer, Set<Area>> kmpartition() {
		int rawCellCount = count * CELL_SCALE;
		Set<PartitionCluster> pcs = new HashSet<PartitionCluster>(rawCellCount);
		Map<Integer, Set<Area>> clusterAssign = new HashMap<Integer, Set<Area>>(rawCellCount);
		Set<Area> inits = RandomSelector.randomSelect(areas, rawCellCount, context.getUniformRandom());
		
		
		int id = 0;
		for (Area b : inits) {
			PartitionCluster pc = new PartitionCluster(id);
			pc.addArea(b);
			pc.convCalc();
			pcs.add(pc);
			id++;
		}
		
		int repeat = 0;
		while (true) {
			boolean conv = true;
			repeat++;
			for (Area b : areas) {
				double min = Double.MAX_VALUE;
				PartitionCluster pick = null;
				for (PartitionCluster pc : pcs) {
					double temp = GeoUtils.manhattanDistance(pc, b);
					if (temp < min) {
						pick = pc;
						min = temp;
					}
				}
				pick.addArea(b);
			}
			
			for (PartitionCluster pc : pcs) {
				double c = pc.convCalc();
				conv = conv && (c <= CONV_DISTANCE);
			}
			
			if (repeat != 1 && conv) break;
		}
		
		// log.debug("Repeat:" + repeat);
		for (Area b : areas) {
			double min = Double.MAX_VALUE;
			PartitionCluster pick = null;
			for (PartitionCluster pc : pcs) {
				double temp = GeoUtils.manhattanDistance(pc, b);
				if (temp < min) {
					pick = pc;
					min = temp;
				}
			}
			if (!clusterAssign.containsKey(pick.getID()))
				clusterAssign.put(pick.getID(), new HashSet<Area>(250));
			clusterAssign.get(pick.getID()).add(b);
		}
		
		id = 0;
		Map<Integer, Set<Area>> ret = new HashMap<Integer, Set<Area>>();
		for (Entry<Integer, Set<Area>> entry : clusterAssign.entrySet()) {
			List<Set<Area>> tmp = split(entry.getValue());
			for (Set<Area> item : tmp) {
				ret.put(id, item);
				id++;
			}
		}
		// log.debug(ret.hashCode());
		return ret;
	}
	
	
	private List<Set<Area>> split(Set<Area> input) {
		ArrayList<Set<Area>> ret = new ArrayList<Set<Area>>();
		HashSet<Area> marked = new HashSet<Area>(input.size());
		LinkedList<Area> queue = new LinkedList<Area>();

		for (Area a : input) {
			if (!marked.contains(a)) {
				Set<Area> part = new HashSet<Area>();
				queue.clear();
				queue.add(a);
				marked.add(a);
				while (!queue.isEmpty()) {
					Area target = queue.pollFirst();
					part.add(target); // visited
					for (EntityID neid : target.getNeighbours()) {
						Area nei = (Area) world.getEntity(neid);
						if (input.contains(nei)) {
							if (!marked.contains(nei)) { // not visited
								queue.add(nei);
								marked.add(nei);
							}
						}
					}
				}
								
				ret.add(part);
			}
		}
		return ret;
	}
	
	@Override
	public int getPartitionCount() {
		return parts.size();
	}

	@Override
	public ArrayList<Partition> getPartitions() {
		return parts;
	}
	
	public int findPartitionIndex(EntityID where) {
		return this.reverse.get(where);
	}

}
