package ant.map.search;

import java.util.ArrayList;
import java.util.LinkedList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import ant.map.objects.IGraphEdge;
import ant.map.objects.LabeledArea;
import ant.map.objects.LabeledGraph;
import ant.map.objects.Route;

/**
 * 
 * @author Ivan
 * 
 */
public class LabeledSPFASearch<V extends LabeledArea, E extends IGraphEdge<V>> extends AbstractShortestPathAlgorithm<V, E> {
	private static final double INFINITY = Double.POSITIVE_INFINITY;
	private static final Log log = LogFactory.getLog(LabeledSPFASearch.class);

	private int size;
	private LabeledGraph<V, E> g;
	private ArrayList<VertexInfo> arr;
	private LinkedList<Integer> queue;

	public LabeledSPFASearch(LabeledGraph<V, E> graph) {
		super(graph);
		this.g = graph;
		this.size = g.getVertexSize();

		queue = new LinkedList<Integer>();
	}

	private void initial() {
		this.arr = new ArrayList<VertexInfo>();
		for (int i = 0; i < this.size; i++) {
			this.arr.add(new VertexInfo(this.g.getVertex(i)));
		}

		queue.clear();
	}

	private void searchInternal(int source) {
		arr.get(source).inQueue = true;
		arr.get(source).sp = 0;
		queue.add(source);
		VertexInfo s, t;
		while (!queue.isEmpty()) {
			s = arr.get(queue.pollFirst());
			s.inQueue = false;
			for (E e : g.edgesOf(s.v)) {
				t = arr.get(e.getTarget().getLabel());
				if (s.sp + e.getWeight() < t.sp) {
					t.sp = s.sp + e.getWeight();
					t.pre = s.label;

					if (!t.inQueue) {
						t.inQueue = true;

						// queue.addLast(t.label);

						// SLF optimize

						if (!queue.isEmpty()) {
							if (t.sp < arr.get(queue.getFirst()).sp) {
								queue.addFirst(t.label);
							} else {
								queue.addLast(t.label);
							}
						} else {
							queue.addLast(t.label);
						}

					}
				}
			}
		}
	}

	@Override
	public Route<V> spSearch(V source, V target) {
		initial();
		searchInternal(source.getLabel());
		return makePath(source, target);
	}

	@Override
	public AbstractSSSPTree<V> ssspSearch(V source) {
		initial();
		searchInternal(source.getLabel());
		return new SimpleSSSPTree(source, this.arr);
	}

	private Route<V> makePath(V source, V target) {
		Route<V> sp = new Route<V>();
		sp.setFrom(source);
		sp.setTo(target);
		sp.setDistance(arr.get(target.getLabel()).sp);

		if (Double.isInfinite(sp.getDistance()))
			return null;

		LinkedList<V> route = new LinkedList<V>();
		V current = target;
		while (true) {
			route.addFirst(current);
			if (current.equals(source) || current == null) {
				break;
			} else {
				current = arr.get(arr.get(current.getLabel()).pre).v;
			}
		}
		sp.setRoute(route);
		return sp;
	}

	private class VertexInfo {
		V v;
		int label = -1;
		boolean inQueue = false;
		double sp = INFINITY;
		int pre = -1;

		public VertexInfo(V v) {
			this.v = v;
			this.label = v.getLabel();
		}
	}

	private class SimpleSSSPTree extends AbstractSSSPTree<V> {

		private ArrayList<VertexInfo> arr;

		public SimpleSSSPTree(V source, ArrayList<VertexInfo> arr) {
			super(source);
			this.arr = arr;
		}

		@Override
		public double getDistance(V target) {
			return arr.get(target.getLabel()).sp;
		}

		@Override
		public V getSPFather(V target) {
			int tmp = arr.get(target.getLabel()).pre;
			if (tmp == -1)
				return null;
			return arr.get(tmp).v;
		}

	}
}
