package ant.map.search;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import ant.map.objects.AreaGraph;
import ant.map.objects.Route;
import ant.map.objects.StandardGraphEdge;

/**
 * 
 * @author Ivan
 * 
 */
public class RangeSPFASearch extends AbstractShortestPathAlgorithm<Area, StandardGraphEdge<Area>> {
	private static final Log log = LogFactory.getLog(RangeSPFASearch.class);

	private static final double INFINITY = Double.POSITIVE_INFINITY;
	private Set<Area> vSet;
	private Map<Area, VertexInfo> mmap;
	private Set<Area> inQueue;
	private LinkedList<Area> queue;
	private StandardWorldModel world;
	private int range;
	private Set<Area> permit;

	public RangeSPFASearch(AreaGraph graph, StandardWorldModel world, int range) {
		super(graph);
		this.world = world;
		this.range = range;
		this.vSet = graph.getVertexSet();
		mmap = new HashMap<Area, VertexInfo>(vSet.size());
		for (Area a : vSet) {
			mmap.put(a, new VertexInfo(a));
		}
		inQueue = new HashSet<Area>();
		queue = new LinkedList<Area>();
		permit = new HashSet<Area>();
	}

	private void initial(Area source) {
		inQueue.clear();
		queue.clear();
		permit.clear();

		long _time_st = System.currentTimeMillis();
		for (StandardEntity se : world.getObjectsInRange(source, range)) {
			if (se instanceof Area) {
				permit.add((Area) se);
			}
		}
		// permit = AreaInConnectedRange.getAreaInConnectedRange(world, source, range);
		long _time_ed = System.currentTimeMillis();
//		if (_time_ed - _time_st > 7)
//			log.debug("RangeSPFASearch.GetObjectsInRange cost: " + (_time_ed - _time_st) + " milliseconds");

		for (Area a : permit) {
			VertexInfo tmp = mmap.get(a);
			tmp.sp = INFINITY;
			tmp.pre = null;
		}
		// log.info("PermitSize:" + permit.size());
	}

	private void search(Area source) {
		mmap.get(source).sp = 0;
		queue.add(source);
		inQueue.add(source);

		VertexInfo s, tmp;
		while (!queue.isEmpty()) {
			Area v = queue.pollFirst();
			s = mmap.get(v);
			inQueue.remove(v);
			for (StandardGraphEdge<Area> edge : graph.edgesOf(v)) {
				tmp = mmap.get(edge.getTarget());
				if (permit.contains(tmp.vertex)) {
					if (s.sp + edge.getWeight() < tmp.sp) {
						tmp.sp = s.sp + edge.getWeight();
						tmp.pre = v;

						if (!inQueue.contains(tmp.vertex)) {
							inQueue.add(tmp.vertex);
							queue.addLast(tmp.vertex);
						}
					}
				}
			}
		}
	}

	private Route<Area> makePath(Area source, Area target) {
		Route<Area> sp = new Route<Area>();
		sp.setFrom(source);
		sp.setTo(target);
		sp.setDistance(mmap.get(target).sp);

		if (Double.isInfinite(sp.getDistance()))
			return null;

		LinkedList<Area> route = new LinkedList<Area>();
		Area current = target;
		while (true) {
			route.addFirst(current);
			if (current.equals(source)) {
				break;
			} else {
				current = mmap.get(current).pre;
			}
		}
		sp.setRoute(route);
		return sp;
	}

	@Override
	public Route<Area> spSearch(Area source, Area target) {
		initial(source);
		if (!permit.contains(target))
			return null;
		search(source);
		return makePath(source, target);
	}

	/**
	 * WARN: transient, don't keep this tree
	 */
	@Override
	public AbstractSSSPTree<Area> ssspSearch(Area source) {
		initial(source);
		search(source);
		return new SimpleSSSPTree(source, mmap, permit);
	}

	private class VertexInfo {
		Area vertex;
		Area pre = null;
		double sp = INFINITY;

		public VertexInfo(Area v) {
			this.vertex = v;
		}
	}

	private static class SimpleSSSPTree extends AbstractSSSPTree<Area> {
		private static final Log log = LogFactory.getLog(RangeSPFASearch.SimpleSSSPTree.class);

		private Map<Area, VertexInfo> mmap;
		private Set<Area> permit;

		public SimpleSSSPTree(Area source, Map<Area, VertexInfo> mmap, Set<Area> permit) {
			super(source);
			this.mmap = mmap;
			this.permit = permit;
			// debug();
		}

		@Override
		public double getDistance(Area target) {
			if (permit.contains(target))
				return mmap.get(target).sp;
			else
				return INFINITY;
		}

		@Override
		public Area getSPFather(Area target) {
			if (permit.contains(target))
				return mmap.get(target).pre;
			else
				return null;
		}

		// public void debug() {
		// for (Area a : permit) {
		// log.info(a + " sp:" + mmap.get(a).sp + ", father:" + mmap.get(a).pre);
		// }
		// }
	}

}