package ant.map.search.astar;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.map.objects.Route;
import ant.map.search.astar.AStarGraphNode.InnerType;
import ant.map.search.astar.AStarGraphNode.SearchInfo;
import ant.utils.FastHeap;
import ant.utils.geo.Point;

/**
 * 
 * @author Ivan
 * 
 */
public class AStarAdvancedSearch {
	private static final Log log = LogFactory.getLog(AStarAdvancedSearch.class);
	private static final boolean BAN_TWIST = false;
	
	private AStarGraph graph;
	private AbstractAStarFunction function;

	private Area originArea;
	private Point originPoint;

	private Area preFailedOriginArea = null;
	private Point preFailedOriginPoint = null;

	private Area destArea;
	private Point destPoint;
	private AgentContext context;
	private FastHeap heap;

	private String debugmsg;
	
	public AStarAdvancedSearch(AgentContext context, AStarGraph graph, AbstractAStarFunction function) {
		this.context = context;
		this.graph = graph;
		this.function = function;
		this.heap = new FastHeap(graph.getVertexSet().size() + 1);
	}

	public void setOrigin(Area originArea, Point originPoint) {
		this.originArea = originArea;
		this.originPoint = originPoint;
	}

	public void setDestination(Area destArea, Point destPoint) {
		this.destArea = destArea;
		this.destPoint = destPoint;
	}

	public Route<AStarGraphNode> search() {
		if (!(originArea.equals(preFailedOriginArea) && originPoint.equals(preFailedOriginPoint))) {
			initialSearch();
			AStarGraphNode x = searchInternal();
			if (x != null) {
				// First Succ Search
				
				debugmsg = "";
				debugmsg += "Not Use AStar Cache and found, Origin= " + originArea + " Dest=" + destArea + "\n";
				debugmsg += " (" + x.toString() + " " + x.searchInfo + ")\n";
				
				resetFailedCache();
				Route<AStarGraphNode> route = new Route<AStarGraphNode>();
				route.setDistance(x.searchInfo.g + function.costEnd(x, destArea, destPoint));
				ArrayList<AStarGraphNode> list = new ArrayList<AStarGraphNode>();
				while (true) {
					list.add(x);
					if (x.searchInfo.pre != null) {
						x = x.searchInfo.pre;
					} else {
						break;
					}
				}
				List<AStarGraphNode> ret = new ArrayList<AStarGraphNode>();
				for (int i = list.size() - 1; i >= 0; i--) {
					ret.add(list.get(i));
				}
				route.setFrom(ret.get(0));
				route.setTo(ret.get(list.size() - 1));
				route.setRoute(ret);
				return route;
			} else {
				debugmsg = "";
				debugmsg += "Not Use AStar Cache but failed, Origin= " + originArea + " Dest=" + destArea + "\n";
				
				preFailedOriginArea = originArea;
				preFailedOriginPoint = originPoint;
				return null;
			}
		} else {
			
			// Use Cache
			List<AStarGraphNode> tmp = graph.getAreaNodeMap().get(destArea.getID());
			double min = Double.POSITIVE_INFINITY;
			AStarGraphNode pick = null;
			for (AStarGraphNode node : tmp) {
				if (node.searchInfo.g + node.searchInfo.h < min) {
					if (function.finish(node, destArea, destPoint)) {
						pick = node;
						min = node.searchInfo.g + node.searchInfo.h;
					}
				}
			}
			
			if (pick == null) {
				debugmsg = "";
				debugmsg += "Use AStar Cache but not found, Origin= " + originArea + " Dest=" + destArea + "\n";
				for (AStarGraphNode node : tmp) {
					debugmsg += "    [[" + node + "(" + node.searchInfo + ")\n";
				}
				return null;
			} else {
				debugmsg = "";
				debugmsg += "Use AStar Cache, Origin= " + originArea + " Dest=" + destArea + "\n";
				debugmsg += " (" + pick.toString() + " " + pick.searchInfo + ")\n";
			}
			
			AStarGraphNode x = pick;
			Route<AStarGraphNode> route = new Route<AStarGraphNode>();
			route.setDistance(x.searchInfo.g + function.costEnd(x, destArea, destPoint));
			ArrayList<AStarGraphNode> list = new ArrayList<AStarGraphNode>();
			while (true) {
				list.add(x);
				if (x.searchInfo.pre != null) {
					x = x.searchInfo.pre;
				} else {
					break;
				}
			}
			List<AStarGraphNode> ret = new ArrayList<AStarGraphNode>();
			for (int i = list.size() - 1; i >= 0; i--) {
				ret.add(list.get(i));
			}
			route.setFrom(ret.get(0));
			route.setTo(ret.get(list.size() - 1));
			route.setRoute(ret);
			return route;
		}
	}

	private AStarGraphNode searchInternal() {
		while (!heap.isEmpty()) {
			AStarGraphNode x = ((SearchInfo) heap.pop()).node;
			x.searchInfo.inOpen = false;
			x.searchInfo.inClose = true;

			if (function.finish(x, destArea, destPoint)) {
				return x;
			}

			for (AStarGraphEdge edge : graph.edgesOf(x)) {
				AStarGraphNode y = edge.getTarget();
				
				if (BAN_TWIST) {
					if (!(x.getType() == InnerType.Building || y.getType() == InnerType.Building)) {
						if (x.searchInfo.pre == null) {
							// First Node
							EntityID sr = AbstractAStarRouteComponent.getSameRoad(x, y);
							if (sr != null) {
								if (originArea.getID().equals(sr)) {
									// GIVEUP
									continue;
								}
							}
						} else {
							AStarGraphNode z = x.searchInfo.pre;
							if (z.getType() != InnerType.Building) {
								// z - x - y
								EntityID a = AbstractAStarRouteComponent.getSameRoad(z, x);
								EntityID b = AbstractAStarRouteComponent.getSameRoad(x, y);
								if (a != null && b != null) {
									if (a.equals(b)) {
										// GIVEUP
										continue;
									}
								}
							}
						}
					}
				}
				
				double ng = x.searchInfo.g + edge.getWeight();

				// log.info("(" + x.searchInfo.g + ">" + ng + ")" + x + " -->" + y);
				if (!Double.isInfinite(ng)) {
					SearchInfo yinfo = y.searchInfo;
					if ((!yinfo.inClose) && (!yinfo.inOpen)) {
						yinfo.g = ng;
						yinfo.inOpen = true;
						yinfo.pre = x;
						heap.add(yinfo);
					} else if (yinfo.inOpen) {
						if (yinfo.g > ng) {
							yinfo.g = ng;
							yinfo.pre = x;
							heap.decadj(yinfo);
						}
					} else if (yinfo.inClose) {
						if (yinfo.g > ng) {
							yinfo.inClose = false;
							yinfo.inOpen = true;
							yinfo.g = ng;
							yinfo.pre = x;
							heap.add(yinfo);
						}
					}
				}
			} // for
				// log.info("HeapSize:" + heap.size());
		}
		return null;
	}

	private void initialSearch() {
		// Process Queue
		this.heap.clear();
		// Process Origin & Dest
		Map<AStarGraphNode, Double> origins = new HashMap<AStarGraphNode, Double>();

		for (AStarGraphNode originNode : graph.getAreaNodeMap().get(originArea.getID())) {
			double c = function.costStart(originNode, originArea, originPoint);
			if (!Double.isInfinite(c))
				origins.put(originNode, c);
		}

		// Process graph
		for (AStarGraphNode vertex : this.graph.getVertexSet()) {
			vertex.searchInfo.initial();
			vertex.searchInfo.h = this.function.heristric(vertex, destArea, destPoint);
			if (origins.containsKey(vertex)) {
				vertex.searchInfo.g = origins.get(vertex);
				vertex.searchInfo.inOpen = true;
				heap.add(vertex.searchInfo);
			} else {
				vertex.searchInfo.g = Double.POSITIVE_INFINITY;
			}
		}
	}

	public void resetFailedCache() {
		preFailedOriginArea = null;
		preFailedOriginPoint = null;
	}
	
	public void printDebug() {
		System.out.println("Adv:" + debugmsg);
	}
}
