package ant.map.search.astar;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.EntityID;
import ant.AgentContext;
import ant.map.component.AntMapBasisComponent;
import ant.map.component.RoadPassObjectsComponent;
import ant.map.objects.RoadSide;
import ant.map.search.astar.AStarGraphNode.InnerType;
import ant.utils.LazyMap;

/**
 * 
 * @author Ivan
 *
 * @depend {@link RoadPassObjectsComponent} {@link AntMapBasisComponent}
 */
public class AStarGraphBuilder {
	private static Log log = LogFactory.getLog(AStarGraphBuilder.class);
	
	private AgentContext context;
	private ArrayList<AStarGraphNode> nodelist;
	private Map<EntityID, List<AStarGraphNode>> mmap;
	private AbstractAStarFunction function;
	private AStarGraph graph;
	private RoadPassObjectsComponent rc;
	private StandardWorldModel world;

	public AStarGraphBuilder(AgentContext context, AbstractAStarFunction function) {
		this.context = context;
		this.function = function;
		this.world = context.getWorld();
		this.rc = context.getAntComponentKit().getRoadPassObjectsComponent();
		this.build();
	}
	
	public AStarGraph getGraph() {
		return graph;
	}
	
	private void build() {
		this.graph = new AStarGraph(this.function);
		buildNodes();
		this.graph.setNodelist(nodelist);
		this.graph.setAreaNodeMap(mmap);
		buildGraph();
		// log.info("AStarGraph node size:" + nodelist.size());
	}
	
	private void buildGraph() {
		// B - B
		for (Building building : context.getAntComponentKit().getAntMapBasisComponent().getBuildingSet()) {
			EntityID from = building.getID();
			for (EntityID to : building.getNeighbours()) {
				if (world.getEntity(to) instanceof Building) {
					AStarGraphNode fromNode = mmap.get(from).get(0);
					AStarGraphNode toNode = mmap.get(to).get(0);
					graph.setEdge(fromNode, toNode);
				}
			}
		}
		
		// R, R - B, B - R
		for (Road road : context.getAntComponentKit().getAntMapBasisComponent().getRoadSet()) {
			List<AStarGraphNode> nodes = mmap.get(road.getID());
			for (AStarGraphNode node : nodes) {
				if (node.getType() == InnerType.Side) {
					EntityID bid = node.getSingleSide().getNeighbourID();
					AStarGraphNode other = mmap.get(bid).get(0);
					if (other.getType() == InnerType.Building) {
						// A-B, B-A
						graph.setEdge(node, other);
						graph.setEdge(other, node);
					} else {
						throw new IllegalStateException();
					}
				}
			}
			
			for (AStarGraphNode a : nodes) {
				for (AStarGraphNode b : nodes) {
					if (!a.equals(b)) {
						graph.setEdge(a, b);
						graph.setEdge(b, a);
					}
				}
			}
		}
	}


	private void buildNodes() {
		Set<AStarGraphNode> nodes = new HashSet<AStarGraphNode>();

		// For buildings
		for (Building building : context.getAntComponentKit().getAntMapBasisComponent().getBuildingSet()) {
			nodes.add(new AStarGraphNode(world, building));
		}

		// For Roads
		for (Road r : context.getAntComponentKit().getAntMapBasisComponent().getRoadSet()) {
			for (RoadSide side : rc.getSides(r.getID())) {
				if (side.isBuildingNeighbour()) {
					nodes.add(new AStarGraphNode(world, side));
				} else if (side.isRoadNeighbour()) {
					// TODO ???
					if (side.getPair() == null) {
						log.info(side);
						Road a = (Road) world.getEntity(side.getEdge().getNeighbour());
						Edge e = a.getEdgeTo(r.getID());
						log.info(a + "," + e);
						log.info("");
					} else {
						nodes.add(new AStarGraphNode(world, side.getPair()));
					}
				}
			}
		}
		
		this.nodelist = new ArrayList<AStarGraphNode>();
		nodelist.addAll(nodes);
		Collections.sort(nodelist);
		for (int i = 0; i < nodelist.size(); i++) {
			nodelist.get(i).setLable(i);
			graph.addVertex(nodelist.get(i));
		}
		
		this.mmap = new LazyMap<EntityID, List<AStarGraphNode>>() {
			@Override
			public List<AStarGraphNode> createValue() {
				return new ArrayList<AStarGraphNode>();
			}
		};
		
		for (AStarGraphNode node : nodelist) {
			if (node.getType() == InnerType.Building) {
				mmap.get(node.getBuilding().getID()).add(node);
			} else if (node.getType() == InnerType.Side) {
				mmap.get(node.getSingleSide().getRoad().getID()).add(node);
			} else if (node.getType() == InnerType.SidePair) {
				mmap.get(node.getSidePair().getLeft().getRoad().getID()).add(node);
				mmap.get(node.getSidePair().getRight().getRoad().getID()).add(node);
			}
		}
	}
}

/*
 * private void buildGraph() {
		this.graph = new AStarGraph(function);
		for (AStarGraphNode node : nodelist) {
			graph.addVertex(node);
		}
		for (AStarGraphNode node : nodelist) {
			if (node.getType() == InnerType.Building) {
				splayBuilding(node);
			} else if (node.getType() == InnerType.Side) {
				splaySingleSide(node);
			} else {
				splaySidePair(node);
			}
		}
	}
	
	private void splaySidePair(AStarGraphNode node) {
		RoadSide left = node.getSidePair().getLeft();
		RoadSide right = node.getSidePair().getRight();
		splay(node, left);
		splay(node, right);
	}

	private void splay(AStarGraphNode source, RoadSide rpe) {
		for (AStarGraphNode target : mmap.get(rpe.getRoad().getID())) {
			if (!target.equals(source)) {
				graph.setEdge(source, target);
			}
		}
	}

	private void splaySingleSide(AStarGraphNode source) {
		RoadSide e = source.getSingleSide();
		splay(source, e);
		StandardEntity next = world.getEntity(e.getNeighbour());
		if (!(next instanceof Building)) {
			throw new RuntimeException();
		} else {
			for (AStarGraphNode target : mmap.get(next.getID())) {
				graph.setEdge(source, target);
			}
		}
	}

	private void splayBuilding(AStarGraphNode source) {
		EntityID buildingId = source.getBuilding().getID();
		for (EntityID nei : source.getBuilding().getNeighbours()) {
			StandardEntity next = world.getEntity(nei);
			if (next instanceof Road) {
				for (AStarGraphNode target : mmap.get(next.getID())) {
					if (target.getType() == InnerType.Side) {
						if (target.getSingleSide().getNeighbour().equals(buildingId)) {
							graph.setEdge(source, target);
						}
					}
				}
			} else if (next instanceof Building) {
				for (AStarGraphNode target : mmap.get(next.getID())) {
					graph.setEdge(source, target);
				}
			}
		}
	}
*/
