package ant.map.search.astar;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Road;
import rescuecore2.standard.entities.StandardWorldModel;
import ant.map.objects.RoadSide;
import ant.map.objects.RoadSidePair;
import ant.utils.FastHeapNode;
import ant.utils.IPositioning;
import ant.utils.LabeledObject;

/**
 * 
 * @author Ivan
 * 
 */
public class AStarGraphNode implements LabeledObject, Comparable<AStarGraphNode>, IPositioning {
	private static final Log log = LogFactory.getLog(AStarGraphNode.class);

	public static enum InnerType {
		Building, SidePair, Side
	}

	private final InnerType type;
	private final Building building;
	private final RoadSide side;
	private final RoadSidePair pair;
	private final int x, y;
	private int label;
	SearchInfo searchInfo = new SearchInfo(this);

	public AStarGraphNode(StandardWorldModel world, Building building) {
		this.type = InnerType.Building;
		this.building = building;
		this.side = null;
		this.pair = null;
		this.x = building.getX();
		this.y = building.getY();
	}

	public AStarGraphNode(StandardWorldModel world, RoadSide side) {
		this.type = InnerType.Side;
		this.side = side;
		this.building = null;
		this.pair = null;
		this.x = side.getX();
		this.y = side.getY();
		if (!(world.getEntity(side.getEdge().getNeighbour()) instanceof Building)) {
			throw new IllegalArgumentException("Bad edge");
		}
	}

	public AStarGraphNode(StandardWorldModel world, RoadSidePair pair) {
		this.type = InnerType.SidePair;
		this.pair = pair;
		this.building = null;
		this.side = null;
		this.x = pair.getX();
		this.y = pair.getY();
	}

	public RoadSide getInnerSide(Road r) {
		if (type == InnerType.SidePair) {
			if (pair.getLeft().getRoad().getID().equals(r.getID()))
				return pair.getLeft();

			if (pair.getRight().getRoad().getID().equals(r.getID()))
				return pair.getRight();

			return null;
		}

		if (type == InnerType.Building)
			return null;

		if (type == InnerType.Side) {
			if (side.getRoad().getID().equals(r.getID())) {
				return side;
			} else {
				return null;
			}
		}
		
		return null;
	}

	public void setLable(int label) {
		this.label = label;
		this.searchInfo.label = label;
	}

	@Override
	public int getLabel() {
		return this.label;
	}

	public InnerType getType() {
		return type;
	}

	public Building getBuilding() {
		return building;
	}

	public RoadSide getSingleSide() {
		return this.side;
	}

	public RoadSidePair getSidePair() {
		return this.pair;
	}

	@Override
	public int getX() {
		return this.x;
	}

	@Override
	public int getY() {
		return this.y;
	}

	@Override
	public int compareTo(AStarGraphNode o) {
		if (type != o.type) {
			return this.type.ordinal() - o.type.ordinal();
		} else {
			if (type == InnerType.Building) {
				return this.building.getID().getValue() - o.building.getID().getValue();
			} else if (type == InnerType.SidePair) {
				return this.pair.compareTo(o.pair);
			} else {
				return this.side.compareTo(o.side);
			}
		}
	}

	static class SearchInfo implements FastHeapNode {
		double h, g;
		boolean inOpen, inClose;
		AStarGraphNode pre;
		final AStarGraphNode node;
		private int label;

		public SearchInfo(AStarGraphNode node) {
			this.node = node;
			this.label = node.label;
			initial();
		}

		public void initial() {
			h = Double.POSITIVE_INFINITY;
			g = 0;
			inOpen = inClose = false;
			pre = null;
		}

		@Override
		public String toString() {
			return "SearchInfo [h=" + h + ", g=" + g + ", inOpen=" + inOpen + ", inClose=" + inClose + ", pre=" + pre + "]";
		}

		@Override
		public int getLabel() {
			return label;
		}

		@Override
		public double getValue() {
			return h + g;
		}

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((building == null) ? 0 : building.hashCode());
		result = prime * result + ((pair == null) ? 0 : pair.hashCode());
		result = prime * result + ((side == null) ? 0 : side.hashCode());
		result = prime * result + ((type == null) ? 0 : type.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AStarGraphNode other = (AStarGraphNode) obj;
		if (building == null) {
			if (other.building != null)
				return false;
		} else if (!building.equals(other.building))
			return false;
		if (pair == null) {
			if (other.pair != null)
				return false;
		} else if (!pair.equals(other.pair))
			return false;
		if (side == null) {
			if (other.side != null)
				return false;
		} else if (!side.equals(other.side))
			return false;
		if (type != other.type)
			return false;
		return true;
	}

	@Override
	public String toString() {
		if (type == InnerType.Building) {
			return "AStarGraphNode" + label + " [building=" + building.getID() + "]";
		}
		if (type == InnerType.Side) {
			return "AStarGraphNode" + label + " [side=" + side + "]";
		}
		if (type == InnerType.SidePair) {
			return "AStarGraphNode" + label + " [pair=" + pair + "]";
		}
		return null;
	}

}
