package ant.map.search.astar;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Road;
import ant.component.universal.ChangeSetDetailComponent;
import ant.liaison.message.IMessage;
import ant.map.component.AntMapBasisComponent;
import ant.map.component.RoadPassObjectsComponent;
import ant.map.objects.RoadSide;
import ant.map.search.astar.AStarGraphNode.InnerType;
import ant.utils.LazyMap;
import ant.utils.blockade.PassableJudgeType;
import ant.utils.blockade.RoadUtils;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Point;

/**
 * Only-My-Insight AStar Algorithm
 * 
 * @author Ivan
 * 
 * @depend {@link RoadPassObjectsComponent} {@link AntMapBasisComponent} {@link ChangeSetDetailComponent}
 */
public class StandardAStarRouteComponent extends AbstractAStarRouteComponent {
	private static final Log log = LogFactory.getLog(StandardAStarRouteComponent.class);
	
	private StandardAStarFunction function;
	private Map<Integer, Set<Integer>> ss;
	private RoadPassObjectsComponent rc;
	
	public StandardAStarRouteComponent() {
		this.ss = new LazyMap<Integer, Set<Integer>>() {
			@Override
			public Set<Integer> createValue() {
				return new HashSet<Integer>();
			}
		};
	}
	
	@Override
	public String getComponentName() {
		return StandardAStarRouteComponent.class.getName();
	}
	
	@Override
	public void initialInternal() {
		super.initialInternal();
		this.rc = context.getAntComponentKit().getRoadPassObjectsComponent();
	}
	
	@Override
	public AbstractAStarFunction getAStarFunction() {
		if (this.function == null) {
			this.function = new StandardAStarFunction();
		}
		return this.function;
	}

	@Override
	public void updateInsightRoad(Road r) {
		List<AStarGraphNode> nodelist = this.graph.getAreaNodeMap().get(r.getID());
		for (int i = 0; i < nodelist.size() - 1; i++) {
			for (int j = i + 1; j < nodelist.size(); j++) {
				AStarGraphNode n1 = nodelist.get(i);
				AStarGraphNode n2 = nodelist.get(j);
				AStarGraphEdge e12 = graph.getEdge(n1, n2);
				AStarGraphEdge e21 = graph.getEdge(n2, n1);
				if (e12 == null || e21 == null) {
					log.error("n1,n2 not connected");
				}
				if (isPassable(n1, n2, r)) {
					ss.get(n1.getLabel()).remove(n2.getLabel());
					ss.get(n2.getLabel()).remove(n1.getLabel());
				} else {
					// log.info("NP:" + n1 + "-" + n2);
					ss.get(n1.getLabel()).add(n2.getLabel());
					ss.get(n2.getLabel()).add(n1.getLabel());
				}
				e12.setWeight(function.initialCost(n1, n2));
				e21.setWeight(function.initialCost(n2, n1));
			}
		}
	}

	private boolean isPassable(AStarGraphNode n1, AStarGraphNode n2, Road r) {
		Edge from = n1.getInnerSide(r).getEdge();
		Edge to = n2.getInnerSide(r).getEdge();
		return RoadUtils.isPassable(context.getAntComponentKit().getRoadInfoCacheComponent(), from, r, to, PassableJudgeType.Default);
	}

	@Override
	public void updateMessage(IMessage message) {
		// TODO
	}

//	@Override
//	public Set<MessageType> getInterestMessageTypes() {
//		// TODO
//		return Collections.EMPTY_SET;
//	}

//	@Override
//	public void laterUpdate() {
//	}
	
	public class StandardAStarFunction extends AbstractAStarFunction {
		
		@Override
		public double heuristic(AStarGraphNode source, AStarGraphNode end) {
			return GeoUtils.manhattanDistance(source, end);
		}

		@Override
		public double heristric(AStarGraphNode source, Area end, Point endPoint) {
			return GeoUtils.manhattanDistance(source.getX(), source.getY(), endPoint.x(), endPoint.y());
		}

		@Override
		public boolean accept(AStarGraphNode source, Area end, Point endPoint) {
			if (end instanceof Road) {
				RoadSide rs = source.getInnerSide((Road) end);
				if (rs == null) {
					log.warn("Bad call");
					return false;
				}
				return RoadUtils.isPassable(context.getAntComponentKit().getRoadInfoCacheComponent(), end, endPoint, rs.getEdge(), PassableJudgeType.Default);
				
			} else {
				if (source.getType() == InnerType.Building) {
					if (source.getBuilding().getID().equals(end.getID())) {
						return true;
					} else {
						log.warn("Bad call");
					}
				}
				return false;
			}
		}

		@Override
		public double costEnd(AStarGraphNode source, Area end, Point endPoint) {
			if (accept(source, end, endPoint)) {
				return GeoUtils.manhattanDistance(source, endPoint);
			} else {
				return Double.POSITIVE_INFINITY;
			}
		}

		@Override
		public double costStart(AStarGraphNode source, Area end, Point endPoint) {
			if (accept(source, end, endPoint)) {
				return GeoUtils.manhattanDistance(source, endPoint);
			} else {
				return Double.POSITIVE_INFINITY;
			}
		}

		@Override
		public double initialCost(AStarGraphNode n1, AStarGraphNode n2) {
			InnerType t1 = n1.getType();
			InnerType t2 = n2.getType();
			if (t1 == InnerType.Building || t2 == InnerType.Building) {
				return GeoUtils.manhattanDistance(n1, n2);
			}
			
			int l1 = n1.getLabel();
			int l2 = n2.getLabel();
			if (ss.get(l1).contains(l2)) 
				return Double.POSITIVE_INFINITY;
			else
				return GeoUtils.manhattanDistance(n1, n2);
		}

		@Override
		public boolean finish(AStarGraphNode source, Area end, Point endPoint) {
			if (end instanceof Road) {
				RoadSide rs = source.getInnerSide((Road) end);
				if (rs == null) {
					return false;
				}
				return RoadUtils.isPassable(context.getAntComponentKit().getRoadInfoCacheComponent(), end, endPoint, rs.getEdge(), PassableJudgeType.Default);
				
			} else {
				if (source.getType() == InnerType.Building) {
					if (source.getBuilding().getID().equals(end.getID())) {
						return true;
					}
				}
				return false;
			}
		}
		
	}

}
