package ant.map.search.astar;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Building;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.Road;
import rescuecore2.worldmodel.EntityID;
import ant.AntComponent;
import ant.component.universal.AgentLocationComponent;
import ant.component.universal.BeliefTrailComponent;
import ant.component.universal.BeliefTrailComponent.BeliefTrail;
import ant.component.universal.ChangeSetDetailComponent;
import ant.liaison.message.IMessage;
import ant.map.component.AntMapBasisComponent;
import ant.map.component.RoadPassObjectsComponent;
import ant.map.objects.RoadSide;
import ant.map.search.astar.AStarGraphNode.InnerType;
import ant.utils.LazyMap;
import ant.utils.blockade.PassableJudgeType;
import ant.utils.blockade.RoadUtils;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Point;

/**
 * 
 * @author Ivan
 * @depend {@link RoadPassObjectsComponent} {@link AntMapBasisComponent}
 *         {@link ChangeSetDetailComponent} {@link AgentLocationComponent}
 *         {@link BeliefTrailComponent}
 */
public class TrailsAStarRouteComponent extends AbstractAStarRouteComponent implements AntComponent {
	private static final Log log = LogFactory.getLog(TrailsAStarRouteComponent.class);
	
	public static final double KNOWN_VISIBILITY = 1;
	public static final double UNKNOWN_VISIBILITY = 0.5;
	public static final double BELIEF_VISIBILITY = 0.95;
	private static final int BELIEF_INFER_STEP = 20;
	private static final double INSIGHT_SCALE = UNKNOWN_VISIBILITY / KNOWN_VISIBILITY;
	private static final double BELIEF_SCALE = UNKNOWN_VISIBILITY / BELIEF_VISIBILITY;
	
	private RoadPassObjectsComponent rpoc;
	private TrailsAStarFunction function;
	// Unable pass
	private LazyMap<Integer, Set<Integer>> ss;
	private BeliefTrailComponent btc;
	private Set<Human> agents;
	
	public TrailsAStarRouteComponent() {
		this.ss = new LazyMap<Integer, Set<Integer>>() {
			@Override
			public Set<Integer> createValue() {
				return new HashSet<Integer>();
			}
		};
	}
	
	@Override
	public String getComponentName() {
		return TrailsAStarRouteComponent.class.getName();
	}
	
	@Override
	public void initialInternal() {
		super.initialInternal();
		this.rpoc = context.getAntComponentKit().getRoadPassObjectsComponent();
		this.btc = (BeliefTrailComponent) context.getAntComponentKit().getAntComponent(BeliefTrailComponent.class);
		this.agents = context.getAntComponentKit().getAntMapBasisComponent().getAgentSet();
	}
	
	@Override
	public AbstractAStarFunction getAStarFunction() {
		if (this.function == null) {
			this.function = new TrailsAStarFunction();
		}
		return this.function;
	}
	
	@Override
	public void updateInsightRoad(Road r) {
		List<AStarGraphNode> nodelist = this.graph.getAreaNodeMap().get(r.getID());
		for (int i = 0; i < nodelist.size() - 1; i++) {
			for (int j = i + 1; j < nodelist.size(); j++) {
				AStarGraphNode n1 = nodelist.get(i);
				AStarGraphNode n2 = nodelist.get(j);
				AStarGraphEdge e12 = graph.getEdge(n1, n2);
				AStarGraphEdge e21 = graph.getEdge(n2, n1);
				if (e12 == null || e21 == null) {
					log.error("n1, n2 not connected");
				} else {
					e12.setVisibility(KNOWN_VISIBILITY);
					e21.setVisibility(KNOWN_VISIBILITY);
					if (isPassable(n1, n2, r)) {
						ss.get(n1.getLabel()).remove(n2.getLabel());
						ss.get(n2.getLabel()).remove(n1.getLabel());
					} else {
						// log.info("NotPass:" + n1 + "-" + n2);
						ss.get(n1.getLabel()).add(n2.getLabel());
						ss.get(n2.getLabel()).add(n1.getLabel());
					}
					e12.setWeight(function.initialCost(n1, n2) * INSIGHT_SCALE);
					e21.setWeight(function.initialCost(n2, n1) * INSIGHT_SCALE);
				}
			}
		}
	}
	
	private boolean isPassable(AStarGraphNode n1, AStarGraphNode n2, Road r) {
		Edge from = n1.getInnerSide(r).getEdge();
		Edge to = n2.getInnerSide(r).getEdge();
		return RoadUtils.isPassable(context.getAntComponentKit().getRoadInfoCacheComponent(), from, r, to, PassableJudgeType.Default);
	}
	
	@Override
	public void updateMessage(IMessage message) {
	}
	
	@Override
	protected void preupdate() {
		super.preupdate();
		if (btc != null) {
			for (Human h : btc.getObstervedAgentSet()) {
				List<BeliefTrail> list = this.btc.getTrails(h.getID());
				if (!list.isEmpty()) {
					BeliefTrail trail = list.get(list.size() - 1);
					if (!trail.isInvalid()) {
						int to = trail.getTrail().size(); // Not included
						int from = to - BELIEF_INFER_STEP;
						if (from < 0) from = 0;
						List<EntityID> tmp = trail.getTrail().subList(from, to);
						processTrail(tmp);
					}
				}
			}
		}
	}
	
	// 结合Clear命令, 可以使start, end两点进一步更新
	private void processTrail(List<EntityID> list) {
		int size = list.size();
		if (list.size() < 2) return;
		for (int i = 1; i < size - 1; i++) {
			EntityID loc0 = list.get(i - 1);
			EntityID loc1 = list.get(i);
			EntityID loc2 = list.get(i + 1);
			if (loc0.equals(loc1) || loc1.equals(loc2) || loc0.equals(loc2)) {
				if (loc0.equals(loc1) || loc1.equals(loc2)) {
					log.error("Dereplication Error");
				}
			} else {
				Area a0 = (Area) world.getEntity(loc0);
				Area a1 = (Area) world.getEntity(loc1);
				Area a2 = (Area) world.getEntity(loc2);
				try {
					process3(a0, a1, a2);
				} catch (Exception e) {
					log.error(e);
				}
			}
		}
	}
	
	private void process3(Area a0, Area a1, Area a2) {
		boolean isb0 = (a0 instanceof Building);
		boolean isb1 = (a1 instanceof Building);
		boolean isb2 = (a2 instanceof Building);

		if (isb1)
			return;

		// log.debug("Process3:" + a0.getID() + " ," + a1.getID() + " ," + a2.getID() + "|" + isb0 + "," + isb1 + "," + isb1);

		List<AStarGraphNode> nodelist = this.graph.getAreaNodeMap().get(a1.getID());
		AStarGraphNode left = null, right = null;
		for (AStarGraphNode node : nodelist) {
			if (node.getType() == InnerType.Side) {
				if (node.getSingleSide().getNeighbourID().equals(a0.getID())) {
					if (isb0) {
						left = node;
					} else {
						throw new IllegalStateException();
					}
				}

				if (node.getSingleSide().getNeighbourID().equals(a2.getID())) {
					if (isb2) {
						right = node;
					} else {
						throw new IllegalStateException();
					}
				}

			} else if (node.getType() == InnerType.SidePair) {
				RoadSide otherSide;
				if (node.getSidePair().getLeft().getRoad().getID().equals(a1.getID())) {
					otherSide = node.getSidePair().getRight();
				} else if (node.getSidePair().getRight().getRoad().getID().equals(a1.getID())) {
					otherSide = node.getSidePair().getLeft();
				} else {
					throw new IllegalStateException("NotFoundOtherSide");
				}

				if (otherSide.getRoad().getID().equals(a0.getID())) {
					if (!isb0) {
						left = node;
					} else {
						throw new IllegalStateException();
					}
				}

				if (otherSide.getRoad().getID().equals(a2.getID())) {
					if (!isb2) {
						right = node;
					} else {
						throw new IllegalStateException();
					}
				}

			}
		} // for
		AStarGraphEdge e12 = graph.getEdge(left, right);
		AStarGraphEdge e21 = graph.getEdge(right, left);
		// log.debug("AStarUpdate:" + left + "," + right);

		if (e12 == null || e21 == null) {
			log.error("n1, n2 not connected");
		} else {
			ss.get(left.getLabel()).remove(right.getLabel());
			ss.get(right.getLabel()).remove(left.getLabel());
			if (e12.getVisibility() < BELIEF_VISIBILITY || e12.getWeight() == Double.POSITIVE_INFINITY) {
				e12.setVisibility(BELIEF_VISIBILITY);
				e12.setWeight(function.initialCost(left, right) * BELIEF_SCALE);
			}
			if (e21.getVisibility() < BELIEF_VISIBILITY || e21.getWeight() == Double.POSITIVE_INFINITY) {
				e21.setVisibility(BELIEF_VISIBILITY);
				e21.setWeight(function.initialCost(right, left) * BELIEF_SCALE);
			}
			
		}
	}
	
	public class TrailsAStarFunction extends AbstractAStarFunction {

		@Override
		public double heuristic(AStarGraphNode source, AStarGraphNode end) {
			return GeoUtils.manhattanDistance(source, end);
		}

		@Override
		public double heristric(AStarGraphNode source, Area end, Point endPoint) {
			return GeoUtils.manhattanDistance(source.getX(), source.getY(), endPoint.x(), endPoint.y());
		}

		@Override
		public boolean accept(AStarGraphNode source, Area end, Point endPoint) {
			if (end instanceof Road) {
				RoadSide rs = source.getInnerSide((Road) end);
				if (rs == null) {
					log.warn("Bad call");
					return false;
				}
				return RoadUtils.isPassable(context.getAntComponentKit().getRoadInfoCacheComponent(), end, endPoint, rs.getEdge(), PassableJudgeType.Default);

			} else {
				if (source.getType() == InnerType.Building) {
					if (source.getBuilding().getID().equals(end.getID())) {
						return true;
					} else {
						log.warn("Bad call");
					}
				}
				return false;
			}
		}

		@Override
		public double costEnd(AStarGraphNode source, Area end, Point endPoint) {
			if (accept(source, end, endPoint)) {
				return GeoUtils.manhattanDistance(source, endPoint);
			} else {
				return Double.POSITIVE_INFINITY;
			}
		}

		@Override
		public double costStart(AStarGraphNode source, Area end, Point endPoint) {
			if (accept(source, end, endPoint)) {
				return GeoUtils.manhattanDistance(source, endPoint);
			} else {
				return Double.POSITIVE_INFINITY;
			}
		}

		@Override
		public double initialCost(AStarGraphNode n1, AStarGraphNode n2) {
			InnerType t1 = n1.getType();
			InnerType t2 = n2.getType();
			if (t1 == InnerType.Building || t2 == InnerType.Building) {
				return GeoUtils.manhattanDistance(n1, n2) / UNKNOWN_VISIBILITY;
			}

			int l1 = n1.getLabel();
			int l2 = n2.getLabel();
			if (ss.get(l1).contains(l2))
				return Double.POSITIVE_INFINITY;
			else
				return GeoUtils.manhattanDistance(n1, n2) / UNKNOWN_VISIBILITY;
		}

		@Override
		public boolean finish(AStarGraphNode source, Area end, Point endPoint) {
			if (end instanceof Road) {
				RoadSide rs = source.getInnerSide((Road) end);
				if (rs == null) {
					return false;
				}
				return RoadUtils.isPassable(context.getAntComponentKit().getRoadInfoCacheComponent(), end, endPoint, rs.getEdge(), PassableJudgeType.Default);

			} else {
				if (source.getType() == InnerType.Building) {
					if (source.getBuilding().getID().equals(end.getID())) {
						return true;
					}
				}
				return false;
			}
		}

		@Override
		public double initialVisibility(AStarGraphNode source, AStarGraphNode target) {
			return UNKNOWN_VISIBILITY;
		}
	}
}
