package ant.utils.blockade;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import rescuecore2.misc.geometry.GeometryTools2D;
import rescuecore2.misc.geometry.Line2D;
import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Edge;
import rescuecore2.worldmodel.EntityID;
import ant.utils.geo.Converter;
import ant.utils.geo.GeoUtils;
import ant.utils.geo.Point;
import ant.utils.geo.Segment;

/**
 * BlockJudger 用于PF判断Blockade是否阻塞通路
 * @author Fun
 * 
 */
public class BlockJudger {
	private static Log log = LogFactory.getLog(BlockJudger.class);

	/**
	 * does the blockade block the area ?
	 * 
	 * @param area
	 * @param blockade
	 * @return
	 */
	public static boolean blockArea(Area area, Blockade blockade,
			int agentRadius) {
		if (area.isBlockadesDefined()) {
			List<EntityID> list = area.getBlockades();
			if (!list.contains(blockade.getID())) {
				return false;
			}
			if (!area.isEdgesDefined()) {
				log.error("Area edge is undefined.");
				return false;
			}
			List<Edge> edges = area.getEdges();
			List<Edge> passableList = new ArrayList<Edge>();
			for (Edge edge : edges) {
				if (edge.isPassable()) {
					passableList.add(edge);
				}
			}
			if (passableList.size() < 2) {
				// ??
				return false;
			}
			boolean ret = false;
			while (passableList.size() >= 2) {
				Edge e1 = passableList.get(passableList.size() - 1);
				Edge e2 = passableList.get(passableList.size() - 2);
				if (checkBlockAreaE2E(area, blockade, e1, e2, agentRadius) == true) {
					ret = true;
					break;
				}
				passableList.remove(e1);
			}
			return ret;

		} else {
			return false;
		}

	}

	/**
	 * does the blockade blocks the area from edge e1 to e2?
	 * 
	 * @param area
	 * @param blockade
	 * @param e1
	 * @param e2
	 * @return
	 */
	public static boolean blockAreaE2E(Area area, Blockade blockade, Edge e1,
			Edge e2, int agentRadius) {
		if (area.isBlockadesDefined()) {
			List<EntityID> list = area.getBlockades();
			if (!list.contains(blockade)) {
				return false;
			}
			if (!area.isEdgesDefined()) {
				log.error("Area edge is undefined.");
				return false;
			}
			if (e1.isPassable() == false || e2.isPassable() == false) {
				log.error(e1 + " or " + e2 + " is not passable.");
				return false;
			}
			return checkBlockAreaE2E(area, blockade, e1, e2, agentRadius);
		} else {
			return false;
		}
	}

	/**
	 * does the blockade blocks the area from point to edge?
	 * 
	 * @param area
	 * @param blockade
	 * @param from
	 *            Point from
	 * @param to
	 *            Edge to
	 * @return
	 */
	public static boolean blockAreaP2E(Area area, Blockade blockade,
			Point from, Edge to, int agentRadius) {
		if (area.isBlockadesDefined()) {
			List<EntityID> bIds = area.getBlockades();
			if (bIds.size() == 0) {
				return false;
			}
			// 判断点是否在b内 最小外接矩形 求交 奇内偶外
			if (blockade.isApexesDefined()) {
				if(GeoUtils.inPolygon(blockade.getApexes(), (int)from.x(), (int)from.y())){
					return true;
				}
				if(GeoUtils.inPolygon(blockade.getApexes(), (int)(to.getStartX()+to.getEndX())/2, (int)(to.getStartY()+to.getEndY())/2)){
					return true;
				}
			}

			if (blockP2E(blockade, from, to, agentRadius)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * 
	 * @param blockade
	 * @param from point
	 * @param to edge
	 * @return if be blocked returns TRUE
	 */
	public static boolean blockP2E(Blockade blockade, Point from,
			Edge to, int agentRadius) {
		// TODO
		Point midPointOfE1, midPointOfE2;
		midPointOfE1 = new Point((int)((to.getStartX()+to.getEndX())/2),(int)((to.getStartY()+to.getEndY())/2));
		midPointOfE2 = new Point(from.x(), from.y());
		// TOFIX
		double direction = (midPointOfE2.x() - midPointOfE1.x())
				/ (double)(midPointOfE1.y() - midPointOfE2.y());
		double incx = agentRadius / Math.sqrt(1 + direction * direction);
		double redir = 1 / direction;
		double incy = agentRadius / Math.sqrt(1 + redir * redir);

		Line2D bline1, bline2;
		if (direction > 0) {
			double x1, x2, y1, y2;
			x1 = midPointOfE1.x() + incx;
			y1 = midPointOfE1.y() + incy;
			bline1 = new Line2D(x1, y1, midPointOfE2.x()
					- midPointOfE1.x(), midPointOfE2.y()
					- midPointOfE1.y());
			x2 = midPointOfE1.x() - incx;
			y2 = midPointOfE1.y() - incy;
			bline2 = new Line2D(x2, y2, midPointOfE2.x()
					- midPointOfE1.x(), midPointOfE2.y()
					- midPointOfE1.y());
		} else {
			double x1, x2, y1, y2;
			x1 = midPointOfE1.x() + incx;
			y1 = midPointOfE1.y() - incy;
			bline1 = new Line2D(x1, y1, midPointOfE2.x()
					- midPointOfE1.x(), midPointOfE2.y()
					- midPointOfE1.y());
			x2 = midPointOfE1.x() - incx;
			y2 = midPointOfE1.y() + incy;
			bline2 = new Line2D(x2, y2, midPointOfE2.x()
					- midPointOfE1.x(), midPointOfE2.y()
					- midPointOfE1.y());
		}

		Line2D cline1, cline2;
		cline1 = new Line2D(to.getStart(), to.getEnd());
		cline2 = new Line2D(from.x(), from.y(), to.getEndX()
				- to.getStartX(), to.getEndY() - to.getStartY());

		Point2D p11 = GeometryTools2D.getIntersectionPoint(bline1, cline1);
		Point2D p12 = GeometryTools2D.getIntersectionPoint(bline1, cline2);
		Point2D p21 = GeometryTools2D.getIntersectionPoint(bline2, cline1);
		Point2D p22 = GeometryTools2D.getIntersectionPoint(bline2, cline2);

		if (p11 == null || p12 == null || p21 == null || p22 == null) {
			return false;
		}

		Line2D boundary1 = new Line2D(p11, p12);
		Line2D boundary2 = new Line2D(p21, p22);

		if (GeoUtils.betweenTwoLines((int) boundary1.getOrigin().getX(),
				(int) boundary1.getOrigin().getY(), (int) boundary1
						.getEndPoint().getX(), (int) boundary1.getEndPoint()
						.getY(), (int) boundary2.getOrigin().getX(),
				(int) boundary2.getOrigin().getY(), (int) boundary2
						.getEndPoint().getX(), (int) boundary2.getEndPoint()
						.getY(), blockade.getX(), blockade.getY())
				&& GeoUtils.betweenTwoLines((int) cline1.getOrigin()
						.getX(), (int) cline1.getOrigin().getY(), (int) cline1
						.getEndPoint().getX(), (int) cline1.getEndPoint()
						.getY(), (int) cline2.getOrigin().getX(), (int) cline2
						.getOrigin().getY(), (int) cline2.getEndPoint().getX(),
						(int) cline2.getEndPoint().getY(), blockade.getX(),
						blockade.getY())) {

			return true;
		}

		List<Line2D> lines = Converter.generateLine2Ds(blockade.getApexes());
		for (Line2D line : lines) {
			if (GeometryTools2D.getSegmentIntersectionPoint(line, boundary1) != null
					|| GeometryTools2D.getSegmentIntersectionPoint(line,
							boundary2) != null) {
				// System.out.println(">t2");
				return true;
			}
		}
		return false;
	}

	/**
	 * does the blockade on the segment start from the Point(x,y) to midpoint of
	 * Edge?
	 * 
	 * @param area
	 * @param blockade
	 * @param x
	 * @param y
	 * @param to
	 * @return if dose return TRUE
	 */
	public static boolean isOnP2ESegment(Area area, Blockade blockade, int x,
			int y, Edge to) {
		Segment s1 = new Segment(x, y,
				(int) (to.getStartX() + to.getEndX()) / 2,
				(int) (to.getStartY() + to.getEndY()) / 2);
		if (area.isBlockadesDefined()) {
			List<EntityID> bIds = area.getBlockades();
			if (bIds.size() == 0 || (!bIds.contains(blockade))) {
				return false;
			}
			if (!blockade.isApexesDefined()) {
				return false;
			}
			List<Segment> ss = Converter.generateSegments(blockade.getApexes());
			for (Segment segment : ss) {
				if (GeoUtils.isSegmentIntersect(s1, segment)) {
					return true;
				}
			}
		}
		return false;
	}
	
	
	public static boolean isOnP2PSegment(Blockade blockade, int fromX, int fromY, int toX, int toY){
		Segment s1 = new Segment(fromX,fromY,toX,toY);
		List<Segment> ss = Converter.generateSegments(blockade.getApexes());
		for (Segment segment : ss) {
			if (GeoUtils.isSegmentIntersect(s1, segment)) {
				return true;
			}
		}
		return false;
	}

	// -------------------------------------------------------------------

	private static boolean checkBlockAreaE2E(Area area, Blockade blockade,
			Edge e1, Edge e2, int agentRadius) {
		// divide edges into two parts
		ArrayList<Edge> sideOne = new ArrayList<Edge>();
		ArrayList<Edge> sideTwo = new ArrayList<Edge>();

		WrappedBlockade b = new WrappedBlockade(blockade);

		ArrayList<Edge> edges = new ArrayList<Edge>(area.getEdges());
		for (int i = 0; i < edges.size(); i++) {
			Edge edge = edges.get(i);
			if (edge.equals(e1) || edge.equals(e2)) {
				edges.remove(i);
			}
		}
		// side one
		Point2D point = e1.getStart();
		while (!(point.equals(e2.getStart()) || point.equals(e2.getEnd()))) {
			for (int i = 0; i < edges.size(); i++) {
				Edge edge = edges.get(i);
				if (point.equals(edge.getStart())
						|| point.equals(edge.getEnd())) {
					sideOne.add(edge);
					edges.remove(edge);
					if (point.equals(edge.getStart())) {
						point = edge.getEnd();
					} else {
						point = edge.getStart();
					}
					break;
				}
			}
		}
		// side two
		point = e1.getEnd();
		while (!(point.equals(e2.getStart()) || point.equals(e2.getEnd()))) {
			for (int i = 0; i < edges.size(); i++) {
				Edge edge = edges.get(i);
				if (point.equals(edge.getStart())
						|| point.equals(edge.getEnd())) {
					sideTwo.add(edge);
					edges.remove(edge);
					if (point.equals(edge.getStart())) {
						point = edge.getEnd();
					} else {
						point = edge.getStart();
					}
					break;
				}
			}
		}

		// side one
		boolean passSideOne = true;

		if (sideOne.size() == 0) {
			if (b.getShortestDistanceToPoint(e1.getStart()) < 2 * agentRadius) {
				passSideOne = false;
			}
		}

		for (Edge sideEdge : sideOne) {
			Line2D segment = b.getShortestSegmetToEdge(sideEdge);
			if (segment == null) {
				continue;
			}
			double distance = GeometryTools2D.getDistance(segment.getOrigin(),
					segment.getEndPoint());
			if (distance < 2 * agentRadius) {
				passSideOne = false;
				break;
			}
		}

		if (passSideOne) {
			return false;
		}

		// side tow
		boolean passSideTwo = true;
		if (sideTwo.size() == 0) {
			if (b.getShortestDistanceToPoint(e1.getEnd()) < 2 * agentRadius) {
				passSideTwo = false;
			}
		}
		for (Edge sideEdge : sideTwo) {
			Line2D segment = b.getShortestSegmetToEdge(sideEdge);
			if (segment == null) {
				continue;
			}
			double distance = GeometryTools2D.getDistance(segment.getOrigin(),
					segment.getEndPoint());
			if (distance < 2 * agentRadius) {
				passSideTwo = false;
				break;
			}
		}

		if (passSideTwo) {
			return false;
		}

		return true;
	}
}
