package ant.utils.geo;

import java.util.ArrayList;
import java.util.List;

/**
 * 
 * @author Fun
 * 
 */
public class Polygon {
	private static final int INVALID_VALUE = -1;
	private List<Segment> edges;
	private List<Point> vertices = null;
	private int[] apexes = null;
	private Point center = null;
	private int radius = INVALID_VALUE;

	public Polygon(List<Segment> edges) {
		this.edges = edges;
	}

	public Polygon(int[] apexes) {
		this.apexes = apexes;
		edges = Polygon.transApexesToEdges(apexes);
	}

	public int[] getApexes() {
		if (apexes == null) {
			apexes = new int[edges.size() * 2];
			for (int i = 0; i < edges.size(); i++) {
				Point point = edges.get(i).getOrigin();
				apexes[2 * i] = point.x();
				apexes[2 * i + 1] = point.y();
			}
		}
		return apexes;
	}

	public List<Segment> getEdges() {
		return edges;
	}

	public List<Point> getVertices() {
		if (vertices == null) {
			vertices = new ArrayList<Point>();
			for (int i = 0; i < edges.size(); i++) {
				Point point = edges.get(i).getOrigin();
				vertices.add(point);
			}
		}
		return vertices;
	}

	public Point getCenter() {
		if (center == null) {
			if (apexes == null) {
				apexes = getApexes();
			}
			int x = 0, y = 0;
			int count = edges.size();
			for (int i = 0; i < count; i++) {
				x += apexes[2 * i];
				y += apexes[2 * i + 1];
			}
			if (count > 0) {
				x /= count;
				y /= count;
			} else {
				x = apexes[0];
				y = apexes[1];
			}

			center = new Point(x, y);
		}
		return center;
	}

	public int getRadius() {
		if (radius == INVALID_VALUE) {
			if (vertices == null) {
				getVertices();
			}
			if (center == null) {
				getCenter();
			}
			double distance = 0;
			for (Point v : vertices) {
				distance += v.distance(center);
			}
			distance /= vertices.size();
			radius = (int) distance;
		}
		return radius;
	}

	@Override
	public String toString() {
		StringBuffer ret = new StringBuffer();
		ret.append("Polygon [");
		List<Point> vs = getVertices();
		for (Point point : vs) {
			ret.append("(" + point.x() + "," + point.y() + "), ");
		}
		ret.append("]");
		return ret.toString();
	}

	@Override
	public int hashCode() {
		return edges.hashCode();
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Polygon other = (Polygon) obj;
		List<Point> selfV = this.getVertices();
		List<Point> otherV = other.getVertices();
		if (selfV.size() != otherV.size()) {
			return false;
		}
		for (int i = 0; i < selfV.size(); i++) {
			if (!otherV.get(i).equals(selfV.get(i))) {
				return false;
			}
		}
		return true;
	}

	public static List<Segment> transApexesToEdges(int[] apexes) {
		List<Segment> ret = new ArrayList<Segment>();
		if (apexes == null) {
			return ret;
		}

		if (apexes.length <= 4) {
			return ret;
		}

		for (int i = 0; i < apexes.length - 3; i += 2) {
			ret.add(new Segment(apexes[i], apexes[i + 1], apexes[i + 2], apexes[i + 3]));
		}

		ret.add(new Segment(apexes[apexes.length - 2], apexes[apexes.length - 1], apexes[0],
				apexes[1]));
		return ret;
	}

	/**
	 * 
	 * @return mid point of nearest edge
	 */
	public static Point getNearestEdgeOfPolygon(int refx, int refy, int[] apexes) {
		int min = Integer.MAX_VALUE;
		int tx = -1;
		int ty = -1;
		for (int i = 0; i < apexes.length; i += 2) {
			int a1 = i;
			int a2 = (i + 2) % apexes.length;
			int cx = (apexes[a1] + apexes[a2]) / 2;
			int cy = (apexes[a1 + 1] + apexes[a2 + 1]) / 2;
			int val = (int) GeoUtils.dist(refx, refy, cx, cy);
			if (val < min) {
				min = val;
				tx = cx;
				ty = cy;
			}
		}
		return new Point(tx, ty);
	}
}
