package ant.utils.traffic;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import rescuecore2.misc.collections.LazyMap;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;

import com.infomatiq.jsi.IntProcedure;
import com.infomatiq.jsi.Rectangle;
import com.infomatiq.jsi.SpatialIndex;
import com.infomatiq.jsi.rtree.RTree;

public class TrafficStepManager {
	private Map<Integer, TrafficStepArea> areaByID;
    private Map<Integer, TrafficStepBlockade> blockadeByID;
    private Map<Area, TrafficStepArea> areas;
    private Map<Blockade, TrafficStepBlockade> blocks;
    private Map<TrafficMove, TrafficStepAgent> agents;
    private Map<TrafficStepArea, Collection<TrafficStepArea>> areaNeighbours;

    private SpatialIndex index;

    /**
       Construct a new TrafficManager.
    */
    public TrafficStepManager() {
        areas = new HashMap<Area, TrafficStepArea>();
        areaByID = new HashMap<Integer, TrafficStepArea>();
        blocks = new HashMap<Blockade, TrafficStepBlockade>();
        blockadeByID = new HashMap<Integer, TrafficStepBlockade>();
        agents = new HashMap<TrafficMove, TrafficStepAgent>();
        areaNeighbours = new LazyMap<TrafficStepArea, Collection<TrafficStepArea>>() {
            @Override
            public Collection<TrafficStepArea> createValue() {
                return new HashSet<TrafficStepArea>();
            }
        };
        index = new RTree();
        index.init(new Properties());
    }

    /**
       Find the area that contains a point.
       @param x The X coordinate.
       @param y The Y coordinate.
       @return The TrafficStepArea that contains the given point, or null if no such area is found.
    */
    public TrafficStepArea findArea(double x, double y) {
        final List<TrafficStepArea> found = new ArrayList<TrafficStepArea>();
        index.intersects(new Rectangle((float)x, (float)y, (float)x, (float)y), new IntProcedure() {
                @Override
                public boolean execute(int id) {
                    found.add(areaByID.get(id));
                    return true;
                }
            });
        for (TrafficStepArea next : found) {
            if (next.contains(x, y)) {
                return next;
            }
        }
        return null;
    }

    /**
       Get the neighbouring areas to a TrafficStepArea.
       @param area The area to look up.
       @return All TrafficAreas that share an edge with the given area.
    */
    public Collection<TrafficStepArea> getNeighbours(TrafficStepArea area) {
        return areaNeighbours.get(area);
    }

    /**
       Get all agents in the same area or a neighbouring area as an agent.
       @param agent The agent to look up.
       @return All agents (except the input agent) that are in the same or a neighbouring area.
    */
    public Collection<TrafficStepAgent> getNearbyAgents(TrafficStepAgent agent) {
        Set<TrafficStepAgent> result = new HashSet<TrafficStepAgent>();
        result.addAll(agent.getArea().getAgents());
        for (TrafficStepArea next : getNeighbours(agent.getArea())) {
            result.addAll(next.getAgents());
        }
        result.remove(agent);
        return result;
    }

    /**
       Remove all objects from this manager.
    */
    public void clear() {
        areas.clear();
        blocks.clear();
        agents.clear();
        areaNeighbours.clear();
        areaByID.clear();
        blockadeByID.clear();
        index = new RTree();
        index.init(new Properties());
    }

    public void clearAgent(){
    	for(TrafficStepAgent agent: agents.values()){
    		TrafficStepArea current = agent.getArea();
    		if(current == null){
    			continue;
    		}
    		current.removeAgent(agent);
    	}
    	agents.clear();
    }
    
    public void clearBlockade(){
    	
    }
    
    
    /**
       Register a new TrafficStepArea.
       @param area The TrafficStepArea to register.
    */
    public void register(TrafficStepArea area) {
        areas.put(area.getArea(), area);
        int id = area.getArea().getID().getValue();
        areaByID.put(id, area);
        index.add(area.getBounds(), id);
    }

    /**
       Register a new TrafficStepAgent.
       @param agent The TrafficStepAgent to register.
    */
    public void register(TrafficStepAgent agent) {
        agents.put(agent.getMove(), agent);
    }

    /**
       Register a new TrafficStepBlockade.
       @param block The TrafficStepBlockade to register.
    */
    public void register(TrafficStepBlockade block) {
        blocks.put(block.getBlockade(), block);
        blockadeByID.put(block.getBlockade().getID().getValue(), block);
    }

    /**
       Remove a blockade.
       @param block The TrafficStepBlockade to remove.
    */
    public void remove(TrafficStepBlockade block) {
        remove(block.getBlockade());
    }

    /**
       Remove a blockade.
       @param block The Blockade to remove.
    */
    public void remove(Blockade block) {
        blocks.remove(block);
        blockadeByID.remove(block.getID().getValue());
    }

    /**
       Get all TrafficAgents.
       @return All TrafficAgents.
    */
    public Collection<TrafficStepAgent> getAgents() {
        return Collections.unmodifiableCollection(agents.values());
    }

    /**
       Get all TrafficAreas.
       @return All TrafficAreas.
    */
    public Collection<TrafficStepArea> getAreas() {
        return Collections.unmodifiableCollection(areas.values());
    }

    /**
       Get all TrafficBlockades.
       @return All TrafficBlockades.
    */
    public Collection<TrafficStepBlockade> getBlockades() {
        return Collections.unmodifiableCollection(blocks.values());
    }

    /**
       Compute pre-cached information about the world. TrafficStepArea and TrafficStepAgent objects must have already been registered with {@link #register(TrafficStepArea)} and {@link #register(TrafficStepAgent)}.
       @param world The world model.
    */
    public void cacheInformation(StandardWorldModel world) {
        areaNeighbours.clear();
        for (StandardEntity next : world) {
            if (next instanceof Area) {
                computeNeighbours((Area)next, world);
            }
        }
    }

    /**
       Get the TrafficStepArea that wraps a given Area.
       @param a The area to look up.
       @return The TrafficStepArea that wraps the given area or null if no such TrafficStepArea exists.
    */
    public TrafficStepArea getTrafficArea(Area a) {
        return areas.get(a);
    }

    /**
       Get the TrafficStepBlockade that wraps a given Blockade.
       @param b The blockade to look up.
       @return The TrafficStepBlockade that wraps the given blockade or null if no such TrafficStepBlockade exists.
    */
    public TrafficStepBlockade getTrafficBlockade(Blockade b) {
        return blocks.get(b);
    }

    /**
       Get the TrafficStepAgent that wraps a given human.
       @param move The human to look up.
       @return The TrafficStepAgent that wraps the given human or null if no such TrafficStepAgent exists.
    */
    public TrafficStepAgent getTrafficAgent(TrafficMove move) {
        return agents.get(move);
    }

    private void computeNeighbours(Area a, StandardWorldModel world) {
        Collection<TrafficStepArea> neighbours = areaNeighbours.get(getTrafficArea(a));
        neighbours.clear();
        for (EntityID id : a.getNeighbours()) {
            Entity e = world.getEntity(id);
            if (e instanceof Area) {
                neighbours.add(getTrafficArea((Area)e));
            }
        }
    }
}
