package ant.utils.traffic;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.uncommons.maths.number.NumberGenerator;
import org.uncommons.maths.random.GaussianGenerator;

import rescuecore2.misc.geometry.Point2D;
import rescuecore2.standard.entities.Area;
import rescuecore2.standard.entities.Blockade;
import rescuecore2.standard.entities.Edge;
import rescuecore2.standard.entities.Human;
import rescuecore2.standard.entities.StandardEntity;
import rescuecore2.standard.entities.StandardWorldModel;
import rescuecore2.worldmodel.Entity;
import rescuecore2.worldmodel.EntityID;
import rescuecore2.worldmodel.WorldModel;
import rescuecore2.worldmodel.WorldModelListener;
import traffic3.simulator.PathElement;
import ant.AgentContext;

public class TrafficStepSimulator {
	private static Log log = LogFactory.getLog(TrafficStepSimulator.class);

	private static final double STEP_TIME_MS = 100;
	private static final double REAL_TIME_S = 60;
	private static final int MICROSTEPS = (int) ((1000.0 / STEP_TIME_MS) * REAL_TIME_S);

	private static final int RESCUE_AGENT_RADIUS = 500;
	private static final int CIVILIAN_RADIUS = 200;
	private static final double RESCUE_AGENT_VELOCITY_MEAN = 0.7;
	private static final double RESCUE_AGENT_VELOCITY_SD = 0.1;
	private static final double CIVILIAN_VELOCITY_MEAN = 0.2;
	private static final double CIVILIAN_VELOCITY_SD = 0.002;

	private TrafficStepManager manager;
	private AgentContext context;
	private StandardWorldModel world;
	private Human me;
	private NumberGenerator<Double> agentVelocityGenerator;
	private NumberGenerator<Double> civilianVelocityGenerator;

	public TrafficStepSimulator(AgentContext context) {
		this.context = context;
		this.world = context.getWorld();
		this.me = context.getHuman();
		this.manager = new TrafficStepManager();
		initial();
	}

	private void initial() {
		TrafficStepConstants.init(context.getConfig());
		manager.clear();

		for (StandardEntity next : world) {
			if (next instanceof Area) {
				convertAreaToTrafficStepArea((Area) next);
			}
		}

		for (StandardEntity next : world) {
			if (next instanceof Blockade) {
				convertBlockade((Blockade) next);
			}
		}

		agentVelocityGenerator = new GaussianGenerator(RESCUE_AGENT_VELOCITY_MEAN,
				RESCUE_AGENT_VELOCITY_SD, context.getConfig().getRandom());
		civilianVelocityGenerator = new GaussianGenerator(CIVILIAN_VELOCITY_MEAN,
				CIVILIAN_VELOCITY_SD, context.getConfig().getRandom());

		world.addWorldModelListener(new WorldModelListener<StandardEntity>() {
			@Override
			public void entityAdded(WorldModel<? extends StandardEntity> model, StandardEntity e) {
				if (e instanceof Blockade) {
					convertBlockade((Blockade) e);
				}
			}

			@Override
			public void entityRemoved(WorldModel<? extends StandardEntity> model, StandardEntity e) {
				if (e instanceof Blockade) {
					Blockade b = (Blockade) e;
					TrafficStepBlockade block = manager.getTrafficBlockade(b);
					block.getArea().removeBlockade(block);
					manager.remove(block);
				}
			}

		});
		manager.cacheInformation(world);
	}

	public static class TrafficSimRes {
		private double x;
		private double y;
		private EntityID location;

		public TrafficSimRes(double x, double y, EntityID location) {
			super();
			this.x = x;
			this.y = y;
			this.location = location;
		}

		public double getX() {
			return x;
		}

		public void setX(double x) {
			this.x = x;
		}

		public double getY() {
			return y;
		}

		public void setY(double y) {
			this.y = y;
		}

		public EntityID getLocation() {
			return location;
		}

		public void setLocation(EntityID location) {
			this.location = location;
		}

	}

	public TrafficSimRes simulate(TrafficMove move) {
		TrafficSimRes ret = null;

		manager.clearAgent();

		for (TrafficStepArea next : manager.getAreas()) { // TODO
			next.clearBlockadeCache();
		}

		// Human human = (Human) world.getEntity(move.getAgentID());

		convertHuman(move, agentVelocityGenerator, civilianVelocityGenerator);

		for (TrafficStepAgent agent : manager.getAgents()) {
			agent.clearPath();
			agent.clearPositionHistory();
			agent.setMobile(true);
		}

		handleMove(move);

		// if (human.isHPDefined() && human.getHP() <= 0) {
		// log.debug("Agent " + human + " is dead");
		// manager.getTrafficAgent(human).setMobile(false);
		// }
		// if (human.isPositionDefined()
		// && (world.getEntity(human.getPosition()) instanceof AmbulanceTeam)) {
		// log.debug("Agent " + human + " is in an ambulance");
		// manager.getTrafficAgent(human).setMobile(false);
		// }
		// if (human.isBuriednessDefined() && human.getBuriedness() > 0) {
		// log.debug("Agent " + human + " is buried");
		// manager.getTrafficAgent(human).setMobile(false);
		// }
		timestep();

		TrafficStepAgent agent = manager.getTrafficAgent(move);

		if (agent != null) {
			TrafficStepArea location = agent.getArea();
			if (location == null) {
				log.error("location:" + location);
			} else {
				ret = new TrafficSimRes(agent.getX(), agent.getY(), location.getArea().getID());
			}
		}

		return ret;
	}

	private void handleMove(TrafficMove move) {
		// Human human = (Human) world.getEntity(move.getAgentID());
		TrafficStepAgent agent = manager.getTrafficAgent(move);
		EntityID current = move.getPath().get(0);
		if (current == null) {
			log.warn("Rejecting move: Agent position is not defined");
			return;
		}
		Entity currentEntity = world.getEntity(current);
		if (!(currentEntity instanceof Area)) {
			log.warn("Rejecting move: Agent position is not an area: " + currentEntity);
			return;
		}
		Area currentArea = (Area) currentEntity;
		List<EntityID> list = move.getPath();
		List<PathElement> steps = new ArrayList<PathElement>();
		// Check that all elements refer to Area instances and build the list of
		// target points
		// Target points between areas are the midpoint of the shared edge
		for (Iterator<EntityID> it = list.iterator(); it.hasNext();) {
			EntityID next = it.next();
			if (next.equals(current)) {
				continue;
			}
			Entity e = world.getEntity(next);
			if (!(e instanceof Area)) {
				log.warn("Rejecting move: Entity ID " + next + " is not an area: " + e);
				return;
			}
			Edge edge = currentArea.getEdgeTo(next);
			if (edge == null) {
				log.warn("Rejecting move: Entity ID " + next + " is not adjacent to " + currentArea);
				return;
			}
			Area a = (Area) e;
			int x = (edge.getStartX() + edge.getEndX()) / 2;
			int y = (edge.getStartY() + edge.getEndY()) / 2;
			Point2D edgePoint = new Point2D(x, y);
			Point2D centrePoint = new Point2D(currentArea.getX(), currentArea.getY());
			steps.add(new PathElement(next, edge.getLine(), edgePoint, centrePoint));
			current = next;
			currentArea = a;
		}
		int targetX = move.getDestinationX();
		int targetY = move.getDestinationY();
		if (targetX == -1 && targetY == -1) {
			targetX = currentArea.getX();
			targetY = currentArea.getY();
		} else if (list.isEmpty()) {
			log.warn("Rejecting move: Path is empty");
			return;
		}
		steps.add(new PathElement(current, null, new Point2D(targetX, targetY)));
		agent.setPath(steps);
	}

	private void convertAreaToTrafficStepArea(Area area) {
		manager.register(new TrafficStepArea(area));
	}

	private void convertBlockade(Blockade b) {
		Area a = (Area) world.getEntity(b.getPosition());
		TrafficStepArea area = manager.getTrafficArea(a);
		if (area == null) {
			convertAreaToTrafficStepArea(a);
			area = manager.getTrafficArea(a);
		}
		TrafficStepBlockade block = new TrafficStepBlockade(b, area);
		manager.register(block);
		area.addBlockade(block);
	}

	private void convertHuman(TrafficMove move, NumberGenerator<Double> agentVelocityGenerator,
			NumberGenerator<Double> civilianVelocityGenerator) {
		double radius = 0;
		double velocityLimit = 0;
		radius = RESCUE_AGENT_RADIUS;
		velocityLimit = agentVelocityGenerator.nextValue();
		TrafficStepAgent agent = new TrafficStepAgent(move, manager, radius, velocityLimit);

		// Set Location
		agent.setLocation(move.getFromX(), move.getFromY());
		manager.register(agent);
	}

	private void timestep() {
		for (TrafficStepAgent agent : manager.getAgents()) {
			agent.beginTimestep();
		}
		for (int i = 0; i < MICROSTEPS; i++) {
			microstep();
		}
		for (TrafficStepAgent agent : manager.getAgents()) {
			agent.endTimestep();
		}
	}

	private void microstep() {
		for (TrafficStepAgent agent : manager.getAgents()) {
			agent.step(STEP_TIME_MS);
		}
	}
}
